#if defined(__MVS__)
/* Enforce old HNLookup for z/OS */
#undef HNLUP2
#endif

#if !defined (MQCD_VERSION_8)
#define MQCD_VERSION_8 8
#endif
/*
* BlockIP2.c -	security channel exit to control connection name of incomming call to a channel,
*		This security exit will only allow connection if the connectionname, and username
*		match the specification on the channel definition or conform to the rules in the
*		specified file.
*
* --------------------------------------------------------------------------
* Description
*
* This exit will check the connection name and see if it matches the pattern of the one
* specified.
* This module was originally based on the IBM support-pack MS05.
*
* This version of the program is a development edition and is without any
* guarantee, and use of it is on your own risk.
*
* Author:
*	   Jrgen Pedersen (c)
*	   email: oz1ccg@mrmq.dk
*
* Copyright:
*	   This program may be used and changed, but may not be sold
*	   or incoperated in any softwarepackage without the authors written permission.
*	   I would like to receive a copy of the improved version, to gather ideas
*	   for further improvements for my exit suite.
*
* --------------------------------------------------------------------------
* History
*
* Written by Jrgen Pedersen								  - 12Dec02 DK	v1.00
*
* Changes
*	   Jrgen H. Pedersen								  - 06Jan04 DK	v1.10
*	   Improved due to inspiration from T.Rob, Sriniddhi and Peter Potkay
*	   Now we allow more patterns specified, seperated by semicolon (;), the number
*	   of patterns is limited ny size of SCYDATA field.
*	   Protection against JMS intruders, or users using mqm or MUSR_MQADMIN.
*
*	   Michael Dag									  - 20Jan04 NL	v1.11
*	   Changed / / comments to / * with ending ... as it wasn't suited for C
*	   compiler on AIX
*	   Changed IFDEF to find out WIN32 or UNIX as UNIX can have 1 CPU aswell
*	   Commented the #pragma statement as this wasn't suited for C compiler
*	   on AIX
*	   Removed * from string and wild in the wildcmp routine and added [0] as
*	   it didn't work on Solaris.
*
*	   Jrgen H. Pedersen								  - 22Jan04 DK	v1.12
*	   Problem with simple pattern (172.22.22.*) this failed due to a design
*	   mistake, this is now solved. Now should the Exit work as designed.
*	   Log improved using real date format, instead of internal formats...
*	   Conditional compilation added, so the same source can be compiled on
*	   differnt platforms. Some other improvements due to nature of UNIX/LINUX
*	   so the log list is fine.
*
*	   Jrgen H. Pedersen								  - 06Feb04 DK	v1.14
*	   Problem with missing timestamp in the log solved. Version 1.14.
*	   Problem with patternmatching without trailing asterix, this problem
*	   was introduced in 1.11.
*
*	   Jrgen H. Pedersen								  - 10Feb04 DK	v1.15
*	   Added funtionallity to verify the connecting userid, based on remote
*	   userid, this matching is also based on wild_cmp(), so wildcards is
*	   allowed. This pattern matching is case sensetive, so was51 is different
*	   from Was51. Check also added for right exit invocation. Any other exit
*	   will just be terminated.
*
*	   inspiration:
*		 SCYDATA('10.1.10.*;-n;userids=zz*,yy*')
*		 SCYDATA('10.1.10.*;10.2.11.*;userids=michael,zz*,yy*')
*
*	   Jrgen H. Pedersen								  - 15Feb04 DK	v1.16
*	   Added functionallity to allow specification rules to be passed in a file,
*	   this file is specified in SCYDATA() like (C:\path\rulefile.txt) or
*	   (/var/mqm/exits/rulefile.txt) depending on your platform.
*	   This allows you to thae a rulefile per channel, and this offers flexibility
*	   to implement various security schemas.
*	   Two extra parameters have been added to SCYDATA: -d & -q. -d is used for
*	   debuging purposes, while -q is q Quitemode, where the output is very
*	   reduced (only one line per attempt to connect).
*
*	   Jrgen H. Pedersen								  - 16Feb04 DK	v1.17
*	   A serious user-validation error was introduced in 1.16 which only allowed usage
*	   of FN= mode. The error is corrected (i hope), and my tests shows it works.
*	   Enhanced error reporting, and control of filters. And still be restrictive
*	   so a nRetCode don't open the connection.
*
*	   Jrgen H. Pedersen								  - 20Feb04 DK	v1.18
*	   The exit is now able to allow blank lines and empty lines.
*
*	   Jrgen H. Pedersen								  - 20Feb04 DK	v1.20
*	   The exit is extented to display SSL properties. Pattern matching for x503
*	   development started. Specifications are passed in configuration file.
*	   no test are done if options != "FN=" due to performance reasons.
*	   This filter option is combined with an DN->User generic mapping function
*	   to ease systems administration, so unknown users with valid certificates
*	   can be kept outside our world.
*	   The way the SSL-specs are searched is top down and with stop on first match,
*	   this gives the ablility to "shop" in the CN's, and just take the ones we
*	   like and block the rest.
*	   Following options are implemented:
*	   SSL=CN=ibmwebspheremqQSJHPT01,MCA=user;	 change MCAUSER for this CN
*	   SSL=CN=ibmwebspheremq*,MCA=*;			 allow all connections from ibmwebsphere*
*	   SSL=CN=*;BLOCK;							 Refuse all connections
*	   SSL=CN=*;MCA=NoBody; 					 set MCAUSER for CNs not matching
*
*	   Jrgen H. Pedersen								  - 05Mar04 DK	v1.21
*	   The logic of the exit have been changed so it's no longer possible to penetrate
*	   this exit using a exit on the client end, to send a sec_msg, to get a connection
*	   accepted status.
*	   This design mistake would leave the channel open to almost anybody, that was capable
*	   of creating a client security exit.
*
*	   Jrgen H. Pedersen								  - 20Mar04 DK	v1.22
*	   Added functions to specify connectionname/userid match, so we can combine
*	   the features of security. This matching feature only works on FN= mode, where
*	   the connectionname/userid is specified using the CON= keyword:
*	   CON=<conname>;<userids>[;MCA={*|userid}];
*	   CON=10.11.12.*;u*;MCA=sysoper;
*	   CON=192.162.1.14;zz*;
*	   The list is searched for first match of connectionname+userid, which means it's very
*	   important to specify the options in the right order.
*	   If you specify:
*	   CON=*;*;MCA=root;
*	   CON=162.12.*;peter;
*	   This will have the result that all incomming connection attemts will have the MCAUSER forced
*	   to "mqm", even "peter", because all connetions+userids will match*;*;. You see the point ?
*
*  2004-03-22  Sid Young Complete re-write and restructure of code.
*          Changed logic to deny all and only accept if all conditions required are met
*          Added enhanced logging capability
*          Changed BlockExit() to call seperate init, parse and test methods.
*
*  2004-03-29  Sid Young Added code to support chanel and user name stamping in log file name.
*	   Found type in ProcessLine(). Tested OK.
*
*	   Jrgen H. Pedersen								  - 30Mar04 DK	v2.10
*	   Retrofitted some code to create a windows version of the program that could
*	   be loaded. And to allow compability with old versions that works without a
*	   rule-file. Restrictions added so connections will be refused if failure in
*	   rule/parameter specifications, because my auto test passed many connection
*	   attempts that should be blocked due to rules.
*	   This is a valid rule spec:127*;userids=u*;-q; resulted in : Warning - Unknown Tag [127*;userids=u*;-q;]
*	   and gives connection accepted .... :o(
*	   Problem with blank rule lines -> gives connection blocked.
*
*	   Exit changed so it respects commented lines containg valid commands like:
*	   #BlockMqmUsers=Y; which in the redesigned version was taken as a command.
*	   This was caused by missing init of variables if the exit was redriven, so the state
*	   was saved between invocations on the same connection.
*
*	   Specification of logfiles changed to conform with the used standard where all specs.
*	   are terminated with a semicolon.
*	   Small changes where numeric constants have been changed to WMQ supplied constants.
*
*	   Jrgen H. Pedersen								  - 27May04 DK	v2.11
*	   Some editorial changes to allow compilation under AIX without problems.
*	   UNIX logging changed, so misspecified log information still will result in logging,
*	   where the spec. is bypassed. LogDirecty removed and is replaced with LogPath, for
*	   complience between environments.
*
*	   Jrgen H. Pedersen								  - 29May04 DK	v2.12
*	   Repporting of refused connections added to log, so we can see who is trying to get in,
*	   this was disapeared on Petterns= and CON= mismatch this is fixed so we can continue.
*	   UseridUpperLowerCase= Introduced. Default is case sensetive to be compatible with older
*	   versions. To ignore case use UseridUpperLowerCase=*. Any other is invalid.
*	   BlockUsers=, added, so we now can use a negativelist, but still with precedence of
*	   CON=, and SSL=, meaning that we can change an incomming user to a blocked user!
*
*	   Jrgen H. Pedersen								  - 07Jun04 DK	v2.13
*	   Validation of SCYDATA field is enhanced, so more errors are reported, for investigation.
*
*	   Neil Casey									  - 03Aug04 	v2.13nwc
*	   No change to functionality at all, but added Windows compiler information so
*	   that external link directives are not required, and added MVS directives to
*	   allow generation of an MVS targetted exit. Files specifications on MVS can be
*	   //DD:JCLNAME or UFS hierarchical names (ie unix names)
*	   Also fixed issues with ConnectionNames. The code was using the EXIT_NAME_LENGTH to
*	   get a value for the length of the ConName. This doesn't work on MVS where the
*	   exit name length is 8. Change all references to use the MQ defined constant
*	   MQ_CONN_NAME_LENGTH.
*
*	   Jrgen H. Pedersen								  - 05Aug04 DK	v2.14
*	   Validation of SCYDATA field is enhanced, so errors on specifying FN=file is
*	   enhanced. If KW FN= not specified, it's mandatory that SCYDATA contains either
*	   a asterix(*), question mark(?) or a pattern starting with 0-9. Due to the fact
*	   that the latest problems reported is in this area of bad specification.
*	   Hard-brackets ([]) is changed in z/OS implementation to (<>), so it becomes
*	   readable in non-english environments.
*
*	   Neil Casey									   - 10Aug04	v2.13nwc2
*	   Extended maximum pattern lengths to 256 characters. Changed logic for building
*	   the pattern strings so the multiple Pattern= or User= etc lines add to the
*	   previously built string, instead of overwriting it. This is a function change
*	   which makes this version behave differently when faced with multiple lines of data
*	   which previously just used the last data found.
*
*	   Jrgen H. Pedersen								  - 11Aug04 DK	v2.15
*	   just a small one... together with a small code change for solaris so that the
*	   timestamp is printed too. This is done using wcsftime...
*
*	   Neil Casey									  - 24Aug04 	v2.16
*	   Remove the wcsftime call. The parameter passed to it (char*) by the code did
*	   not match the expected parameter (wchar*) The timestamping works fine on Solaris
*	   without this mod. Remove some commented out code which was for debugging.
*          Reformat source indenting.
*
*	   Jrgen H. Pedersen								  - 08Feb05 DK	v2.17
*	   A small change about setting MCAUSER based on CON= control, z/OS
*	   will have a field filled with trailing spaces. Added RespectMCA keyword.
*
*	   Jrgen H. Pedersen								  - 21Apr05 DK	v2.18
*	   Handling of SEC_PARMS added as part of WMQ 6.0 support.
*
*	   Jrgen H. Pedersen								  - 24Apr05 DK	v2.20
*	   Added support for max connections on a given channel.
*	   new keyword: MAXCHL=ChannelName;MaxConnections; was added.
*	   This is currently not supported on WebSphere MQ for Z/OS.
*
*	   Jrgen H. Pedersen								  - 03May05 DK	v2.21
*	   Problem with userid check in CON= Fixed.
*	   Wildcard compare extented with character range, numeric, alpha patterns,
*	   to enhance filtering capabilities.
*	   AllowBlankUserID added, Default is changed so we don't allow blank userids
*          anymore.
*
*          Jrgen H. Pedersen								  - 17May05 DK	v2.22
*          MAXCHL=ChannelName;MaxConnections; is now supported on
*          WebSphere MQ for Z/OS prior to version 6.0. using MQSC:
*          "DISPLAY CHS(<channel name>) CURRENT" command.
*          Explanation on exit-reason and channel type enhanced, to give the user
*          a better understanding of what is happening.
*          We have fixed the known problem in the userid-list of the CON= specification
*          so it now works as designed.
*          On z/OS we're also blocking connection attempts using the STCID from the xxxxCHIN task,
*          this should prevent unauth. access.
*
*          Jrgen H. Pedersen								  - 1Jul05 DK	v2.30
*          Problem with reentant code changed, so it should be reentrant.
*          Syntax check of parameter specification is improved.
*
*          Jrgen H. Pedersen								  - 11Sep05 DK	v2.31
*          strtok changed to strtok_r for non windows implementations to deal with
*          reentant code problems. This should help make it more stable.
*
*          Jrgen H. Pedersen								  - 20Sep05 DK	v2.32
*          MAXCHL for unix changed to use shrmem for performance reasons.
*
*          Jrgen H. Pedersen								  - 25Oct05 DK	v2.35
*          OS/400 support added (bypass strtok_r) and handling a new print model.
*          New OS/400 entrypoint (int main() added with conditional compile.
*
*          Jrgen H. Pedersen								  - 14Nov05 DK	v2.36
*          Handling of semaphores in LockSpecSem() fixed.
*
*          Jrgen H. Pedersen								  - 14Nov05 DK	v2.37
*          Connection refused message enhanced to conname and channelname
*
*          Jrgen H. Pedersen								  - 12Dec05 DK	v2.38
*          Connection refused, Pattern string is too long, max. This message wasn't
*          showing the resulting message length, just the current one. So when concat
*          patterns we had a issue.  This was also fixed for Userids
*
*          Jrgen H. Pedersen								  - 25Jan06 DK	v2.39
*          Porting for Linux AMD64. various definitions changed.
*
*          Jrgen H. Pedersen								  - 5Feb06 DK	v2.40
*          Problem with semaphore addressing fixed. And a bit about logging on UNIX.
*          shmdt(shm); chaned to shmdt((char*)shm); due to changes on Solaris x86
*
*          Jrgen H. Pedersen								  - 13Feb06 DK	v2.41
*          Problem when a config file contains garbage and MQXR_TERM is
*          invoked and it returns a failure .... This was seen WIN2K3.
*
*          Jrgen H. Pedersen								  - 14Feb06 DK	v2.42
*          Logging to EventLog added on windows. Currently all events are logged.
*          Switch will be added to SCYDATA to control this feature.
*
*          Jrgen H. Pedersen								  - 18Feb06 DK	v2.43
*          Needed cleanup procedure for getActualCurrentNumberOfRunningChannels to cleanup
*          allocated "databags" mqDeleteBag must be included in the windows edition to prevent
*          storage leak.
*
*          Jrgen H. Pedersen								  - 1Mar06 DK	v2.44
*          Added support for shared memory on windows. This is implemented using a seperate
*          program BlockIP2S, that initializes the Shared Memory Segment(SMS). There are a
*          detatched BlockIP2S per queue manager running with BlockIP2. (when the channel limitter
*          is activated).
*          Storage Leak from pre. version 2.44 is also solved. This appled only to the delivered
*          windows DLL, due to a compilation problem. Serilasation was added in windows to handle
*          logfile contention. This is also included in the control of SMS and BlockIP2S.
*
*          Jrgen H. Pedersen								  - 19Mar06 DK	v2.45
*          Shared memory naming changed to support HACMP and other MA special settings.
*          The shm name under *NIX is now based on /var/mqm/mqs.init
*
*          Jrgen H. Pedersen								  - 4Apr06 DK	v2.46
*          Added logic to handle cycling of logfiles, to prevent fill up of filesystems.
*          Controlled by:
*          LogCount=nn; # of versions (between 3 and 99).
*          LogSize=nnnnnnn; Size og the logfile before switching. min 100KB.
*          Default LogDrive og LogPath log path is changed on windows to conformw with the
*          UNIX implementation, so we use the windows settings. And FileName is extented with "001"
*          to allow circular logging.
*
*          Jrgen H. Pedersen								  - 5Apr06 DK	v2.48
*          Problem with Log file rotation solved. Error messages was enhanced.
*
*          Jrgen H. Pedersen								  - 20Apr06 DK	v2.50
*          Pattern matching extented to allow imbedded ** this means that the generic specs may
*          look like this:  Patterns=123.*.123; SSL=CN=ibmwebspheremq*T01,MCA=user;
*          ASC=Y/AllowSelfSignedCertificates=Y added.
*          SSL=[C=,][L=,][O=,][OU=,]CN=;[[MCA={userid|role|*};]|[BLOCK;]] some examples:
*          TERM=N/Y for controlling print of termination message.
*          MAX_SSL raised to 256 and MAX_PL to 1024*4.
*
*          Jrgen H. Pedersen								  - 20Apr06 DK	v2.51
*          wildcmplist() problem fixed.
*
*          Jrgen H. Pedersen								  - 20Apr06 DK	v2.52
*          messlen and buflen changed from long to MQLONG in deductStatusQ, getActualCurrentNumberOfRunningChannelszOS
*
*          Jrgen H. Pedersen								  - 10May06 DK	v2.53
*          Support for BLANK_USERID added, changes was made in CheckUserId() and CheckCONList()
*          to obtain the wanted functionality. Testcase: test25a-b, test39a-b.
*          Implemetation of hostname support done in CheckConnectionPattern()
*          This implementation does not support IPv6.
*          Deleted many compile warnings to get a clean compilation list without warnings.
*
*          Jrgen H. Pedersen								  - 28May06 DK	v2.55
*          Storage leak on z/os solved together with DNS lookup on z/os.
*
*          Jrgen H. Pedersen								  - 10Aug06 DK	v2.56
*          Problem with errors in spec. files solved. AMQ9190 The user exit ... invoked for ..
*          with id '11' and reason '12', returned values that are not valid, as reported in the preceding messages.
*          The channel stops.
*          Detection for IY86343. added.
*
*          Jrgen H. Pedersen								  - 10Aug06 DK	v2.57
*          Problem with accepted CON= where there are no CON= that should give the auth.
*
*          Hubert Kleinmanns								  - 10Oct06 D	v2.56a
*          Fixed a problem in function 'CheckSSLList'. BlockIP2 exited in this function with a
*          zero length 'SSLRemCertIssNamePtr' in structure 'pChannelExitParams'. This problem
*          occured on Solaris Sparc systems with MQv6.
*
*          Jrgen H. Pedersen								  - 10Dec06 DK	v2.60
*          Impl. of additional specification file for configuration.
*          BlockIP2 look default for /var/mqm/exits/BlockIP2.ini or ExitPath\BlockIP2.ini on the
*          distributed platforms if FN= is not specified. This is implemented to ease distribution
*          in complex installations so a generix specification can be used. The NEW syntax on the
*          BlockIP2.ini is documented in the manual.
*          If the file is not present or cannot be read due to access violation it's trated as not
*          found and BlockiP2 will continues as before.
*
*          Jrgen H. Pedersen								  - 12Jan07 DK v2.61
*          Differentiated debug modes added (-d0, -d1 and -d8) and -i for ignore looking for BlockIP2.ini
*          -d0 is typicly used for debug of the spec file.
*          -d  old mode, lists all exitpoint invocations.
*          -d1 shows functions called (technical debug).
*          -d8 Code devlopment debugging mode)
*          LOGCYCLE={N|_Y}; added to get rid of 001 logfiles to back to behaviour pre 2.46
*          -l option added to help LOGCYCLE, from SCYDATA.
*
*          Jrgen H. Pedersen								  - 16mar07 DK v2.62
*          *NIX problem in ProcessMQSiniFile_qm_ini solved when looking for BlockIP2.ini
*          Problem with file cycle solved for dist. platforms.
*
*          Jrgen H. Pedersen								  - 05Apr07 DK v2.64
*          Complex SSL filtering problems solved. Added ST= and PC= in the SSL filtering
*          to be complient with gsk7cmd and runmqckm and their capabilities.
*          Problem with FNx= fixed. Reason was premature release or storage.
*
*          Jrgen H. Pedersen								  - 06May07 DK v2.65
*          Added MCA=xx(nn,nn) capabilities on SSL= to enhance the extraction of an userid
*          that can be used for authentication. Any DN keyword can be used as key....
*          The function works "substring like" starting with index 1 as first position and
*          with "end of string" when next seperator "," are reached.
*          SCYDATA() written to log when using logmode -d0 and up.
*          *NIX storage leak fixed.  localtime() changed to localtime_r() to prevent highly
*          used system to breakdown. The use of gethostbyname() changed to gethostbyname_r()
*          to prevent SIGSEGV.
*
*          Jrgen H. Pedersen								  - 17May07 DK v2.66
*          Problem with channellimitter solved when using BlockIP2.ini
*
*          Jrgen H. Pedersen								  - 06Jun07 DK v2.67
*          Problem with FN= and file not found abend solved.
*          Problem with QMGR= and CHANNEL= logic solved. CHANNEL= stmt didn't lock for previous
*          accepted channels.
*          ENV changed to show platform information. Like MVS, AIX, Linux, WIN etc.
*          Added #pragma to disable _POSIX warning 4996 for fileno.
*
*          Jrgen H. Pedersen								  - 24Jun07 DK	v2.68
*          Support for AS/400 implemented... And it's working....
*          Extented to use shared storage and DNS Support.
*
*          Jrgen H. Pedersen								  - 11Jul07 DK	v2.69
*          Sporadic errors solved for Solaris
*          Limit on CON_MAX enforced BLOCKIP-68E added.
*
*          Jrgen H. Pedersen								  - 12Mar08 DK	v2.70
*          Security problem with windows 2003 solved, changed in in BlockIP2 and BlockIP2S.
*          Extented error reporting in starting windows SHM and registry load.
*          SYSLOGFCLTY= and SYSLOGPRTY= added for UNIX to control syslog().
*          syslog_r() added for AIX for better thread safe.
*
*          Consider changing logpath for *NIX-64 to exits64, by adding cond. compile stmt.
*          Show the name of the BlockIP2.ini in -d option, to ease configuration trouble.
*
*          Report open reason for failed file operations..... (fopen()  errno)
*          BlockIP2 failed to open the specified logfile [] and other.
*
*          Jrgen H. Pedersen								  - 20Oct08 DK	v2.71
*          Changed sprintf() to snprintf() to avoid stack crashes.
*          Added thread id to identify the correspondent requests and answers.
*
*          Jrgen H. Pedersen								  - 21Oct08 DK	v2.72
*          Added CloseHandle(mutex) to prevent loss of handles, and to keep the stuff running
*          for long periods of time.
*
*          Jrgen H. Pedersen								  - 25Oct08 DK	v2.73
*          Changed size of connection table size to 64KB from 2, giving room for 1500 channels.
*
*          Jrgen H. Pedersen								  - 30Jan09 DK	v2.74
*          Changed LOGDEBUG1 to LOGDATA for nospace in chltable for UNIX impl.
*          Channel struct changed to BlockIP2.2a to assure complience with old version, and
*          contains now SHMMAX_CHL.
*
*          Jrgen H. Pedersen								  - 12Mar09 DK	v2.75
*          Suppress TERM note when quiet mode selected.
*          AllowBlankUserID=N added to comply with the book....
*
*          Jrgen H. Pedersen								  - 20Mar09 DK	v2.76
*          Added support for longer WTO messages on z/os.  Problem with BLOCKIP2-I50 and
*          CONNAME print solved.
*          Added support for MSGPFX under z/OS for WTO's to comply with CA tools...
*
*          Jrgen H. Pedersen								  - 22Jun09 DK	v2.77
*          Support for allowing MCAUSER specified on the channel to be overridden by the incomming
*          userid, this is done with the MCA=* on the CON= statement.
*
*          Jrgen H. Pedersen								  - 07Jul09 DK	v2.78
*          BlockMqmUsers=N added.
*          Problem with CON= and hostnames solved. It username and MCA was removed due to
*          logic error.
*          Need for mqm linkage removed for most platforms.
*
*          Matt Batterham   						                  - 06Oct09 AUS	v2.79
*          Leading to fix dealing with NONE; in SCYDATA(), to help problem solving when Patterns= is missing in BlockIP2,ini
*
*          Jrgen H. Pedersen								  - 07DEC09 DK	v2.80
*          Problem with MQ 7.0.1.0 where conname can contain the source port: 10.1.1.1(1490)
*
*          LogFormat=M (QueueManager) added to include the queue manager in the logfile name.
*          QMGR_SHORT_NAME=<shortname>; up to 8 characters shortname of qmgr to include in the logfile,
*          usefull when long queue manager are used...  By default it takes the first 8 characters
*          from queue manager name.
*          INCQMGR=Y; added to include a short qmgr id in the logfile. Highly usefull for large systems.
*          Not so usefull when used with LogFormat=M.
*
*          Introduced USESECPARMS=Y; that allows the user to set the MCAUSER based on the content in
*          SecurityParms. This allows users to use BlockIP2 with MQExplorer to do authentication using an
*          OAM exit. **NOTE** If not used together with an Object Authority Manager (OAM) this leads
*          to security exposure because t.x. MQExplorer (and other applications) allow the user to
*          type in an userid aka mqm.
*          Default is USESECPARMS=N; so it's your own choise to enable the feature.
*          However it requires the invoking client application to utilize MQCSP_AUTH_USER_ID_AND_PWD.
*          From the manual: This value indicates that MQCSP user ID and password fields will be used
*          by the Object Authority Manager (OAM) to perform authentication on a MQCONNX call.
*          When this is specified, the MQCSP structure is passed to the OAM Authenticate User function,
*          which can set appropriate identity context fields.
*
*          Jrgen H. Pedersen								  - 14DEC09 DK	v2.81
*          Problem with MQ 7.0.1.0 where conname can contain the source port: 10.1.1.1(1490)
*          Added compile flag TEST281 to test it.
*
*          Jrgen H. Pedersen								  - 03FEB10 DK	v2.82
*          Problem with channel limitter on heavely used multithreaded UNIX where shared memory
*          was exhaused. Changes made to wildcmplist() to solve problem with "BLANK_USERID" in a
*          multiline specification: "CON=*;BLANK_USERID,QMQM,QMQMADM,QSECOFR;MCA=NoBody;"
*          A dedicated test is included to deal with this test.
*          LogCycle= added to ease configuration.
*
*          Added debug level on ver=x.xx line for ease support.
*
*          Jrgen H. Pedersen								  - 04MAR10 DK	v2.83
*          Added support for multi reply from DNSservers
*          Added support for standby qmgrs. (DataPath on UNIX)
*
*          Paul Giordano                                                                  - 06May10 PG  V2.84
*          Add logic to z/OS to load process parameters into memory for efficiency. Keyed with
*          nZOSLoadMem and parm -m. Added Silent flag to 50I and 99I WTOs. Removed unused __4kmalc()
*          and #if BLOCKIM logic.  RetroFitted by MrMQ.
*
*          Jrgen H. Pedersen								  - 16AUG10 DK	v2.85
*          Problem with SIGSEGV when processing mqs.ini solved. Was caused by storage overlay.
*          Problem with missing logfiles solved (was also causing SIGSEGV) thanks to Anshul Rastogi
*          and others.
*
*          Jrgen H. Pedersen								  - 18AUG10 DK	v2.86
*          More changes regarding SIGSEGV done.
*
*          Jrgen H. Pedersen								  - 22JAN11 DK	v2.90
*          More More changes to support standby queue managers. Monitoring failover.
*          UNIX Shared memory anchor point moved from mq path to /tmp/.<qmgr>_BlockIP2_<fid> to prevent
*          further problems with MQ product changes.  This can be disabled with option -t, or
*          TMPPATH=No; in configfile. TMPPATH=Yes; Is default.
*
*          Falk Dressler                                                                  - 14MAR11 FD  v2.91
*          Avoid to refer to invalid memory when checking for '\n' in config file lines
*
*          Jrgen H. Pedersen								  - 12APR11 DK	v2.92
*          Problem with CLUSSDR channels and "empty" patterns solved.
*
*          Jrgen H. Pedersen								  - 13SEP11 DK	v2.93
*          Enhancement of debugging for option USESECPARMS=Y
*
*          Jrgen H. Pedersen								  - 20SEP11 DK	v2.94
*          Enhancement of debugging for blank userids and exit response.
*          Additional logic added to control the behaviour of SEC_PARMS, to allow DotNet applications
*          to connect to MQ without formal authentication and to respect the MCAUSER settings.
*
*          Jrgen H. Pedersen								  - 26MAY12 DK	v2.95
*          Changed determination of ENV to have right ident. of env. Added ZLINUX conditional
*          definition.
*
*          Jrgen H. Pedersen								  - 15JUN12 DK	v2.96
*          Removed need for including CSQXSTUB in linkage steb to prevent IEW2456E complaining
*          about missing MQOPEN/MQCLOSE/MQGET and MQPUT1 modules to remove the need for using a
*          PDSE as CSQXLOAD. Perhaps all this old stuff should be removed to clean up the source.
*
*          STB								  		  - xxXXX12 xx	v2.97
*          Replaced LoadRegistry() to automatically support multi install under windows.
*          Retrofitted by MrMQ.
*
*          Jrgen H. Pedersen								  - 03JAN13 DK	v2.98
*          Added FAP= check parameter to block connections from older MQ devices, like MQ 5.x clients.
*
*          Jrgen H. Pedersen								  - 01APR13 DK	v2.99
*          Forced initiation of nZOSLoadMem assure configuration is read on z/OS V2R1
*
*          Jrgen H. Pedersen								  - 09MAR15 DK	v3.00
*          Windoes x64 migration.
*          Adding filtering on MQCXP for PartnerName to allow additional queue manager filtering.
*/

/*
* Known FAP levels:
*
*    HP NSS V5.x - FAP 7
*    MQ V6       - FAP 8
*    MQ V7.0.0   - FAP 9
*    MQ V7.0.1   - FAP 10
*    MQ V7.1     - FAP 11
*    MQ V7.5     - FAP 12
*    MQ V8.0     - FAP 13
*/


/*
* Needed changes:
*
*           Enhance SSL filtering: So L= is different from MAIL=...
*
*          Timer feature to discover "failback" so shared memory is re-initialized if it's reused.
*          This can cause lost connection in case where the queue manager crashes.
*/


/*
* Planned changes
*          Possibility of specifying MCAUSER, to help on JMS connection difficulties,
*          done without SSL or other encrytion. This is just meant to be a small
*          security imoprovement, not a 100% solution. Could be spec. like:
*          JMS=<channel_name>;<conname>;<MCAUSER>;[DEF|ONLYMCA;]
*          Implemented 80% in version 1.22
*
*          Filtering on Certificate issuer ? Can it be used for anything ?
*          Maybe just for blocking, but that's up to crl.
*
*          Improve performance, so we only read configuration file once, and use this information without
*          having to reading, parsing, tokenizing it. 040522.
*
*          Adding error message when trying to read a line that is longer than MAX_LEN....
*
*/

/*
* Version no
*/

#define VERSION "3.00"


/* --------------------------------------------------------------------------
* Build hints
*
*       For Solaris:
*         cc -c -KPIC BlockIP2.c
*         ld -G BlockIP2.o -o Block
*         cp BlockIP2 /var/mqm/exits
*
*
* Compiled on windows using Microsoft (R) 32-bit C/C++ Optimizing Compiler Version 14.00.50727.42 for 80x86
*          cl BlockIP2.c /MT /DWIN32 /DWIN32_SHM /DHNLUP /D_CRT_SECURE_NO_DEPRECATE /D__BI2_78__ /LD advapi32.lib netapi32.lib ws2_32.lib
*
* Compiled on Linux using GNU C:
*          gcc -fPIC -shared  -o BlockIP2 BlockIP2.c -lmqm_r -DUNIX -D_RENTRANT -DHNLUP -DLINUX
*
* Compiled on Linux AMD_64 using GNU C:
*          gcc -m64 -o BlockIP2 BlockIP2.c -fPIC -ansi -shared -D_REENTRANT -DUNIX -DHNLUP -DLINUX -D_XOPEN_SOURCE -L/opt/mqm/lib64
*
* Compiled on RHEL6 (64) Using gcc:
*          gcc -m64 -o BlockIP2 BlockIP2.c -fPIC -ansi -shared -D_REENTRANT -DUNIX -DHNLUP -DLINUX -D_XOPEN_SOURCE -L/opt/mqm/lib64 -I /opt/mqm/inc
*
* Compiled on Linux s390x using GNU C:
*          gcc -m64 -o BlockIP2 BlockIP2.c -fPIC -ansi -shared -D_REENTRANT -DZLINUX -DUNIX -D_XOPEN_SOURCE -DHNLUP -Wall -I /opt/mqm/inc
*
* Compilation parameters supplied by Michael Dag:
* Compiled on Solaris using GNU C    : gcc -G -I/opt/include -I/opt/mqm/inc -o BlockIP2 BlockIP2.c -D_REENTRANT -DUNIX -DHNLUP -DSOLARIS
* Compiled on AIX 5.3 using C for AIX: cc  -G -I/usr/include -I/usr/mqm/inc -o BlockIP2 BlockIP2.c  -D_REENTRANT -DUNIX -DHNLUP
*
* Compilation parameters supplied by techno:
* Compiled on HP-UX: cc -c +z BlockIP2.c -DUNIX -DHNLUP
*       Linked on HP-UX: ld -b BlockIP2.o BlockIP2 -lmrm_r
*
* Compilation on OS400: CRTCMOD MODULE(MYLIB/BLOCKIP2) SRCSTMF('/home/mrmq/BlockIP2.c') OUTPUT(*PRINT) OPTION(*GEN)  TERASPACE(*YES *TSIFC) INCDIR('/QIBM/ProdData/mqm/inc') SYSIFCOPT(*IFSIO) DEFINE(OS400 HNLUP)
*                       CRTPGM PGM(QGPL/BLOCKIP2) MODULE(MYLIB/BLOCKIP2) BNDSRVPGM(QMQM/LIBMQM_R) ACTGRP(*CALLER) OPTION(*GEN) DETAIL(*FULL)
*
* Compiled on Solaris X86_64 using GNU C:
*          gcc -c -m64 BlockIP2.c -ansi -fPIC -DUNIX -D_REENTRANT -DSOLARIS -DHNLUP -I/opt/mqm/inc
*          /usr/ccs/bin/amd64/ld -64 -mt -G -o BlockIP2 BlockIP2.o -R/usr/lib/64 -lsocket -lnsl -ldl
*
*
*
* --------------------------------------------------------------------------
*/

/* 2.78 change */
#if defined(UNIX)
#define __BI2_78__
#endif

#if defined WIN32
#define _WIN32_WINNT 0x0500
#endif

#define SEM_SHM    0
#define SEM_CHL    1
#define SEM_LOG    2
#define SEM_RENAME 3


#if defined(__MVS__)
/* NWC Cater for MVS operation */
#define ENV "MVS"
#define HNLUP     /* currently don't enable the DNS code for z/os until the stubs are finished. */
#if defined HNLUP
#define _XOPEN_SOURCE_EXTENDED 1
#define _OPEN_SYS_SOCK_IPV6
#include <netdb.h>
#include <errno.h>
#endif

#else
#if defined(WIN32)
#pragma warning(disable: 4996)
#endif

#if !defined(SOLARIS) && !defined(WIN32) && !defined(AIX) && !defined(LINUX)
#if MQAT_DEFAULT==MQAT_OS400
#define OS400      1
#endif
#endif

#endif   /* ! __MVS__ */

/*
* standard headers
*/
#if !defined(WIN32)
#include <pthread.h>
#endif

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>

#if defined(WIN32)
#ifdef HNLUP
#include <Winsock2.h>               /* 2.53 */
#endif
#include <windows.h>
/* Added by STB */
#include <psapi.h>
/* End of added by STB */
#ifndef snprintf
#define snprintf _snprintf
#endif
#include <sddl.h>                   /* 2.70 */
#include <sys\stat.h>
#define __EXT_LOG_FILENAME__ 1
#endif
/*
* MQ headers
*/
#include <cmqc.h>
#include <cmqxc.h>

/* start chl. inq */
#if !defined(__MVS__)
#include <cmqbc.h>
#include <cmqcfc.h>
#endif

#if defined(__MVS__)
#define HNLUP
#include <spc.h>
#endif   /* __MVS__ */

#if MQAT_DEFAULT==MQAT_OS400
#define ENV "OS400"
#endif

#if MQAT_DEFAULT==MQAT_OS390
#define ENV "z/OS"
#endif

#if MQAT_DEFAULT==MQAT_WINDOWS
#define ENV "WIN"
#endif

#if MQAT_DEFAULT==MQAT_WINDOWS_NT
#define ENV "WIN"
#endif

#if MQAT_DEFAULT==MQAT_AIX
#if defined(AIX)
#define ENV "AIX"
#endif

#if defined(SOLARIS)
#define ENV "Solaris"
#endif

#if defined(UNIX)
#define ENV "Linux"
#endif

#if defined(ZLINUX)
#undef ENV
#define ENV "s390x"
#endif
#endif            /* MQAT_AIX (or UNIX) */


#ifndef MQXR_SEC_PARMS                /* Help version 5.x compilation to run on 6.0 qmgrs */
#define MQXR_SEC_PARMS 29
#endif

#if defined(UNIX) || defined (OS400)
#if MQAT_DEFAULT!=MQAT_OS400          /* Not for OS400 */
#define __UNIX__ 1
#define __EXT_LOG_FILENAME__ 1
#define __USE_MISC 1
#endif   /* MQAT_DEFAULT!=MQAT_OS400 */

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <signal.h>
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>        /* testing solaris */

#if MQAT_DEFAULT!=MQAT_OS400          /* Not for OS400 */
#include <syslog.h>         /* using UNIX syslogs too */
#endif

#ifdef HNLUP
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#endif    /* HNLUP */

#if !defined(_THREAD_SAFE_JHP)
#define _THREAD_SAFE_JHP        /* 2.66 */
#endif    /* ! _THREAD_SAFE_JHP */
#endif    /* end if unix like defines */


#if MQAT_DEFAULT==MQAT_OS400
#define _REENTRANT 1
#endif

/*
* Common Defines
*/

#define SEC_FAILURE 1
#define SEC_OK      2

#define MAX_SSL   256
#define SSL_LEN    80
#define SSL_LEN_PEER 512 /* 80 */
#define MAX_CON    64
#define CON_LEN    80
#define MAX_PL 1024*4
/*#define MAX_PL  256*/
/*#define MAX_PL  128 */
#define MAX_LEN   400
#define MAX_MCAUSER_LEN  64                              /* max len of MCAUSERids 2.94 */

#if defined (__MVS__)
#define LOGMODE "a"
#else
#define LOGMODE "a+"
#endif


/* character coding indicator */
#define CC_ASCII     1
#define CC_EBCDIC    2

/* Size of AgentBuffer */
#define WSIZE 80

/*
* Bit mask of Logging options
*/
#define LOGNAME    1                                       /* always used but defined for consistency */
#define LOGDATE    2
#define LOGCHANNEL 4
#define LOGUSER    8
#define LOGQMANAGER 16                                     /* insert queue manager into filename 2.80 */


#define LOGDATA   0                                        /* normal data */
#define LOGDEBUG0 1                                        /* Debug information minimized mostly for spec file */
#define LOGDEBUG  2                                        /* Debug information */
#define LOGDEBUG1 3                                        /* Debug information tech */
#define LOGDEBUG8 8                                        /* Debug information full tech */

#define MXC_RC_NONE  0
#define MXC_RC_ERROR 1

#if defined (WIN32)
#define LOG_FILEONLY    1
#define LOG_EVTLOGONLY  2
#define LOG_BOTH        3
#endif

#if defined (UNIX) || defined (OS400)
#define LOG_FILEONLY    1
#define LOG_EVTLOGONLY  2
#define LOG_BOTH        3
#endif

#define INI_IGNORE      0
#define INI_PROCESS     1

#if defined (__MVS__)                                      /* define communication to BLOCKIPW */

/* Start of inet_ntop fixup... */
#define SPRINTF(x) ((size_t)sprintf x)
#if !defined u_int32_t
typedef unsigned int u_int32_t ;
#endif
#ifndef NI_MAXHOST
# define NI_MAXHOST      1025
#endif
#ifndef NI_MAXSERV
# define NI_MAXSERV        32
#endif
#ifndef NI_NUMERICHOST
# define NI_NUMERICHOST     1                              /* Don't try to look up hostname.  */
#endif
#ifndef NI_NUMERICSERV
# define NI_NUMERICSERV     2                              /* Don't convert port number to name.  */
#endif
#ifndef NI_NOFQDN
# define NI_NOFQDN          4                              /* Only return nodename portion.  */
#endif
#ifndef NI_NAMEREQD
# define NI_NAMEREQD        8                              /* Don't return numeric addresses.  */
#endif
#ifndef NI_DGRAM
# define NI_DGRAM          16                              /* Look up UDP service rather than TCP.  */
#endif
#ifndef NS_IN6ADDRSZ
#define NS_IN6ADDRSZ       16
#endif
#ifndef NS_INT16SZ
#define NS_INT16SZ          2
#endif
/* end og inet_ntop fixup */

#define WTOLENGTH 160

typedef struct _bi2wto
{                                                          /* communication structure */
    char msg[WTOLENGTH];
} bi2wto_type;

typedef struct _bi2dataspace
{                                                          /* communication structure */
    long version;
    char function[1];
    char channel[MQ_CHANNEL_NAME_LENGTH];
    long limit;
    long current;
    long maxrunning;
    char status[7];
} bi2ds_type;

#pragma linkage (BLOCKIPS,OS)
#pragma linkage (BLOCKIPW,OS)
#pragma linkage(IEANTCR,OS)
#pragma linkage(IEANTRT,OS)

long BLOCKIPS(bi2ds_type *);
long BLOCKIPW(bi2wto_type *);
void IEANTCR(int *, char *, long *, int *, int *);
void IEANTRT(int *, char *, long *, int *);

#pragma linkage(BLOCKIPN,OS)                               /* gethostbyname()  */
void BLOCKIPNblockipn(char *,long,char *,long*);
#pragma linkage(BLOCKIPM,OS)                               /* GETMAIN/FREEMAIN */
void BLOCKIPM(long *,long *,char *,long*);

/*
 * Name/Token Level Constants
 */
#define IEANT_HOME_LEVEL             2

/*
 * Name/Token Persistence Constants
 */
#define IEANT_NOPERSIST              0

/*
 * Name/Token Return Code Constants
 */
#define IEANT_OK                     0
#define IEANT_DUP_NAME               4
#define IEANT_NOT_FOUND              4

#endif

#if defined UNIX || defined (OS400)
typedef struct _shmem_channels
{
    int  iNextChl;
    char szChannelName[MQ_CHANNEL_NAME_LENGTH+1];
    int  iActiveChannelLimit;
    int  iCurrentActiveConnections;
    int  iMaxActiveConnections;
} chlstruc;

/*#define SHMSZ 2048*/
#define SHMSZ 2048*32                                      /* to give more room for 1500 channels           2.73 */
#define SHMMAX_CHL ((SHMSZ-63) / sizeof(chlstruc)) -1
typedef struct _shmem_dataspace
{                                                          /* communication structure                            */
    int  iPid;                                             /* who created/owns the shmem                         */
    int  iNumOfChannels;
    int  iMaxChlInTable;                                   /* SHMMAX_CHL when allocated                     2.74 */
    char master_time[15];                                  /* Last time master file has changed             2.90 */
    char Reserved[30];
    MQLONG lTmpTime;                                       /* Timer so we know when to check for moved qmgr 2.90 */
    chlstruc chnl[SHMMAX_CHL];
} shmem_type;

#endif   /* defined UNIX || defined (OS400) */

#if defined WIN32
typedef struct _shmem_channels
{
    int  iNextChl;
    char szChannelName[MQ_CHANNEL_NAME_LENGTH+1];
    int  iActiveChannelLimit;
    int  iCurrentActiveConnections;
    int  iMaxActiveConnections;
} chlstruc;
/*#define SHMSZ 2048*/
#define SHMSZ 2048*32                                      /* to give more room for 1500 channels, 2.73 */
#define SHMMAX_CHL ((SHMSZ-51) / sizeof(chlstruc)) -1

typedef struct _shmem_dataspace
{                                                          /* communication structure */
    int iPid;                                              /* who created/owns the shmem */
    int iNumOfChannels;
    char Reserved[43];

    chlstruc chnl[SHMMAX_CHL];
} shmem_type;

#endif


typedef struct CURRCHLDATA
{
    long limCurrChl;
    long CurrChannels;
    long MaxChannels;
} currChl_type;

/* Storage that needs to be allocated to pass information to MQXR_TERM */
typedef struct WS
{
    int      bFirstTime;
    long     lMaxChannelActive;
#if defined (UNIX) || defined (OS400)
    char     szQmgrDirectory[255];
    char     szQmgrPrefix[255];
    char     szDataPath[255];                              /* NFS mounted qmgr (standby) 2.83  */
#endif
    char     ConName[MQ_CONN_NAME_LENGTH+1];               /* Local copy of the connectionname */
    int      bLogTERM;
    char     LongMCAUser[MAX_MCAUSER_LEN+1];               /* Extented MCAUSER field     2.94  */
} ws_type;                                                 /* WS Buffer                        */

typedef struct WORKDATA
{
    /* start inq chl */
    MQHCONN  hConn;
    MQLONG   HobjStat;
    int      Version;                                      /* MQCXP version data for later compare */
    char     szStatusQ[MQ_OBJECT_NAME_LENGTH+1];
    char     szQMgr[MQ_Q_MGR_NAME_LENGTH+1];
    char     szChannelName[MQ_CHANNEL_NAME_LENGTH+1];
    long     lMaxChannelActive;
    char     szLogPath[128];
    char     szLogDrive[3];                                /* only applicable to Windows systems */
    int      nLogFormat;                                   /* Bit mask 0-255 values */
    char     szLogChannel[64];
    char     szLogFileName[64];
    int      NoLogFileName;                                /* Omit BlockIP2 in logfile      2.95 */
    char     szLogFileExtension[32];
    char     szTmpBufff[256];
    char     szWTOBuffErr[256];
    char     szWTOBuffOk[256];
    char     szSCYDATA[MQ_EXIT_DATA_LENGTH+1];
    char     szSCYNAME[MQ_EXIT_NAME_LENGTH+1];
    /*
    * Flags
    */
    int      nDebugFlag;
    int      nZOSconv;                                     /* convert brackets on z/OS to handle national chars */
    int      nZOSLoadMem;                                  /* Load zOS parms into memory for greater efficiency 2.84 */
    int      bRespectMCA;                                  /* we don't respect MCAsettings */
    int      bAllowBlankUserID;                            /* We will not accept blank uids */
    int      bBlockMqmUsers;
    int      bDebugMode;
    int      SilentMode;
    int      bLogByDate;
    int      bUseridUpperLowerCase;                        /* be case sensitive */
    int      ReportLogFileError;                           /* only report logfileerror once */
    int      lExpireValue;                                 /* let channel record live for 10 minutes (in secs) */
    int      lRefuseReason;                                /* send refuse reason back to client exit.. */
    int      bInteractive;                                 /* is the client requesting feedback ? */

    char     CONList[MAX_CON] [CON_LEN];
    char     SSLPeer[SSL_LEN_PEER+1];
    char     SSLList [MAX_SSL] [SSL_LEN];
    int      SSLCount;
    int      CONCount;

    char     RemUID[MQ_USER_ID_LENGTH+1];                  /* remote user identifier */
    char     ConName[MQ_CONN_NAME_LENGTH+1];               /* Local copy of the connectionname */
    char     Flags[64];
    char     U2[MAX_PL];

    char     Users[MAX_PL];                                /* list of userids */
    char     BlockUsers[MAX_PL];                           /* list of Blocked userids */
    char     Pattern[MAX_PL];                              /* The pattern to match against */
    char     szClientUser[WSIZE];                          /* user in the other end */
    int      bFileSizeCheck;                               /* default is 0 switched on */
    int      bFileSizeDone;                                /* default is not done */
    long     lLogFileCount;                                /* Number of logfiles to maintain default =3 */
    long     lMaxlogFileSize;                              /* max size of logfile before cycling defailt 200kb */

#if defined(__MVS__)
    MQLONG Hobj;
    char     szReplyQ[MQ_OBJECT_NAME_LENGTH+1];
    char     szWTOPfx[21];
#endif
#if defined (UNIX) || defined (OS400)
    int      iCurrentActiveConnections;
    key_t    KeySemaphores;
    key_t    KeyShmem;
    char     szQmgrDirectory[255];
    char     szQmgrPrefix[255];
    char     szDataPath[255];                              /* NFS mounted qmgr (standby) 2.83 */
    int      lWinLog;
    int      lTmpPath;                                     /* use new shared anchor path                    2.90 */
    char     szMasterFile[255];                            /* Masterfile v. 7.0.1                           2.90 */
    int      iStbyMode;                                    /* Standby checks supported                      2.90 */
#endif
#if defined(UNIX)
    int      iSyslogPrty;                                  /* priority to use when writing to syslog */
    int      iSyslogFclty;                                 /* facility to log as */
#endif

#if defined (WIN32)
    int       iCurrentActiveConnections;
    int      lWinLog;
    HANDLE   hMutex;
    HANDLE   hMutexShm;
    char     szMutexKey[256];
#endif

#if defined(MQCXP_VERSION_6)
    int      AllowASC;
#endif
    int      bLogTERM;
    int      bIgnoreBI2ini;

#if defined (UNIX) || defined (WIN32) || defined (OS400)
    char     szConfigFile[255];
    char     ExitPath[128];                                /* where exit DLL are loaded from */
    char     szShortQMName[9];                             /* sort Queue Manager name to add to log 2.80 */
    int      bIncludeQmgrNameInLog;
#endif
    int      bThisQmgr;                                    /* Used in 2.60 */
    int      bChannelPatterns;
    int      bThisChannel;
    int      bThisChannelOK;
    int      bThisChannelDone;
    int      ExitReason;
    int      bUseSecurityParms;                            /* 2.80 take MCA user from SecurityParms */
    int      bForceSecurityParmsOff;                       /* 2.94 bypass usage of SecurityParms    */
    ws_type *ppWS1;                                        /* 2.94 pointer to pWS1                  */
    MQLONG   lFAPLevel;                                    /* 2.98 FAP level check                  */
	char     PartnerName[MAX_PL];                          /* 3.00 Partnername check                */
} workingStorage_type;

#ifdef HNLUP2
typedef struct DNSReply
{
    int      replies;
    char     szIpAddr[16][64];
} DNSReply;
#endif

#define DEBUGMSG 0
#if DEBUGMSG == 1
  /* debug messages to stdout */
  #define OST stdout
  #define DBGPRINTF(x) fprintf x
#elif DEBUGMSG == 2
  /* debug messages to file */
  FILE * OST = 0;
#if MQAT_DEFAULT==MQAT_WINDOWS_NT
  #define MSGFILE   "c:\\msgexitEFPI.log" /* <-- tailor for environment */
#endif
#if MQAT_DEFAULT==MQAT_ZOS
  #define MSGFILE   "//DD:SYSPRINT"   /* <-- tailor for environment */
#endif
#if !defined MSGFILE
  #define MSGFILE   "/var/mqm/exits/plog.log"   /* <-- tailor for environment */
#endif
  #define DBGPRINTF(x) { \
                       OST=fopen(MSGFILE,"a+"); \
                       if(OST) fprintf x; \
                       fclose(OST); }
#else
  /* no debug messages */
  #define DBGPRINTF(x)
#endif /* DEBUGMSG */


/*
* 22-03-2004 Sid Young Additional Defs for Enhanced Logging
*
* Add to Spec File specified in FN= parameter

 LogFormat=
 N	 - For name only or
 ND  - Name and Date or
 NDC - Name Date Channel format
 NCD - Name Channel Date format or any variation
 #
 # File Name to use (Default BlockIP2)
 #
 LogFileName=BlockLog
 LogExt=txt
 LogFormat=ND
 # example of use on Windows path
 #
 LogPath=\qml\log
 LogDrive=c:
 #
 # on Linux systems (LogDrive ignored)
 #
 LogPath=/var/spool/blockip2/logs
*/


/*
* Function Definitions
*/

int  wildcmp(const char *wild, const char *string, workingStorage_type *pWS);
int  wildcmplist(char *wild, char *string, int flg, workingStorage_type *pWS);  /* 2.82 */
/*int  wildcmplist(char *wild, char *string, workingStorage_type *pWS);*/
void LogData(int iMode, char * szData, workingStorage_type *pWS);
int  CheckInvalidUsers(PMQCD pChannelDefinition, workingStorage_type *pWS);
int  CheckBlankUser(PMQCD pChannelDefinition, workingStorage_type *pWS);
int  CheckUserId(PMQCD pChannelDefinition, workingStorage_type *pWS );
int  CheckBlockedUserId(PMQCD pChannelDefinition, workingStorage_type *pWS );
int  CheckConnectionPattern( PMQCD pChannelDefinition, workingStorage_type *pWS );
int  CheckCONList( MQCD *pChannelDefinition, MQLONG ExitReason, workingStorage_type *pWS);
int  CheckFAP(PMQCXP pChannelExitParams, workingStorage_type *pWS);    /* 2.98 */
int  CheckPartnerName( MQCD *pChannelDefinition, PMQCXP pChannelExitParams, workingStorage_type *pWS );

char * clean(char *bf, int len);

#ifdef MQCD_VERSION_7
int  CheckSSLList( MQCD *pChannelDefinition, PMQCXP pChannelExitParams, workingStorage_type *pWS );
int  getDNMCAUser(char * key, char * dn, char * resultstr, int len,workingStorage_type *pWS);
#endif

int  CopyUsers(char *p, workingStorage_type *pWS);
int  Index1(char *p, long buflen, const char *s);
int  GetParam(char *pb, char *pt,const char *t1,const char *t2);

int  ProcessData(char *line, workingStorage_type *pWS );
int  ProcessFile( char *p, workingStorage_type *pWS);
int  InitialiseData(PMQCD pChannelDefinition, workingStorage_type *pWS, ws_type *pWS1, int ExitReason);
void InitEmerg(workingStorage_type *pWS);
void i_strupr(char *p);
char * Strip (char * s);   /* 2.45 */
void translate (char *src, char *ny, char *old, workingStorage_type *pWS);
void translate2 (char *src, char *ny, char *old);

#if defined(__MVS__)
void WTODataE(char * szData, workingStorage_type *pWS);
void WTODataOk(char * szData, workingStorage_type *pWS);
#endif

long UpdateStatusQ(MQCD *pChannelDefinition,  long* what2do, workingStorage_type *pWS);
long deductStatusQ(MQCD *pChannelDefinition, workingStorage_type *pWS);

void nullDelimit(char[],char[],long);
long connectToQMgr(workingStorage_type *pWS);
int  CheckRunningChannels(MQCD *pChannelDefinition, workingStorage_type *pWS);
char * MQExitReason(PMQCXP pChannelExitParms);
char * MQExitId(PMQCXP pChannelExitParms);
char * MQChannelType(PMQCD pChannelDefinition);
char * MQExitResponseType(PMQCXP pChannelExitParams);
char * MQAuthenticationType(PMQCXP pChannelExitParams);

void SetLongMCAUser(char *szMCAUser, workingStorage_type *pWS);
void InsertLongMCAUser(PMQCD pChannelDefinition, workingStorage_type *pWS, ws_type *pWS1);

/* start of inq chl defs */
#if !defined(__MVS__) && !defined(__BI2_78__)
long bagSetup(MQHBAG*const,MQHBAG*const,MQHBAG*const, workingStorage_type *pWS);
long getActualCurrentNumberOfRunningChannels(long*, workingStorage_type *pWS);
long openStatusQ(workingStorage_type *pWS);
long closeStatusQ(workingStorage_type *pWS);
#endif

#if defined(__MVS__)
#if !defined(__BI2_78__)
long chsSetup(workingStorage_type *pWS);
long getActualCurrentNumberOfRunningChannelszOS(long*, workingStorage_type *pWS);
long chsSendDisplayCHStatus(char * cmdq, char * replyq, workingStorage_type *pWS);
long closeReplyQ(workingStorage_type *pWS);
#endif
void GetZOSUserId(char *buffer);
#endif

#if defined(UNIX) || defined (OS400)
int  LockAllSem(workingStorage_type *pWS);
int  FreeAllSem(workingStorage_type *pWS);
int  LockChannelSem(workingStorage_type *pWS);
int  FreeChannelSem(workingStorage_type *pWS);
int  LockRenameSem(workingStorage_type *pWS);
int  FreeRenameSem(workingStorage_type *pWS);
int  LockLogSem(workingStorage_type *pWS);
int  FreeLogSem(workingStorage_type *pWS);
int  FreeSpecSem(int sem_number, workingStorage_type *pWS);
int  LockSpecSem(int sem_number, workingStorage_type *pWS);
long UpdateShrMem(MQCD *pChannelDefinition,  long* what2do, workingStorage_type *pWS);
int  ProcessMQSiniFile(char *szQmgrName, workingStorage_type *pWS, ws_type *pWS1);
int  ProcessMQSiniFile_qm_ini(char *szQmgrName, workingStorage_type *pWS, ws_type *pWS1);
void SendLogMessage(char *szData, workingStorage_type *pWS);
int HasMasterChanged(char *ts, workingStorage_type *pWS);
#endif


#if !defined (_REENTRANT)
char *strtok_r(char * s, const char * delim , char ** ptrptr );
#endif

#if defined (WIN32)
int  LockChannelSem(workingStorage_type *pWS);
int  FreeChannelSem(workingStorage_type *pWS);
long UpdateShrMem(MQCD *pChannelDefinition,  long* what2do, workingStorage_type *pWS);
void SendLogMessage(char *szData, workingStorage_type *pWS);
void OpenMTX(workingStorage_type *pWS, char *szLogFile);
int LoadRegistry(workingStorage_type *pWS);
void OpenMTXBlockIP2S(workingStorage_type *pWS, HANDLE *MutexBlockIP2S);
int LockBlockIP2S(workingStorage_type *pWS, HANDLE *MutexBlockIP2S);
int FreeBlockIP2S(workingStorage_type *pWS, HANDLE *MutexBlockIP2S);
void WINFmtMsg(DWORD eNum, char *ErrorMsg, int lenErrMsg);

BOOL CreateMyDACL(SECURITY_ATTRIBUTES * pSA, workingStorage_type *pWS);
PSID SpcLookupName(LPCTSTR lpszSystemName, LPCTSTR lpszAccountName, workingStorage_type *pWS);
int PrintSid( SID *ps, char * resstring, int len );
#endif
void _Lsplitpath(const char* path, char* drv, char* dir, char* name, char* ext);
char * GenerateLogFileName(char * szFile, workingStorage_type *pWS);

#if defined (__EXT_LOG_FILENAME__)
long CycleFiles(workingStorage_type *pWS);
long getFileSize(char* szFileName, workingStorage_type *pWS);
long HandleCycleFiles(workingStorage_type *pWS);
#endif

#if defined (UNIX) || defined (WIN32) || defined (OS400)
int ProcessConfigFile(MQCHAR48  qMgrName, workingStorage_type *pWS);
#endif


#ifdef HNLUP
#ifdef HNLUP2
MQLONG GetHostIP(char *connectionName, DNSReply *DNSReplyPtr, workingStorage_type *pWS);
#else
MQLONG GetHostIP(char *connectionName, char *connectionAddrPtr, workingStorage_type *pWS);
#endif
#if defined __MVS__
/* Create local support for the inet_ntop, that fails when calling USS. */
char * inet_ntop(int af, char *src, char *dst, int size);
char * inet_ntop6(char *src, char *dst, int size);
char * inet_ntop4(char *src, char *dst, int size);
#endif /* __MVS__ */

#endif  /*HNLUP */
void HexDump(void *pAddressIn, long  lSize, workingStorage_type *pWS);


/**
* rlnts()
*
* Remove trailing spaces
*
*
* @return length, string has no trailing spaces
*/
static int rlnts(int l, char * s)
{
    while( l && s[l-1] == ' ') --l;
    return l;
}




/********************************************************************************************
 *
 * InitEmerg()
 *
 *
 * Very basic setup. Just to able to handle the very first error  reporting
 *
 ********************************************************************************************/
void  InitEmerg(workingStorage_type *pWS)
{
    /*
    * Process some debugging parameters first
    */
    memset(pWS,'\00',sizeof(pWS));
    pWS->nDebugFlag=LOGDATA;

    /*
    * Setup logging information
    */
    pWS->NoLogFileName=0;                                                     /* 2.95 */
#if defined (__MVS__)
    /* NWC New defaults for MVS implementation */
    strcpy(pWS->szLogFileName,"//DD:SYSPRINT");
    strcpy(pWS->szLogFileExtension,"");
    strcpy(pWS->szLogDrive,"");
    pWS->nLogFormat=LOGNAME;
    strcpy(pWS->szLogChannel,"");
    strcpy(pWS->szLogPath,"");
    pWS->nZOSLoadMem=0;                      /* JHP 2014*/
#else /* not MVS */
    strcpy(pWS->szLogFileName,"BlockIP2");
    strcpy(pWS->szLogFileExtension,"log");
    strcpy(pWS->szLogDrive,"C:");
    pWS->nLogFormat=LOGNAME;
    strcpy(pWS->szLogChannel,"");
    memset(pWS->szShortQMName,'\00',sizeof(pWS->szShortQMName));              /* 2.80 */
    pWS->bIncludeQmgrNameInLog=0;                                             /* 2.80 */
#if defined (WIN32)
    strcpy(pWS->szLogPath,"\\");
    pWS->hMutex = (void*)ERROR_INVALID_HANDLE;
    pWS->hMutexShm = (void*)ERROR_INVALID_HANDLE;
    strcpy(pWS->szMutexKey,"");
    strcpy(pWS->ExitPath,"");
    if (LoadRegistry(pWS) == 0)
    {
        memcpy(pWS->szLogDrive,pWS->ExitPath,2);
        memcpy(pWS->szLogPath,pWS->ExitPath+2,strlen(pWS->ExitPath)-1);  /* take '\00' too */
    }

#else       /* should be unix file system */
    strcpy(pWS->szLogPath,"/var/mqm/exits/");
#endif
#endif

#if MQAT_DEFAULT==MQAT_OS400
    strcpy(pWS->szLogFileName,"BLOCKIP2");
    strcpy(pWS->szLogFileExtension,"LOG");
    strcpy(pWS->szLogPath,"/QIBM/Userdata/mqm/qmgrs/");
    strcat(pWS->szLogPath,pWS->szQMgr);
    strcat(pWS->szLogPath,"/errors/");
    strcpy(pWS->szLogDrive,"");
    pWS->lTmpPath=0;                                   /* Don't use /tmp/ as default for shared storage on iSeries */
    strcpy(pWS->szMasterFile,"");                      /* where to located the master file   2.90 */
    pWS->iStbyMode=0;                                  /* Standby not enabled                2.90 */
#endif

#if defined(UNIX)
    pWS->iSyslogPrty=LOG_INFO;                         /* priority to use when writing to syslog  */
    pWS->iSyslogFclty=LOG_DAEMON;                      /* facility to log as */
    pWS->lTmpPath=1;                                   /* use /tmp/ as default for shared storage */
    strcpy(pWS->szMasterFile,"");                      /* where to located the master file   2.90 */
    pWS->iStbyMode=0;                                  /* Standby not enabled                2.90 */
#endif

#if defined (UNIX) || defined (OS400)
    pWS->KeySemaphores=-1;
    pWS->KeyShmem=-1;
    strcpy(pWS->szQmgrPrefix,"");
    strcpy(pWS->szQmgrDirectory,"");
    strcpy(pWS->szDataPath,"");                            /* 2.83 */
#endif


    pWS->bUseSecurityParms=0;                          /* 2.80 Don't use SecurityParms    */
    pWS->bForceSecurityParmsOff=0;                     /* 2.94 Don't bypass SecurityParms */
    pWS->lFAPLevel=-1;                                 /* 2.98 FAP Check not active */
}


/**
* InitialiseData()
*
*
* Setup required fields from Exit data and
* cleanup any lists.
*
*
* @pre	  Require exit parameters to be set by MQ
* @return int SEC_OK if if no failures else SEC_FAILURE
*/
int  InitialiseData(PMQCD pChannelDefinition, workingStorage_type *pWS, ws_type *pWS1, int ExitReason)
{
    int i;
    char *p;

    pWS->nDebugFlag=LOGDATA;
    pWS->ExitReason=ExitReason;
    pWS->ppWS1=pWS1;                                       /* Tell us where to locate to stored area 2.94 */

    strcpy(pWS->Pattern,"");                               /* The pattern to match against */
    strncpy(pWS->Pattern,pChannelDefinition->SecurityUserData,MQ_EXIT_DATA_LENGTH);
    pWS->Pattern[MQ_EXIT_DATA_LENGTH] = 0x00;              /* force end of string */
    for (i=0; i<MQ_EXIT_DATA_LENGTH; i++)
    {
        if(pWS->Pattern[i] == ' ')
        {
            pWS->Pattern[i] = 0x00;
            break;
        }
    }
    strcpy(pWS->szSCYDATA,pWS->Pattern);
    for (i=0; i<MQ_EXIT_DATA_LENGTH; i++)
    {
        if(pWS->Pattern[i] == '-')
        {
            pWS->Pattern[i] = 0x00;
            break;
        }
    }

    /* find name of this security exit, usually BlockIP but could have a different name */
    /* if there is a blank or parenthesis char, the remaining part after does not belong to the name of the exit */
    strncpy(pWS->szSCYNAME,pChannelDefinition->SecurityExit,MQ_EXIT_NAME_LENGTH);
    pWS->szSCYNAME[MQ_EXIT_NAME_LENGTH] = 0x00;              /* force end of string */
    for (i=0; i<MQ_EXIT_NAME_LENGTH; i++)
    {
        if((pWS->szSCYNAME[i] == ' ') || (pWS->szSCYNAME[i] == '('))
        {
            pWS->szSCYNAME[i] = 0x00;
            break;
        }
    }

	strcpy(pWS->PartnerName,"");                                /* No partnername check  3.00 */

    p = strstr(pWS->szSCYDATA,"-i");
    if( p != 0 ) pWS->bIgnoreBI2ini = INI_IGNORE;
    else pWS->bIgnoreBI2ini = INI_PROCESS;

#if defined (UNIX) || defined (OS400)
    /*
    * Turn off the usage of /tmp as anchor for shared storage and mutex
    */
    p = strstr(pWS->szSCYDATA,"-t");                            /* Location override       2.90 */
    if( p != 0 ) pWS->lTmpPath = 0;                             /* use the old location    2.90 */
    else pWS->lTmpPath = 1;                                     /* use new location /tmp   2.90 */
#endif

    /*
    * Process some debugging parameters first
    */
    p = strstr(pWS->szSCYDATA,"-d");
    if( p != 0 )
    {
        p = strstr(pWS->szSCYDATA,"-d1");
        if( p != 0 )
        {
            pWS->nDebugFlag=LOGDEBUG1;
        }
        else
        {
            p = strstr(pWS->szSCYDATA,"-d0");
            if( p != 0 )
            {
                pWS->nDebugFlag=LOGDEBUG0;
            }
            else
            {
                pWS->nDebugFlag=LOGDEBUG;
            }
        }
        p = strstr(pWS->szSCYDATA,"-d8");
        if( p != 0 )
        {
            pWS->nDebugFlag=LOGDEBUG8;
        }
    }


#if defined (UNIX) || defined (OS400) || defined (WIN32)
    pWS->lWinLog=LOG_FILEONLY;
#endif
    pWS->bLogTERM=1;                                       /* By default log TERM call */
    /* Initialize working storage.. */
    strcpy(pWS->szStatusQ,"SYSTEM.ADMIN.BIP2.QUEUE");
    pWS->lMaxChannelActive=-1;
    memset(pWS->szQMgr,'\00',sizeof(pWS->szQMgr));
    memcpy(pWS->szQMgr,pChannelDefinition->QMgrName,rlnts(MQ_Q_MGR_NAME_LENGTH, pChannelDefinition->QMgrName));
#if defined (UNIX) || defined (OS400) || defined (WIN32)
    memset(pWS->szShortQMName,'\00',sizeof(pWS->szShortQMName));              /* 2.80 */
    strncpy(pWS->szShortQMName,pWS->szQMgr,sizeof(pWS->szShortQMName)-1);     /* 2.80 */
    pWS->szShortQMName[sizeof(pWS->szShortQMName)-1]='\00';                   /* 2.80 */
#endif


    memset(pWS->szChannelName,'\00',sizeof(pWS->szChannelName));
    memcpy(pWS->szChannelName,pChannelDefinition->ChannelName,rlnts(MQ_CHANNEL_NAME_LENGTH, pChannelDefinition->ChannelName));

    pWS->nZOSconv=0;                                       /* convert brackets on z/OS to handle national chars */
    pWS->bFileSizeCheck=0;                                 /* We haven't played with fileSizeCHeck yet */
    pWS->bFileSizeDone=0;                                  /* We havent cheched filesize yet 2.62 */
    pWS->lLogFileCount=3;                                  /* Number of logfiles to maintain default =3 */
    pWS->lMaxlogFileSize=200*1024;                         /* max size of logfile before cycling defailt 200kb */

    pWS->bRespectMCA=0;                                    /* we don't respect MCAsettings */

    if (pChannelDefinition->ChannelType==MQCHT_SVRCONN)
    {
        pWS->bAllowBlankUserID=0;                          /* We will not accept blank uids */
    }
    else
    {
        pWS->bAllowBlankUserID=1;                          /* We will accept blank uids on non SVRCONN channels */
    }
    pWS->bBlockMqmUsers=0;
    pWS->bDebugMode=0;
    pWS->SilentMode=0;
    pWS->bLogByDate=0;
    pWS->bUseridUpperLowerCase=0;                          /* be case sensitive */
    pWS->ReportLogFileError=0;                             /* only report logfileerror once */
    pWS->lExpireValue=600*10;                              /* let channel record live for 10 minutes (in secs) */
    pWS->lRefuseReason=0;                                  /* send refuse reason back to client exit.. */
    pWS->bInteractive=0;                                   /* is the client requesting feedback ? */

    pWS->SSLCount=0;
    pWS->CONCount = 0;

    strcpy(pWS->RemUID,"");                                /* remote user identifier */
    strcpy(pWS->ConName,"");                               /* Local copy of the connectionname */
    strcpy(pWS->Flags,"");
    strcpy(pWS->U2,"");
    strcpy(pWS->Users,"");                                 /* list of userids */
    strcpy(pWS->BlockUsers,"");                            /* list of Blocked userids */
    strcpy(pWS->szClientUser,"");                          /* user in the other end */

#if defined (UNIX) || defined (OS400)
    pWS->KeySemaphores=-1;
    pWS->KeyShmem=-1;
    strcpy(pWS->szQmgrPrefix,"");
    strcpy(pWS->szQmgrDirectory,"");
    strcpy(pWS->szDataPath,"");                            /* 2.83 */
#endif

    /*
    * Setup logging information
    */
    pWS->NoLogFileName=0;                                  /* 2.95 */

#if defined (__MVS__)
    /* NWC New defaults for MVS implementation */
    strcpy(pWS->szLogFileName,"//DD:SYSPRINT");
    strcpy(pWS->szLogFileExtension,"");
    strcpy(pWS->szLogDrive,"");
    pWS->nLogFormat=LOGNAME;
    strcpy(pWS->szLogChannel,"");
    strcpy(pWS->szLogPath,"");

#else /* not MVS */
    strcpy(pWS->szLogFileName,"BlockIP2");
    strcpy(pWS->szLogFileExtension,"log");
    strcpy(pWS->szLogDrive,"C:");
    pWS->nLogFormat=LOGNAME;
    strcpy(pWS->szLogChannel,"");
#if defined (WIN32)
    strcpy(pWS->szLogPath,"\\");
    pWS->hMutex = (void*)ERROR_INVALID_HANDLE;
    pWS->hMutexShm = (void*)ERROR_INVALID_HANDLE;
    strcpy(pWS->szMutexKey,"");
    strcpy(pWS->ExitPath,"");
    if (LoadRegistry(pWS) == 0)
    {
        memcpy(pWS->szLogDrive,pWS->ExitPath,2);
        memcpy(pWS->szLogPath,pWS->ExitPath+2,strlen(pWS->ExitPath)-1);  /* take '\00' too */
    }

#else       /* should be unix file system */
    strcpy(pWS->szLogPath,"/var/mqm/exits/");
#endif
#endif

#if MQAT_DEFAULT==MQAT_OS400
    strcpy(pWS->szLogFileName,"BLOCKIP2");
    strcpy(pWS->szLogFileExtension,"LOG");
    strcpy(pWS->szLogPath,"/QIBM/Userdata/mqm/qmgrs/");
    strcat(pWS->szLogPath,pWS->szQMgr);
    strcat(pWS->szLogPath,"/errors/");
    strcpy(pWS->szLogDrive,"");
#endif

#if defined UNIX || defined (OS400)
    /* functions added under 2.45 */
    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData( LOGDEBUG1, "Looking for BlockIP2.ini processing", pWS);

    if (ProcessMQSiniFile(pWS->szQMgr, pWS, pWS1) !=SEC_OK) {
        if (pWS->ExitReason==MQXR_INIT) {
            if (pWS->nDebugFlag >= LOGDEBUG1)
                LogData(LOGDEBUG1,"  ProcessMQSiniFile() requested connection broken.", pWS);
        }
        return SEC_FAILURE;
    }
    /* func added under 2.60 */
    if (ProcessMQSiniFile_qm_ini(pWS->szQMgr, pWS, pWS1) !=SEC_OK)
    {
        if (pWS->ExitReason==MQXR_INIT)
        {
            if (pWS->nDebugFlag >= LOGDEBUG1)
                LogData(LOGDEBUG1,"  ProcessMQSiniFile_qm_ini() requested connection broken.", pWS);
        }
        return SEC_FAILURE;
    }
    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData( LOGDEBUG1, "Looking for BlockIP2.ini was done with success.", pWS);
#endif

    switch(ExitReason)
    {
    case MQXR_INIT:
        pWS1->bFirstTime=1;
        pWS1->lMaxChannelActive=pWS->lMaxChannelActive;
        pWS1->bLogTERM=pWS->bLogTERM;
        strcpy(pWS1->LongMCAUser,"");
        break;
    default:
        pWS->lMaxChannelActive=pWS1->lMaxChannelActive;
        pWS->bLogTERM=pWS1->bLogTERM;
#if defined (UNIX) || defined (OS400)
        strcpy(pWS->szQmgrDirectory,pWS1->szQmgrDirectory);
        strcpy(pWS->szQmgrPrefix,pWS1->szQmgrPrefix);
        strcpy(pWS->szDataPath,pWS1->szDataPath);          /* 2.83 */
#endif
    }

    /*
    * Empty CONList and SSList tables
    */
    for(i=0; i < MAX_SSL; i++)
        strcpy(pWS->SSLList[i], "");
    for(i=0; i < MAX_CON; i++)
        strcpy(pWS->CONList[i], "");

#ifdef MQCXP_VERSION_6
    if (pWS->Version<MQCXP_VERSION_6)
    {
        pWS->AllowASC = 1;                                 /* By default allow Self Signd Certs on non supported */
    }
    else
    {
        pWS->AllowASC = 0;                                 /* By default don't allow Self Signd Certs */
    }
#endif

    pWS->CONCount = 0;
    pWS->SSLCount = 0;
    pWS->RemUID[0]=0x00;                                   /* remote user identifier */
    pWS->ConName[0]= 0x00;                                 /* Local copy of the connectionname */
    pWS->szLogChannel[0]=0x00;
    pWS->bRespectMCA = 0;
    strncpy(pWS->ConName,pChannelDefinition->ConnectionName,MQ_CONN_NAME_LENGTH); /* only take the first 32 bytes ! */
#ifdef TEST281
    strncpy(pWS->ConName,"127.0.0.1(12345)",MQ_CONN_NAME_LENGTH);   /* JHP 2.81 test */
#endif
    for (i=0; i<MQ_CONN_NAME_LENGTH; i++)
    {
        /* if(pWS->ConName[i] == ' ') JHP 2.80 changed */
        if((pWS->ConName[i] == ' ') || (pWS->ConName[i] == '('))
        {
            pWS->ConName[i] = 0x00;
            break;
        }
    }

#ifdef TEST281
    HexDump(pWS->ConName, MQ_CONN_NAME_LENGTH, pWS);
#endif
    if (ExitReason==MQXR_INIT)
    {
        strcpy(pWS1->ConName,pWS->ConName);
        if (pWS->nDebugFlag >= LOGDEBUG8)
        {
            strcpy(pWS->szTmpBufff,"");
#if MQAT_DEFAULT==MQAT_OS400
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"ConName is now <%s>",   pWS1->ConName);
#elif defined(__MVS__)
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"ConName is now <%s>",   pWS1->ConName);
#else
            snprintf(pWS->szTmpBufff, sizeof(pWS->szTmpBufff), "ConName is now [%s]",   pWS1->ConName);
#endif
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDEBUG8,pWS->szTmpBufff, pWS);
        }

        if (!strcmp(pWS->ConName,"0.0.0.0"))
        {                                                                                                   /* 2.56 */
            LogData(LOGDATA,"Warning: It seems like your queue manager needs fix 6.0.2.1 or later", pWS);   /* 2.56 */
        }
        if (!strcmp(pWS->ConName,""))
        {                                                                                                   /* 2.67 */
            LogData(LOGDATA,"Warning: It seems like your queue manager needs fix 6.0.2.1 or later", pWS);   /* 2.67 */
        }
    }
    else
    {
        strcpy(pWS->ConName,pWS1->ConName);                /* restore the right one */
    }
    strncpy(pWS->RemUID,pChannelDefinition->RemoteUserIdentifier,MQ_USER_ID_LENGTH);
    pWS->RemUID[MQ_USER_ID_LENGTH]=0x00;                   /* force last char */

    for(i=0; i<MQ_USER_ID_LENGTH; i++)
    {
        if(pWS->RemUID[i] == ' ')
        {
            pWS->RemUID[i] = 0x00;
            break;
        }
    }

    strncpy(pWS->szLogChannel, pChannelDefinition->ChannelName,MQ_CHANNEL_NAME_LENGTH);
    pWS->szLogChannel[MQ_CHANNEL_NAME_LENGTH]=0x00;
    for (i=0; i<MQ_CHANNEL_NAME_LENGTH; i++)
    {
        if(pWS->szLogChannel[i] == ' ')
        {
            pWS->szLogChannel[i] = 0x00;
            break;
        }
    }

#if defined (WIN32)
    p = strstr(pWS->szSCYDATA,"-w");
    if( p != 0 ) pWS->lWinLog=LOG_EVTLOGONLY;
#endif
#if defined (UNIX)
    p = strstr(pWS->szSCYDATA,"-w");
    if( p != 0 ) pWS->lWinLog=LOG_EVTLOGONLY;
#endif
    /*
    * Process some debugging parameters first
    */

#if defined(MQCXP_VERSION_6)
    p=strstr(pWS->szSCYDATA,"-s");                         /* allow selfsigned certs */
    if( p != 0 ) pWS->AllowASC=1;
#endif

    /* start JHP 040330 */
    p = strstr(pWS->szSCYDATA,"FN=");                      /* Filename specified .. */
    if( p == NULL )
    {
        /* check pattern syntax JHP 040805 */
        if ((*pWS->Pattern != '*') && (*pWS->Pattern != '?') /*&& *pWS->Pattern != 'N'*/ && ((*pWS->Pattern > '9') || (*pWS->Pattern < '0' )))
        {
            /* 2.80 start */
            if ((*pWS->Pattern==' ') || (*pWS->Pattern=='\00'))  /* we forgot to fill in the SCYDATA, change it to "NONE;" */
            {
                memcpy(pWS->Pattern,"NONE;",5);
                if (pWS->nDebugFlag >= LOGDEBUG)
                    LogData(LOGDEBUG, "adjusting SCYDATA to \"NONE;\"", pWS);
            }
            /* 2.80 end   */
            if (memcmp(pWS->Pattern,"NONE",4)) {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Errors SCYDATA specification - Connection refused for Channel [%.*s]",
                         rlnts(MQ_CHANNEL_NAME_LENGTH, pChannelDefinition->ChannelName),
                         pChannelDefinition->ChannelName);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData( LOGDATA, pWS->szTmpBufff, pWS);
#if defined(__MVS__)
                WTODataE("01E", pWS);
#endif
                return SEC_FAILURE;
            }
        }
        p = strstr(pWS->szSCYDATA,"-n");
        if( p != 0 ) pWS->bBlockMqmUsers=1;

        p = strstr(pWS->szSCYDATA,"+b");
        if( p != 0 ) pWS->bAllowBlankUserID=1;

        p = strstr(pWS->szSCYDATA,"userids=");
        if( p != 0 )
        {
            if( CopyUsers(p,pWS)==SEC_FAILURE)
            {
                return SEC_FAILURE;
            }
        }
#if defined (UNIX) || defined (WIN32) || defined (OS400)
        if (pWS->bIgnoreBI2ini==INI_PROCESS)
        {
            if(pWS->ExitReason==MQXR_INIT)
            {
                if (pWS->nDebugFlag >= LOGDEBUG1)
                    LogData( LOGDEBUG1, "Starting on BlockIP2.ini processing", pWS);        /* 2.60 start */
            }
            if (ProcessConfigFile(pChannelDefinition -> QMgrName, pWS) > 0)
            {
                if (ExitReason==MQXR_INIT) {               /* Only report it on MQXR_INIT */
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Errors found in BlockIP2.ini - Connection refused for Channel [%.*s]",
                             rlnts(MQ_CHANNEL_NAME_LENGTH, pChannelDefinition->ChannelName),
                             pChannelDefinition->ChannelName);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData( LOGDATA, pWS->szTmpBufff, pWS);
#if defined(__MVS__)
                    WTODataE("01E", pWS);
#endif
                }
                pWS1->lMaxChannelActive=pWS->lMaxChannelActive;               /* Update flag 2.66 */
                pWS1->bLogTERM=pWS->bLogTERM;
                return SEC_FAILURE;
            }
            if(pWS->ExitReason==MQXR_INIT)
            {
                if (pWS->nDebugFlag >= LOGDEBUG1)
                    LogData( LOGDEBUG1, "Finished BlockIP2.ini processing", pWS);          /* 2.60 end */
            }
            pWS1->lMaxChannelActive=pWS->lMaxChannelActive;               /* Update flag 2.66 */
            pWS1->bLogTERM=pWS->bLogTERM;
        }
        else
            if(pWS->ExitReason==MQXR_INIT)
            {
                if (pWS->nDebugFlag >= LOGDEBUG1)
                    LogData( LOGDEBUG1, "Processing of BlockIP2.ini Bypassed.",pWS);
            }
#endif

    }

    return SEC_OK;
    /* END JHP 040330 */
}

void i_strupr(char *p)
{
    while (*p)
    {
        *p = toupper(*p);
        p++;
    }
}

/*#if defined (WIN32)*/
#if !defined (_REENTRANT)
char *strtok_r(char * s, const char * delim , char ** ptrptr )
{
    return(strtok(s, delim));
}
#endif

/*
* BlockExit - Channel security exit, entry point
*
* PARAMETERS:
*	 pChannelExitParams (PMQCXP)  - input/output
*		Channel exit parameter block
*
*	 pChannelDefinition (PMQCD)   - input/output
*		Channel definition
*
*	 pDataLength		(PMQLONG) - input/output
*		On input:  length of AgentBuffer
*		On output: length of returned data (AgentBuff/ErrorBuff)
*
*	 pAgentBufferLength (PMQLONG) - input
*		Agent Buffer length
*
*	 AgentBuffer		(PMQBYTE) - input/output
*		Agent buffer
*
*	 pExitBufferLength	(PMQLONG) - input/output
*		Exit buffer length
*
*	 pExitBufferAddr	(PMQPTR)  - input/output
*		Exit buffer
*
* RETURN:
*	 void
*
* NOTES:
*
*	 For more details on the operation of security exits, see:
*		'MQSeries - Distributed Queue Management Guide'
*
* --------------------------------------------------------------------------
* This security exit does NOT perform any security validation, it only solves
* is to show how to log connections made to a particular channel.
* --------------------------------------------------------------------------
*
*/
/**
*  BlockExit()
*
*
* main entry point for security exit
*
*
* @return void
*/

#if MQAT_DEFAULT==MQAT_OS400
int main (int argc, char * argv[])

#else

/* NWC MVS linkage requirements added */
#if defined(__MVS__)
/* Define code CSECT name */
#pragma csect(code,"BLOCK2")
/* Define static CSECT name */
#pragma csect(static,"BLOCK2D")

#pragma runopts(trap(off),PLIST(MVS))

#pragma environment(BLOCKIP2)

int BLOCKIP2 (
    PMQCXP	 pChannelExitParams,
    PMQCD 	 pChannelDefinition,
    PMQLONG	 pDataLength,
    PMQLONG	 pAgentBufferLength,
    PMQBYTE	 AgentBuffer,
    PMQLONG	 pExitBufferLength,
    PMQPTR	 pExitBufferAddr)

/*  MVS */
#else
/*
* dummy function used as entry point to exit, only needed for *IX boxes
*/

extern void MQENTRY MQStart(void) {;}

#if defined (WIN32)

__declspec (dllexport) void MQENTRY BlockExit (
    PMQCXP	  pChannelExitParams,
    PMQCD	  pChannelDefinition,
    PMQLONG	  pDataLength,
    PMQLONG	  pAgentBufferLength,
    PMQBYTE	  AgentBuffer,
    PMQLONG	  pExitBufferLength,
    PMQPTR	  pExitBufferAddr);
#endif

extern void MQENTRY BlockExit(
        PMQCXP	 pChannelExitParams,
        PMQCD 	 pChannelDefinition,
        PMQLONG	 pDataLength,
        PMQLONG	 pAgentBufferLength,
        PMQBYTE	 AgentBuffer,
        PMQLONG	 pExitBufferLength,
        PMQPTR	 pExitBufferAddr)

    /* Not MVS */
#endif
/* not OS400 */
#endif
{
    int i;
    char *p;
#if defined(__MVS__)
    long len, reason, tPtr;
#endif
    workingStorage_type ws;
    workingStorage_type *pWS;

#if MQAT_DEFAULT!=MQAT_OS400
    ws_type ** ppExitUserArea = (ws_type**) &pChannelExitParams->ExitUserArea;
    ws_type *  pExitUserArea  = *ppExitUserArea;
#endif

#if MQAT_DEFAULT==MQAT_OS400
    PMQCXP	 pChannelExitParams;   /* Channel Exit Parameter pointer */
    PMQCD 	 pChannelDefinition;   /* Channel Definition pointer     */
    PMQLONG	 pDataLength;          /* Data Length                    */
    PMQLONG	 pAgentBufferLength;   /* Agent Buffer Length            */
    PMQBYTE	 AgentBuffer;          /* Agent Buffer Pointer           */
    PMQLONG	 pExitBufferLength;    /* Exit Buffer Length             */
    PMQPTR	 pExitBufferAddr;      /* Exit Buffer Address Pointer    */

    ws_type ** ppExitUserArea; /* = (ws_type**) &pChannelExitParams->ExitUserArea;*/
    ws_type *  pExitUserArea; /*  = *ppExitUserArea;*/

    if (argc !=8)
    {
        fprintf(stdout,"\nBlockIP2 version %s was invoked without MQ.\n",VERSION);
        fprintf(stdout,
                "\nBlockIP2 - ERROR wrong number of parameters\n");
        return 8 ;
    }

    /* Assign values on argv[] to the channel exit parameter list
       as defined in the WebSphere MQ Intercommunications manual  */
    pChannelExitParams  = (PMQCXP)   argv[1];
    pChannelDefinition  = (PMQCD)    argv[2];
    pDataLength         = (PMQLONG)  argv[3];
    pAgentBufferLength  = (PMQLONG)  argv[4];
    AgentBuffer         = (PMQBYTE)  argv[5];
    pExitBufferLength   = (PMQLONG)  argv[6];
    pExitBufferAddr     = (PMQPTR)   argv[7];

    ppExitUserArea = (ws_type**) &pChannelExitParams->ExitUserArea;
    pExitUserArea  = *ppExitUserArea;
#endif

    pExitBufferAddr   = pExitBufferAddr;
    pExitBufferLength = pExitBufferLength;
    pWS=&ws;

    /* Emergency init just to be shure that error reporting works */
    InitEmerg(pWS);

#if defined(__MVS__)
    strcpy(pWS->szWTOPfx,"BLOCKIP-");   /* jhp 2.76 */
    snprintf(pWS->szWTOBuffErr,sizeof(pWS->szWTOBuffErr),"Conn, REFUSED Chl.=%.*s CONN=%.*s USER=%.*s",
             rlnts(MQ_CHANNEL_NAME_LENGTH, pChannelDefinition->ChannelName),
             pChannelDefinition->ChannelName,
             rlnts(MQ_CONN_NAME_LENGTH, pWS->ConName),
             pWS->ConName,
             rlnts(MQ_USER_ID_LENGTH, pChannelDefinition->RemoteUserIdentifier),
             pChannelDefinition->RemoteUserIdentifier);
    pWS->szWTOBuffErr[sizeof(pWS->szWTOBuffErr)-1]='\00';
    snprintf(pWS->szWTOBuffOk,sizeof(pWS->szWTOBuffOk),"Conn. ACCEPTED Chl.=%.*s CONN=%.*s USER=%.*s",
             rlnts(MQ_CHANNEL_NAME_LENGTH, pChannelDefinition->ChannelName),
             pChannelDefinition->ChannelName,
             rlnts(MQ_CONN_NAME_LENGTH, pWS->ConName),
             pWS->ConName,
             rlnts(MQ_USER_ID_LENGTH, pChannelDefinition->RemoteUserIdentifier),
             pChannelDefinition->RemoteUserIdentifier);
    pWS->szWTOBuffOk[sizeof(pWS->szWTOBuffOk)-1]='\00';
#endif

    /*
    * check if security exit has been invoked
    */
    if(pChannelExitParams->ExitId!=MQXT_CHANNEL_SEC_EXIT)
    {
        pChannelExitParams->ExitResponse = MQXCC_SUPPRESS_FUNCTION;
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"ERROR!! - Unexpected invocation.  ExitId=[%ld]",   (long)pChannelExitParams->ExitId);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDATA,pWS->szTmpBufff, pWS);
#if MQAT_DEFAULT==MQAT_OS400
        return 0;
#else
        return;
#endif
    }
    if (pChannelExitParams->ExitReason==MQXR_INIT)
    {
        /*       LogData(LOGDEBUG,"Before calloc", pWS);*/

#if !defined(__MVS__)
        pExitUserArea = (ws_type *)calloc( 1, sizeof(ws_type) );
#else
        len = sizeof(ws_type);
        BLOCKIPM(&tPtr, &len, "G", &reason);                        /* GETMAIN */
        pExitUserArea=(ws_type*)tPtr;
#endif
        if (pExitUserArea)
        {
            /*            if (pWS->nDebugFlag >= LOGDEBUG1)
                            LogData(LOGDEBUG1,"after calloc", pWS);*/
            *ppExitUserArea = pExitUserArea;

            pExitUserArea->bFirstTime=0;

            /*            if (pWS->nDebugFlag >= LOGDEBUG1)
                            LogData(LOGDEBUG1,"calloc success", pWS);*/

        }
        else
        {
            pChannelExitParams->ExitResponse = MQXCC_SUPPRESS_FUNCTION;
            LogData(LOGDATA,"calloc failed", pWS);
            return;
        }
    }
    else
    {

        /*    sprintf(pWS->szTmpBufff,"szLogFileName [%s]",pExitUserArea->szLogFileName);
            LogData(LOGDATA,pWS->szTmpBufff, pWS);
        */
    }

    /*
    * initialize some structures
    */
    i=pWS->lMaxChannelActive;
    pWS->Version =pChannelExitParams->Version;
    if (InitialiseData(pChannelDefinition, pWS, pExitUserArea, pChannelExitParams->ExitReason)==SEC_FAILURE)
    {
        if (pChannelExitParams->ExitReason == MQXR_TERM)		  /* JHP 0403030 */
        {
            if (pExitUserArea->lMaxChannelActive!=-1) {                       /* Be ready to store shm info here */
                if (deductStatusQ(pChannelDefinition, pWS) == 0)
                {}
                if (pWS->nDebugFlag >= LOGDEBUG)
                    LogData( LOGDEBUG, "======= TERM ======", pWS);
            }
#if defined(__MVS__)
            len = sizeof(ws_type);
            tPtr=(long)pExitUserArea;
            BLOCKIPM(&tPtr, &len, "F", &reason);
#else
            free( pExitUserArea );
#endif
            pExitUserArea = 0;

            if (pWS->nDebugFlag >= LOGDEBUG1)
                LogData(LOGDEBUG1,"Free success", pWS);
        }
        if (pChannelExitParams->ExitReason == MQXR_TERM)
        {		  /* JHP 0403030 */
            pChannelExitParams->ExitResponse  = MQXCC_OK;      /* Allways ok on MQXR_TERM JHP 2.60 */
        }
        else
        {
            pChannelExitParams->ExitResponse = MQXCC_SUPPRESS_FUNCTION; /* Allways error on not TERM 2.60 */
        }
#if MQAT_DEFAULT==MQAT_OS400
        return 0;
#else
        return;
#endif
    }

#if defined(__MVS__)
    snprintf(pWS->szWTOBuffErr,sizeof(pWS->szWTOBuffErr),"Conn, REFUSED Chl.=%.*s CONN=%.*s USER=%.*s",
             rlnts(MQ_CHANNEL_NAME_LENGTH, pChannelDefinition->ChannelName),
             pChannelDefinition->ChannelName,
             rlnts(MQ_CONN_NAME_LENGTH, pWS->ConName),
             pWS->ConName,
             rlnts(MQ_USER_ID_LENGTH, pChannelDefinition->RemoteUserIdentifier),
             pChannelDefinition->RemoteUserIdentifier);
    pWS->szWTOBuffErr[sizeof(pWS->szWTOBuffErr)-1]='\00';
    snprintf(pWS->szWTOBuffOk,sizeof(pWS->szWTOBuffOk),"Conn. ACCEPTED Chl.=%.*s CONN=%.*s USER=%.*s",
             rlnts(MQ_CHANNEL_NAME_LENGTH, pChannelDefinition->ChannelName),
             pChannelDefinition->ChannelName,
             rlnts(MQ_CONN_NAME_LENGTH, pWS->ConName),
             pWS->ConName,
             rlnts(MQ_USER_ID_LENGTH, pChannelDefinition->RemoteUserIdentifier),
             pChannelDefinition->RemoteUserIdentifier);
    pWS->szWTOBuffOk[sizeof(pWS->szWTOBuffOk)-1]='\00';
#endif


    /*
    *  Only accept z/os option on z/OS implementation
    */
#if MQAT_DEFAULT==MQAT_OS400
    p = strstr(pWS->szSCYDATA,"-z");
    if( p != 0 ) pWS->nZOSconv=1;
#endif
#if defined(__MVS__)
    p = strstr(pWS->szSCYDATA,"-z");
    if( p != 0 ) pWS->nZOSconv=1;
    p = strstr(pWS->szSCYDATA,"-m");                                    /* 2.84 */
    if( p != 0 ) {                                                      /* 2.84 */
        pWS->nZOSLoadMem=1;                                             /* 2.84 */
    }                                                                   /* 2.84 */
#endif
    p = strstr(pWS->szSCYDATA,"-q");
    if( p != 0 )
    {
        pWS->SilentMode=1;
        pWS->bLogTERM=0;
    }

    p = strstr(pWS->szSCYDATA,"FN=");	/* Filename specified .. */
    if( p != NULL )
    {
        p=p+3;
        if(ProcessFile(p, pWS)==SEC_FAILURE)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Errors in parameter file - Connection refused for Channel [%.*s]",
                     rlnts(MQ_CHANNEL_NAME_LENGTH, pChannelDefinition->ChannelName),
                     pChannelDefinition->ChannelName);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData( LOGDATA, pWS->szTmpBufff, pWS);
#if defined(__MVS__)
            WTODataE("02E", pWS);
#endif
            /* Send exit reponse ok if it's a MQXR_TERM invocation, to avoid MQ error complaining about worng exit reason */
            if (pChannelExitParams->ExitReason == MQXR_TERM)		  /* JHP 060810 */
            {
                pChannelExitParams->ExitResponse = MQXCC_OK; /* JHP 060810 */
            }
            else
            {
                pChannelExitParams->ExitResponse = MQXCC_SUPPRESS_FUNCTION; /* JHP 060810 */
            }
#if MQAT_DEFAULT==MQAT_OS400
            return 0;
#else
            return;
#endif
        }
        pExitUserArea->lMaxChannelActive=pWS->lMaxChannelActive;               /* Update flag */
        pExitUserArea->bLogTERM=pWS->bLogTERM;
    }


    /*
    * set default exit responses at this point
    */
    pChannelExitParams->ExitResponse  = MQXCC_OK;
    if (pChannelExitParams->ExitReason == MQXR_TERM)		  /* JHP 0403030 */
    {
        LogData( LOGDEBUG, "======= TERM ======", pWS);
        if (pExitUserArea->lMaxChannelActive!=-1) {                       /* Be ready to store shm info here */
            if (deductStatusQ(pChannelDefinition, pWS) == 0)
            {}
        }

        if (pWS->lMaxChannelActive != -1)       /* Max active channels was specified, check it.. */
        {
            /*           if (deductStatusQ(pChannelDefinition, pWS) == 0)
                       {}*/
        }
        if (pWS->bLogTERM==1 && pWS->SilentMode!=1)   /* 2.75 */
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Channel closed [%.*s] Connection Name [%s]",
                     rlnts(MQ_CHANNEL_NAME_LENGTH, pChannelDefinition->ChannelName),
                     pChannelDefinition->ChannelName,
                     pWS->ConName);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData( LOGDATA, pWS->szTmpBufff, pWS);
            strcpy(pWS->szWTOBuffOk,pWS->szTmpBufff);
#if defined(__MVS__)
            if(pWS->SilentMode==0)
                WTODataOk("99I", pWS);
#endif
        }
        if (pWS->nDebugFlag >= LOGDEBUG1)
            LogData(LOGDEBUG1,"Before Free", pWS);
#if defined(__MVS__)
        len = sizeof(ws_type);
        tPtr=(long)pExitUserArea;
        BLOCKIPM(&tPtr, &len, "F", &reason);
#else
        free( pExitUserArea );
#endif
        ppExitUserArea = 0;
        if (pWS->nDebugFlag >= LOGDEBUG1)
            LogData(LOGDEBUG1,"Free success", pWS);
#if MQAT_DEFAULT==MQAT_OS400
        return 0;
#else
        return;
#endif
    }

    /* begin 2.46 */
#if defined (__EXT_LOG_FILENAME__)
#if defined (WIN32)
    if (pWS->lWinLog!=LOG_EVTLOGONLY)
    {
#endif
        if (pWS->bFileSizeDone==0)    /* We haven't cheched filesize yet 2.62 */
        {
            HandleCycleFiles(pWS);
            pWS->bFileSizeDone=1;
        }
#if defined (WIN32)
    }
#endif
#endif   /* __EXT_LOG_FILENAME__ */
    /* end 2.46 */


    /*
    * switch on ExitReason to see why exit was invoked
    */
    switch(pChannelExitParams->ExitReason)
    {
    case MQXR_INIT: 	   /* Initialization (all exits) */
        LogData( LOGDEBUG, "======= INIT ======", pWS);
        /*#if !defined(__MVS__)*/
        if (pWS->nDebugFlag >= LOGDEBUG1) {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"lMaxChannelActive is now %ld (before channel check) ",
                     pWS->lMaxChannelActive);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData( LOGDEBUG1, pWS->szTmpBufff, pWS);
        }
        if (pWS->lMaxChannelActive != -1)       /* Max active channels was specified, check it.. */
        {
            if (CheckRunningChannels(pChannelDefinition, pWS) != SEC_OK)
            {
                if (pWS->nDebugFlag >= LOGDEBUG1)
                    LogData( LOGDEBUG1, "CheckRunningChannels - refused...", pWS);
#if defined(__MVS__)
                WTODataE("32E", pWS);
#endif
                pChannelExitParams->ExitResponse = MQXCC_SUPPRESS_FUNCTION;
#if MQAT_DEFAULT==MQAT_OS400
                return 0;
#else
                return;
#endif
            }
        }
        /*#endif*/
        break;

    case MQXR_TERM: 	   /* Termination  (all exits) */
        if (pWS->nDebugFlag >= LOGDEBUG)
            LogData( LOGDEBUG, "======= TERM ======", pWS);
#if defined(__MVS__)
        len = sizeof(ws_type);
        tPtr=(long)pExitUserArea;
        BLOCKIPM(&tPtr, &len, "F", &reason);
#else
        free( pExitUserArea );
#endif
        ppExitUserArea = 0;
        if (pWS->nDebugFlag >= LOGDEBUG1)
            LogData(LOGDEBUG1,"Free success", pWS);
        break;

#ifdef MQXR_SEC_PARMS
    case MQXR_SEC_PARMS: 	   /* Security message exchange  (SCY exits) */
        if (pWS->nDebugFlag >= LOGDEBUG)
            LogData( LOGDEBUG, "======= SEC_PARMS ======", pWS);

        if (pChannelExitParams->Version >= MQCXP_VERSION_5)          /* Should be 6 according to the book 2.81 */
        {
            /* 2.80 start */
            /* make sure they came in using the MQCSP, if so, set the MCAUSER to the MQCSP userid */
            /* if (pWS->nDebugFlag >= LOGDEBUG0)  2.81 changed to */
            if (pWS->nDebugFlag >= LOGDEBUG8)
                HexDump(pChannelExitParams, sizeof(MQCXP), pWS);

            if(pChannelExitParams->SecurityParms==0)
            {
                if (pWS->nDebugFlag >= LOGDEBUG0)
                    LogData(LOGDEBUG0,"SecurityParms is empty",pWS);
            }
            else
            {
                if (pWS->nDebugFlag >= LOGDEBUG0)
                {

                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"pWS->bUseSecurityParms %d AuthenticationType %s (%d) Bypass Sec. Parms %d",
                             pWS->bUseSecurityParms,
                             MQAuthenticationType(pChannelExitParams),
                             pChannelExitParams->SecurityParms->AuthenticationType,
                             pWS->bForceSecurityParmsOff);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';

                    LogData( LOGDEBUG0, pWS->szTmpBufff, pWS);

                }
                if (pWS->bUseSecurityParms==1)
                {
                    if (pWS->bForceSecurityParmsOff==1 &&
                             pChannelExitParams->SecurityParms->AuthenticationType == MQCSP_AUTH_USER_ID_AND_PWD)  /* 2.94 start */
                    {
                        pChannelExitParams->SecurityParms->AuthenticationType=MQCSP_AUTH_NONE;   /* Force security checks off */
                        if (pWS->nDebugFlag >= LOGDEBUG)
                        {
                            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Security change, AuthenticationType forced to MQCSP_AUTH_NONE");
                            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                            LogData( LOGDEBUG0, pWS->szTmpBufff, pWS);
                        }
                    }
                    else                                                      /* 2.94 end */
                    {
                        if (pChannelExitParams->ExitReason == MQXR_SEC_PARMS && pChannelExitParams->SecurityParms!=0 &&
                                pChannelExitParams->SecurityParms->AuthenticationType == MQCSP_AUTH_USER_ID_AND_PWD)
                        {
                            if (pChannelDefinition->LongMCAUserIdLength>0)
                            {
                                if (pWS->nDebugFlag >= LOGDEBUG)
                                {
                                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Setting MCAUserId from CSPUserId [%.*s]",
                                    pChannelExitParams->SecurityParms->CSPUserIdLength,
                                    pChannelExitParams->SecurityParms->CSPUserIdPtr);
                                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                                    LogData( LOGDEBUG0, pWS->szTmpBufff, pWS);
                                }
                                pChannelDefinition->LongMCAUserIdPtr = pChannelExitParams->SecurityParms->CSPUserIdPtr;
                                pChannelDefinition->LongMCAUserIdLength = pChannelExitParams->SecurityParms->CSPUserIdLength;
                            }
                        }
                    }
                }
                else
                {
#ifdef MQCD_VERSION_6
                    InsertLongMCAUser(pChannelDefinition, pWS, pExitUserArea); /* Set long MCAUSER */

                    if (pWS->nDebugFlag>=LOGDEBUG8)
                    {
                        sprintf(pWS->szTmpBufff,"Long MCA LongMCAUserIdPtr: [%.*s]  %ld LongRemoteUserIdPtr: [%.*s]  %ld MCA: [%.*s]",
                                pChannelDefinition->LongMCAUserIdLength,
                                (char*)pChannelDefinition->LongMCAUserIdPtr,
                                (long int)pChannelDefinition->LongMCAUserIdLength,
                                pChannelDefinition->LongRemoteUserIdLength,
                                (char*)pChannelDefinition->LongRemoteUserIdPtr,
                                (long int)pChannelDefinition->LongRemoteUserIdLength,
                                rlnts(sizeof(pChannelDefinition->MCAUserIdentifier), pChannelDefinition->MCAUserIdentifier),
                                pChannelDefinition->MCAUserIdentifier);
                        LogData(LOGDEBUG8,pWS->szTmpBufff, pWS);
                    }
                    if (pWS->bForceSecurityParmsOff==1 &&
                             pChannelExitParams->SecurityParms->AuthenticationType == MQCSP_AUTH_USER_ID_AND_PWD)  /* 2.94 start */
                    {
                        pChannelExitParams->SecurityParms->AuthenticationType=MQCSP_AUTH_NONE;   /* Force security checks off */
                        if (pWS->nDebugFlag >= LOGDEBUG)
                        {
                            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Security change, AuthenticationType forced to MQCSP_AUTH_NONE");
                            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                            LogData( LOGDEBUG0, pWS->szTmpBufff, pWS);
                        }
                    }
#endif
                }
            }
            /* 2.80 end */
        }
        break;
        /* we handle both INIT_SEC and SEC_MSG the same way to keep connecting clients with security exits secure */
#endif

    case MQXR_INIT_SEC:    /* Initiate security exchange (SCTY exits only) */
    case MQXR_SEC_MSG:	   /* Init security message  */
        if(pChannelExitParams->ExitReason==MQXR_INIT_SEC)
        {
            if (pWS->nDebugFlag >= LOGDEBUG)
                LogData( LOGDEBUG, "======= INIT_SEC ======", pWS);
        }
        else
        {
            if (pWS->nDebugFlag >= LOGDEBUG)
                LogData( LOGDEBUG, "======= SEC_MSG ======", pWS);
        }
        /*
        * this stmt is placed here to help testing.
        */
        /*
        printf( "BlockExit QMgr=[%.*s] ChannelName=[%.*s] ConnName=[%.*s] Uid=[%.*s]\n",
        rlnts(MQ_Q_MGR_NAME_LENGTH, pChannelDefinition->QMgrName),
        pChannelDefinition->QMgrName,
        rlnts(MQ_CHANNEL_NAME_LENGTH, pChannelDefinition->ChannelName),
        pChannelDefinition->ChannelName,
        rlnts(MQ_CONN_NAME_LENGTH, pChannelDefinition->ConnectionName),
        pChannelDefinition->ConnectionName,
        rlnts(MQ_USER_ID_LENGTH, pChannelDefinition->RemoteUserIdentifier),
        pChannelDefinition->RemoteUserIdentifier);
        */

        /*
        * Record debug info
        */
        if(pWS->SilentMode==0)
        {
            /*			snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"ver=%s env=%s ExitId=%ld ExitReason=%ld ChannelType=%ld",
            				VERSION, ENV,
            				pChannelExitParams->ExitId,
            				pChannelExitParams->ExitReason,
            				pChannelDefinition->ChannelType);
            			pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            			LogData( LOGDEBUG, pWS->szTmpBufff);*/

            if (pWS->nDebugFlag >= LOGDEBUG)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"ver=%s env=%s ExitId=%s ExitReason=%s FAP=%d ChannelType=%s",
                         VERSION, ENV,
                         MQExitId(pChannelExitParams),
                         MQExitReason(pChannelExitParams),
                         pChannelExitParams->FAPLevel,
                         MQChannelType(pChannelDefinition));
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData( LOGDEBUG, pWS->szTmpBufff, pWS);
            }
            if (pWS->nDebugFlag >= LOGDEBUG0)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"BlockExit QMgr=[%.*s] ChannelName=[%.*s] ConnName=[%.*s] Uid=[%.*s] pDataLength=%ld",
                         rlnts(MQ_Q_MGR_NAME_LENGTH, pChannelDefinition->QMgrName),
                         pChannelDefinition->QMgrName,
                         rlnts(MQ_CHANNEL_NAME_LENGTH, pChannelDefinition->ChannelName),
                         pChannelDefinition->ChannelName,
                         rlnts(MQ_CONN_NAME_LENGTH, pChannelDefinition->ConnectionName),
                         pChannelDefinition->ConnectionName,
                         rlnts(MQ_USER_ID_LENGTH, pChannelDefinition->RemoteUserIdentifier),
                         pChannelDefinition->RemoteUserIdentifier,
                         (long)*pDataLength);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData( LOGDEBUG0, pWS->szTmpBufff, pWS);

                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"BlockExit SCYDATA=[%.*s]",
                         rlnts(MQ_EXIT_DATA_LENGTH, pChannelDefinition->SecurityUserData),
                         pChannelDefinition->SecurityUserData);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData( LOGDEBUG0, pWS->szTmpBufff, pWS);
            }

#ifdef MQCD_VERSION_7										  /* for compilers */
            if(pChannelDefinition->Version == MQCD_VERSION_7)
            {  /* version 5.3 */
                if(pChannelDefinition->SSLPeerNameLength > 0 )
                {  /* eval in two.. */
                    if (pWS->nDebugFlag >= LOGDEBUG)
                    {
                        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"QMgr=[%.*s] Channel=[%.*s] SSLPeerName=[%.*s] SSLCipherSpec=[%.*s] SSLClientAuth=[%ld]",
                                 rlnts(MQ_Q_MGR_NAME_LENGTH, pChannelDefinition->QMgrName),
                                 pChannelDefinition->QMgrName,
                                 rlnts(MQ_CHANNEL_NAME_LENGTH, pChannelDefinition->ChannelName),
                                 pChannelDefinition->ChannelName,
                                 (int)pChannelDefinition->SSLPeerNameLength,
                                 (char*)pChannelDefinition->SSLPeerNamePtr,
                                 rlnts(MQ_SSL_CIPHER_SPEC_LENGTH, pChannelDefinition->SSLCipherSpec),
                                 pChannelDefinition->SSLCipherSpec,
                                 (long)pChannelDefinition->SSLClientAuth);
                        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                        LogData( LOGDEBUG, pWS->szTmpBufff, pWS);
                    }
                }
            }
#endif
        }
        /*
        * Deny all connections automatically
        */
        pChannelExitParams->ExitResponse = MQXCC_SUPPRESS_FUNCTION;
        pChannelExitParams->ExitResponse2 = MQXR2_USE_AGENT_BUFFER;

        if (pWS->bUseridUpperLowerCase==1)		   /* jhp 2.12 */
        {
            if (pWS->nDebugFlag >= LOGDEBUG1)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"RemUID Before strupr [%s]",pWS->RemUID);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
            }
            i_strupr(pWS->RemUID);
            if (pWS->nDebugFlag >= LOGDEBUG1)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"RemUID After strupr [%s]",pWS->RemUID);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
            }
        }


        /*
        * If FN specified - parse file or parse exit data
        */
        p = strstr(pWS->Pattern,"FN=");	/* Filename specified .. */
        if( p != NULL )
        {
            p=p+3;
            if(ProcessFile(p, pWS)==SEC_FAILURE)
            {
                if (pWS->nDebugFlag >= LOGDEBUG)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Errors in parameter file - Connection refused");
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData( LOGDEBUG, pWS->szTmpBufff, pWS);
                }
            }
        }
        /*
        * this part is removed to allow compliance with the previous version of the exit
        */
        /*						  else
        if (ProcessData(Pattern)==SEC_FAILURE)
        {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Errors in parameters - Connection refused");
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData( LOGDEBUG, pWS->szTmpBufff);
        						};*/
        /*								  ProcessData(Pattern);*/
        /*
        * file parsed OK (or was not specified) - cleanup "Pattern" data
        */
        for (i=0; i<MAX_PL; i++)
        {
            if(pWS->Pattern[i] == ',')
            {
                pWS->Pattern[i] = ';';
            }
            if(pWS->Pattern[i] == ' ')
            {
                pWS->Pattern[i] = 0x00;
                break;
            }
        }
        if(pWS->Pattern[0] != 0)
        {
            if(CheckConnectionPattern(pChannelDefinition, pWS)==SEC_FAILURE)
            {
                if (pWS->nDebugFlag >= LOGDEBUG)
                    LogData(LOGDEBUG,"Connection failure!", pWS);
#if MQAT_DEFAULT==MQAT_OS400
                return 0;
#else
                return;
#endif
            }
        }
        /*
        * Copy and report on user list extracted from file
        */
        strcpy(pWS->U2,pWS->Users);
        if (pWS->nDebugFlag >= LOGDEBUG)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Users: [%s] len [%ld] %ld", pWS->U2,(long)strlen(pWS->U2),(long)strlen(pWS->Users));
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData( LOGDEBUG, pWS->szTmpBufff, pWS);
        }
        /*
        * If a Blocked user list was given, we must check it now
        */
        if(strlen(pWS->BlockUsers) > 0)
        {
            if(CheckBlockedUserId(pChannelDefinition, pWS)==SEC_FAILURE)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, Channel [%.*s] ConName [%.*s] User [%.*s] is in negative list [%s]",
                         rlnts(MQ_CHANNEL_NAME_LENGTH, pChannelDefinition->ChannelName),
                         pChannelDefinition->ChannelName,
                         rlnts(MQ_CONN_NAME_LENGTH, pWS->ConName),	/* NWC */
                         pWS->ConName,
                         rlnts(MQ_USER_ID_LENGTH, pChannelDefinition->RemoteUserIdentifier),
                         pChannelDefinition->RemoteUserIdentifier,
                         pWS->BlockUsers);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData( LOGDATA, pWS->szTmpBufff, pWS );
#if defined(__MVS__)
                WTODataE("03E", pWS);
#endif
#if MQAT_DEFAULT==MQAT_OS400
                return 0;
#else
                return;
#endif
            }
        }
        /*
        * If a user list was given, we must check it now
        */
        if(strlen(pWS->Users) > 0)
        {
            if(CheckUserId(pChannelDefinition, pWS)==SEC_FAILURE)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, Channel [%.*s] ConName [%.*s] User [%.*s] not in positive list [%s]",
                         rlnts(MQ_CHANNEL_NAME_LENGTH, pChannelDefinition->ChannelName),
                         pChannelDefinition->ChannelName,
                         rlnts(MQ_CONN_NAME_LENGTH, pWS->ConName),	/* NWC */
                         pWS->ConName,
                         rlnts(MQ_USER_ID_LENGTH, pChannelDefinition->RemoteUserIdentifier),
                         pChannelDefinition->RemoteUserIdentifier,
                         pWS->Users);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData( LOGDATA, pWS->szTmpBufff, pWS );
#if defined(__MVS__)
                WTODataE("04E", pWS);
#endif
#if MQAT_DEFAULT==MQAT_OS400
                return 0;
#else
                return;
#endif
            }
        }

        /*  added under 3.00
        * Check the PartnerName= command, if not specified we return safely
        */

        if(CheckPartnerName(pChannelDefinition, pChannelExitParams, pWS)==SEC_FAILURE)
        {
            if (pChannelDefinition->ChannelType==MQCHT_SVRCONN)
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, Channel [%.*s] ConName [%.*s] User [%.*s] PartnerName [%.*s] requested by CheckPartnerName()",
                     rlnts(MQ_CHANNEL_NAME_LENGTH, pChannelDefinition->ChannelName),
                     pChannelDefinition->ChannelName,
                     rlnts(MQ_CONN_NAME_LENGTH, pWS->ConName),
                     pWS->ConName,
                     rlnts(MQ_USER_ID_LENGTH, pChannelDefinition->RemoteUserIdentifier),
                     pChannelDefinition->RemoteUserIdentifier,
                     rlnts(MQ_USER_ID_LENGTH, pChannelExitParams->PartnerName),
                     pChannelExitParams->PartnerName);
            else
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, Channel [%.*s] ConName [%.*s] User [%.*s] PartnerName [%.*s] requested by CheckPartnerName()",
                     rlnts(MQ_CHANNEL_NAME_LENGTH, pChannelDefinition->ChannelName),
                     pChannelDefinition->ChannelName,
                     rlnts(MQ_CONN_NAME_LENGTH, pWS->ConName),
                     pWS->ConName,
                     rlnts(MQ_USER_ID_LENGTH, pChannelDefinition->RemoteUserIdentifier),
                     pChannelDefinition->RemoteUserIdentifier,
                     rlnts(MQ_Q_MGR_NAME_LENGTH, pChannelExitParams->PartnerName),
                     pChannelExitParams->PartnerName);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData( LOGDATA, pWS->szTmpBufff, pWS );
#if defined(__MVS__)
            WTODataE("35E", pWS);
#endif
#if MQAT_DEFAULT==MQAT_OS400
            return 0;
#else
            return;
#endif
        }
        /*
        * Check the FAP= command, if not specified we return safely
        * added in 2.98
        */
        if(CheckFAP(pChannelExitParams,pWS) == SEC_FAILURE)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, FAP check failed for Chl=[%s] Req. level=%ld, pres. lvl=%ld",
                     pWS->szChannelName,
                     pWS->lFAPLevel,
                     pChannelExitParams->FAPLevel);

            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            /* reported from CheckFAP()
            LogData( LOGDATA, pWS->szTmpBufff,pWS );
            */
#if defined(__MVS__)
            WTODataE("47E",pWS);
#endif
#if MQAT_DEFAULT==MQAT_OS400
            return 0;
#else
            return;
#endif
        }
        /*
        * Check the CON= command, if not specified we return safely
        */
        if(CheckCONList(pChannelDefinition,pChannelExitParams->ExitReason,pWS) == SEC_FAILURE)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, Channel [%.*s] ConName [%.*s] User [%.*s] was not accepted in CON=",
                     rlnts(MQ_CHANNEL_NAME_LENGTH, pChannelDefinition->ChannelName),
                     pChannelDefinition->ChannelName,
                     rlnts(MQ_CONN_NAME_LENGTH, pWS->ConName),
                     pWS->ConName,
                     rlnts(MQ_USER_ID_LENGTH, pChannelDefinition->RemoteUserIdentifier),
                     pChannelDefinition->RemoteUserIdentifier);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData( LOGDATA, pWS->szTmpBufff,pWS );
#if defined(__MVS__)
            WTODataE("05E",pWS);
#endif
#if MQAT_DEFAULT==MQAT_OS400
            return 0;
#else
            return;
#endif
        }

#ifdef MQCD_VERSION_7
        /*
        * Check the SSL= command, if not specified we return safely
        */

        if(CheckSSLList(pChannelDefinition, pChannelExitParams, pWS)==SEC_FAILURE)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, Channel [%.*s] ConName [%.*s] User [%.*s] SSLPeer [%s] requested by CheckSSLList()",
                     rlnts(MQ_CHANNEL_NAME_LENGTH, pChannelDefinition->ChannelName),
                     pChannelDefinition->ChannelName,
                     rlnts(MQ_CONN_NAME_LENGTH, pWS->ConName),	/* NWC */
                     pWS->ConName,
                     rlnts(MQ_USER_ID_LENGTH, pChannelDefinition->RemoteUserIdentifier),
                     pChannelDefinition->RemoteUserIdentifier,
                     pWS->SSLPeer);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData( LOGDATA, pWS->szTmpBufff, pWS );
#if defined(__MVS__)
            WTODataE("06E", pWS);
#endif
#if MQAT_DEFAULT==MQAT_OS400
            return 0;
#else
            return;
#endif
        }
#endif


        /*
        * Now check for excluded user account names
        */
        if(pWS->bAllowBlankUserID==0)
        {
            if(CheckBlankUser(pChannelDefinition, pWS)==SEC_FAILURE)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, Channel [%.*s] ConName [%.*s] User [%.*s]",
                         rlnts(MQ_CHANNEL_NAME_LENGTH, pChannelDefinition->ChannelName),
                         pChannelDefinition->ChannelName,
                         rlnts(MQ_CONN_NAME_LENGTH, pWS->ConName),	/* NWC */
                         pWS->ConName,
                         rlnts(MQ_USER_ID_LENGTH, pChannelDefinition->RemoteUserIdentifier),
                         pChannelDefinition->RemoteUserIdentifier);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData( LOGDATA, pWS->szTmpBufff, pWS );
#if MQAT_DEFAULT==MQAT_OS400
                return 0;
#else
                return;
#endif
            }
        }
        if (pWS->bBlockMqmUsers==1)
        {
            if(CheckInvalidUsers(pChannelDefinition, pWS)==SEC_FAILURE)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, Channel [%.*s] ConName [%.*s] User [%.*s]",
                         rlnts(MQ_CHANNEL_NAME_LENGTH, pChannelDefinition->ChannelName),
                         pChannelDefinition->ChannelName,
                         rlnts(MQ_CONN_NAME_LENGTH, pWS->ConName),	/* NWC */
                         pWS->ConName,
                         rlnts(MQ_USER_ID_LENGTH, pChannelDefinition->RemoteUserIdentifier),
                         pChannelDefinition->RemoteUserIdentifier);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData( LOGDATA, pWS->szTmpBufff, pWS );
#if MQAT_DEFAULT==MQAT_OS400
                return 0;
#else
                return;
#endif
            }
        }
        strcpy(pWS->Flags,"");
        if(pWS->bBlockMqmUsers==1) strcpy(pWS->Flags,"BlockMqmUsers=Y ");
        if(pWS->bAllowBlankUserID==1) strcat(pWS->Flags,"AllowBlankUserID=Y ");
        if(pWS->bUseridUpperLowerCase==1) strcat(pWS->Flags,"UseridUpperLowerCase=* ");
#ifdef MQCXP_VERSION_6
        if(pWS->AllowASC==1) strcat(pWS->Flags,"ASC=Y ");
#endif
#if defined (UNIX) || defined (WIN32) || defined (OS400)
        if(pWS->bIncludeQmgrNameInLog==1) strcat(pWS->Flags,"IQNiL=1 ");
        if(pWS->bIncludeQmgrNameInLog==0) strcat(pWS->Flags,"IQNiL=0 ");
#endif
        if(pWS->SilentMode==0)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection accepted, Channel [%.*s] ConName [%.*s] Flags [%.*s] User [%s]",
                     rlnts(MQ_CHANNEL_NAME_LENGTH, pChannelDefinition->ChannelName),
                     pChannelDefinition->ChannelName,
                     rlnts(MQ_CONN_NAME_LENGTH, pWS->ConName),	/* NWC */
                     pWS->ConName,
                     rlnts((int)strlen(pWS->Flags),pWS->Flags),
                     pWS->Flags,
                     pWS->RemUID);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA, pWS->szTmpBufff, pWS);
#if defined(__MVS__)
            WTODataOk("50I", pWS);
#endif
        }
        /*
        * If we get here then all required parameters were correctly met - so set exit to OK
        */
        /*
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"debug: Connection accepted QMgr=[%.*s] Channel=[%.*s] ConName=[%.*s] Uid=[%.*s]",
        	rlnts(MQ_Q_MGR_NAME_LENGTH, pChannelDefinition->QMgrName),
        	pChannelDefinition->QMgrName,
        	rlnts(MQ_CHANNEL_NAME_LENGTH, pChannelDefinition->ChannelName),
        	pChannelDefinition->ChannelName,
        	rlnts(MQ_CONN_NAME_LENGTH, pChannelDefinition->ConnectionName),
        	pChannelDefinition->ConnectionName,
        	rlnts(MQ_USER_ID_LENGTH, pChannelDefinition->RemoteUserIdentifier),
        	pChannelDefinition->RemoteUserIdentifier);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG, pWS->szTmpBufff);
        */

        pChannelExitParams->ExitResponse  = MQXCC_OK;
        if (pWS->nDebugFlag >= LOGDEBUG1)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Return status Exitresponse=%s (%ld)",
                     MQExitResponseType(pChannelExitParams),
                     (long)pChannelExitParams->ExitResponse);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData( LOGDEBUG1, pWS->szTmpBufff, pWS);
        }
#if MQAT_DEFAULT==MQAT_OS400
        return 0;
#else
        return;
#endif
        break;
        /*
        * Unsupported exit reason - abort security exchange
        */
    default:
        pChannelExitParams->ExitResponse = MQXCC_SUPPRESS_FUNCTION;
        LogData(LOGDATA,"error - unsupported exit reason", pWS);
        if (pWS->nDebugFlag >= LOGDEBUG)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"ver=%s env=%s ExitId=%ld ExitReason=%ld ChannelType=%ld",
                     VERSION, ENV,
                     (long)pChannelExitParams->ExitId,
                     (long)pChannelExitParams->ExitReason,
                     (long)pChannelDefinition->ChannelType);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData( LOGDEBUG, pWS->szTmpBufff, pWS);
        }
#if MQAT_DEFAULT==MQAT_OS400
        return 0;
#else
        return;
#endif
        break;
    }  /* switch - ExitReason */
    if (pWS->nDebugFlag >= LOGDEBUG1)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Return status Exitresponse=%s (%ld)",
                 MQExitResponseType(pChannelExitParams),
                 (long)pChannelExitParams->ExitResponse);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData( LOGDEBUG1, pWS->szTmpBufff, pWS);
    }
#if MQAT_DEFAULT==MQAT_OS400
    return 0;
#else
    return;
#endif
} /* BlockExit */




/**
* CheckUserId()
*
*
* Check List of Users to see if RemUID is among them
*
* @pre	List of users already filled
* @return int SEC_OK if user found else SEC_FAILURE
*/
int CheckUserId(PMQCD pChannelDefinition, workingStorage_type *pWS )
{
    char *p;
    char *u;
    char *strtok_cb;
    char szUserList[33];
    int i,j;
    /*
    * Scan through list of user ID's
    */

    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1,"CheckUserId()", pWS);

    p = pWS->U2;
    u = (char*)strtok_r(pWS->U2 , "," , &strtok_cb);
    i=(int)strlen(pWS->U2);
    if (pWS->nDebugFlag >= LOGDEBUG1)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Users=[%s] first u=[%s] totlen %i", pWS->U2,u,i);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
    }
    memset(szUserList,0,sizeof(szUserList));
    do
    {
        j=0;
        for(i=0; i<(int)strlen(u); i++)
        {
            if(u[i] != ' ')
            {
                szUserList[j]=u[i];
                j=j+1;
            }
        }
        szUserList[j] = (char)NULL;	  /* Added (char) as without it, it didn't work on Solaris MD 20-JAN-2004 */

        if (pWS->nDebugFlag >= LOGDEBUG1)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Users=[%s] extract=[%s]  u=[%s]", pWS->U2, szUserList,u);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
        }
        if(pWS->bAllowBlankUserID==1) {
            if (!strcmp(szUserList,"BLANK_USERID"))
            {
                if (!strcmp(pWS->RemUID,"")) {
                    if (pWS->nDebugFlag >= LOGDEBUG0)
                    {
                        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"User accepted [%s]",szUserList);
                        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                        LogData(LOGDEBUG0,pWS->szTmpBufff, pWS);
                    }
                    return SEC_OK;
                }
            }
            else
            {
                if(wildcmp((char *)&szUserList,(char *)&pWS->RemUID, pWS))
                {
                    if (pWS->nDebugFlag >= LOGDEBUG0)
                    {
                        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"User accepted [%s]",szUserList);
                        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                        LogData(LOGDEBUG0,pWS->szTmpBufff, pWS);
                    }
                    return SEC_OK;
                }
            }
        }
        else
        {
            if(wildcmp((char *)&szUserList,(char *)&pWS->RemUID, pWS))
            {
                if (pWS->nDebugFlag >= LOGDEBUG0)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"User accepted [%s]",szUserList);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDEBUG0,pWS->szTmpBufff, pWS);
                }
                return SEC_OK;
            }
        }
    } while ((u = (char*)strtok_r(NULL, ",", &strtok_cb)) != NULL);

    if (pWS->nDebugFlag >= LOGDEBUG0)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, User [%.*s] not in list [%s]",
                 rlnts(MQ_USER_ID_LENGTH, pChannelDefinition->RemoteUserIdentifier),
                 pChannelDefinition->RemoteUserIdentifier,
                 pWS->U2);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG0,pWS->szTmpBufff, pWS);
        if (rlnts(MQ_USER_ID_LENGTH, pChannelDefinition->RemoteUserIdentifier)==0)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"An empty userid was presented. This can be filtered with special keyword \'BLANK_USERID\' see manual for more info.");
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDEBUG0,pWS->szTmpBufff, pWS);
        }
    }
    return SEC_FAILURE;
}

/**
* CheckBlockedUserId()
*
*
* Check List of BlockedUsers to see if RemUID is among them
*
* @pre	List of users already filled
* @return int SEC_OK if user found else SEC_FAILURE
*/
int CheckBlockedUserId(PMQCD pChannelDefinition, workingStorage_type *pWS )
{
    char *p;
    char *u;
    char *strtok_cb;
    char szUserList[33];
    int i,j;
    /*
    * Scan through list of user ID's
    */

    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1,"CheckBlockedUserId()", pWS);

    p = pWS->BlockUsers;
    u = (char*)strtok_r(pWS->BlockUsers,",", &strtok_cb);

    if (pWS->nDebugFlag >= LOGDEBUG0)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"BlockedUsers=[%s] first u=[%s]", pWS->BlockUsers,u);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG0, pWS->szTmpBufff, pWS);
    }

    memset(szUserList,0,sizeof(szUserList));
    do
    {
        j=0;
        for(i=0; i<(int)strlen(u); i++)
        {
            if(u[i] != ' ')
            {
                szUserList[j]=u[i];
                j=j+1;
            }
        }
        szUserList[j] = (char)NULL;	  /* Added (char) as without it, it didn't work on Solaris MD 20-JAN-2004 */

        if (pWS->nDebugFlag >= LOGDEBUG0)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Users=[%s] extract=[%s]  u=[%s]", pWS->BlockUsers, szUserList,u);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDEBUG0, pWS->szTmpBufff, pWS);
        }

        if(wildcmp((char *)&szUserList,(char *)&pWS->RemUID, pWS))
        {
            if (pWS->nDebugFlag >= LOGDEBUG0)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"User Blocked [%s]",szUserList);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDEBUG0,pWS->szTmpBufff, pWS);
            }
            return SEC_FAILURE;
        }
    } while ((u = (char*)strtok_r(NULL, ",", &strtok_cb)) != NULL);

    if (pWS->nDebugFlag >= LOGDEBUG1)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"User not found in BlockUsers [%s]",pWS->BlockUsers);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1,pWS->szTmpBufff, pWS);
    }

    return SEC_OK;
}



/**
* CheckConnectionPattern()
*
*
* Check the list of Patterns against the Connname
*
* @pre	All , changed to ; in list so we tokenise correctly
* @return int - SEC_OK or SEC_FAILURE
*/
int CheckConnectionPattern(PMQCD pChannelDefinition, workingStorage_type *pWS )
{
    int j;
    int i;
    char ipstring[33];
#ifdef HNLUP
#ifdef HNLUP2
    DNSReply dnsReply;
    int reply;
#else
    char szIpWork[64];
#endif
    MQLONG rc;
#endif

    strncpy(pWS->ConName,pChannelDefinition->ConnectionName,MQ_CONN_NAME_LENGTH);
#ifdef TEST281
    strncpy(pWS->ConName,"127.0.0.1(12345)",MQ_CONN_NAME_LENGTH);   /* JHP 2.81 test */
#endif
    for (i=0; i<MQ_CONN_NAME_LENGTH; i++)
    {
        /* if(pWS->ConName[i] == ' ') JHP 2.81 changed */
        if((pWS->ConName[i] == ' ') || (pWS->ConName[i] == '('))
        {
            pWS->ConName[i] = 0x00;
            break;
        }
    }

    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1,"CheckConnectionPattern() Entered", pWS);

    if(!strcmp(pWS->Pattern,"NONE;"))
    {
        sprintf(pWS->szTmpBufff,"Connection refused. Pattern NONE; was supplied, please check BlockIP2.ini");
        LogData(LOGDATA, pWS->szTmpBufff, pWS);
#if defined(__MVS__)
        WTODataE("70E", pWS);
#endif
        return SEC_FAILURE;
    }



    j=0;

/*    for (i=0;i <= MAX_PL; i++) 2.92 changed to */
    for (i=0;i <= strlen(pWS->Pattern); i++)
    {
        if((pWS->Pattern[i] != ';') && (pWS->Pattern[i] != 0x00))
        { 		  /* jhp 04-01-21 single pattern incl. */
            /* NWC separated by ';' because ',' altered to ';' earlier */
            ipstring[j] = pWS->Pattern[i];
            if (j<sizeof(ipstring)-1)                  /* prevent storage overlay 2.92 */
                j=j+1;
        }
        else
        {
            /* if (j == 0) break;  2.92 changed to */
            if (j == 0) continue;
            ipstring[j] = (char)NULL; 	/* Added (char) as without it, it didn't work on Solaris MD 20-JAN-2004 */

            DBGPRINTF((OST, "CheckConnectionPattern() Pattern [%s]  ip[%s] j %i \n", pWS->Pattern, ipstring ,j ));

            if (strlen(pWS->Pattern) < 100) {   /* 2.60 fix */
                if (pWS->nDebugFlag >= LOGDEBUG0)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  Pattern [%s]  ip[%s] j %i ", pWS->Pattern, ipstring,j);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDEBUG0, pWS->szTmpBufff, pWS);
                }
            }
            j=0;
            /* 2.53 start */
#ifdef HNLUP
            /* Add gethostbyname call here */
            if (isalpha(ipstring[0])) {   /* host address is a name, do address conversion */
#ifdef HNLUP2
                rc= GetHostIP(ipstring, (DNSReply*)&dnsReply, pWS);
                if (pWS->nDebugFlag >= LOGDEBUG0)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  GetHostIP() on hostname [%s] replies %ld returned [%s] rc (0=OK) %ld", ipstring, dnsReply.replies, dnsReply.szIpAddr[0],(long)rc);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDEBUG0, pWS->szTmpBufff, pWS);
                }
                if (rc==0)
                {
                    for (reply=0; reply< dnsReply.replies; reply++)
                    {
                        if (wildcmp((char *)&dnsReply.szIpAddr[reply],(char *)&pWS->ConName,pWS))
                        {
                            if (pWS->nDebugFlag >= LOGDEBUG0)
                            {
                                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  Pattern [%s], ConName [%s] passed test..",
                                         dnsReply.szIpAddr[reply], pWS->ConName);
                                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                                LogData(LOGDEBUG0,pWS->szTmpBufff, pWS);
                            }
                            if (pWS->nDebugFlag >= LOGDEBUG1)
                                LogData(LOGDEBUG1,"CheckConnectionPattern() Done", pWS);
                            return SEC_OK;
                        }
                    }
                }
                else
                    strcpy(ipstring,"Not-found");
#else
                rc= GetHostIP(ipstring, (char*)&szIpWork, pWS);
                if (pWS->nDebugFlag >= LOGDEBUG0)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  GetHostIP() on hostname [%s] returned [%s] rc (0=OK) %ld", ipstring, szIpWork,(long)rc);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDEBUG0, pWS->szTmpBufff, pWS);
                }
                if (rc==0)
                    strcpy(ipstring,szIpWork);
                else
                    strcpy(ipstring,"Not-found");
#endif
            }
#endif
            /* 2.53 end */

            if (pWS->nDebugFlag >= LOGDEBUG1)
            {
                 snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  Pattern [%s], ConName [%s] to be checked next",
                          ipstring, pWS->ConName);
                 pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                 LogData(LOGDEBUG8,pWS->szTmpBufff, pWS);
            }

            if (wildcmp((char *)&ipstring,(char *)&pWS->ConName,pWS))
            {
                if (pWS->nDebugFlag >= LOGDEBUG0)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  Pattern [%s], ConName [%s] passed test..",
                             ipstring, pWS->ConName);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDEBUG0,pWS->szTmpBufff, pWS);
                }
                if (pWS->nDebugFlag >= LOGDEBUG1)
                    LogData(LOGDEBUG1,"CheckConnectionPattern() Done", pWS);
                return SEC_OK;
            }
            if (pWS->Pattern[i] == 0x00) /* NWC terminate at end of string JHP040811 */
            {
                i = MAX_PL;
            }
        }
    }
    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  Connection refused for pattern [%s] ChannelName=[%s] user=[%.*s] ConnName=[%s]",
             pWS->Pattern,
             pWS->szChannelName,
             rlnts(MQ_USER_ID_LENGTH, pChannelDefinition->RemoteUserIdentifier),
             pChannelDefinition->RemoteUserIdentifier,
             pWS->ConName);
    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
    LogData(LOGDATA,pWS->szTmpBufff, pWS);
#if defined(__MVS__)
    WTODataE("07E", pWS);
#endif
    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1,"CheckConnectionPattern() Done", pWS);
    return SEC_FAILURE;
}

/**
*
*/
int  CheckFAP(PMQCXP pChannelExitParams, workingStorage_type *pWS)
{
	int rc=SEC_OK;		/* No errors yet */

	if (pWS->lFAPLevel==-1)
	{
		if (pWS->nDebugFlag >= LOGDEBUG1)
			LogData(LOGDEBUG1,"CheckFAP() Skipped", pWS);
		return 0;		/* don't spend moretime here, not activated */
	}

	if (pWS->nDebugFlag >= LOGDEBUG1)
		LogData(LOGDEBUG1,"CheckFAP() Entered", pWS);

	if (pChannelExitParams->FAPLevel < pWS->lFAPLevel)  /* Not at acceptable level */
	{
		snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"FAP Level check failed for ChannelName=[%s] Req. level=%ld, presented FAP level=%ld",
				 pWS->szChannelName,
				 pWS->lFAPLevel,
				 pChannelExitParams->FAPLevel);
		pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
		LogData(LOGDATA,pWS->szTmpBufff, pWS);
		rc=SEC_FAILURE;
	}

	if (pWS->nDebugFlag >= LOGDEBUG1)
		LogData(LOGDEBUG1,"CheckFAP() Done", pWS);
	return rc;
}


/**
* wildcmplist()
*
*
* Usage:
* This is a lightweight and simple pattern matching function.
*
* if(wildcmplist("sans*,bl?h.*", "blah.jpg")) {
* 	 we have a match!
* } else {
* 	 no match =(
*}
*
* @return  int  0=nRetCode to match  1=match
*/
/*int wildcmplist(char *wild, char *string, workingStorage_type *pWS)*/
int  wildcmplist(char *wild, char *string, int flg, workingStorage_type *pWS)
{
    char work[255];
    int  i=0;

    i=0;
    memset(work,'\00',sizeof(work));
    do
    {
        work[i++] = *wild;
        if((*wild == ',') | (*wild == '\00') | (*wild == ';'))
        {
            work[--i]='\00';
            if (flg==1)   /* 2.82 start special handling for "BLANK_USER" */
            {
                if ((!strncmp(work,"BLANK_USERID",12)) && (strlen(string)==0))
                {
                    if (pWS->nDebugFlag >= LOGDEBUG0)
                    {
                        LogData(LOGDEBUG0,"  BLANK_USER match", pWS);
                    }
                    return 1;
                }
            }            /* 2.82 end */
            if (wildcmp((char *)work,string, pWS))
            {
                return 1;
            }
            memset(work,'\00',sizeof(work));
            i=0;
        }
        wild = wild + 1;
    }  while (*wild != '\00')  ;
    if (i >0)
        if (wildcmp((char *)work,string,pWS))
        {
            return 1;
        }
    return 0;
}

/**
* wildcmp()
*
*
* Usage:
* This is a lightweight and simple pattern matching function.
*
* if(wildcmp("bl?h.*", "blah.jpg")) {
* 	 we have a match!
* } else {
* 	 no match =(
*}
*
* @return  int  0=nRetCode to match  1=match
*/
int wildcmp (const char *wild, const char *string, workingStorage_type *pWS)
{
    int  ichfrom,ichto;
    while (*wild)
    {
        if (*wild == '*')
        {
            wild++;
            while (*string)
                if (wildcmp(wild, string,pWS))
                    return 1;
                else
                    string++;
        }
        else if (*wild == *string)
        {
            wild++;
            string++;
        }
        else
        {
            if (*string)
            {
                switch (*wild)
                {

                case '?':
                    wild++;
                    string++;
                    break;
                case '#': 	   /* Single numeric position */
                    if (isdigit(*string) == 0)
                        return 0;
                    wild++;
                    string++;
                    break;
                case '$': 	   /* Single alphabetic position */
                    if (isalpha(*string) == 0)
                        return 0;
                    wild++;
                    string++;
                    break;

                case '[': 	   /* specified range */
                    if ((*(wild+2) != '-') | (*(wild+4) != ']') | (isalnum(*(wild+1)) ==0) | (isalnum(*(wild+3)) ==0))
                        return 0;
                    else
                    {
                        ichfrom=*(wild+1);
                        ichto=*(wild+3);
                        if (pWS->nDebugFlag >= LOGDEBUG0)
                        {
                            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Testing..... from %i to %i string %i",ichfrom, ichto, *string);
                            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                            LogData(LOGDEBUG1,pWS->szTmpBufff, pWS);
                        }
#if defined (__MVS__off)
                        if (((isdigit(ichfrom) != 0) & (isdigit(ichto) != 0)) | (isalpha(ichfrom) != 0) & (isalpha(ichto) != 0))
                        {                                                                /* same type... */
                            if (((int)*string < ichfrom) | ((int)*string > ichto))
                                return 0;
                        }
                        else
                        {
                            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"wildcmp don't support mixed patterns yet. (0-J) J-0");
                            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                            LogData(LOGDATA,pWS->szTmpBufff);
#if defined(__MVS__)
                            WTODataE("99E");
#endif
                            return 0;
                        }
#else
                        if (((int)*string < ichfrom) | ((int)*string > ichto))
                            return 0;
#endif
                    }
                    wild = wild + 5;
                    string++;
                    break;
                case '(': 	   /* specified range */
                    if ((*(wild+2) != '-') | (*(wild+4) != ')') | (isalnum(*(wild+1)) ==0) | (isalnum(*(wild+3)) ==0))
                        return 0;
                    else
                    {
                        ichfrom=*(wild+1);
                        ichto=*(wild+3);
                        if (pWS->nDebugFlag >= LOGDEBUG0)
                        {
                            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Testing..... from %i to %i string %i",ichfrom, ichto, *string);
                            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                            LogData(LOGDEBUG1,pWS->szTmpBufff, pWS);
                        }
#if defined (__MVS__off)
                        if (((isdigit(ichfrom) != 0) & (isdigit(ichto) != 0)) | (isalpha(ichfrom) != 0) & (isalpha(ichto) != 0))
                        {                                                                /* same type... */
                            if (((int)*string < ichfrom) | ((int)*string > ichto))
                                return 0;
                        }
                        else
                        {
                            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"wildcmp don't support mixed patterns yet. (0-J) J-0");
                            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                            LogData(LOGDATA,pWS->szTmpBufff);
#if defined(__MVS__)
                            WTODataE("99E");
#endif
                            return 0;
                        }
#else
                        if (((int)*string < ichfrom) | ((int)*string > ichto))
                            return 0;
#endif
                    }
                    wild = wild + 5;
                    string++;
                    break;


                default:
                    return 0;
                }
            }
            else return 0;
        }
    }
    return (!*wild) && (!*string);
}

/*	Strip white space off a string
**	------------------------------
**	Return value points to first non-white character, or to 0 if none.
**	All trailing white space is OVERWRITTEN with zero.
*/
char * Strip (char * s)
{
    if (s)
    {
        char * p=s;
        for(p=s;*p;p++);		/* Find end of string */
        for(p--;p>=s;p--)
        {
            if (isspace((int) *p))
                *p=0;			/* Zap trailing blanks */
            else
                break;
        }
        while (isspace((int) *s)) s++;	/* Strip leading blanks */
    }
    return s;
}


int Index1(char *p, long buflen, const char *s){
    /* search a buffer for the supplied string,
    the buffer may include non-character data even nulls */
    auto int k;
    auto long i;
    k=(int)strlen(s);  if (k<=0) return(-1);  /* invalid search string */
    for (i=0;i<buflen;i++)
    {
        if (!memcmp(p+i,s,k))
            return(i+1);
    }
    return(0);
}

int GetParam(char *pb, char *pt,const char *t1,const char *t2) {
    /* this will extract the string found within pb delimtied
    by t1 and then t2.  pt is set to this string and
    excludes the the tokens.  The prefix token (t1) MUST be unique
    to then entire line. If not, the first is always retreived. */
    int i,j;
    j=Index1(pb,(int)strlen(pb),t1);
    if (j<1) return(1);
    i=Index1(pb+j-1,(int)strlen(pb)-j+1,t2);
    if (i<1) return(1);
    strncpy(pt,pb+j-1+strlen(t1),(i-(strlen(t2)+strlen(t1))));
    memset(pt+(i-(strlen(t2)+strlen(t1))),'\00',1);
    return(0);
}


#ifdef MQCD_VERSION_7

/**
* getDNMCAUser()
*
*
* Retrieves the specified argument from the parsed DN and returns this in resultstr
* If none specified then exit with SEC_OK
*
*
* @return int SEC_OK	SEC_FAILURE
*/

int getDNMCAUser(char * key, char * dn, char * resultstr, int len,workingStorage_type *pWS)
{

    int  rc=SEC_FAILURE;    /* mark error */
    char work[80];
    char kw[8];
    char key2[80];
    char *p;
    char *u;
    char *strtok_cb;
    int  i,j,substrfrom,substrlen;

    if (pWS->nDebugFlag >= LOGDEBUG)
        LogData(LOGDEBUG,"getDNMCAUser()", pWS);

    strcpy(resultstr,"");

    strcpy(key2,key);

    p = strstr(key2,"(");
    if (p==NULL)
    {
        LogData(LOGDATA,"Invalid MCA Specification was spefified. Format like: MCA=CN(15,8)",pWS);
        return SEC_FAILURE;
    }
    u=(char*)strtok_r(p,",", &strtok_cb);
    if (u==NULL)
    {
        LogData(LOGDATA,"Invalid MCA Specification was spefified. Format like: MCA=CN(15,8)",pWS);
        return SEC_FAILURE;
    }
    u=u+1;
    substrfrom=atoi(u)-1;
    if (substrfrom < 0 || substrfrom > 99)
    {
        LogData(LOGDATA,"Invalid MCA Specification was spefified. From shall be positive (1 thru 99)",pWS);
        return SEC_FAILURE;
    }
    u=(char*)strtok_r(NULL,")", &strtok_cb);
    if (u==NULL)
    {
        LogData(LOGDATA,"Invalid MCA Specification was spefified. Format like: MCA=CN(15,8)",pWS);
        return SEC_FAILURE;
    }
    substrlen=atoi(u);
    if (substrlen < 1 || substrlen > len)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Invalid MCA Specification was spefified. From shall be positive (1 thru %ld)",len-1);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDATA,pWS->szTmpBufff,pWS);
        return SEC_FAILURE;
    }

    /* Now find the keyword */
    strcpy(key2,key);
    p=(char*)strtok_r(key2,"(", &strtok_cb);
    if (p != NULL)
    {
        if (strlen(p)< (sizeof(kw)-1))
        {
            strcpy(kw,p);
            strcat(kw,"=");

            p = strstr(dn,kw);                      /*  allow C= patterns */
            u=(char*)strtok_r(p,",", &strtok_cb);   /* Just take the argument */
            u=u+strlen(kw);                         /* u now contains the value, now substring it */
            if (substrfrom > strlen(u))
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Specified MCA=%s pointed points behind the value",key);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDATA,pWS->szTmpBufff,pWS);
                return SEC_FAILURE;
            }
            if (substrlen > len)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Specified MCA=%s length is too long, max %ld",key,len);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDATA,pWS->szTmpBufff,pWS);
                return SEC_FAILURE;
            }
            if (substrlen > (sizeof(work)-1))
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Specified MCA=%s length is too long, max %ld",key,sizeof(work));
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDATA,pWS->szTmpBufff,pWS);
                return SEC_FAILURE;
            }
            strncpy(resultstr,u+substrfrom,substrlen);
            resultstr[substrlen]='\00';
            if (strstr(resultstr," "))
            {
                memset(work,'\00',sizeof(work));
                j=0;
                if (strlen(resultstr) > (sizeof(work)-1)) return 16;
                for (i=0;i<strlen(resultstr); i++)
                {
                    if (resultstr[i] != ' ')
                    {
                        work[j]=resultstr[i];
                        j++;
                    }
                }
                strcpy(resultstr,work);
            }

            if (pWS->nDebugFlag >= LOGDEBUG)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  Extracted MCA User [%s]",resultstr);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDEBUG,pWS->szTmpBufff,pWS);
            }
            rc=SEC_OK;
        }
    }
    return rc;
}


/**
* CheckPartnerName()
*
*
* Check the PartnerName= command and process accordingly
* If none specified then exit with SEC_OK
*
*
* @return int SEC_OK	SEC_FAILURE
*/
int  CheckPartnerName( MQCD *pChannelDefinition, PMQCXP pChannelExitParams, workingStorage_type *pWS )
{
    char PartnerName[MQ_Q_MGR_NAME_LENGTH+1];

    memset(PartnerName,'\00',sizeof(PartnerName));
    if (pChannelDefinition->ChannelType==MQCHT_SVRCONN)
        memcpy(PartnerName,pChannelExitParams->PartnerName,rlnts(MQ_USER_ID_LENGTH, pChannelExitParams->PartnerName));
    else
        memcpy(PartnerName,pChannelExitParams->PartnerName,rlnts(MQ_Q_MGR_NAME_LENGTH, pChannelExitParams->PartnerName));

    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1,"CheckPartnerName() Entered", pWS);

    if (strlen(pWS->PartnerName)>0)
    {
        if (pWS->nDebugFlag >= LOGDEBUG1)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  Remote PartnerName [%s] list [%s]",
                     PartnerName, pWS->PartnerName);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDEBUG1,pWS->szTmpBufff,pWS);
        }
        if (!wildcmplist((char *)&pWS->PartnerName,PartnerName,1,pWS))
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"PartnerName check failed [%s] list [%s]",
                     PartnerName, pWS->PartnerName);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA,pWS->szTmpBufff,pWS);
#if defined(__MVS__)
            WTODataE("BLOCKIP2-xxE", pWS);
#endif
            return SEC_FAILURE;
        }
    }

    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1,"CheckPartnerName() Done", pWS);
    return SEC_OK;
}



/**
* CheckSSLList()
*
*
* Check the SSL= commands and process accordingly
* If none specified then exit with SEC_OK
*
*
* @return int SEC_OK	SEC_FAILURE
*/
int  CheckSSLList( MQCD *pChannelDefinition, PMQCXP pChannelExitParams, workingStorage_type *pWS )

{
    char work[33];
    int i, iErr;
    char *p;
    char *u;
    char *strtok_cb;
    char work_OU[64]="";     /* Org unit */
    char work_C[64] ="";     /* country */
    char work_O[64] ="";     /* organisation */
    char work_L[64] ="";     /* location */
    char work_CN[64]="";     /* name */
    char work_ST[64]="";     /* state */
    char work_PC[64]="";     /* Postal code */
    /*    char work_filter[SSL_LEN+1];          */  /* 2.64 */
    /*    char work_strtok[SSL_LEN+1];          */
    char work_filter[SSL_LEN_PEER+1];           /* 2.70 */
    char work_strtok[SSL_LEN_PEER+1];           /* 2.70 */

    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1,"CheckSSLList()", pWS);

    if(pChannelDefinition->Version < MQCD_VERSION_7) return SEC_OK;  /* SSL not supported */

#ifdef MQCXP_VERSION_6
    if (pWS->AllowASC==0)                                            /* Dis-allow self signed certificates */
    {
        /*       if (pChannelExitParams->Version >= MQCXP_VERSION_6)*/
        if (pChannelExitParams->Version >= MQCXP_VERSION_5)          /* Should be 6 according to the book */
        {
            /* Moved the 'sprintf' statement into the 'if' block, it didn't work on Solaris with MQv6 and gcc, HK 10-OCT-2006 */
            if ((pChannelExitParams->SSLRemCertIssNameLength > 0) | (pChannelDefinition->SSLPeerNameLength>0))
            {
                if (pWS->nDebugFlag >= LOGDEBUG0)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Certificate issued by [%.*s] len(%ld) SSLPeer [%.*s] len(%ld)",
                             rlnts(pChannelExitParams->SSLRemCertIssNameLength, pChannelExitParams->SSLRemCertIssNamePtr),
                             (char*)pChannelExitParams->SSLRemCertIssNamePtr,
                             (long)pChannelExitParams->SSLRemCertIssNameLength,
                             rlnts(pChannelDefinition->SSLPeerNameLength, pChannelDefinition->SSLPeerNamePtr),
                             (char*)pChannelDefinition->SSLPeerNamePtr,
                             (long)pChannelDefinition->SSLPeerNameLength);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDEBUG0,pWS->szTmpBufff, pWS);
                }
            }

            if (pChannelDefinition->SSLPeerNameLength==pChannelExitParams->SSLRemCertIssNameLength &&
                    pChannelExitParams->SSLRemCertIssNameLength > 0)
            {
                if (!memcmp(pChannelDefinition->SSLPeerNamePtr,pChannelExitParams->SSLRemCertIssNamePtr,
                            rlnts(pChannelExitParams->SSLRemCertIssNameLength, pChannelExitParams->SSLRemCertIssNamePtr)))
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, Self signed Certificates is not allowed. Certificate was issued by [%.*s]",
                             rlnts(pChannelExitParams->SSLRemCertIssNameLength, pChannelExitParams->SSLRemCertIssNamePtr),
                             (char*)pChannelExitParams->SSLRemCertIssNamePtr);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDATA,pWS->szTmpBufff, pWS);
#if defined(__MVS__)
                    WTODataE("BLOCKIP2-45E", pWS);
#endif
                    if (pWS->nDebugFlag >= LOGDEBUG1)
                        LogData(LOGDEBUG1,"CheckSSLList() Done", pWS);
                    return SEC_FAILURE;
                }
            }
            else   /* Length is different on cert issuer and SSLPeer */
            {
            }
        }      /* MQCXP_VERSION_6 lower issuer not present */
        else
        {
            LogData(LOGDATA,"Connection refused, Self signed Certificates filtering is not supported in pre version 6.0" , pWS);
#if defined(__MVS__)
            WTODataE("BLOCKIP2-46E", pWS);
#endif
            if (pWS->nDebugFlag >= LOGDEBUG1)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, CSQXP.Version [%ld] v6[%d] current[%d]",
                         (long)pChannelExitParams->Version,
                         MQCXP_VERSION_6,
                         MQCXP_CURRENT_VERSION);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDEBUG1,pWS->szTmpBufff, pWS);
            }
            if (pWS->nDebugFlag >= LOGDEBUG1)
                LogData(LOGDEBUG1,"CheckSSLList() Done", pWS);
            return SEC_FAILURE;
        }
    }         /* Check for self signed certificates */
#endif


    if(pWS->SSLCount == 0)
    {
        if (pWS->nDebugFlag >= LOGDEBUG1)
            LogData(LOGDEBUG1,"CheckSSLList() Done", pWS);
        return SEC_OK;
    }

    if(pChannelDefinition->Version >= MQCD_VERSION_7)  /* changed in 2.50 to >= from == */
    {  /* version 5.3 */
        if(pChannelDefinition->SSLPeerNameLength > 0 && pWS->SSLCount > 0)
        {  /* eval in two.. */
            if(pChannelDefinition->SSLPeerNameLength >= SSL_LEN_PEER)     /* 2.70 */
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, SSLPeerName too long max [%d] was [%ld]",
                         SSL_LEN_PEER,(long)pChannelDefinition->SSLPeerNameLength);   /* 2.70 */
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDATA,pWS->szTmpBufff, pWS);
#if defined(__MVS__)
                WTODataE("08E", pWS);
#endif
                if (pWS->nDebugFlag >= LOGDEBUG1)
                    LogData(LOGDEBUG1,"CheckSSLList() Done", pWS);
                return SEC_FAILURE;
            } 					 /* spec too long */
            memcpy(pWS->SSLPeer,pChannelDefinition->SSLPeerNamePtr,pChannelDefinition->SSLPeerNameLength);
            pWS->SSLPeer[pChannelDefinition->SSLPeerNameLength]=0x00;
            strcat(pWS->SSLPeer,",");
            /* start of 2.50 change */
            i=GetParam(pWS->SSLPeer,work_CN,"CN=",",");
            i=GetParam(pWS->SSLPeer,work_C,"C=",",");
            i=GetParam(pWS->SSLPeer,work_OU,"OU=",",");
            i=GetParam(pWS->SSLPeer,work_L,"L=",",");
            i=GetParam(pWS->SSLPeer,work_O,"O=",",");
            i=GetParam(pWS->SSLPeer,work_ST,"ST=",",");
            i=GetParam(pWS->SSLPeer,work_PC,"PC=",",");
            if (pWS->nDebugFlag >= LOGDEBUG0)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"CheckSSLList() MQCD CN=[%s],O=[%s],C=[%s]",work_CN,work_O,work_C);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDEBUG0,pWS->szTmpBufff, pWS);
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"CheckSSLList() MQCD OU=[%s],L=[%s],ST=[%s],PC=[%s]",work_OU,work_L,
                         work_ST,work_PC);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDEBUG0,pWS->szTmpBufff, pWS);
            }
            /* end 2.50 chnage */


            p = strstr(pWS->SSLPeer,"CN=");		/*	allow CN patterns */
            if(p != NULL)
            {
                strcpy(pWS->SSLPeer,p);			/* we're just taking CN */
                pWS->SSLPeer[strlen(p)]=0x00;
                if (pWS->nDebugFlag >= LOGDEBUG0)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"SSLPeer [%s]",p);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDEBUG0,pWS->szTmpBufff, pWS);
                }
                for(i=0; i<pWS->SSLCount; i++)
                {   /* run thru SSL spec */
                    /* we need a bit of redesign here ... 2.64 start */
                    strcpy(work_strtok,pWS->SSLList[i]);
                    p=(char*)strtok_r(work_strtok,";", &strtok_cb);
                    if(p != NULL)
                    {
                        strcpy(work_filter,p);
                    }
                    else
                        strcpy(work_filter,"");
                    /* 2.64 end */
                    if (pWS->nDebugFlag >= LOGDEBUG0)
                    {
                        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"SSLSpec [%s] wf[%s]",pWS->SSLList[i],work_filter);
                        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                        LogData(LOGDEBUG0,pWS->szTmpBufff, pWS);
                    }
                    /* start 2.50 */
                    iErr=0;  /* No errors yet */
                    if (strlen(work_C) > 0)
                    {
                        /*                        strcpy(work_strtok,pWS->SSLList[i]);*/
                        strcpy(work_strtok,work_filter);
                        p = strstr(work_strtok,"C=");                     /*  allow C= patterns */
                        if(p != NULL)
                        {
                            u=(char*)strtok_r(p,",", &strtok_cb);	  /* Just take the C */
                            if(u != NULL )
                            {                                             /* It's  time to ask wildcard.. */
                                u=u+2;
                                if (pWS->nDebugFlag >= LOGDEBUG)
                                {
                                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"C= Cert=[%s] mask=[%s]",work_C, u);
                                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                                    LogData(LOGDEBUG,pWS->szTmpBufff, pWS);
                                }
                                if (wildcmp(u,(char *)&work_C,pWS))
                                {
                                    if(pWS->SilentMode ==0 )
                                    {
                                        if (pWS->nDebugFlag >= LOGDEBUG0)
                                        {
                                            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"SSL Pattern matched C=[%s]",u);
                                            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                                            LogData(LOGDEBUG0, pWS->szTmpBufff, pWS);
                                        }
                                    }
                                }
                                else {
                                    if (pWS->nDebugFlag >= LOGDEBUG)
                                        LogData(LOGDEBUG, "SSL entry skipped. Didn't match C=.", pWS);
                                    iErr=1;
                                }
                            }
                        }
                    }
                    if (strlen(work_O) > 0 && iErr==0) {
                        /*                        strcpy(work_strtok,pWS->SSLList[i]);*/
                        strcpy(work_strtok,work_filter);
                        p = strstr(work_strtok,"O=");                     /*  allow O= patterns */
                        if(p != NULL)
                        {
                            u=(char*)strtok_r(p,",", &strtok_cb);	  /* Just take the O */
                            if(u != NULL )
                            {                                             /* It's  time to ask wildcard.. */
                                u=u+2;
                                if (pWS->nDebugFlag >= LOGDEBUG)
                                {
                                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"O= Cert=[%s] mask=[%s]",work_O, u);
                                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                                    LogData(LOGDEBUG,pWS->szTmpBufff, pWS);
                                }
                                if (wildcmp(u,(char *)&work_O,pWS))
                                {
                                    if(pWS->SilentMode ==0 )
                                    {
                                        if (pWS->nDebugFlag >= LOGDEBUG0)
                                        {
                                            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"SSL Pattern matched O=[%s]",u);
                                            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                                            LogData(LOGDEBUG0, pWS->szTmpBufff, pWS);
                                        }
                                    }
                                }
                                else {
                                    if (pWS->nDebugFlag >= LOGDEBUG)
                                        LogData(LOGDEBUG, "SSL entry skipped. Didn't match O=.", pWS);
                                    iErr=1;
                                }
                            }
                        }
                    }
                    if (strlen(work_OU) > 0 && iErr==0)
                    {
                        /*                        strcpy(work_strtok,pWS->SSLList[i]);*/
                        strcpy(work_strtok,work_filter);
                        p = strstr(work_strtok,"OU=");                    /*  allow OU= patterns */
                        if(p != NULL)
                        {
                            u=(char*)strtok_r(p,",", &strtok_cb);	  /* Just take the OU */
                            if(u != NULL )
                            {                                             /* It's  time to ask wildcard.. */
                                u=u+3;
                                if (pWS->nDebugFlag >= LOGDEBUG0)
                                {
                                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"OU= Cert=[%s] mask=[%s]",work_OU, u);
                                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                                    LogData(LOGDEBUG0,pWS->szTmpBufff, pWS);
                                }
                                if (wildcmp(u,(char *)&work_OU,pWS))
                                {
                                    if(pWS->SilentMode ==0 ) {
                                        if (pWS->nDebugFlag >= LOGDEBUG0)
                                        {
                                            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"SSL Pattern matched OU=[%s]",u);
                                            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                                            LogData(LOGDEBUG0, pWS->szTmpBufff, pWS);
                                        }
                                    }
                                }
                                else {
                                    if (pWS->nDebugFlag >= LOGDEBUG)
                                        LogData(LOGDEBUG, "SSL entry skipped. Didn't match OU=.", pWS);
                                    iErr=1;
                                }
                            }
                        }
                    }
                    if (strlen(work_L) > 0 && iErr==0) {
                        /*                        strcpy(work_strtok,pWS->SSLList[i]);*/
                        strcpy(work_strtok,work_filter);
                        p = strstr(work_strtok,"L=");                     /*  allow L= patterns */
                        if(p != NULL)
                        {
                            u=(char*)strtok_r(p,",", &strtok_cb);	  /* Just take the L */
                            if(u != NULL )
                            {                                             /* It's  time to ask wildcard.. */
                                u=u+2;
                                if (pWS->nDebugFlag >= LOGDEBUG0)
                                {
                                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"L= Cert=[%s] mask=[%s]",work_L, u);
                                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                                    LogData(LOGDEBUG0,pWS->szTmpBufff, pWS);
                                }
                                if (wildcmp(u,(char *)&work_L,pWS))
                                {
                                    if(pWS->SilentMode ==0 )
                                    {
                                        if (pWS->nDebugFlag >= LOGDEBUG0)
                                        {
                                            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"SSL Pattern matched L=[%s]",u);
                                            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                                            LogData(LOGDEBUG0, pWS->szTmpBufff, pWS);
                                        }
                                    }
                                }
                                else {
                                    if (pWS->nDebugFlag >= LOGDEBUG)
                                        LogData(LOGDEBUG, "SSL entry skipped. Didn't match L=.", pWS);
                                    iErr=1;
                                }
                            }
                        }
                    }
                    if (strlen(work_CN) > 0 && iErr==0)
                    {
                        /*                        strcpy(work_strtok,pWS->SSLList[i]);*/
                        strcpy(work_strtok,work_filter);
                        p = strstr(work_strtok,"CN=");          /*  allow CN= patterns */
                        if(p != NULL) {
                            u=(char*)strtok_r(p,",", &strtok_cb);	  /* Just take the CN= */
                            if(u != NULL ) {                        /* It's  time to ask wildcard.. */
                                u=u+3;
                                if (pWS->nDebugFlag >= LOGDEBUG0)
                                {
                                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"CN= Cert=[%s] mask=[%s]",work_CN, u);
                                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                                    LogData(LOGDEBUG0,pWS->szTmpBufff, pWS);
                                }
                                if (wildcmp(u,(char *)&work_CN,pWS)) {
                                    if(pWS->SilentMode ==0 )
                                    {
                                        if (pWS->nDebugFlag >= LOGDEBUG0)
                                        {
                                            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"SSL Pattern matched CN=[%s]",u);
                                            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                                            LogData(LOGDEBUG0, pWS->szTmpBufff, pWS);
                                        }
                                    }
                                }
                                else {
                                    if (pWS->nDebugFlag >= LOGDEBUG)
                                        LogData(LOGDEBUG, "SSL entry skipped. Didn't match CN=.", pWS);
                                    iErr=1;
                                }
                            }
                        }
                        else {
                            LogData(LOGDATA, "SSL entry skipped. Didn't match CN=.", pWS);
                            iErr=1;  /* Block this one, CN= is mandatory */
                        }
                    }

                    /* end 2.50 */

                    /* Start 2.64 */
                    if (strlen(work_ST) > 0 && iErr==0)
                    {
                        strcpy(work_strtok,work_filter);
                        p = strstr(work_strtok,"ST=");          /*  allow ST= patterns */
                        if(p != NULL) {
                            u=(char*)strtok_r(p,",", &strtok_cb);	  /* Just take the ST= */
                            if(u != NULL ) {                        /* It's  time to ask wildcard.. */
                                u=u+3;
                                if (pWS->nDebugFlag >= LOGDEBUG0)
                                {
                                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"ST= Cert=[%s] mask=[%s]",work_ST, u);
                                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                                    LogData(LOGDEBUG0,pWS->szTmpBufff, pWS);
                                }
                                if (wildcmp(u,(char *)&work_ST,pWS)) {
                                    if(pWS->SilentMode ==0 )
                                    {
                                        if (pWS->nDebugFlag >= LOGDEBUG0)
                                        {
                                            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"SSL Pattern matched ST=[%s]",u);
                                            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                                            LogData(LOGDEBUG0, pWS->szTmpBufff, pWS);
                                        }
                                    }
                                }
                                else {
                                    if (pWS->nDebugFlag >= LOGDEBUG)
                                        LogData(LOGDEBUG, "SSL entry skipped. Didn't match ST=.", pWS);
                                    iErr=1;
                                }
                            }
                        }
                    }
                    if (strlen(work_PC) > 0 && iErr==0)
                    {
                        strcpy(work_strtok,work_filter);
                        p = strstr(work_strtok,"PC=");          /*  allow PC= patterns */
                        if(p != NULL) {
                            u=(char*)strtok_r(p,",", &strtok_cb);	  /* Just take the PC= */
                            if(u != NULL ) {                        /* It's  time to ask wildcard.. */
                                u=u+3;
                                if (pWS->nDebugFlag >= LOGDEBUG0)
                                {
                                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"PC= Cert=[%s] mask=[%s]",work_PC, u);
                                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                                    LogData(LOGDEBUG0,pWS->szTmpBufff, pWS);
                                }
                                if (wildcmp(u,(char *)&work_PC,pWS)) {
                                    if(pWS->SilentMode ==0 )
                                    {
                                        if (pWS->nDebugFlag >= LOGDEBUG0)
                                        {
                                            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"SSL Pattern matched PC=[%s]",u);
                                            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                                            LogData(LOGDEBUG0, pWS->szTmpBufff, pWS);
                                        }
                                    }
                                }
                                else {
                                    if (pWS->nDebugFlag >= LOGDEBUG)
                                        LogData(LOGDEBUG, "SSL entry skipped. Didn't match PC=.", pWS);
                                    iErr=1;
                                }
                            }
                        }
                    }

                    /* end 2.64 */

                    if (iErr==0) {
                        if(pWS->SilentMode ==0 )
                        {
                            if (pWS->nDebugFlag >= LOGDEBUG0)
                            {
                                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Full SSL Pattern matched [%s]",work_filter);
                                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                                LogData(LOGDEBUG0, pWS->szTmpBufff, pWS);
                            }
                        }
                        /* Reload u-pointer */
                        u=(char*)strtok_r(pWS->SSLList[i],";", &strtok_cb);
                        i=pWS->SSLCount;				        /* We're thru pattern matching */
                        u=(char*)strtok_r(NULL,";", &strtok_cb);		/* Now we have to find the result */
                        if(u != NULL )
                        { 			                                /* Specified ?? */
                            if (pWS->nDebugFlag >= LOGDEBUG)
                            {
                                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"SSL Action [%s]", u);
                                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                                LogData(LOGDEBUG, pWS->szTmpBufff, pWS);
                            }
                            if(!strcmp(u,"BLOCK"))
                            {
                                LogData(LOGDATA, "Connection refused, CN requested blocked", pWS);
#if defined(__MVS__)
                                WTODataE("09E", pWS);
#endif
                                if (pWS->nDebugFlag >= LOGDEBUG1)
                                    LogData(LOGDEBUG1,"CheckSSLList() Done", pWS);
                                return SEC_FAILURE;
                            }
                            if(!strncmp(u,"MCA=",4))
                            {
                                u=u+4;
                                memset(work,' ',sizeof(work));
                                strcpy(work,u);
                                if(strcmp(work,"*"))
                                {

                                    /* 2.65 start */
                                    p = strstr(u,"(");           /* is there a subspec CN(15,8) ? */
                                    if (p != NULL)
                                    {
                                        if (getDNMCAUser(u,pWS->SSLPeer,work,sizeof(work),pWS)!=SEC_OK)
                                        {
                                            if (pWS->nDebugFlag >= LOGDEBUG1)
                                                LogData(LOGDEBUG1,"CheckSSLList() Done", pWS);
                                            return SEC_FAILURE;
                                        }
                                    }
                                    /* 2.65 end */
                                    SetLongMCAUser(work, pWS);     /* 2.94 */
                                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"SSL Set MCA userid to: [%s]",work);
                                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                                    LogData(LOGDATA, pWS->szTmpBufff, pWS);
#if defined(__MVS__)
                                    WTODataOk("51I", pWS);
                                    work[strlen(work)] = ' ';
                                    memset(pChannelDefinition->MCAUserIdentifier,' ',MQ_MCA_USER_ID_LENGTH);
#endif
                                    strncpy(pChannelDefinition->MCAUserIdentifier,work,MQ_MCA_USER_ID_LENGTH);
                                    if (pWS->nDebugFlag >= LOGDEBUG1)
                                        LogData(LOGDEBUG1,"CheckSSLList() Done", pWS);
                                    return SEC_OK;
                                }
                                else
                                {
                                    if (pWS->nDebugFlag >= LOGDEBUG)
                                    {
                                        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"SSL leave MCA userid unchanged.");
                                        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                                        LogData(LOGDEBUG, pWS->szTmpBufff, pWS);
                                    }
                                    if (pWS->nDebugFlag >= LOGDEBUG1)
                                        LogData(LOGDEBUG1,"CheckSSLList() Done", pWS);
                                    return SEC_OK;
                                }
                            }
                        }
                    }
                } /* for()*/
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"No Matching SSL settings found for certificate: [%s]",pWS->SSLPeer);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDATA, pWS->szTmpBufff, pWS);
#if defined(__MVS__)
                WTODataE("99E", pWS);
#endif
                if (pWS->nDebugFlag >= LOGDEBUG1)
                    LogData(LOGDEBUG1,"CheckSSLList() Done", pWS);
                return SEC_FAILURE;
            }
        }
    }
    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1,"CheckSSLList() Done", pWS);
    return SEC_OK;
}
#endif







/**
 * CheckCONList()
 *
 * Check the CON= parameters
 * If no CON= spec was supplied, so return just "OK"
 *
 * @return int 0=ignore 1=error
 *
 *
 * Notes 2.83 we issue an extra
 */
int CheckCONList(MQCD *pChannelDefinition,MQLONG ExitReason, workingStorage_type *pWS)
{
    int  i;
    char *u;
    char *strtok_cb;
    char work[33]; /* just working storage */
    int  bUserIdOK;
#ifdef HNLUP
#ifdef HNLUP2
    DNSReply dnsReply;
    int reply;
    int ipFilter=0;
#endif
    /*    MQLONG rc;*/
    char szIpWork[64];
#endif
    MQLONG rc;

    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1,"CheckCONList() Entered", pWS);

    if(pWS->CONCount == 0)
    {
        if (pWS->nDebugFlag >= LOGDEBUG1)
            LogData(LOGDEBUG1,"CheckCONList() Done", pWS);
        return SEC_OK;
    }

    for(i=0; i<pWS->CONCount; i++)
    {
        if (pWS->nDebugFlag >= LOGDEBUG)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  CONList[%ld] = [%s]",i,pWS->CONList[i]);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDEBUG,pWS->szTmpBufff, pWS);
        }
        u=(char*)strtok_r(pWS->CONList[i],";", &strtok_cb); /* Just take the connection pattern	*/
        if(u != NULL )
        {                                                  /* It's  time to ask wildcard.. */
            if (pWS->nDebugFlag >= LOGDEBUG0)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  CON Pattern: [%s]",u);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDEBUG0,pWS->szTmpBufff, pWS);
            }
            /* 2.53 start */
            strcpy(szIpWork,u);
#ifdef HNLUP
            if (isalpha(u[0]))
            {                           /* host address is a name */
                if (pWS->nDebugFlag >= LOGDEBUG1)
                    LogData(LOGDEBUG1,"  Need hostname lookup",pWS);
#ifdef HNLUP2
                ipFilter=0;             /* ipfiltering not done yet */
                rc= GetHostIP(u, (DNSReply*)&dnsReply, pWS);
                if (pWS->nDebugFlag >= LOGDEBUG0)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  GetHostIP() on hostname [%s] replies %ld returned [%s] rc (0=OK) %ld", u, dnsReply.replies, dnsReply.szIpAddr[0],(long)rc);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDEBUG0, pWS->szTmpBufff, pWS);
                }
                if (rc==0)
                {
                    for (reply=0; reply< dnsReply.replies; reply++)
                    {
                        if (wildcmp((char *)&dnsReply.szIpAddr[reply],(char *)&pWS->ConName,pWS))
                        {
                            if (pWS->nDebugFlag >= LOGDEBUG0)
                            {
                                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  Pattern [%s], ConName [%s] passed test..",
                                         dnsReply.szIpAddr[reply], pWS->ConName);
                                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                                LogData(LOGDEBUG0,pWS->szTmpBufff, pWS);
                            }
                            strcpy(szIpWork,dnsReply.szIpAddr[reply]);
                            break;
                        }
                        else
                        {
                            if (pWS->nDebugFlag >= LOGDEBUG0)
                            {
                                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  Pattern [%s], ConName [%s] failed test..",
                                         dnsReply.szIpAddr[reply], pWS->ConName);
                                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                                LogData(LOGDEBUG0,pWS->szTmpBufff, pWS);
                            }
                        }
                    }
                }
                else
                    strcpy(szIpWork,"Not-found");
#else
                rc= GetHostIP(u, (char*)&szIpWork, pWS);

                if (pWS->nDebugFlag >= LOGDEBUG)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  GetHostIP() on hostname [%s] returned [%s] rc (0=OK) %ld", u, szIpWork,(long)rc);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDEBUG, pWS->szTmpBufff, pWS);
                }
#endif
                /*                if (rc==0)             removed 2.78
                                    strcpy(u,szIpWork);
                                else
                                    strcpy(u,"Not-found");
                removed under 2.78 */
            }
#endif
            /* 2.53 end */
            /*            if(wildcmp(u,(char *)&pWS->ConName,pWS))  rem changed to 2.78 */

            if (pWS->nDebugFlag >= LOGDEBUG8)                        /* 2.81 */
                HexDump(pWS->ConName, MQ_CONN_NAME_LENGTH, pWS);     /* 2.81 */

            /*#ifdef HNLUP2
                        if(ipFilter==1)
            #else*/
            if(wildcmp(szIpWork,(char *)&pWS->ConName,pWS))
                /*#endif*/
            {
                if (pWS->nDebugFlag >= LOGDEBUG0)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  CON Pattern matched [%s] CON name [%s]", szIpWork,(char*)&pWS->ConName);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDEBUG0,pWS->szTmpBufff, pWS);
                }
                u=(char*)strtok_r(NULL,";", &strtok_cb);		/* Now we have to find the userid */
                if(u != NULL )
                { 			  /* Specified userid ?? */
                    if (pWS->bUseridUpperLowerCase==1) 			/* JHP 2.12 */
                    {
                        if (pWS->nDebugFlag >= LOGDEBUG1)
                        {
                            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  ConUserid Before strupr [%s]",u);
                            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                            LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
                        }
                        i_strupr(u);
                        if (pWS->nDebugFlag >= LOGDEBUG1)
                        {
                            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  ConUserid After strupr [%s]",u);
                            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                            LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
                        }
                    }
                    if (pWS->nDebugFlag >= LOGDEBUG0)
                    {
                        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  CON Userid: [%s]",u);
                        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                        LogData(LOGDEBUG0,pWS->szTmpBufff, pWS);
                    }
                    /* we need to check the userid as a list (CON=127.*;peter,post,sonny;MCA=mrmqo1;) */
                    /* 2.53 start*/
                    bUserIdOK=0;
                    if(pWS->bAllowBlankUserID==0)
                    {
                        if(wildcmplist(u,(char *)&pWS->RemUID,1,pWS))
                            bUserIdOK=1;          /* This user was ok */
                    }
                    else
                    {  /* We allow blank userids */
                        if (!strcmp(u,"BLANK_USERID")) {
                            if (!strcmp(pWS->RemUID,"")) {
                                bUserIdOK=1;          /* This user was ok */
                            }
                        }
                        else {
                            if(wildcmplist(u,(char *)&pWS->RemUID,1,pWS))
                                bUserIdOK=1;          /* This user was ok */
                        }
                    }

                    if (bUserIdOK==1)
                        /* if(wildcmplist(u,(char *)&pWS->RemUID,pWS))*/
                        /* 2.53 end */
                    {
                        if (pWS->nDebugFlag >= LOGDEBUG0)
                        {
                            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  CON/RemUid Pattern matched [%s] RemUID [%s]",u, (char*)&pWS->RemUID);
                            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                            LogData(LOGDEBUG0,pWS->szTmpBufff, pWS);
                        }
                        i=pWS->CONCount;					/* We're thru pattern matching */
                        u=(char*)strtok_r(NULL,";", &strtok_cb);		/* Now we have to find the result */
                        if(u != NULL )
                        { 		 /* Specified what keyword do we have ? */
                            if(!strcmp(u,"BLOCK"))   /* Block this combination ? */
                            {
                                LogData(LOGDATA, "Connection refused, CON was requested blocked", pWS);
#if defined(__MVS__)
                                WTODataE("09E", pWS);
#endif
                                return SEC_FAILURE;
                            }
                            if(u != NULL )
                            { 			  /* Specified userid ?? */
                                if(!strncmp(u,"MCA=",4))
                                {
                                    if (pWS->nDebugFlag >= LOGDEBUG)
                                        LogData(LOGDEBUG, "  CON MCA= specified", pWS);
                                    u=u+4;
                                    memset(work,' ',sizeof(work));
                                    strcpy(work,u);
                                    if (pWS->bRespectMCA==1)
                                    {
                                        if (pWS->nDebugFlag >= LOGDEBUG1)
                                        {
                                            LogData(LOGDEBUG1, "  RespectMCA was specified..", pWS);
                                        }
                                    }
                                    else
                                    {
                                        if (pWS->nDebugFlag >= LOGDEBUG1)
                                            LogData(LOGDEBUG1, "  RespectMCA was not specified...", pWS);
                                    }

                                    if(strcmp(work,"*"))
                                    {
                                        if ((pWS->bRespectMCA==1) && (memcmp(pChannelDefinition->MCAUserIdentifier,
                                                                             work,
                                                                             sizeof(pChannelDefinition->MCAUserIdentifier))) != 0 )
                                        {
                                            LogData(LOGDATA, "Connection refused, RespectMCA specified", pWS);
#if defined(__MVS__)
                                            WTODataE("23E", pWS);
#endif
                                            return SEC_FAILURE;
                                        }
                                        if (pWS->SilentMode !=1)   /* 2.84 */
                                        {
                                            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  CON Set MCA userid to [%s] from [%s]",u,pWS->RemUID);
                                            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                                            LogData(LOGDATA,pWS->szTmpBufff, pWS);
                                        }
#if defined (__MVS__)
                                        work[strlen(work)] = ' ';
                                        memset(pChannelDefinition->MCAUserIdentifier,' ',MQ_MCA_USER_ID_LENGTH);
#endif
                                        strncpy(pChannelDefinition->MCAUserIdentifier,work,MQ_MCA_USER_ID_LENGTH);
                                        SetLongMCAUser(work, pWS);     /* 2.94 */
                                    }
                                    else                                            /* 2.77 starting */
                                    {
#if defined (__MVS__)
                                        work[strlen(work)] = ' ';
                                        memset(pChannelDefinition->MCAUserIdentifier,' ',MQ_MCA_USER_ID_LENGTH);
#endif
                                        strncpy(pChannelDefinition->MCAUserIdentifier,pWS->RemUID,MQ_MCA_USER_ID_LENGTH);
                                        SetLongMCAUser(pWS->RemUID, pWS);     /* 2.94 */
                                        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  CON Set MCA userid to [%s] from received remote user",pWS->RemUID);
                                        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                                        LogData(LOGDATA,pWS->szTmpBufff, pWS);
                                    }                                               /* 2.77 ending */
                                }
                            }
                        }
                        LogData(LOGDEBUG0,"  CheckCONList leave return 0 (OK)", pWS);
                        if (pWS->nDebugFlag >= LOGDEBUG1)
                            LogData(LOGDEBUG1,"CheckCONList() Done", pWS);
                        return SEC_OK;	 /* Successful check */
                    }  /* if check userid .. */
                }
                else
                {
                    LogData(LOGDEBUG,"  CheckCONList User was not specified.", pWS);  /* Should not happen 2.78 */
                }
            }
        }
    }
    if (pWS->nDebugFlag >= LOGDEBUG0)
        LogData(LOGDEBUG0, "  CheckCONList leave return 1 (Failed)", pWS);
    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1,"CheckCONList() Done", pWS);
    return SEC_FAILURE; /* Check failed */
}





/**
* CheckInvalidUsers()
*
*
* Run through a list of possible invalid user names and fail if present
* otherwise drop through on success.
*
* Expansion Possibilities
* -----------------------
* Simply add additional names to end of list
*
*
* @return int SEC_OK or SEC_FAILUR
*/
int CheckInvalidUsers(PMQCD pChannelDefinition, workingStorage_type *pWS)
{
#if defined(__MVS__)
    char szZosUserId[9];
#endif

    if (pWS->nDebugFlag >= LOGDEBUG0)
        LogData(LOGDEBUG1,"CheckInvalidUsers() Entered", pWS);

    if (pWS->bAllowBlankUserID==0)
    {
        if(pChannelDefinition->RemoteUserIdentifier[0] == ' ')
        {
            LogData(LOGDATA,"Connection refused for blank user identifier", pWS);
#if defined(__MVS__)
            WTODataE("10E", pWS);
#endif
            if (pWS->nDebugFlag >= LOGDEBUG0)
                LogData(LOGDEBUG1,"CheckInvalidUsers() Done", pWS);
            return SEC_FAILURE;
        }
    }
    /*
    * Check for system uids
    */
    if(!strcmp(pWS->RemUID,"mqm"))
    {
        LogData(LOGDATA,"Connection refused for system user identifier -1", pWS);
#if defined(__MVS__)
        WTODataE("11E", pWS);
#endif
        if (pWS->nDebugFlag >= LOGDEBUG0)
            LogData(LOGDEBUG1,"CheckInvalidUsers() Done", pWS);
        return SEC_FAILURE;
    }
    if(!strcmp(pWS->RemUID,"MUSR_MQADMIN"))
    {
        LogData(LOGDATA,"Connection refused for system user identifier.-2", pWS);
#if defined(__MVS__)
        WTODataE("12E", pWS);
#endif
        if (pWS->nDebugFlag >= LOGDEBUG0)
            LogData(LOGDEBUG1,"CheckInvalidUsers() Done", pWS);
        return SEC_FAILURE;
    }
    if(!strcmp(pWS->RemUID,"musr_mqadmin"))
    {
        LogData(LOGDATA,"Connection refused for system user identifier.-3", pWS);
#if defined(__MVS__)
        WTODataE("13E", pWS);
#endif
        if (pWS->nDebugFlag >= LOGDEBUG0)
            LogData(LOGDEBUG1,"CheckInvalidUsers() Done", pWS);
        return SEC_FAILURE;
    }
    if(!strcmp(pWS->RemUID,"QMQM"))
    {
        LogData(LOGDATA,"Connection refused for system user identifier.-4", pWS);
#if defined(__MVS__)
        WTODataE("24E", pWS);
#endif
        if (pWS->nDebugFlag >= LOGDEBUG0)
            LogData(LOGDEBUG1,"CheckInvalidUsers() Done", pWS);
        return SEC_FAILURE;
    }
#if defined (__MVS__)
    memset(szZosUserId,'\00',sizeof(szZosUserId));
    GetZOSUserId((char*)&szZosUserId);
    if(!strcmp(pWS->RemUID,szZosUserId))
    {
        LogData(LOGDATA,"Connection refused for system user identifier(CHIN task).-5", pWS);
        WTODataE("25E", pWS);
        if (pWS->nDebugFlag >= LOGDEBUG0)
            LogData(LOGDEBUG1,"CheckInvalidUsers() Done", pWS);
        return SEC_FAILURE;
    }
#endif
        if (pWS->nDebugFlag >= LOGDEBUG0)
            LogData(LOGDEBUG1,"CheckInvalidUsers() Done", pWS);
    return SEC_OK;
}

/**
* CheckBlankUser()
*
*
* Check if userid is present and fail if not present
* otherwise drop through on success.
*
*
* @return int SEC_OK or SEC_FAILUR
*/
int CheckBlankUser(PMQCD pChannelDefinition, workingStorage_type *pWS)
{

    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1,"CheckBlankUser() Entered", pWS);

    if(pChannelDefinition->RemoteUserIdentifier[0] == ' ')
    {
        LogData(LOGDATA,"Connection refused for blank user identifier", pWS);
#if defined(__MVS__)
        WTODataE("10E", pWS);
#endif
        if (pWS->nDebugFlag >= LOGDEBUG0)
            LogData(LOGDEBUG1,"CheckBlankUser() Done", pWS);
        return SEC_FAILURE;
    }
    if (pWS->nDebugFlag >= LOGDEBUG0)
        LogData(LOGDEBUG1,"CheckBlankUser() Done", pWS);
    return SEC_OK;
}


/**
*  ProcessConfigFile()
*
*
* ProcessConfigFile looks for the default BlockIP2.ini configuration file
* in the exitPath library of the queuemanager. Path is resolved using registry
* or the mqs.ini / qm.ini files.
*
* Open file and call the ProcessData() method to process the options.
* Return failure for data processing errors.
*
*
* @return int  SEC_OK or SEC_FAILURE
*/
#if defined (UNIX) || defined (WIN32) || defined (OS400)

int ProcessConfigFile(MQCHAR48  qMgrName, workingStorage_type *pWS)
{
    int  rc=0;
    int  i;
    char *result;
    FILE *phInFile;			/* Declare file handles for input and output   */
    char line[MAX_LEN];
    int  bFound;
    char szTemp[300];
    long ErrorReported;

    pWS->bThisQmgr=1;
    pWS->bThisChannelOK=1;

    strcpy(pWS->szConfigFile,pWS->ExitPath);
#if defined (WIN32)
    strcat(pWS->szConfigFile,"\\");
#else
    strcat(pWS->szConfigFile,"/");
#endif
    strcat(pWS->szConfigFile,"BlockIP2.ini");
    if (pWS->ExitReason==MQXR_INIT) {
        if (pWS->nDebugFlag >= LOGDEBUG8)
        {
            snprintf(szTemp,sizeof(szTemp),"Info: BlockIP2.ini [%s] to look for.",pWS->szConfigFile);
            szTemp[sizeof(szTemp)-1]='\00';
            LogData(LOGDEBUG8,szTemp,pWS);
        }
    }

    phInFile = fopen( pWS->szConfigFile, "r" );
    if( phInFile == NULL )
    {
        ErrorReported=errno;
        if (pWS->ExitReason==MQXR_INIT) {
            if (pWS->nDebugFlag >= LOGDEBUG)
            {
                snprintf(szTemp,sizeof(szTemp),"Info: BlockIP2.ini [%s] was not found. (%s)",pWS->szConfigFile, strerror(ErrorReported));
                szTemp[sizeof(szTemp)-1]='\00';
                LogData(LOGDEBUG,szTemp,pWS);
                /***                snprintf(szTemp,sizeof(szTemp),"Info: Reason: %s", strerror(ErrorReported));
                                szTemp[sizeof(szTemp)-1]='\00';
                                LogData(LOGDEBUG,szTemp,pWS);
                ***/
            }
        }
    }
    else {
        do {
            result = fgets(line,MAX_LEN,phInFile);
            if(result != NULL)
            {
                bFound=0;
                i=(int)strlen(line);
                /* 2.91 */
                if(i==0) { continue; }
                if(line[i-1] == '\n')				/* last char is a new-line */
                {
                    line[i-1]  = '\0';				/* replace new-line with null */
                    --i;							/* reduce buffer length */
                }
                /*
                * Skip comment and empty lines - just process for possible data
                */
                if((line[0] != '#') && (strlen(line) > 0))
                {
                    if(ProcessData( line, pWS )==SEC_FAILURE)
                        rc=8;
                }
            }
        } while (result != NULL);
        fclose( phInFile );
    }
    if (pWS->ExitReason==MQXR_INIT) {
        if (pWS->nDebugFlag >= LOGDEBUG8)
        {
            snprintf(szTemp,sizeof(szTemp),"BlockIP2.ini [%s] processing ended with rc %d.",pWS->szConfigFile,rc); /* MB 2.79 Print the name of the actual config file in use */
            szTemp[sizeof(szTemp)-1]='\00';
            LogData(LOGDEBUG8,szTemp,pWS);
        }
    }
    return rc;
}

#endif

/**
* ProcessData()
*
*
*
* Process a line of data for all possible tags and setup required structures.
* Can be called from ProcessFile() method or direct from exit data field.
* If tag is unknow then return OK anyway (no harm in doing so).
*
*
* Expansion Possibilities
* -----------------------
* Add additional commands to the end of the method.
*
*
* @return int SEC_OK if line was processed OK else SEC_FAILURE if an error occured
*/
int ProcessData(char *pszData, workingStorage_type *pWS)
{
    int bFound=0;
    char *p;
    char *u;
    char *strtok_cb;
    int i;
    char work[49];  /* max obj len + 1 */

    /*
    * Test for command termination
    */
    p = (char*)strtok_r(pszData," ", &strtok_cb);
    if(p != NULL)
    {
        /*       p=p-1;*/
        if(p[strlen(p)-1] != ';')
            /*       if(p[0] != ';')*/
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, invalid specification, parm not terminated with \";\" [%s]" ,pszData);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA, pWS->szTmpBufff, pWS);
            return SEC_FAILURE;
        }

    }

    p = strstr(pszData,"CHANNEL=");
    if(p != NULL)
    {
        if (pWS->bThisQmgr==1) {
            p=p+8;
            u=(char*)strtok_r(p,";",&strtok_cb);
            if(u != NULL) {
                pWS->bThisChannel=0;
                pWS->bThisChannelOK=0;             /* Channel match is not ok  */
                pWS->bChannelPatterns=1;
                if(wildcmplist(u,(char *)Strip(pWS->szChannelName),0,pWS)) {
                    if (pWS->ExitReason==MQXR_INIT)
                    {
                        if (pWS->nDebugFlag >= LOGDEBUG0)
                        {
                            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"CHANNEL= [%s] matched [%s]." ,(char *)Strip(pWS->szChannelName),u);
                            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                            LogData(LOGDEBUG0, pWS->szTmpBufff, pWS);
                        }
                    }
                    pWS->bThisChannel=1;
                    pWS->bThisChannelOK=1;
                    pWS->bThisChannelDone=1;
                }
            }
        }
        bFound=1;
    }

    p = strstr(pszData,"QMGR=");
    if((p != NULL)   && (p == pszData)) {
        p=p+5;
        p=(char*)strtok_r(p," ",&strtok_cb);
        u=(char*)strtok_r(p,";",&strtok_cb);
        if(u != NULL) {
            pWS->bChannelPatterns=0;
            pWS->bThisQmgr=0;
            pWS->bThisChannel=0;
            pWS->bThisChannelOK=0;             /* Channel match is not ok  */
            if(wildcmplist(u,(char *)Strip(pWS->szQMgr),0,pWS))
            {
                if (pWS->ExitReason==MQXR_INIT) {
                    if (pWS->nDebugFlag >= LOGDEBUG0)
                    {
                        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"QMGR= [%s] matched [%s].",(char *)Strip(pWS->szQMgr) ,u);
                        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                        LogData(LOGDEBUG0, pWS->szTmpBufff, pWS);
                    }
                }
                pWS->bThisQmgr=1;
                pWS->bThisChannel=1;               /* All channels accepted */
                pWS->bThisChannelOK=1;             /* All channels accepted */
            }
        }
        bFound=1;
    }

    /*
    * Test for logging command
    */
    p = strstr(pszData,"LogFormat=");
    if(p != NULL)
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            pWS->bLogByDate=1;
            p=p+10;
            u=(char*)strtok_r(p,";", &strtok_cb);
            if(u!=NULL)
            {
                for(i=0;i<(int)strlen(u);i++)
                {
                    if(*(u+i) == 'N') pWS->nLogFormat |= LOGNAME;
                    if(*(u+i) == 'D') pWS->nLogFormat |= LOGDATE;
                    if(*(u+i) == 'C') pWS->nLogFormat |= LOGCHANNEL;
                    if(*(u+i) == 'U') pWS->nLogFormat |= LOGUSER;
                    if(*(u+i) == 'M') pWS->nLogFormat |= LOGQMANAGER;   /* 2.80 */
                }
            }
        }
    }

    p = strstr(pszData,"NoLogFileName=Y");             		/* Omit BlockIP2 in in Logfile 2.95 */
    if((p!=NULL) && (bFound==0))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            u=(char*)strtok_r(p,";", &strtok_cb);
            if(u != NULL)
            {
                pWS->NoLogFileName=1;
            }
        }
    }

    p = strstr(pszData,"NoLogFileName=N");                      /* Keep BlockIP2 in in Logfile 2.95 */
    if((p!=NULL) && (bFound==0))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            u=(char*)strtok_r(p,";", &strtok_cb);
            if(u != NULL)
            {
                pWS->NoLogFileName=0;		/* Default */
            }
        }
    }

    p = strstr(pszData,"LogFileName=");
    if((p!=NULL) && (bFound==0))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            p=p+12;
            u=(char*)strtok_r(p,";", &strtok_cb);
            if(u != NULL)
            {
                strcpy(pWS->szLogFileName,u);
            }
        }
    }

    p = strstr(pszData,"LogExt=");
    if((p!=NULL) && (bFound==0))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            p=p+7;
            u=(char*)strtok_r(p,";", &strtok_cb);
            if(u != NULL)
            {
                strcpy(pWS->szLogFileExtension,u);
            }
        }
    }

    p = strstr(pszData,"LogPath=");
    if((p!=NULL) && (bFound==0))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            p=p+8;
            u=(char*)strtok_r(p,";", &strtok_cb);
            if(u != NULL)
            {
                strcpy(pWS->szLogPath,u);
            }
        }
    }

    p = strstr(pszData,"LogDrive=");
    if((p !=NULL) && (bFound==0))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            p=p+9;
            u=(char*)strtok_r(p,";", &strtok_cb);
            if(u != NULL)
            {
                strcpy(pWS->szLogDrive,u);
            }
        }
    }

    /* 2.46 start */
#if defined (__EXT_LOG_FILENAME__)             /* Only UNIX, LINUX and Windows */
    p = strstr(pszData,"LogCount=");
    if((p !=NULL) && (bFound==0))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            p=p+9;
            u=(char*)strtok_r(p,";", &strtok_cb);
            if(u != NULL)
            {
                pWS->lLogFileCount=atol(u);
                if (pWS->nDebugFlag >= LOGDEBUG)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Number of logfiles set [%ld]",pWS->lLogFileCount);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDEBUG, pWS->szTmpBufff, pWS);
                }
                if ((pWS->lLogFileCount < 1) | (pWS->lLogFileCount > 99)) {
                    pWS->lLogFileCount=3;
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Invalid value was specified for LogCount, value reset to default [%ld]",pWS->lLogFileCount);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDATA, pWS->szTmpBufff, pWS);
                }
            }
        }
    }

    p = strstr(pszData,"LogSize=");
    if((p !=NULL) && (bFound==0))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            p=p+8;
            u=(char*)strtok_r(p,";", &strtok_cb);
            if(u != NULL)
            {
                pWS->lMaxlogFileSize=atol(u);
                if (pWS->nDebugFlag >= LOGDEBUG)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Size of logfiles set [%ld]",pWS->lMaxlogFileSize);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDEBUG, pWS->szTmpBufff, pWS);
                }
                if (pWS->lMaxlogFileSize < 100*1024) {
                    pWS->lMaxlogFileSize=200*1024;
                    if (pWS->nDebugFlag >= LOGDEBUG)
                    {
                        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Invalid value was specified for LogSize, value reset to default [%ld]",pWS->lMaxlogFileSize);
                        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                        LogData(LOGDEBUG, pWS->szTmpBufff, pWS);
                    }
                }
            }
        }
    }

    /*
    * Control cycling of logfile
    */

    p = strstr(pszData,"LOGCYCLE=");    /* 2.61 */
    if((p !=NULL) && (bFound==0))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            p=p+9;
            u=(char*)strtok_r(p,";", &strtok_cb);
            if(u != NULL)
            {
                if (!strcmp(u,"Y"))
                    pWS->bFileSizeCheck=0;            /* cycle files, this is default */
                else
                    if (!strcmp(u,"N"))
                        pWS->bFileSizeCheck=1;            /* switch off to maintain transperacy with older releases */
                    else
                    {
                        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, LOGCYCLE= invalid [%s]. Legal arguments are Y or N",u);
                        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                        LogData(LOGDATA, pWS->szTmpBufff, pWS);
#if defined(__MVS__)
                        WTODataE("99E", pWS);
#endif
                        return SEC_FAILURE;
                    }
            }
            else
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, LOGCYCLE= semicolon(;) is missing");
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDATA, pWS->szTmpBufff, pWS);
#if defined(__MVS__)
                WTODataE("99E", pWS);
#endif
                return SEC_FAILURE;
            }
        }
    }
    p = strstr(pszData,"LogCycle=");    /* 2.61 */
    if((p !=NULL) && (bFound==0))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            p=p+9;
            u=(char*)strtok_r(p,";", &strtok_cb);
            if(u != NULL)
            {
                if (!strcmp(u,"Y"))
                    pWS->bFileSizeCheck=0;            /* cycle files, this is default */
                else
                    if (!strcmp(u,"N"))
                        pWS->bFileSizeCheck=1;            /* switch off to maintain transperacy with older releases */
                    else
                    {
                        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, LogCycle= invalid [%s]. Legal arguments are Y or N",u);
                        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                        LogData(LOGDATA, pWS->szTmpBufff, pWS);
#if defined(__MVS__)
                        WTODataE("99E", pWS);
#endif
                        return SEC_FAILURE;
                    }
            }
            else
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, LogCycle= semicolon(;) is missing");
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDATA, pWS->szTmpBufff, pWS);
#if defined(__MVS__)
                WTODataE("99E", pWS);
#endif
                return SEC_FAILURE;
            }
        }
    }


#endif /* __EXT_LOG_FILENAME__ defined */
    /* 2.46 end */

    /*
    * Test for Block system accounts command
    */
    p = strstr(pszData,"BlockMqmUsers=Y");
    if(p != NULL)
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            pWS->bBlockMqmUsers=1;
        }
    }

    /*
    * Test for UnBlock system accounts command
    */
    p = strstr(pszData,"BlockMqmUsers=N");
    if(p != NULL)
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            pWS->bBlockMqmUsers=0;
        }
    }

    /*
    * Save pattern structure away if applicable
    */
    p = strstr(pszData,"Patterns=");
    if((p != NULL) && (bFound==0))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            p=p+9;
            u=(char*)strtok_r(p,";", &strtok_cb);
            if(u != NULL)
            {
                if (!memcmp(pWS->Pattern,"NONE",4))
                {  /* Remove the crap */
                    strcpy(pWS->Pattern,"");
                }
                if (!memcmp(pWS->Pattern,"FN=",3))
                {  /* Remove the crap */
                    strcpy(pWS->Pattern,"");
                }
                if( (strlen(u) + 1 + strlen(pWS->Pattern)) >= MAX_PL) /*NWC allow multiple lines */
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, Pattern string is too long, max[%ld] was [%ld]",(long)MAX_PL-1,(long)(strlen(u) + 1 + strlen(pWS->Pattern)));
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDATA, pWS->szTmpBufff, pWS);
#if defined(__MVS__)
                    WTODataE("14E", pWS);
#endif
                    return SEC_FAILURE;
                }

                strcat(pWS->Pattern,u);
                strcat(pWS->Pattern,";");	/* NWC put ; on the end to allow multiple lines */
            }
        }
    }


    /*  3.00 start
    * Save Userids list away if applicable
    */
    p = strstr(pszData,"PartnerName=");
    if((p != NULL) && (bFound==0))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1)
		{
            p=p+12;
            u=(char*)strtok_r(p,";", &strtok_cb);
            if(u != NULL)
			{
				if(strlen(u) > (sizeof(pWS->PartnerName)-1))
				{
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, PartnerName string is too long, max[%d], was [%d]",sizeof(pWS->PartnerName)-1,(int)(strlen(u)));
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDATA, pWS->szTmpBufff, pWS);
#if defined(__MVS__)
                    WTODataE("xxE", pWS);
#endif
                    return SEC_FAILURE;
				}
			memset(pWS->PartnerName,'\00',sizeof(pWS->PartnerName));
			strncpy(pWS->PartnerName,u,MQ_Q_MGR_NAME_LENGTH);

			}
		}
	}

    /*
    * Save Userids list away if applicable
    */
    p = strstr(pszData,"Userids=");
    if((p != NULL) && (bFound==0))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            p=p+8;
            u=(char*)strtok_r(p,";", &strtok_cb);
            if(u != NULL)
            {
                if( (strlen(u) +1 + strlen(pWS->Users)) >= MAX_PL) /* NWC allow multiple lines */
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, Users string is too long, max[%d], was [%d]",MAX_PL-1,(int)(strlen(u) +1 + strlen(pWS->Users)));
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDATA, pWS->szTmpBufff, pWS);
#if defined(__MVS__)
                    WTODataE("15E", pWS);
#endif
                    return SEC_FAILURE;
                }
                if (pWS->Users[strlen(pWS->Users)] == ';') {  /* 2.50 change */
                    if (pWS->nDebugFlag >= LOGDEBUG1)
                        LogData(LOGDEBUG1,"Need to remove ;",pWS);
                }
                strcat(pWS->Users,u);
                strcat(pWS->Users,","); /* NWC allow multiple lines */

                if (pWS->nDebugFlag >= LOGDEBUG1)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Length of pWS->Users [%ld]",(long)strlen(pWS->Users));
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
                }

                pWS->Users[strlen(pWS->Users)] = 0x00;              /* JHP 2.50 */
                if (pWS->bUseridUpperLowerCase==1) 			/* JHP 2.12 */
                {
                    if (strlen(pWS->Users) < 200) {
                        if (pWS->nDebugFlag >= LOGDEBUG1)
                        {
                            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Users Before strupr [%s]",pWS->Users);
                            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                            LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
                        }
                    }
                    i_strupr(pWS->Users);
                    if (strlen(pWS->Users) < 200) {
                        if (pWS->nDebugFlag >= LOGDEBUG1)
                        {
                            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Users After strupr [%s]",pWS->Users);
                            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                            LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
                        }
                    }
                }
            }
        }
    }

    /*
    * Save Connection list (CON=) options (if applicable).
    */
    p = strstr(pszData,"CON=");
    if((p != NULL) && (bFound==0))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            p=p+4;
            u=(char*)strtok_r(p," ", &strtok_cb);          /* Just take the data, no imbedded blanks */
            if(u != NULL)
            {
                if(u[strlen(u)-1] != ';')
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, CON Spec invalid, not terminated with \";\" [%s]" ,p-4);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDATA, pWS->szTmpBufff, pWS);
                    return SEC_FAILURE;
                }
                if(strlen(u) >= CON_LEN)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, CON Spec string is too long, max [%d], was [%ld]" ,SSL_LEN-1,(long)strlen(u));
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDATA, pWS->szTmpBufff, pWS);
#if defined(__MVS__)
                    WTODataE("16E", pWS);
#endif
                    return SEC_FAILURE;
                }
                /*
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"CON spec added [%s]",u);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDEBUG, pWS->szTmpBufff);
                */
                if (pWS->CONCount>=MAX_CON)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, CON Too many CON= specs active. Max [%d]" ,MAX_CON);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDATA, pWS->szTmpBufff, pWS);
#if defined(__MVS__)
                    WTODataE("68E", pWS);
#endif
                    return SEC_FAILURE;
                }
                strcpy(pWS->CONList[pWS->CONCount++],u);
                if (pWS->nDebugFlag >= LOGDEBUG1)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"CON= inserted CONList[%ld] [%s]",pWS->CONCount,u);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
                }
            }
        }
    }

    /*
    * Test for TERM= pattern and set flags
    */
    p = strstr(pszData,"TERM=N");
    if((p != NULL) && (bFound==0)  && (p == pszData))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            pWS->bLogTERM=0;                               /* Don't print TERM message include ConName */
        }
    }
    p = strstr(pszData,"TERM=Y");
    if((p != NULL) && (bFound==0)  && (p == pszData))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            pWS->bLogTERM=1;                               /* Print TERM message include ConName (Default) */
        }
    }

#ifdef MQCXP_VERSION_6
    /*
    * Test for ASC= pattern and set flags
    */
    if (pWS->Version>=MQCXP_VERSION_6) {
        p = strstr(pszData,"ASC=Y");
        if((p != NULL) && (bFound==0)  && (p == pszData))
        {
            bFound=1;
            if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
                pWS->AllowASC=1;                           /* Allow self signed certificates */
            }
        }

        /*
        * Test for ASC= pattern and set flags
        */
        p = strstr(pszData,"ASC=N");
        if((p != NULL) && (bFound==0)  && (p == pszData))
        {
            bFound=1;
            if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
                pWS->AllowASC=0;                           /* Dis-allow self signed certificates */
            }
        }

        /*
        * Test for ASC= pattern and set flags
        */
        p = strstr(pszData,"AllowSelfSignedCertificate=Y");
        if((p != NULL) && (bFound==0)  && (p == pszData))
        {
            bFound=1;
            if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
                pWS->AllowASC=1;                           /* Allow self signed certificates */
            }
        }
        p = strstr(pszData,"AllowSelfSignedCertificates=Y");
        if((p != NULL) && (bFound==0)  && (p == pszData))
        {
            bFound=1;
            if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
                pWS->AllowASC=1;                           /* Allow self signed certificates */
            }
        }

        /*
        * Test for ASC= pattern and set flags
        */
        p = strstr(pszData,"AllowSelfSignedCertificate=N");
        if((p != NULL) && (bFound==0)  && (p == pszData))
        {
            bFound=1;
            if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
                pWS->AllowASC=0;                           /* Dis-allow self signed certificates */
            }
        }
        p = strstr(pszData,"AllowSelfSignedCertificates=N");
        if((p != NULL) && (bFound==0)  && (p == pszData))
        {
            bFound=1;
            if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
                pWS->AllowASC=0;                           /* Dis-allow self signed certificates */
            }
        }

        /* 2.80 start */
        /*
        * Test for USESECPARMS= pattern and set flags
        */
        p = strstr(pszData,"USESECPARMS=N");
        if((p != NULL) && (bFound==0)  && (p == pszData))
        {
            bFound=1;
            if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
                pWS->bUseSecurityParms=0;                  /* Disable security Parms usage */
            }
        }
        p = strstr(pszData,"USESECPARMS=Y");
        if((p != NULL) && (bFound==0)  && (p == pszData))
        {
            bFound=1;
            if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
                pWS->bUseSecurityParms=1;                  /* Enable security Parms usage */
            }
        }
        /* 2.80 end */

        /* 2.94 start */
        /*
        * Test for USESECPARMS= pattern and set flags
        */
        p = strstr(pszData,"SECPARMSOFF=N");
        if((p != NULL) && (bFound==0)  && (p == pszData))
        {
            bFound=1;
            if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
                pWS->bForceSecurityParmsOff=0;                  /* Enable security checks (MQCSP_AUTH_USER_ID_AND_PWD) */
            }
        }
        p = strstr(pszData,"SECPARMSOFF=Y");
        if((p != NULL) && (bFound==0)  && (p == pszData))
        {
            bFound=1;
            if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
                pWS->bForceSecurityParmsOff=1;                  /* Disable security checks (MQCSP_AUTH_NONE) */
            }
        }
        /* 2.94 end */

    }
#endif

    /*
    * Test for SSL= pattern and save parameters in List
    */
    p = strstr(pszData,"SSL=");
    if((p != NULL) && (bFound==0))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            p=p+4;
            u=(char*)strtok_r(p," ", &strtok_cb);          /* Just take the data, no imbedded blanks */
            if(u != NULL)
            {
                if(strlen(u) >= SSL_LEN)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, SSL Spec string is too long, max[%d], was [%ld]",SSL_LEN-1,(long)strlen(u));
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDATA, pWS->szTmpBufff, pWS);
#if defined(__MVS__)
                    WTODataE("17E", pWS);
#endif
                    return SEC_FAILURE;
                }
                if(pWS->SSLCount >= MAX_SSL)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, number of SSL exceeds [%d]",MAX_SSL);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDATA, pWS->szTmpBufff, pWS);
#if defined(__MVS__)
                    WTODataE("33E", pWS);
#endif
                    return SEC_FAILURE;
                }

                if (pWS->nDebugFlag >= LOGDEBUG0)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"SSL CN spec added [%s]",u);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDEBUG0, pWS->szTmpBufff,pWS);
                }

                strcpy(pWS->SSLList[pWS->SSLCount++],u);
            }
        }
    }

    /*
    * Test for Case sensitivity flag
    */
    p = strstr(pszData,"UseridUpperLowerCase=*");
    if((p != NULL)  && (p == pszData))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            pWS->bUseridUpperLowerCase=1;
        }
    }


    /*
    * Save BlockUsers list away if applicable
    */
    p = strstr(pszData,"BlockUsers=");
    if((p != NULL) && (bFound==0))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            p=p+11;
            u=(char*)strtok_r(p,";", &strtok_cb);
            if(u != NULL)
            {
                if( (strlen(u) + 1 + strlen(pWS->BlockUsers)) >= MAX_PL)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, BlockUsers string is too long, max[%d], was [%ld]",MAX_PL-1,(long)strlen(u));
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDATA, pWS->szTmpBufff, pWS);
#if defined(__MVS__)
                    WTODataE("18E", pWS);
#endif
                    return SEC_FAILURE;
                }
                strcat(pWS->BlockUsers,u);
                strcat(pWS->BlockUsers,";");
                pWS->BlockUsers[strlen(u)] = 0x00;
                if (pWS->bUseridUpperLowerCase==1)         /* JHP 2.12 */
                {
                    if (pWS->nDebugFlag >= LOGDEBUG1)
                    {
                        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"BlockUsers Before strupr [%s]",pWS->BlockUsers);
                        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
                    }
                    i_strupr(pWS->BlockUsers);
                    if (pWS->nDebugFlag >= LOGDEBUG1)
                    {
                        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"BlockUsers After strupr [%s]",pWS->BlockUsers);
                        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
                    }
                }
            }
        }
    }

    /*
    * Test for Respect channel definition MCA userid
    */
    p = strstr(pszData,"RespectMCA=Y");
    if((p != NULL) && (bFound==0)  && (p == pszData))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            pWS->bRespectMCA=1;
        }
    }

    /* Start inq chl */
    /*
    * Test for MAX active channels
    */
    p = strstr(pszData,"MAXCHL=");
    if((p != NULL) && (bFound==0))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            p=p+7;
            p=(char*)strtok_r(p," ", &strtok_cb);
            u=(char*)strtok_r(p,";", &strtok_cb);
            if(u != NULL)
            {
                if (wildcmp((char *)u, (char *)&pWS->szChannelName, pWS))
                {
                    strcpy(work,u);
                    u=(char*)strtok_r(NULL,";", &strtok_cb);  /* Now we have to find max number of channels */
                    if(u == NULL )
                    {
                        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"MAX active channels set on [%s], number missing",pWS->szChannelName);
                        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                        LogData(LOGDATA, pWS->szTmpBufff, pWS);
                        return SEC_FAILURE;
                    }
                    pWS->lMaxChannelActive=atol(u);
                    if (pWS->nDebugFlag >= LOGDEBUG0)
                    {
                        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"MAX active channels set on [%s] set to [%ld] Pattern used[%s]",pWS->szChannelName,pWS->lMaxChannelActive, work);
                        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                        LogData(LOGDEBUG0, pWS->szTmpBufff, pWS);
                    }
                }
                else
                {
                    if (pWS->nDebugFlag >= LOGDEBUG0)
                    {
                        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"MAX active channels set on [%s], was skipped.",u);
                        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                        LogData(LOGDEBUG0, pWS->szTmpBufff, pWS);
                    }
                }
            }
        }
    }
    /* End inq chl */
    /*
    * Test for Allow blank userid
    */
    p = strstr(pszData,"AllowBlankUserID=Y");
    if((p != NULL)  && (p == pszData))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            pWS->bAllowBlankUserID=1;
        }
    }
    /*
    * Test for Dis-Allow blank userid  2.75
    */
    p = strstr(pszData,"AllowBlankUserID=N");
    if((p != NULL)  && (p == pszData))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            pWS->bAllowBlankUserID=0;
        }
    }

    /*
    * Test for Expiry settings
    */
    p = strstr(pszData,"Expiry=");
    if((p != NULL) && (bFound==0))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            p=p+7;
            p=(char*)strtok_r(p," ", &strtok_cb);
            u=(char*)strtok_r(p,";", &strtok_cb);
            if(u != NULL)
            {
                pWS->lExpireValue=atol(u)*10;
                if (pWS->nDebugFlag >= LOGDEBUG)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Expiry for channel rescan [%d]",pWS->lExpireValue);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDEBUG, pWS->szTmpBufff, pWS);
                }
            }
        }
    }

    /*
    * Test for Expiry settings
    */
    p = strstr(pszData,"ConfigQueue=");
    if((p != NULL) && (bFound==0))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            p=p+12;
            p=(char*)strtok_r(p," ", &strtok_cb);
            u=(char*)strtok_r(p,";", &strtok_cb);
            if(u != NULL)
            {
                if (strlen(u) < MQ_OBJECT_NAME_LENGTH) {
                    strcpy(pWS->szStatusQ,u);
                    if (pWS->nDebugFlag >= LOGDEBUG)
                    {
                        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Configuration queue is set to [%s]",pWS->szStatusQ);
                        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                        LogData(LOGDEBUG, pWS->szTmpBufff, pWS);
                    }
                }
                else {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Configuration queue is invalid [%s] name too long(%ld). max %d",u,(long)strlen(u),MQ_OBJECT_NAME_LENGTH);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDATA, pWS->szTmpBufff, pWS);
#if defined(__MVS__)
                    WTODataE("99E", pWS);
#endif
                    return SEC_FAILURE;                    /* Newer ever accept bad parameters */

                }
            }
        }
    }


#if defined (WIN32)
    /*
    * Test for Log control
    */
    p = strstr(pszData,"WinLog=ONLY");
    if(p != NULL  && (p == pszData))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            pWS->lWinLog=LOG_EVTLOGONLY;
        }
    }
    p = strstr(pszData,"WinLog=FILE");
    if(p != NULL  && (p == pszData))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            pWS->lWinLog=LOG_FILEONLY;
        }
    }
    p = strstr(pszData,"WinLog=BOTH");
    if(p != NULL  && (p == pszData))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            pWS->lWinLog=LOG_BOTH;
        }
    }
#endif


#if defined (UNIX)   /* not enabled for OS400 */
    /*
    * Test for Log control
    */
    p = strstr(pszData,"SysLog=ONLY");
    if(p != NULL  && (p == pszData))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            pWS->lWinLog=LOG_EVTLOGONLY;
        }
    }
    p = strstr(pszData,"SysLog=FILE");
    if(p != NULL  && (p == pszData))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            pWS->lWinLog=LOG_FILEONLY;
        }
    }
    p = strstr(pszData,"SysLog=BOTH");
    if(p != NULL  && (p == pszData))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            pWS->lWinLog=LOG_BOTH;
        }
    }


    /*
    * Test for syslog facility control
    */
    p = strstr(pszData,"SYSLOGFCLTY=LOG_DAEMON");
    if((p != NULL) && (bFound==0) && (p == pszData))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1)
        {
            pWS->iSyslogFclty=LOG_DAEMON;
        }
    }
    /*
    * Test for syslog facility control
    */
    p = strstr(pszData,"SYSLOGFCLTY=LOG_USER");
    if((p != NULL) && (bFound==0) && (p == pszData))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1)
        {
            pWS->iSyslogFclty=LOG_USER;
        }
    }
    /*
    * Test for syslog level control
    */
    p = strstr(pszData,"SYSLOGPRTY=LOG_ERR");
    if((p != NULL) && (bFound==0) && (p == pszData))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1)
        {
            pWS->iSyslogPrty=LOG_ERR;
        }
    }
    /*
    * Test for syslog level control
    */
    p = strstr(pszData,"SYSLOGPRTY=LOG_INFO");
    if((p != NULL) && (bFound==0) && (p == pszData))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1)
        {
            pWS->iSyslogPrty=LOG_INFO;
        }
    }
    /*
    * Test for syslog level control
    */
    p = strstr(pszData,"SYSLOGPRTY=LOG_NOTICE");
    if((p != NULL) && (bFound==0) && (p == pszData))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1)
        {
            pWS->iSyslogPrty=LOG_NOTICE;
        }
    }
    /*
    * Test for syslog level control
    */
    p = strstr(pszData,"SYSLOGPRTY=LOG_WARNING");
    if((p != NULL) && (bFound==0) && (p == pszData))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1)
        {
            pWS->iSyslogPrty=LOG_WARNING;
        }
    }

    /*
    * Test for TMPPATH control  added under 2.90
    */
    p = strstr(pszData,"TMPPATH=Yes");
    if((p != NULL) && (bFound==0) && (p == pszData))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1)
        {
            pWS->lTmpPath=1;
        }
    }
    p = strstr(pszData,"TMPPATH=No");
    if((p != NULL) && (bFound==0) && (p == pszData))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1)
        {
            pWS->lTmpPath=0;
        }
    }
#endif

#if defined(__MVS__)
    p = strstr(pszData,"WTOPFX=");
    if((p!=NULL) && (bFound==0))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            p=p+7;
            u=(char*)strtok_r(p,";", &strtok_cb);
            if(u != NULL)
            {
                strncpy(pWS->szWTOPfx,u,sizeof(pWS->szWTOPfx)-1);
                pWS->szWTOPfx[sizeof(pWS->szWTOPfx)-1]='\00';
            }
        }
    }
#endif


#if !defined(__MVS__)
    p = strstr(pszData,"INCQMGR=Y");
    if((p != NULL) && (bFound==0) && (p == pszData))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            pWS->bIncludeQmgrNameInLog=1;
        }
    }
    p = strstr(pszData,"INCQMGR=N");
    if((p != NULL) && (bFound==0)  && (p == pszData))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            pWS->bIncludeQmgrNameInLog=0;
        }
    }
    p = strstr(pszData,"QMGR_SHORT_NAME=");
    if((p!=NULL) && (bFound==0))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            p=p+16;
            u=(char*)strtok_r(p,";", &strtok_cb);
            if(u != NULL)
            {
                strncpy(pWS->szShortQMName,u,sizeof(pWS->szShortQMName)-1);
                pWS->szShortQMName[sizeof(pWS->szShortQMName)-1]='\00';
            }
        }
    }
#endif
   /*
    * Handle FAP check paramter here 2.98
    */
    p = strstr(pszData,"FAP=");
    if((p !=NULL) && (bFound==0))
    {
        bFound=1;
        if (pWS->bThisQmgr==1 && pWS->bThisChannelOK==1) {
            p=p+4;
            u=(char*)strtok_r(p,";", &strtok_cb);
            if(u != NULL)
            {
                pWS->lFAPLevel=atol(u);
                if (pWS->nDebugFlag >= LOGDEBUG)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"FAP Level check set to [%ld]",pWS->lFAPLevel);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDEBUG, pWS->szTmpBufff, pWS);
                }
            }
        }
    }

    /*
    * Add additional tags here

     ---->

    */

    /*
    * Record a warning - lets not worry if parameter is wrong
    */
    if(bFound==0)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Error - Unknown Tag [%s], connection refused", pszData);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        if (pWS->ExitReason==MQXR_INIT) {                  /* we just report it on MQXR_INTI */
            LogData(LOGDATA, pWS->szTmpBufff, pWS);
#if defined(__MVS__)
            WTODataE("19E", pWS);
#endif
        }
        return SEC_FAILURE;                                /* Newer ever accept bad parameters */
    }
    return SEC_OK;
}

/**
*  ProcessFile()
*
*
*
* Open file and call the ProcessData() method to process the options.
* Return failure for file errors or data processing errors.
*
*
* @return int  SEC_OK or SEC_FAILURE
*/
int ProcessFile( char *p, workingStorage_type *pWS)
{
    int  bFound;
    int  nRetCode=SEC_OK;                                  /* parameters are OK */
    int  i;
    char *u;
    char *strtok_cb;
    char szFileName[32];                                   /* Input Data file name */
    FILE *phInFile;                                        /* Declare file handles for input and output   */
    char *result;
    char line[MAX_LEN];
    char cline[MAX_LEN];
    long ErrorReported;

#if defined(__MVS__)
    long len, reason, tPtr;                                /* 2.84 start */
    int level=IEANT_HOME_LEVEL;
    int rc;
    long token[4]={0,0,0,0};                               /* 16 byte */
    int persist=IEANT_NOPERSIST;
    int nTokenLoad=0;
    char username[]="BLOCKIP2         ";
    char *p1, *p2;
    struct line_list {
        char *line;
        struct line_list *next;
    };
    typedef struct line_list line_item;
    line_item *lPtr;
#endif                                                     /* 2.84 end */

    /* snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"ProcessFile() Entered");
       pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
       LogData(LOGDEBUG, pWS->szTmpBufff);
    */

    pWS->bThisQmgr=1;                                      /* Assure compability with pre 2.60 config files */
    pWS->bThisChannelOK=1;                                 /* Assure compability with pre 2.60 config files */


    szFileName[0]  = 0;
    if( strchr(p,';') == 0 )
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, invalid file name pattern specified: FN=<filename...>; [%s], trailing ';' missing.", pWS->Pattern);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData( LOGDATA, pWS->szTmpBufff, pWS );
#if defined(__MVS__)
        WTODataE("20E", pWS);
#endif
        return SEC_FAILURE;
    }

    u=(char*)strtok_r(p,";", &strtok_cb);
    memcpy(szFileName,u,strlen(u));
    szFileName[strlen(u)] = 0x00;

#if defined(__MVS__)
    if (pWS->nZOSLoadMem == 1)
    {                                                       /* Use memory efficiency */
        p1 = strchr(szFileName,'(') + 1;                    /* Pull member name */
        p2 = strchr(szFileName,')');
        len = p2 - p1;
        memcpy(username+8, p1, len);                        /* place in for internal name */
        username[16]=0x00;
        IEANTRT(&level, username, &token, &rc);             /* try to retrieve token */
        if (rc == IEANT_NOT_FOUND)
        {                                                   /* token not found, create it */
            len = sizeof(line_item);
            BLOCKIPM((long*)&lPtr, &len, "G", &reason);     /* GETMAIN token line_item */
            BLOCKIPM((long*)&lPtr->next, &len, "G", &reason); /* GETMAIN real line_item */
            lPtr->line = NULL;
            memcpy(&token, &lPtr, sizeof(lPtr));            /* move pointer of dummy line_item into token */
            IEANTCR(&level, username, &token, &persist, &rc); /* save it as an address space token */
            if (rc != IEANT_OK)
            {                                               /* if we fail, abend it */
                WTODataE("26E", pWS);
                return SEC_FAILURE;
            }
            nTokenLoad = 1;                                 /* indicate new memory load */
            if (pWS->nDebugFlag >= LOGDEBUG8)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Opened new token [%s]",username);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDEBUG8, pWS->szTmpBufff, pWS);
            }
            lPtr = lPtr->next;                              /* Load up lPtr to be first line */
            lPtr->next = NULL;                              /* Null up next pointer */
            lPtr->line = NULL;                              /* Null ip line pointer */
        }
        else if (rc == IEANT_OK)
        {                                                   /* Have address space token already */
            memcpy(&lPtr, &token, sizeof(lPtr));            /* get pointer to first line */
            lPtr = lPtr->next;
            nTokenLoad = 0;                                 /* flag in memory data */
            if (pWS->nDebugFlag >= LOGDEBUG8)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Opened old token [%s]",username);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDEBUG8, pWS->szTmpBufff, pWS);
            }
        }
        else
        {                                                   /* bad return code from initial retrieve */
            WTODataE("27E", pWS);
            return SEC_FAILURE;
        }
    }

    if (pWS->nZOSLoadMem == 0 || nTokenLoad == 1)
    {                                                       /* not efficiency or loading initial memory */
#endif
        if (pWS->nDebugFlag >= LOGDEBUG)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Opening file [%s]",szFileName);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDEBUG, pWS->szTmpBufff, pWS);
        }

        phInFile = fopen( szFileName, "r" );
        if( phInFile == NULL )
        {
            ErrorReported=errno;
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, Specified file [%s] was not found.", szFileName);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData( LOGDATA, pWS->szTmpBufff, pWS);
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, Reason: %s", strerror(ErrorReported));
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData( LOGDATA, pWS->szTmpBufff, pWS);

#if defined(__MVS__)
            WTODataE("21E", pWS);
#endif
            /*        fclose( phInFile ); 2.67 removal */
            return SEC_FAILURE;
        }

#if defined(__MVS__)                                       /* 2.84 */
    } /* not efficiency or loading initial memory             2.84 */
#endif                                                     /* 2.84 */

    pWS->Pattern[0] = 0x00;                                /* NWC start with empty pattern, not *; */
    pWS->Users[0] = 0x00;
    pWS->BlockUsers[0] = 0x00;                             /* NWC allow for multiple lines */
    do
    {
#if defined(__MVS__)
        if (pWS->nZOSLoadMem == 1)
        {                                                   /* Memory efficiency requested */
            if (nTokenLoad == 0)
            {                                               /* token already, use memory data */
                if (lPtr == NULL)
                {                                           /* Current line_item NULL, end of chain */
                    result = NULL;                          /* indicate via result */
                }
                else
                {
                    strncpy(line, lPtr->line, MAX_LEN);     /* copy line                  */
                    result = line;                          /* point result (fgets)       */
                    lPtr = lPtr->next;                      /* move to next line          */
                    if (pWS->nDebugFlag >= LOGDEBUG8)
                    {
                        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Read cached line: [%s]",line);
                        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                        LogData(LOGDEBUG8, pWS->szTmpBufff, pWS);
                    }
                }
            }
            else
            {                                               /* we're loading memory       */
                result = fgets(line,MAX_LEN,phInFile);      /* get line from file         */
                if (result != NULL)
                {
                    i=strlen(line);                         /* Tidy up lines              */
                    /* 2.91 */
                    if(i==0) { continue; }
                    if(line[i-1] == '\n')                   /* last char is a new-line    */
                        line[i-1]  = '\0';                  /* replace new-line with null */
                    memset(cline,' ',i);
                    cline[i]=0x00;
                    /*
                    * Skip comment and empty lines - just process for possible data
                    */
                    if((line[0] != '#') && (strcmp(line,cline) != 0))
                    {
                        len = MAX_LEN;
                        BLOCKIPM((long*)&lPtr->line, &len, "G", &reason); /* GETMAIN a line for the chain */
                        strncpy(lPtr->line, line, MAX_LEN);     /* copy line data */
                        len = sizeof(line_item);                /* get length of line linker */
                        BLOCKIPM((long*)&lPtr->next, &len, "G", &reason); /* get memory for next line_item */
                        lPtr = lPtr->next;                      /* point at the next memory area */
                        lPtr->next = NULL;                      /* null out next pointer */
                        lPtr->line = NULL;                      /* null out line pointer */
                        if (pWS->nDebugFlag >= LOGDEBUG8)
                        {
                            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Cached new line: [%s]",line);
                            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                            LogData(LOGDEBUG8, pWS->szTmpBufff, pWS);
                        }
                    }
                }
            }
        }
        else
        {                                                   /* not caching to memory */
#endif
            result = fgets(line,MAX_LEN,phInFile);
#if defined(__MVS__)
        }
#endif
        if(result != NULL)
        {
            bFound=0;
            i=(int)strlen(line);
            /* 2.91 */
            if(i==0) { continue; }
            if(line[i-1] == '\n')                          /* last char is a new-line */
            {
                line[i-1]  = '\0';                         /* replace new-line with null */
                --i;                                       /* reduce buffer length */
            }
            memset(cline,' ',i);
            cline[i]=0x00;
            /*
            * Skip comment and empty lines - just process for possible data
            */
            if((line[0] != '#') && (strcmp(line,cline) != 0))
            {
                if(ProcessData( line, pWS )==SEC_FAILURE)
                    nRetCode=SEC_FAILURE;
            }
            else
                bFound=1;                                  /* Comment lines are allways processed */
        }
    } while (result != NULL);
    if (pWS->Pattern[0] == 0x00)                           /* NWC If pattern is NULL, use *; */
    {
        strcpy(pWS->Pattern,"*;");
    }
    fclose( phInFile );
    return nRetCode;
}


#if defined(UNIX) || defined (OS400)
/**
*  ProcessMQSiniFile()
*
*
*
* Open MQS.ini file and process the configuration specifications.
* Return failure for file errors or data processing errors.
*
*
* @return int  SEC_OK or SEC_FAILURE
* function added under 2.45
*/
int ProcessMQSiniFile(char *szQmgrName, workingStorage_type *pWS, ws_type *pWS1)
{
    int  bFound;
    int  nRetCode=SEC_OK;                                  /* parameters are OK */
    int  i;
    char szFileName[32];                                   /* Input Data file name */
    FILE *phInFile;                                        /* Declare file handles for input and output   */
    char *result;
    char line[MAX_LEN];                                    /* Changed to MAX_LEN 2.85 */
    char sline[MAX_LEN];                                   /* Changed to MAX_LEN 2.85 */
    char cline[MAX_LEN];                                   /* Changed to MAX_LEN 2.85 */
    char szQmgrDirectory[255] = "";
    char szQmgrPrefix[255] = "";
    char szDataPath[255] = "";                             /* 2.83 */
    char * p;
    int  groupid=0;
    int  bfoundQmgr=0;
    int  bQMFound=0;
    long ErrorReported;

    char szQmgrDirectoryW[255] = "";                       /* 2.83 */
    char szQmgrPrefixW[255] = "";                          /* 2.83 */
    char szDataPathW[255] = "";                            /* 2.83 */


#define MQINI_GROUP_NotDefined       0
#define MQINI_GROUP_AllQueueManagers 1
#define MQINI_GROUP_ClientExitPath   0
#define MQINI_GROUP_QueueManager     4
#define MQINI_GROUP_LogDefaults      0

    if (pWS->nDebugFlag >= LOGDEBUG8)
        LogData(LOGDEBUG8, "  ProcessMQSiniFile() Entered", pWS);
#if MQAT_DEFAULT==MQAT_OS400
    strcpy(szFileName,"/QIBM/UserData/mqm/mqs.ini");
#else
    strcpy(szFileName,"/var/mqm/mqs.ini");
#endif
    phInFile = fopen( szFileName, "r" );
    if( phInFile == NULL )
    {
        ErrorReported=errno;
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, Specified file [%s] was not found.", szFileName);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData( LOGDATA, pWS->szTmpBufff, pWS);
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, Reason: %s", strerror(ErrorReported));
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData( LOGDATA, pWS->szTmpBufff, pWS);
        fclose( phInFile );
        return SEC_FAILURE;
    }
    if (pWS->ExitReason==MQXR_INIT)
    {
        if (pWS->nDebugFlag >= LOGDEBUG8)
            LogData(LOGDEBUG8, "    mqs.ini opened successful.", pWS);
    }
    do
    {
        result = fgets(sline,MAX_LEN,phInFile);
        if(result != NULL)
        {
            p=Strip(sline);
            strcpy(line,p);
            bFound=0;
            i=strlen(line);
            /* 2.91 */
            if(i==0) { continue; }
            if(line[i-1] == '\n')                          /* last char is a new-line */
            {
                line[i-1]  = '\0';                         /* replace new-line with null */
                --i;                                       /* reduce buffer length */
            }
            memset(cline,' ',i);
            cline[i]=0x00;
            /*
            * Skip comment and empty lines - just process for possible data
            */
            if((line[0] != '#') && (strcmp(line,cline) != 0))
            {
                if (line[strlen(line)-1]==':')
                {
                    /* 2.83 start */
                    if (bfoundQmgr==1)                     /* we have retrieved the qmgr info, now move it */
                    {
                        strcpy(szQmgrDirectory,szQmgrDirectoryW);
                        strcpy(szQmgrPrefix,szQmgrPrefixW);
                        strcpy(szDataPath,szDataPathW);
                    }
                    /* 2.83 end   */
                    bFound=0;
                    bfoundQmgr=0;
                    groupid=MQINI_GROUP_NotDefined;
                }
                if (!strcmp(line,"AllQueueManagers:"))
                {
                    bFound=1;
                    groupid=MQINI_GROUP_AllQueueManagers;
                }
                if (!strcmp(line,"ClientExitPath:"))
                {
                    bFound=1;
                    groupid=MQINI_GROUP_ClientExitPath;
                }
                if (!strcmp(line,"LogDefaults:"))
                {
                    bFound=1;
                    groupid=MQINI_GROUP_LogDefaults;
                }
                if (!strcmp(line,"QueueManager:"))
                {
                    bFound=1;
                    groupid=MQINI_GROUP_QueueManager;
                    strcpy(szQmgrDirectoryW,"");                /*                          2.90 */
                    strcpy(szQmgrPrefixW,"");                   /*                          2.90 */
                    strcpy(szDataPathW,"");                     /*                          2.90 */
                }
                if (pWS->ExitReason==MQXR_INIT)
                {
                    if (pWS->nDebugFlag >= LOGDEBUG8)
                    {
                        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"      Line read %i %s",groupid, line);
                        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                        LogData( LOGDEBUG8, pWS->szTmpBufff, pWS);
                    }
                }

                /*if (groupid==MQINI_GROUP_QueueManager && bfoundQmgr==1) 2.83 chanhed to : */
                if (groupid==MQINI_GROUP_QueueManager)
                {
                    p = strstr(line,"Prefix=");
                    if(p != NULL )
                    {
                        p=p+7;
                        strcpy(szQmgrPrefixW,p);
                        if (pWS->ExitReason==MQXR_INIT)
                        {
                            if (pWS->nDebugFlag >= LOGDEBUG8)
                                LogData(LOGDEBUG8,"      QmgrPrefix found",pWS);
                        }
                    }
                    p = strstr(line,"Directory=");
                    if(p != NULL )
                    {
                        p=p+10;
                        strcpy(szQmgrDirectoryW,p);
                        if (pWS->ExitReason==MQXR_INIT)
                        {
                            if (pWS->nDebugFlag >= LOGDEBUG8)
                                LogData(LOGDEBUG8,"      QmgrDirectory found",pWS);
                        }
                    }
                    p = strstr(line,"DataPath=");                  /* 2.83 start */
                    if(p != NULL )
                    {
                        p=p+9;
                        strcpy(szDataPathW,p);
                        if (pWS->ExitReason==MQXR_INIT)
                        {
                            if (pWS->nDebugFlag >= LOGDEBUG8)
                                LogData(LOGDEBUG8,"      DataPath found",pWS);
                                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"      DataPath: %s",szDataPathW);
                                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                                LogData( LOGDEBUG8, pWS->szTmpBufff, pWS);
                        }
                    }                                              /* 2.83 end */
                }

                if (groupid==MQINI_GROUP_QueueManager && bFound==0)
                {
                    if (pWS->ExitReason==MQXR_INIT)
                    {
                        if (bQMFound==0)
                        {
                            if (pWS->nDebugFlag >= LOGDEBUG8)
                                LogData(LOGDEBUG8,"      Now we have the queueManager token...",pWS);
                            bQMFound=1;
                        }
                    }
                    p = strstr(line,"Name=");
                    if(p != NULL )
                    {
                        p=p+5;
                        if (!strcmp(p,szQmgrName)) {
                            bfoundQmgr=1;
                            if (pWS->ExitReason==MQXR_INIT) {

                                if (pWS->nDebugFlag >= LOGDEBUG8)
                                    LogData(LOGDEBUG8,"    Wanted qmgr found",pWS);
                            }
                        }
                    }
                }
            }
            else
                bFound=1;                                  /* Comment lines are allways processed */
        }
    } while (result != NULL);
    fclose( phInFile );
    /* 2.83 start */
    if (bfoundQmgr==1)                     /* we have retrieved the qmgr info (last qmgr in list), now move it */
    {
        if (pWS->ExitReason==MQXR_INIT)
        {
            if (pWS->nDebugFlag >= LOGDEBUG8)
                LogData(LOGDEBUG8, "    Last qmgr....", pWS);
        }
        strcpy(szQmgrDirectory,szQmgrDirectoryW);
        strcpy(szQmgrPrefix,szQmgrPrefixW);
        strcpy(szDataPath,szDataPathW);
    }
    /* 2.83 end   */

    if (pWS->ExitReason==MQXR_INIT) {
        if (pWS->nDebugFlag >= LOGDEBUG8)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"    Resulting path: %s/qmgrs/%s",szQmgrPrefix,szQmgrDirectory);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData( LOGDEBUG8, pWS->szTmpBufff, pWS);
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"    Resulting DataPath: %s",szDataPath);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData( LOGDEBUG8, pWS->szTmpBufff, pWS);
        }
    }

#if defined UNIX || defined (OS400)
    if (strlen(szDataPath)>0)                          /* Standby is configured....      2.90 */
    {
        pWS->iStbyMode=1;                              /* Standby enabled                2.90 */
    }
#endif

    strcpy(pWS->szQmgrPrefix,szQmgrPrefix);
    strcpy(pWS->szQmgrDirectory,szQmgrDirectory);
    strcpy(pWS->szDataPath,szDataPath);                    /* 2.83 */
    strcpy(pWS1->szQmgrPrefix,szQmgrPrefix);
    strcpy(pWS1->szQmgrDirectory,szQmgrDirectory);
    strcpy(pWS1->szDataPath,szDataPath);                   /* 2.83 */

    if (pWS->nDebugFlag >= LOGDEBUG8)
        LogData(LOGDEBUG8, "  ProcessMQSiniFile() Done", pWS);

    return nRetCode;
}

/**
*  ProcessMQSiniFile_qm_ini()
*
*
*
* Open MQS.ini file and process the configuration specifications.
* Return failure for file errors or data processing errors.
*
*
* @return int  SEC_OK or SEC_FAILURE
* function added under 2.60
*/
int ProcessMQSiniFile_qm_ini(char *szQmgrName, workingStorage_type *pWS, ws_type *pWS1)
{
    int  bFound;
    int  nRetCode=SEC_OK;                                  /* parameters are OK */
    int  i;
    char szFileName[528];                                  /* Input Data file name 2.62 changed to 528 */
    FILE *phInFile;                                        /* Declare file handles for input and output   */
    char *result;
    char line[MAX_LEN];                                    /* Changed to MAX_LEN 2.86 */
    char sline[MAX_LEN];                                   /* Changed to MAX_LEN 2.86 */
    char cline[MAX_LEN];                                   /* Changed to MAX_LEN 2.86 */
    char szExitPath[255] = "";
    char szExitPath64[255] = "";
    char * p;
    int  groupid=0;
    long ErrorReported;

#define MQINI_GROUP_ExitPath         1
#define MQINI_GROUP_Log              0
#define MQINI_GROUP_Service          0
#define MQINI_GROUP_ServiceComponent 0


    if (pWS->nDebugFlag >= LOGDEBUG8)
        LogData(LOGDEBUG8, "  ProcessMQSiniFile_qm_ini() Entered", pWS);

    strcpy(szFileName,pWS->szQmgrPrefix);
    strcat(szFileName,"/qmgrs/");
    strcat(szFileName,pWS->szQmgrDirectory);
    if (strlen(pWS->szDataPath) > 0 )                      /* DataPath was used, use it 2.83 */
    {                                                      /*                           2.83 */
        strcpy(szFileName,pWS->szDataPath);                /*                           2.83 */
    }                                                      /*                           2.83 */
#if defined(UNIX) || defined (OS400)
    strcpy(pWS->szMasterFile,szFileName);                  /* Build the master path     2.90 */
    strcat(pWS->szMasterFile,"/master");                   /*                           2.90 */
#endif
    strcat(szFileName,"/qm.ini");

    phInFile = fopen( szFileName, "r" );
    if( phInFile == NULL )
    {
        if (pWS->ExitReason==MQXR_INIT) {
            ErrorReported=errno;
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, Specified file [%s] was not found.", szFileName);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData( LOGDATA, pWS->szTmpBufff, pWS);
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, Reason: %s", strerror(ErrorReported));
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData( LOGDATA, pWS->szTmpBufff, pWS);
        }
        fclose( phInFile );
        if (pWS->nDebugFlag >= LOGDEBUG8)
            LogData(LOGDEBUG8, "  ProcessMQSiniFile_qm_ini() Done", pWS);
        return SEC_FAILURE;
    }
    do
    {
        result = fgets(sline,MAX_LEN,phInFile);
        if(result != NULL)
        {
            p=Strip(sline);
            strcpy(line,p);
            bFound=0;
            i=strlen(line);
            /* 2.91 */
            if(i==0) { continue; }
            if(line[i-1] == '\n')                          /* last char is a new-line */
            {
                line[i-1]  = '\0';                         /* replace new-line with null */
                --i;                                       /* reduce buffer length */
            }
            memset(cline,' ',i);
            cline[i]=0x00;
            /*
            * Skip comment and empty lines - just process for possible data
            */
            if((line[0] != '#') && (strcmp(line,cline) != 0))
            {
                if (line[strlen(line)-1]==':')
                {
                    bFound=0;
                    groupid=MQINI_GROUP_NotDefined;
                }
                if (!strcmp(line,"ExitPath:"))
                {
                    bFound=1;
                    groupid=MQINI_GROUP_ExitPath;
                }
                if (!strcmp(line,"Log:"))
                {
                    bFound=1;
                    groupid=MQINI_GROUP_Log;
                }
                if (!strcmp(line,"Service:"))
                {
                    bFound=1;
                    groupid=MQINI_GROUP_Service;
                }
                if (!strcmp(line,"ServiceComponent:"))
                {
                    bFound=1;
                    groupid=MQINI_GROUP_ServiceComponent;
                }

                if (pWS->ExitReason==MQXR_INIT) {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"    Line read %i %s",groupid, line);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    if (pWS->nDebugFlag >= LOGDEBUG8)
                        LogData( LOGDEBUG8, pWS->szTmpBufff, pWS);
                }

                if (groupid==MQINI_GROUP_ExitPath) {
                    p = strstr(line,"ExitsDefaultPath=");
                    if(p != NULL )
                    {
                        p=p+17;
                        strcpy(szExitPath,p);
                        if (pWS->ExitReason==MQXR_INIT) {
                            if (pWS->nDebugFlag >= LOGDEBUG8)
                                LogData(LOGDEBUG8,"      ExitPath found",pWS);
                        }
                    }
                    p = strstr(line,"ExitsDefaultPath64=");
                    if(p != NULL )
                    {
                        p=p+19;
                        strcpy(szExitPath64,p);
                        if (pWS->ExitReason==MQXR_INIT) {
                            if (pWS->nDebugFlag >= LOGDEBUG8)
                                LogData(LOGDEBUG8,"      ExitsDefaultPath64 found",pWS);
                        }
                    }
                }

            }
            else
                bFound=1;                                  /* Comment lines are allways processed */
        }
    } while (result != NULL);
    fclose( phInFile );
    if (pWS->ExitReason==MQXR_INIT) {
        if (pWS->nDebugFlag >= LOGDEBUG8)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"    Resulting ExitPath paths: %s 64: %s",szExitPath,szExitPath64);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData( LOGDEBUG8, pWS->szTmpBufff, pWS);
        }
    }
    strcpy(pWS->ExitPath,szExitPath);

    if (pWS->nDebugFlag >= LOGDEBUG8)
        LogData(LOGDEBUG8, "  ProcessMQSiniFile_qm_ini() Done", pWS);

    return nRetCode;
}


#endif

/**
*  CopyUsers()
*  added JHP 2004-03-30
*
*
* Copy userids from the passed pointer into Users[], this is to help
*	   with backward compability
* Return failure if problem with parameter.
*
*
* @return int  SEC_OK or SEC_FAILURE
*/

int CopyUsers(char *p, workingStorage_type *pWS)
{
    char *u;
    char *strtok_cb;

    p=p+8;
    u=(char*)strtok_r(p,";", &strtok_cb);
    if(u != NULL)
    {
        if(strlen(u) >= MAX_PL)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, Users string is too long, max[%d], was [%ld]",MAX_PL-1,(long)strlen(u));
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA, pWS->szTmpBufff, pWS);
#if defined(__MVS__)
            WTODataE("22E", pWS);
#endif
            return SEC_FAILURE;
        }
        strcpy(pWS->Users,u);
        pWS->Users[strlen(u)] = 0x00;
        if (pWS->bUseridUpperLowerCase==1)                 /* 2.12 */
        {
            if (pWS->nDebugFlag >= LOGDEBUG1)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Users Before strupr [%s]",pWS->Users);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
            }
            i_strupr(pWS->Users);
            if (pWS->nDebugFlag >= LOGDEBUG1)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Users After strupr [%s]",pWS->Users);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
            }
        }
    }
    return SEC_OK ;
}


/**
* GenerateLogFileName(char * Buffer)
*
*
* Method to geterate logfile name
*
*
* @return char *
*/
char * GenerateLogFileName(char * szFile, workingStorage_type *pWS)
{
    struct tm *timenow;
#ifdef _THREAD_SAFE_JHP
    struct tm  wk_localtime;
#endif

    time_t longtime;

    char szBuff1[64];
    char szDate[32];
    /*    char szFile[256];*/

    char *p = NULL;
    int  iFirst=1;

    time(&longtime);
#ifdef _THREAD_SAFE_JHP
    memset(&wk_localtime, '\0', sizeof(wk_localtime));
    timenow = localtime_r(&longtime,&wk_localtime);
#else
    timenow = localtime(&longtime);
#endif
    snprintf(szDate,sizeof(szDate),"%04d-%02d-%02d", timenow->tm_year+1900, timenow->tm_mon+1,timenow->tm_mday);
    szDate[sizeof(szDate)-1]='\00';

    if (pWS->nLogFormat==0)
    {
       if (pWS->NoLogFileName==1)
       {
          pWS->NoLogFileName=0;                         /* Can only be used with fmt. 2.95 */
       }
    }
    memset(szBuff1,'\00',sizeof(szBuff1));              /* Clear buffer               2.95 */
    if (pWS->NoLogFileName==0)                                /* Include LogFileName  2.95 */
    {
    	snprintf(szBuff1,sizeof(szBuff1),"%s", pWS->szLogFileName);
    	iFirst=0;                                      /* Not first anymore           2.95 */
    }
    szBuff1[sizeof(szBuff1)-1]='\00';
#if ! defined (__MVS__)
    /* 2.80 start */
    if(pWS->nLogFormat & LOGQMANAGER)
    {
        if (iFirst==0)                                 /* First ?                     2.95 */
        {
            strcat(szBuff1, "_");
            iFirst=0;                                  /* Not first anymore           2.95 */
        }
        strcat(szBuff1,(char *)Strip(pWS->szQMgr));
    }
    /* 2.80 end */
    if(pWS->nLogFormat & LOGCHANNEL)
    {
        if (iFirst==0)                                 /* First ?                     2.95 */
        {
            strcat(szBuff1, "_");
            iFirst=0;                                  /* Not first anymore           2.95 */
        }
        strcat(szBuff1,pWS->szLogChannel);
    }
    if(pWS->nLogFormat & LOGUSER)
    {
        if (iFirst==0)                                 /* First ?                     2.95 */
        {
            strcat(szBuff1, "_");
            iFirst=0;                                  /* Not first anymore           2.95 */
        }
        strcat(szBuff1,pWS->RemUID);
    }
    if(pWS->nLogFormat & LOGDATE)
    {
        if (iFirst==0)                                 /* First ?                     2.95 */
        {
            strcat(szBuff1, "_");
            iFirst=0;                                  /* Not first anymore           2.95 */
        }
        strcat(szBuff1,szDate);
    }
#if defined (__EXT_LOG_FILENAME__)                         /* Only UNIX, LINUX and Windows */
    if (pWS->bFileSizeCheck==0)
        strcat(szBuff1,"001");                             /* make room for file rename */
#endif
    strcat(szBuff1, ".");
    strcat(szBuff1, pWS->szLogFileExtension);
#endif
    /*
    * Now have file to save to - now check path
    */
    strcpy(szFile,"");

#if defined (WIN32)
    strcat(szFile,pWS->szLogDrive);
    if( szFile[1] != ':') strcat(szFile, ":");
    strcat(szFile,pWS->szLogPath);
    strcat(szFile,"\\");
#elif defined (__MVS__)
    /* NWC No drive or path required on MVS */
    strcpy(szFile,"");
#else
    if (pWS->szLogPath[0] != '/')                          /* Fix bad spec JHP 040522 */
    {
        strcpy(szFile,"/");
    }
    strcat(szFile,pWS->szLogPath);
    if (szFile[strlen(szFile)-1] != '/')                   /* Fix bad spec JHP 040522 */
    {
        strcat(szFile,"/");
    }
#endif

    strcat(szFile,szBuff1);

#if MQAT_DEFAULT==MQAT_OS400
#ifdef JHP
    strcat(szFile,pWS->szLogPath);
    if (szFile[strlen(szFile)-1] != '/')                   /* insert seperator */
    {
        strcat(szFile,"/");
    }
    strcat(szFile, ".");
    strcat(szFile, pWS->szLogFileExtension);
#endif
#endif

    p=szFile;
    return p;
}


/**
* LogData()
*
*
* Method to log data in a specific way
*
*	  iMode 0 = Normal 1= Debug String
*	 szData ASCII data to report
*
*
* @return void
*/
void LogData(int iMode, char * szData, workingStorage_type *pWS)
{
    FILE *fptr;
#ifdef _THREAD_SAFE_JHP
    struct tm  wk_localtime;
#endif
    struct tm *timenow;
    time_t longtime;

    char   szBuff1[64];

    char szDate[32];
    char szTime[32];
    char szFile[256];
    char szFile2[256];

#if defined (WIN32)
    DWORD dwWaitResult;
    long lCount;
#endif
    long ErrorReported;

    /* skip debug messages if not in debug mode*/
    /*    if((iMode==LOGDEBUG) &&
                (pWS->nDebugFlag!=LOGDEBUG)) return;*/

    if ((iMode > (int)LOGDATA) && (pWS->nDebugFlag < iMode)) return; /*2.62*/

    /* Start JHP060215 */
#if defined (WIN32)
    /* Send the message to the event viewer */
    if ((pWS->lWinLog==LOG_BOTH) | (pWS->lWinLog==LOG_EVTLOGONLY))
    {
        SendLogMessage(szData, pWS);
        if (pWS->lWinLog==LOG_EVTLOGONLY)
        {
            return;
        }
    }
#endif
    /*End JHP060214 */

    /* Start 2.49 */
#if defined (UNIX)
    /* Send the message to the event viewer */
    if ((pWS->lWinLog==LOG_BOTH) | (pWS->lWinLog==LOG_EVTLOGONLY))
    {
        SendLogMessage(szData, pWS);
        if (pWS->lWinLog==LOG_EVTLOGONLY)
        {
            return;
        }
    }
#endif
    /*End 2.49 */

#if defined (__MVS__)
    translate (szData, "","", pWS);                        /* remove [] from the list JHP 040805 */
#endif

    /*
    * Process normal string or debug string with debug flag active
    */

    time(&longtime);
#ifdef _THREAD_SAFE_JHP
    memset(&wk_localtime, '\0', sizeof(wk_localtime));
    timenow = localtime_r(&longtime,&wk_localtime);
#else
    timenow = localtime(&longtime);
#endif
    sprintf(szDate,"%04d-%02d-%02d", timenow->tm_year+1900, timenow->tm_mon+1,timenow->tm_mday);
    sprintf(szTime,"%02d:%02d:%02d",timenow->tm_hour,timenow->tm_min,timenow->tm_sec);
    sprintf(szBuff1,"%s", pWS->szLogFileName);

#if MQAT_DEFAULT==MQAT_OS400  /* test added */
    translate (szData, "","", pWS);	   /* remove [] from the list JHP 040805 */
    strcpy(szFile,GenerateLogFileName(szFile2,pWS));
    if((fptr = fopen( szFile,LOGMODE))!=NULL)
    {
        fprintf(fptr,szDate);
        fprintf(fptr,"|");
        fprintf(fptr,szTime);
        fprintf(fptr,"|");
        if (pWS->bIncludeQmgrNameInLog==1)
        {
            fprintf(fptr,pWS->szShortQMName);
            fprintf(fptr,"|");
        }
        fprintf(fptr,szData);
        fprintf(fptr,"\n");
        fclose(fptr);
    }

#else  /* End of OS400 */

    strcpy(szFile,GenerateLogFileName(szFile2,pWS));

#if defined (WIN32)
    OpenMTX(pWS,szFile);                                   /* See if the Mutex is in place */
    /* Wait for the mutex before writing the logevent, to prevent problems with the file. Just one is allowed.*/
    lCount=0;
    while( ( dwWaitResult = WaitForSingleObject( pWS->hMutex, 5L ) ) == WAIT_TIMEOUT)
        lCount++;
#endif
#if defined (UNIX) || defined (OS400)
/*    DBGPRINTF((OST, "  LLS -- Debugger %d\n",__LINE__));*/
    LockLogSem(pWS);
#endif
    if((fptr = fopen( szFile,LOGMODE))!=NULL)
    {
        fprintf(fptr,szDate);
        fprintf(fptr, "|");
        fprintf(fptr,szTime);
#if defined(WIN) || defined(UNIX)
        fprintf(fptr, "|");
        {
            char w[32];
            /*        	sprintf(w,"%d",GetCurrentProcessId());*/
#if defined(WIN32)
            sprintf(w,"%5d",GetCurrentThreadId());
#elif defined(UNIX)
    sprintf(w,"%5ld",(long unsigned int) pthread_self());
#endif
            fprintf(fptr,w);
        }
#endif
        fprintf(fptr, "|");

#if defined (UNIX) || defined (WIN32) || defined (OS400)
        if (pWS->bIncludeQmgrNameInLog ==1)
        {
            fprintf(fptr,pWS->szShortQMName);
            fprintf(fptr,"|");
        }
#endif
        fprintf(fptr, szData);
#if defined (WIN32)
        if (lCount >0 )
            fprintf(fptr, " Waited %ld ms on the logfile",lCount*5);
#endif
        fprintf(fptr, "\n");
        fclose(fptr);
    }
    else                                                   /* If a failure in specified path, use default specs. */
    {
        ErrorReported=errno;
#if defined (WIN32)
        strcpy(szFile2,"c:\\BlockIP2.log");
#elif ! defined (__MVS__)
    strcpy(szFile2,"/var/mqm/exits/BlockIP2.log");
#else
    /* NWC Use DD statement format to hook to JCL defined files on MVS */
    strcpy(szFile2,"//DD:BLOCKLOG");
#endif
#if MQAT_DEFAULT==MQAT_OS400
        strcpy(szFile2,"QIBM/UserData/BLOCKIP2.LOG");
#endif

        if((fptr = fopen( szFile2,LOGMODE))!=NULL)
        {
            if (pWS->ReportLogFileError==0)
            {
                pWS->ReportLogFileError=1;                 /* not first anymore */
                fprintf(fptr,szDate);
                fprintf(fptr, "|");
                fprintf(fptr,szTime);
                fprintf(fptr, "|");
#if ! defined (__MVS__)
                fprintf(fptr, "BlockIP2 failed to open the specified logfile [%s], used default. Reason: %s\n",szFile,strerror(ErrorReported));
#else
    if (pWS->nZOSconv==1)
        fprintf(fptr, "BlockIP2 failed to open the specified logfile <%s>, used default. Reason: %s\n",szFile,strerror(ErrorReported));
    else
        fprintf(fptr, "BlockIP2 failed to open the specified logfile [%s], used default. Reason: %s\n",szFile,strerror(ErrorReported));
#endif
            }

            fprintf(fptr,szDate);
            fprintf(fptr, "|");
            fprintf(fptr,szTime);
            fprintf(fptr, "|");
            fprintf(fptr, szData);
#if defined (WIN32)
            fprintf(fptr, " Waited %ld ms on the logfile",lCount*5);
#endif
            fprintf(fptr, "\n");
            fclose(fptr);
        }
        else                                               /* If a failure in default path, use stdout specs. */
        {
#if defined (__TO_BE_FIXED_)
            if (pWS->ReportLogFileError==0)
            {
                pWS->ReportLogFileError=1;                 /* not first anymore */
                fprintf(fptr,szDate);
                fprintf(fptr, "|");
                fprintf(fptr,szTime);
                fprintf(fptr, "|");
#if ! defined (__MVS__)
                fprintf(fptr, "BlockIP2 failed to open the specified logfile [%s], used default. Reason: %s\n",szFile,strerror(ErrorReported));
#else
    if (pWS->nZOSconv==1)
        fprintf(fptr, "BlockIP2 failed to open the specified logfile <%s>, used default. Reason: %s\n",szFile,strerror(ErrorReported));
    else
        fprintf(fptr, "BlockIP2 failed to open the specified logfile [%s], used default. Reason: %s\n",szFile,strerror(ErrorReported));
#endif
            }

            printf(szDate);
            printf("|");
            printf(szTime);
            printf("|");
            printf(szData);
            printf("\n");
            fclose(fptr);
#endif
        }
    }
#if defined (UNIX) || defined (OS400)
/*    DBGPRINTF((OST, "  FLS -- Debugger %d\n",__LINE__));*/
    FreeLogSem(pWS);
#endif
#if defined (WIN32)
    ReleaseMutex(pWS->hMutex);
    CloseHandle(pWS->hMutex);
    pWS->hMutex = (void*)ERROR_INVALID_HANDLE;
#endif
#endif    /* NOT AS/400 */
}

#if defined(__MVS__)
/**
* WTODataE()
*
*
* Method to log data to WTO
*
*	 szData ASCII data to report
*
*
* @return void
*
* added JHP 040806
*/
void WTODataE(char * szData, workingStorage_type *pWS)
{
    char * ptr;
    char szData1[512];

    bi2wto_type  wtomsg;                                   /* jhp 040805 */
    bi2wto_type  * pwtomsg;                                /* jhp 040805 */


    ptr = strcpy(szData1,pWS->szWTOPfx);                   /* jhp 2.76 */
    ptr = strcat(szData1,szData);
    ptr = strcat(szData1, " ");
    ptr = strcat(szData1, pWS->szWTOBuffErr );

    pwtomsg = &wtomsg;                                     /* jhp 040805 */

    memset(pwtomsg -> msg,' ',WTOLENGTH);
    if (strlen(szData1) > WTOLENGTH)
        memcpy(pwtomsg -> msg,
               szData1,
               WTOLENGTH) ;
    else
        memcpy(pwtomsg -> msg,
               szData1,
               strlen(szData1));
    BLOCKIPW(pwtomsg);
}

/**
* WTODataOk()
*
*
* Method to log data to WTO
*
*	 szData ASCII data to report
*
*
* @return void
*
* added JHP 040806
*/
void WTODataOk(char * szData, workingStorage_type *pWS)
{
    char * ptr;
    char szData1[512];

    bi2wto_type  wtomsg;                                   /* jhp 040805 */
    bi2wto_type  * pwtomsg;                                /* jhp 040805 */

    ptr = strcpy(szData1,pWS->szWTOPfx);                   /* jhp 2.76 */
    ptr = strcat(szData1,szData);
    ptr = strcat(szData1, " ");
    ptr = strcat(szData1, pWS->szWTOBuffOk );

    pwtomsg = &wtomsg;                                     /* jhp 040805 */

    memset(pwtomsg -> msg,' ',WTOLENGTH);
    if (strlen(szData1) > WTOLENGTH)
        memcpy(pwtomsg -> msg,
               szData1,
               WTOLENGTH) ;
    else
        memcpy(pwtomsg -> msg,
               szData1,
               strlen(szData1));
    BLOCKIPW(pwtomsg);
}
#endif

/**
* SendLogMessage()
*
*
* Method to log data to EventList on windows
*
*	 szData ASCII data to report
*
*
* @return void
*
* added JHP 060214
*/
#if defined (WIN32)
void SendLogMessage(char *szData, workingStorage_type *pWS)
{
    HANDLE  hSource;
    char    *szList[1];

    szList[0]=szData;
    hSource = RegisterEventSource(NULL, "WMQ BlockIP2");
    if (hSource != NULL)
    {
        ReportEvent(hSource,EVENTLOG_INFORMATION_TYPE,
                    0,1,NULL,1,0,(const unsigned char**)szList,NULL);
        DeregisterEventSource(hSource);
    }
}
#endif

/**
* SendLogMessage()
*
*
* Method to log data to EventList on windows
*
*	 szData ASCII data to report
*
*
* @return void
*
* added JHP 060214
*/
#if defined (UNIX)
void SendLogMessage(char *szData, workingStorage_type *pWS)
{
#if defined(AIX)
    struct syslog_data sdata = SYSLOG_DATA_INIT;

    openlog_r("BlockIP2", LOG_PID, pWS->iSyslogFclty, &sdata);
    syslog_r(pWS->iSyslogPrty,  &sdata, "%s", szData);
    closelog_r(&sdata);
#else
    openlog("BlockIP2", LOG_PID, pWS->iSyslogFclty);
    syslog(pWS->iSyslogPrty, "%s", szData);
    closelog();
#endif
}
#endif

void translate (char *src, char *ny, char *old, workingStorage_type *pWS)
{
    if (pWS->nZOSconv==1)
    {
        while (*src)
        {
            if (*src=='[')
                *src='<';
            if (*src==']')
                *src='>';
            src++;
        }
    }
}

void translate2 (char *src, char *ny, char *old)
{
    while (*src)
    {
        if (*src==':')
            *src='_';
        if (*src=='\\')
            *src='_';
        if (*src==' ')
            *src='_';
        src++;
    }
}

#if !defined(__MVS__)
/* Check the active connections */
int CheckRunningChannels(MQCD *pChannelDefinition, workingStorage_type *pWS)
{
    long lCurrChl;

    if (pWS->nDebugFlag >= LOGDEBUG1)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"CheckRunningChannels() Entered");
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
    }

#if defined UNIX || defined (OS400)
    lCurrChl = UpdateShrMem(pChannelDefinition,(long*)1,pWS);
    if (lCurrChl < 0 )                                     /* number of channels over limit */
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, Channel [%.*s] ConName [%.*s] User [%.*s] requested by CheckRunningChannels()",
                 rlnts(MQ_CHANNEL_NAME_LENGTH, pChannelDefinition->ChannelName),
                 pChannelDefinition->ChannelName,
                 rlnts(MQ_CONN_NAME_LENGTH, pWS->ConName),  /* NWC */
                 pWS->ConName,
                 rlnts(MQ_USER_ID_LENGTH, pChannelDefinition->RemoteUserIdentifier),
                 pChannelDefinition->RemoteUserIdentifier);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData( LOGDATA, pWS->szTmpBufff, pWS);
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"MAX active channels set on [%s] was exceeded, limit was [%ld] Active was [%d]. Connection was refused.",
                 pWS->szChannelName,pWS->lMaxChannelActive,pWS->iCurrentActiveConnections);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDATA, pWS->szTmpBufff, pWS);

        if (pWS->nDebugFlag >= LOGDEBUG1)
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"CheckRunningChannels() Done");

        return SEC_FAILURE;
    }
    if (lCurrChl == 0 )                                    /* Number of channels below limit, just return */
    {
        if (pWS->nDebugFlag >= LOGDEBUG1)
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"CheckRunningChannels() Done");
        return SEC_OK;
    }

#else   /* ! UNIX (and not MVS) */
#if defined WIN32_SHM
    lCurrChl = UpdateShrMem(pChannelDefinition,(long*)1,pWS);
    if (lCurrChl < 0 )                                     /* number of channels over limit */
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, Channel [%.*s] ConName [%.*s] User [%.*s] requested by CheckRunningChannels()",
                 rlnts(MQ_CHANNEL_NAME_LENGTH, pChannelDefinition->ChannelName),
                 pChannelDefinition->ChannelName,
                 rlnts(MQ_CONN_NAME_LENGTH, pWS->ConName),  /* NWC */
                 pWS->ConName,
                 rlnts(MQ_USER_ID_LENGTH, pChannelDefinition->RemoteUserIdentifier),
                 pChannelDefinition->RemoteUserIdentifier);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData( LOGDATA, pWS->szTmpBufff, pWS);
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"MAX active channels set on [%s] was exceeded, limit was [%ld] Active was [%ld]. Connection was refused.",
                 pWS->szChannelName,pWS->lMaxChannelActive,pWS->iCurrentActiveConnections);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDATA, pWS->szTmpBufff, pWS);


        return SEC_FAILURE;
    }
    if (lCurrChl == 0 )                                    /* Number of channels below limit, just return */
    {
        return SEC_OK;
    }

#else
    if (connectToQMgr(pWS) != MXC_RC_NONE)
    {
        return SEC_FAILURE;
    }
    if (pWS->nDebugFlag >= LOGDEBUG1)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connected to QueueManager [%s]",pWS->szQMgr);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
    }

    if (openStatusQ(pWS) != MXC_RC_NONE)
    {
        return SEC_FAILURE;
    }
    lCurrChl =  UpdateStatusQ(pChannelDefinition,(long*)-1,pWS);
    if (lCurrChl < 0 )                                     /* number of channels over limit */
    {
        if (closeStatusQ(pWS) != MXC_RC_NONE)
        {
            return SEC_FAILURE;
        }
        return SEC_FAILURE;
    }
    if (lCurrChl == 0 )                                    /* Number of channels below limit, just return */
    {
        if (closeStatusQ(pWS) != MXC_RC_NONE)
        {
            return SEC_FAILURE;
        }
        return SEC_OK;
    }
#endif
#endif

#if !defined(__BI2_78__)
    /* Else it's time to update.. */
    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1,"CheckRunningChannels() Get... update", pWS);
#if !defined(__MVS__)
    if (getActualCurrentNumberOfRunningChannels(&lCurrChl, pWS) != MXC_RC_NONE)
#else
    if (getActualCurrentNumberOfRunningChannelszOS(&lCurrChl, pWS) != MXC_RC_NONE)
#endif
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, Channel [%.*s] ConName [%.*s] User [%.*s] requested by CheckRunningChannels()",
                 rlnts(MQ_CHANNEL_NAME_LENGTH, pChannelDefinition->ChannelName),
                 pChannelDefinition->ChannelName,
                 rlnts(MQ_CONN_NAME_LENGTH, pWS->ConName),  /* NWC */
                 pWS->ConName,
                 rlnts(MQ_USER_ID_LENGTH, pChannelDefinition->RemoteUserIdentifier),
                 pChannelDefinition->RemoteUserIdentifier);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData( LOGDATA, pWS->szTmpBufff, pWS);
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"getActualCurrentNumberOfRunningChannels failed.");
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDATA, pWS->szTmpBufff, pWS);

        lCurrChl =  UpdateStatusQ(pChannelDefinition,(long*)lCurrChl,pWS);
        if (closeStatusQ(pWS) != MXC_RC_NONE)
        {
            return SEC_FAILURE;
        }
        return SEC_FAILURE;
    }
    else
    {
        if (lCurrChl+1 > pWS->lMaxChannelActive)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connection refused, Channel [%.*s] ConName [%.*s] User [%.*s] requested by CheckRunningChannels()",
                     rlnts(MQ_CHANNEL_NAME_LENGTH, pChannelDefinition->ChannelName),
                     pChannelDefinition->ChannelName,
                     rlnts(MQ_CONN_NAME_LENGTH, pWS->ConName), /* NWC */
                     pWS->ConName,
                     rlnts(MQ_USER_ID_LENGTH, pChannelDefinition->RemoteUserIdentifier),
                     pChannelDefinition->RemoteUserIdentifier);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData( LOGDATA, pWS->szTmpBufff, pWS);
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"MAX active channels set on [%s] was exceeded, limit was [%ld] Active was [%ld]. Connection was refused.",
                     pWS->szChannelName,pWS->lMaxChannelActive,lCurrChl);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA, pWS->szTmpBufff, pWS);
            lCurrChl =  UpdateStatusQ(pChannelDefinition,(long*)lCurrChl,pWS);
            if (closeStatusQ(pWS) != MXC_RC_NONE)
            {
                return SEC_FAILURE;
            }
            return SEC_FAILURE;
        }
        lCurrChl =  UpdateStatusQ(pChannelDefinition,(long*)lCurrChl,pWS);
    }
    if (closeStatusQ(pWS) != MXC_RC_NONE)
    {
        return SEC_FAILURE;
    }
    if (pWS->nDebugFlag >= LOGDEBUG)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Max Active channels check passed. Number of active channels [%ld]",lCurrChl);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG, pWS->szTmpBufff, pWS);
    }
    return SEC_OK;
#endif
    return SEC_FAILURE;   /* we should not come here */
}
#endif


#if defined(__MVS__)
/* Check the active connections */
int CheckRunningChannels(MQCD *pChannelDefinition, workingStorage_type *pWS)
{
    long lCurrChl;
    lCurrChl =  UpdateStatusQ(pChannelDefinition,(long*)lCurrChl,pWS);
    if (pWS->nDebugFlag >= LOGDEBUG1)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"UpdateStatusQ gave [%ld]",lCurrChl);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
    }
    if (lCurrChl < 0)
    {
        return SEC_FAILURE;
    }
    else
    {
        return SEC_OK;
    }
}
#endif



#if !defined(__MVS__)  && !defined(__BI2_78__)

/* start of inquire channel usage
**********************************************************************************************************
* Use the admin interface to query the channel status of instances of the channel in question, and count
* the number with MQCHS_RUNNING
*
* arg_plNumberOfChannels number of extant channels with the relevant SvrConn name
**********************************************************************************************************
*/
long getActualCurrentNumberOfRunningChannels( long* arg_plNumberOfChannels, workingStorage_type *pWS)
{

    MQLONG reasonCode;
    MQLONG completionCode;
    long   lRc=MXC_RC_NONE;

    MQHBAG adminBag = MQHB_UNUSABLE_HBAG;
    MQHBAG responseBag = MQHB_UNUSABLE_HBAG;
    MQHBAG qAttrsBag = MQHB_UNUSABLE_HBAG;


    if (pWS->nDebugFlag >= LOGDEBUG1)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Entered: getActualCurrentNumberOfRunningChannels()");
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
    }

    *arg_plNumberOfChannels = 0;

    do
    {
        lRc = bagSetup(&adminBag,&responseBag,&qAttrsBag, pWS);
        if (lRc != MXC_RC_NONE)
        {
            if (pWS->nDebugFlag >= LOGDEBUG1)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"bagSetup gave. Reason=%ld",lRc);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
            }
            break;
        }

        mqAddString(adminBag, MQCACH_CHANNEL_NAME, MQBL_NULL_TERMINATED, pWS->szChannelName, &completionCode, &reasonCode);
        if (reasonCode != MQRC_NONE)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Error with mqAddString for MQCACH_CHANNEL_NAME of %s. Reason=%ld",pWS->szChannelName,(long)reasonCode);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA, pWS->szTmpBufff, pWS);
            lRc = MXC_RC_ERROR;
            break;
        }

        mqAddInquiry(adminBag, MQIACH_CHANNEL_STATUS, &completionCode, &reasonCode);
        if (reasonCode != MQRC_NONE)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Error with mqAddInquiry for MQIACH_CHANNEL_STATUS. Reason=%ld",(long)reasonCode);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA, pWS->szTmpBufff, pWS);
            lRc = MXC_RC_ERROR;
            break;
        }

        mqExecute(pWS->hConn,
                  MQCMD_INQUIRE_CHANNEL_STATUS,
                  MQHB_NONE,
                  adminBag,
                  responseBag,
                  MQHO_NONE,
                  MQHO_NONE,
                  &completionCode,
                  &reasonCode);

        if (reasonCode != MQRC_NONE)
        {
            if (reasonCode == MQRCCF_COMMAND_FAILED)
            {
                MQLONG mqExecuteCC;
                MQLONG mqExecuteRC;
                MQHBAG errorBag;

                mqInquireBag(responseBag, MQHA_BAG_HANDLE, 0, &errorBag, &completionCode, &reasonCode);
                if (reasonCode != MQRC_NONE)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Error with mqInquireBag . Reason=%ld",(long)reasonCode);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDATA, pWS->szTmpBufff, pWS);
                    lRc = MXC_RC_ERROR;
                    break;
                }
                mqInquireInteger(errorBag,
                                 MQIASY_COMP_CODE,
                                 MQIND_NONE,
                                 &mqExecuteCC,
                                 &completionCode,
                                 &reasonCode );
                if (reasonCode != MQRC_NONE)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Error with mqInquireInteger (CC). Reason=%ld",(long)reasonCode);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDATA, pWS->szTmpBufff, pWS);
                    lRc = MXC_RC_ERROR;
                    break;
                }

                mqInquireInteger(errorBag,
                                 MQIASY_REASON,
                                 MQIND_NONE,
                                 &mqExecuteRC,
                                 &completionCode,
                                 &reasonCode);
                if (reasonCode != MQRC_NONE)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Error with mqInquireInteger (RC). Reason=%ld",(long)reasonCode);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDATA, pWS->szTmpBufff, pWS);
                    lRc = MXC_RC_ERROR;
                    break;
                }

                if (mqExecuteRC == MQRCCF_CHL_STATUS_NOT_FOUND)
                {
                    *arg_plNumberOfChannels = 0;
                }
                else
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Error with mqExecute. ExecuteReason=%ld",(long)mqExecuteRC);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDATA, pWS->szTmpBufff, pWS);
                    lRc = MXC_RC_ERROR;
                    break;
                }
            }
            else
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Error with mqExecute. Reason=%ld",(long)reasonCode);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDATA, pWS->szTmpBufff, pWS);
                lRc = MXC_RC_ERROR;
                break;
            }
        }

        if (lRc == MXC_RC_NONE)
        {
            MQLONG lNumberOfResponseBags;
            mqCountItems(responseBag, MQHA_BAG_HANDLE, &lNumberOfResponseBags, &completionCode, &reasonCode);
            if (reasonCode != MQRC_NONE)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Error with mqCountItems. Reason=%ld",(long)reasonCode);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDATA, pWS->szTmpBufff, pWS);
                lRc = MXC_RC_ERROR;
                break;
            }
            if (lNumberOfResponseBags)
            {
                long l1;
                MQLONG	lItemCount;

                /* one bag per channel instance - loop thru them */
                for (l1=0;l1< lNumberOfResponseBags;l1++)
                {
                    mqInquireBag(responseBag, MQHA_BAG_HANDLE, l1, &qAttrsBag, &completionCode, &reasonCode);
                    if (reasonCode != MQRC_NONE)
                    {
                        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Error with mqInquireBag. Reason=%ld",(long)reasonCode);
                        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                        LogData(LOGDATA, pWS->szTmpBufff, pWS);
                        lRc = MXC_RC_ERROR;
                        break;
                    }
                    mqCountItems(qAttrsBag,MQSEL_ALL_SELECTORS ,&lItemCount,&completionCode, &reasonCode);
                    if (reasonCode != MQRC_NONE)
                    {
                        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Error with mqCountItems . Reason=%ld",(long)reasonCode);
                        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                        LogData(LOGDATA, pWS->szTmpBufff, pWS);
                        lRc = MXC_RC_ERROR;
                        break;
                    }
                    if (lItemCount)
                    {
                        MQLONG	lStatus;
                        mqInquireInteger(qAttrsBag, MQIACH_CHANNEL_STATUS, MQIND_NONE,&lStatus,
                                         &completionCode, &reasonCode);
                        if (reasonCode != MQRC_NONE)
                        {
                            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Error with mqInquireInteger MQIACH_CHANNEL_STATUS. Reason=%ld",(long)reasonCode);
                            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                            LogData(LOGDATA, pWS->szTmpBufff, pWS);
                            lRc = MXC_RC_ERROR;
                            break;
                        }

                        /* If the status is RUNNING, then increment counter */
                        if (lStatus == MQCHS_RUNNING)
                        {
                            (*arg_plNumberOfChannels)++;
                        }
                    }
                }
            }
        }
    }while(0);

    /* delete the alloacted dataBags JHP 060218 */
    if (adminBag != MQHB_UNUSABLE_HBAG)
    {
        mqDeleteBag (&adminBag, &completionCode, &reasonCode);
        if (reasonCode != MQRC_NONE)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Error mqDeleteBag(adminBag). Reason=%ld",(long)reasonCode);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA, pWS->szTmpBufff, pWS);
        }
    }
    if (responseBag != MQHB_UNUSABLE_HBAG)
    {
        mqDeleteBag (&responseBag, &completionCode, &reasonCode);
        if (reasonCode != MQRC_NONE)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Error mqDeleteBag(responseBag). Reason=%ld",(long)reasonCode);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA, pWS->szTmpBufff, pWS);
        }
    }
    /* end of change delete the alloacted dataBags JHP 060218 */
    return lRc;
}

/*
**********************************************************************************************************
* setup admin interface objects
**********************************************************************************************************
*/
long bagSetup(MQHBAG*const arg_pAdminBag,MQHBAG*const arg_pResponseBag,MQHBAG*const arg_pQAttrsBag, workingStorage_type *pWS)
{

    long lRc = MXC_RC_NONE;

    MQLONG reasonCode;
    MQLONG completionCode;

    do
    {
        if (*arg_pAdminBag != MQHB_UNUSABLE_HBAG)
        {
            mqClearBag(*arg_pAdminBag, &completionCode, &reasonCode);
            if (reasonCode != MQRC_NONE)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Error with mqClearBag . Reason=%ld",(long)reasonCode);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDATA, pWS->szTmpBufff, pWS);
                lRc = MXC_RC_ERROR;
                break;
            }
        }
        mqCreateBag(MQCBO_ADMIN_BAG, arg_pAdminBag, &completionCode, &reasonCode);
        if (reasonCode != MQRC_NONE)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Error with mqCreateBag . Reason=%ld",(long)reasonCode);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA, pWS->szTmpBufff, pWS);
            lRc = MXC_RC_ERROR;
            break;
        }

        if (*arg_pResponseBag != MQHB_UNUSABLE_HBAG){
            mqClearBag(*arg_pResponseBag, &completionCode, &reasonCode);
            if (reasonCode != MQRC_NONE)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Error with mqClearBag . Reason=%ld",(long)reasonCode);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDATA, pWS->szTmpBufff, pWS);
                lRc = MXC_RC_ERROR;
                break;
            }
        }
        mqCreateBag(MQCBO_ADMIN_BAG, arg_pResponseBag, &completionCode, &reasonCode);
        if (reasonCode != MQRC_NONE)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Error with mqCreateBag . Reason=%ld",(long)reasonCode);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA, pWS->szTmpBufff, pWS);
            lRc = MXC_RC_ERROR;
            break;
        }

        if (*arg_pQAttrsBag != MQHB_UNUSABLE_HBAG)
        {
            mqClearBag(*arg_pQAttrsBag, &completionCode, &reasonCode);
            if (reasonCode != MQRC_NONE)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Error with mqClearBag . Reason=%ld",(long)reasonCode);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDATA, pWS->szTmpBufff, pWS);
                lRc = MXC_RC_ERROR;
                break;
            }
            *arg_pQAttrsBag = MQHB_UNUSABLE_HBAG;
        }
    }while(0);

    return lRc;

}


#endif

#if defined(__MVS__) && !defined(__BI2_78__)

/* start of inquire channel usage
**********************************************************************************************************
* Use the admin interface to query the channel status of instances of the channel in question, and count the number with MQCHS_RUNNING
*
* arg_plNumberOfChannels number of extant channels with the relevant SvrConn name
**********************************************************************************************************
*/
long getActualCurrentNumberOfRunningChannelszOS( long* arg_plNumberOfChannels, workingStorage_type *pWS)
{

    MQLONG   reasonCode=0;
    MQLONG   completionCode;

    MQGMO    gmo = {MQGMO_DEFAULT};
    MQMD     md = {MQMD_DEFAULT};

    long     lOpenRc,lRc=MXC_RC_NONE;
    MQLONG   buflen, messlen;
    char     buffer[500];
    char    *p;
    char    *u;
    char    *strtok_cb;

    if (pWS->nDebugFlag >= LOGDEBUG1)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Entered: getActualCurrentNumberOfRunningChannelszOS()",pWS);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1, pWS->szTmpBufff,pWS);
    }
    *arg_plNumberOfChannels = 0;

    do{
        lRc = chsSetup(pWS);
        lOpenRc=lRc;

        /* LogData(LOGDEBUG1, "Back from chlSetup()");*/

        if (lRc == MXC_RC_NONE)
        {
            lRc = chsSendDisplayCHStatus("SYSTEM.COMMAND.INPUT",pWS->szReplyQ,pWS);
        }
        if (lRc == MXC_RC_NONE)
        {
            gmo.Options = MQGMO_WAIT +                     /* wait for new messages           */
                          MQGMO_NO_SYNCPOINT;              /* no cyncpointing.. */
            gmo.WaitInterval = 500;                        /* .5 second limit for waiting     */

            do
            {
                memcpy(md.MsgId, MQMI_NONE, sizeof(md.MsgId));
                memcpy(md.CorrelId, MQCI_NONE, sizeof(md.CorrelId));
                buflen = sizeof(buffer) - 1;               /* buffer size available for GET   */

                MQGET(pWS->hConn,                          /* connection handle                 */
                      pWS->Hobj,                           /* object handle                     */
                      &md,                                 /* message descriptor                */
                      &gmo,                                /* get message options               */
                      buflen,                              /* buffer length                     */
                      buffer,                              /* message buffer                    */
                      &messlen,                            /* message length                    */
                      &completionCode,                     /* completion code                   */
                      &reasonCode);                        /* reason code                       */
                if (reasonCode == 0)
                {
                    if (memcmp(buffer,"CSQ9022I",8)==0)    /* NORMAL COMPLETION -> we're done */
                    {
                        continue;
                    }
                    if (memcmp(buffer,"CSQM425I",8)==0)    /* Server-connection Channel status */
                    {
                        buffer[messlen] = '\00';
                        if (pWS->nDebugFlag >= LOGDEBUG1)
                        {
                            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"message received %s",buffer);
                            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                            LogData(LOGDEBUG1, pWS->szTmpBufff,pWS);
                        }
                        p = strstr(buffer," STATUS(");
                        if (p != NULL)
                        {
                            p=p+8;
                            u=(char*)strtok_r(p,")",&strtok_cb);
                            if(u != NULL)
                            {
                                if (strlen(u) > 6)
                                {
                                    if (memcmp(u,"RUNNING",7) == 0)
                                        (*arg_plNumberOfChannels)++;
                                }
                            }

                        }
                        /* If the status is RUNNING, then increment counter */
                        /* if (lStatus == MQCHS_RUNNING){*/
                        /*    (*arg_plNumberOfChannels)++;*/
                        /*    }*/
                    }
                }

            }  while (reasonCode == 0);
        }
        if (lOpenRc == MXC_RC_NONE)
        {
            lRc = closeReplyQ(pWS);
        }

    } while(0);

    return lRc;
}


/*
**********************************************************************************************************
* setup Reply queue.
**********************************************************************************************************
*/
long chsSetup(workingStorage_type *pWS)
{

    long     lRc = MXC_RC_NONE;

    MQLONG   reasonCode;
    MQLONG   completionCode;
    MQLONG   O_options;                                    /* MQOPEN options                */
    MQOD     od = {MQOD_DEFAULT};                          /* Object Descriptor             */

    O_options = MQOO_INPUT_SHARED                          /* open queue for input          */
                | MQOO_FAIL_IF_QUIESCING                   /* but not if MQM stopping       */
                ;
    /*   LogData(LOGDEBUG, "Before MQOPEN");
    */
    strcpy(od.ObjectName,"SYSTEM.COMMAND.REPLY.MODEL");
    /*	strcpy(odReplyQ.DynamicQName,"SYSTEM.BlockIP2.*"); */
    strcpy(od.DynamicQName,"SYSTEM.CSQOREXX.*");   /* test only */

    MQOPEN(pWS->hConn,                                     /* connection handle            */
           &od,                                            /* object descriptor for queue  */
           O_options,                                      /* open options                 */
           &pWS->Hobj,                                     /* object handle                */
           &completionCode,                                /* MQOPEN completion code       */
           &reasonCode);                                   /* reason code                  */

    /* report reason, if any; stop if failed      */
    if (reasonCode != MQRC_NONE)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"MQOPEN ended with reason code %d", reasonCode);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDATA, pWS->szTmpBufff,pWS);
        lRc = MXC_RC_ERROR ;
    }

    if (completionCode == MQCC_FAILED)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"unable to open queue for output");
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDATA, pWS->szTmpBufff,pWS);
        lRc = MXC_RC_ERROR ;
    }
    /*    nullDelimit(pWS->szReplyQ,od.ObjectName,MQ_OBJECT_NAME_LENGTH+1);*/
    memset(pWS->szReplyQ,'\00',sizeof(pWS->szReplyQ));
    memcpy(pWS->szReplyQ,od.ObjectName,rlnts(MQ_OBJECT_NAME_LENGTH, od.ObjectName));

    /* snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"chsSetup() Hobj is %ld ReplyQ %s",Hobj,szReplyQ);
       pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
       LogData(LOGDEBUG, pWS->szTmpBufff);
    */
    return lRc;

}

/*
**********************************************************************************************************
* setup Reply queue.
**********************************************************************************************************
*/
long chsSendDisplayCHStatus(char * cmdq, char * replyq, workingStorage_type *pWS)
{

    MQLONG   reason;
    MQLONG   compCode;

    MQOD     od = {MQOD_DEFAULT};
    MQMD     md = {MQMD_DEFAULT};
    MQPMO    pmo = {MQPMO_DEFAULT};

    long     BufferLength;
    char     Buffer[200];
    long     lRc=MXC_RC_NONE;

    memcpy(od.ObjectName,cmdq,MQ_OBJECT_NAME_LENGTH);
    memcpy(md.ReplyToQ,replyq,MQ_OBJECT_NAME_LENGTH);

    memcpy(md.Format,                                      /* character string format            */
           MQFMT_STRING, (size_t)MQ_FORMAT_LENGTH);
    md.MsgType=MQMT_REQUEST;

    strcpy(Buffer,"DISPLAY CHSTATUS(");
    strcat(Buffer,pWS->szChannelName);
    strcat(Buffer,") CURRENT");
    BufferLength = strlen(Buffer);
    /*
        strcpy(Buffer,"DISPLAY CHSTATUS(TO.QM.QSAT) CURRENT");
        BufferLength = strlen(Buffer);*/

    pmo.Options= MQPMO_NO_SYNCPOINT;

    if (pWS->nDebugFlag >= LOGDEBUG1)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"MQPUT1 Sending [%s]", Buffer);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1, pWS->szTmpBufff,pWS);
    }

    MQPUT1 (pWS->hConn,
            &od,
            &md,
            &pmo,
            BufferLength,
            Buffer,
            &compCode,
            &reason);
    /* report reason, if any; stop if failed      */
    if (reason != MQRC_NONE)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"MQPUT1 ended with reason code %d", reason);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDATA, pWS->szTmpBufff,pWS);
        lRc = MXC_RC_ERROR ;
    }

    if (compCode == MQCC_FAILED)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"unable to put on COMMAND queue");
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDATA, pWS->szTmpBufff,pWS);
        lRc = MXC_RC_ERROR ;
    }
    return lRc;

}
/*
**********************************************************************************************************
* Close reply queue
*
**********************************************************************************************************
*/
long closeReplyQ(workingStorage_type *pWS)
{
    long     lRc = MXC_RC_NONE;

    MQLONG   reason;
    MQLONG   compCode;
    MQLONG   C_options;                                    /* MQCLOSE options                */

    /* snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Close and delete the Reply queue");
       pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
       LogData(LOGDEBUG, pWS->szTmpBufff);
    */
    C_options = MQCO_DELETE_PURGE;                         /* delete the queue and its data */

    MQCLOSE(pWS->hConn,
            &pWS->Hobj,
            C_options,
            &compCode,
            &reason);

    if (reason == MQRC_NONE)
    {
    }
    else
    {
        lRc = MXC_RC_ERROR;
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Error closing reply queue %s. Reason=%ld",pWS->szReplyQ,reason);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDATA, pWS->szTmpBufff,pWS);
    }
    /* LogData(LOGDEBUG, "Close reply Q done....");*/

    return lRc;
}

#endif

#if !defined(__BI2_78__)
/*
**********************************************************************************************************
* Connect
*
**********************************************************************************************************
*/
long connectToQMgr(workingStorage_type *pWS)
{
    long     lRc = MXC_RC_NONE;

    MQLONG   reason;
    MQLONG   compCode;

    if (pWS->nDebugFlag >= LOGDEBUG)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Connect to default Qmgr %s.",pWS->szQMgr);
        LogData(LOGDEBUG, pWS->szTmpBufff, pWS);
    }

    MQCONN(pWS->szQMgr,
           &pWS->hConn,
           &compCode,
           &reason);

    if (pWS->nDebugFlag >= LOGDEBUG1)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Debug connecting to %s. Reason=%ld hConn=%ld",pWS->szQMgr,(long)reason,(long)pWS->hConn);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
    }

    if (reason == MQRC_ALREADY_CONNECTED || reason == MQRC_NONE)
    {
    }
    else
    {
        lRc = MXC_RC_ERROR;
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Error connecting to %s. Reason=%ld",pWS->szQMgr,(long)reason);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDATA, pWS->szTmpBufff, pWS);
    }
    /* LogData(LOGDEBUG, "Connect to QMgr done....");*/

    return lRc;
}
#endif

#if !defined(__MVS__) && !defined(__BI2_78__)

/*
**********************************************************************************************************
* Open Status queue.
**********************************************************************************************************
*/
long openStatusQ(workingStorage_type *pWS) {

    long     lRc = MXC_RC_NONE;

    MQLONG   reasonCode;
    MQLONG   completionCode;
    MQLONG   O_options;                                    /* MQOPEN options                */
    MQOD     od = {MQOD_DEFAULT};                          /* Object Descriptor             */

    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1,"In openStatusQ", pWS);

    O_options = MQOO_INPUT_EXCLUSIVE /*SHARED*/	           /* open queue for input, exclusive, and output */
                | MQOO_OUTPUT
                | MQOO_FAIL_IF_QUIESCING                   /* but not if MQM stopping       */
                ;
    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1, "Before MQOPEN status Q", pWS);

    strcpy(od.ObjectName,pWS->szStatusQ);
    do
    {
        MQOPEN(pWS->hConn,                                 /* connection handle             */
               &od,                                        /* object descriptor for queue   */
               O_options,                                  /* open options                  */
               &pWS->HobjStat,                             /* object handle                 */
               &completionCode,                            /* MQOPEN completion code        */
               &reasonCode);                               /* reason code                   */
    } while (reasonCode==MQRC_OBJECT_IN_USE);

    /* report reason, if any; stop if failed      */
    if (reasonCode != MQRC_NONE)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"MQOPEN status Q ended with reason code %ld on %s hConn=%ld", (long)reasonCode,pWS->szStatusQ,(long)pWS->hConn);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDATA, pWS->szTmpBufff, pWS);
        lRc = MXC_RC_ERROR ;
    }
    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1,"In leave openStatusQ", pWS);

    return lRc;
}

/*
**********************************************************************************************************
* Close status queue
*
**********************************************************************************************************
*/
long         closeStatusQ(workingStorage_type *pWS)
{
    long     lRc = MXC_RC_NONE;

    MQLONG   reason;
    MQLONG   compCode;
    MQLONG   C_options;                                    /* MQCLOSE options               */

    if (pWS->nDebugFlag >= LOGDEBUG1)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Close Status queue");
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
    }

    C_options = MQCO_NONE;                                 /* just close the queue          */

    MQCLOSE(pWS->hConn,
            &pWS->HobjStat,
            C_options,
            &compCode,
            &reason);

    if (reason == MQRC_NONE){
    }
    else
    {
        lRc = MXC_RC_ERROR;
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Error closing status queue %s. Reason=%ld",pWS->szStatusQ,(long)reason);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDATA, pWS->szTmpBufff, pWS);
    }
    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1, "Close status Q done....", pWS);

    return lRc;
}
#endif

#if !defined(__MVS__)
#if !defined(__BI2_78__)
/*
**********************************************************************************************************
* Update status queue
*
**********************************************************************************************************
*/
long UpdateStatusQ(MQCD *pChannelDefinition,  long* what2do, workingStorage_type *pWS)
{
    long     lRc = -1;
    MQLONG   buflen;
    MQLONG   messlen;

    MQLONG   reason;
    MQLONG   compCode;
    char     buffer[30];
    currChl_type *b;

    MQGMO    gmo = {MQGMO_DEFAULT};
    MQPMO    pmo = {MQPMO_DEFAULT};
    MQMD     md = {MQMD_DEFAULT};

    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1, "Update status Q entered....", pWS);


    gmo.Options = MQGMO_WAIT |                             /* wait for new messages         */
                  MQGMO_NO_SYNCPOINT;                      /* no cyncpointing..             */
    gmo.WaitInterval = 50;                                 /* nil second limit for waiting  */

    memcpy(md.MsgId, pChannelDefinition->ChannelName, sizeof(pChannelDefinition->ChannelName));
    memcpy(md.CorrelId, MQCI_NONE, sizeof(md.CorrelId));
    buflen = sizeof(buffer) - 1;                           /* buffer size available for GET */

    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1, "Before MQ Get....", pWS);
    MQGET(pWS->hConn,                                      /* connection handle             */
          pWS->HobjStat,                                   /* object handle                 */
          &md,                                             /* message descriptor            */
          &gmo,                                            /* get message options           */
          buflen,                                          /* buffer length                 */
          buffer,                                          /* message buffer                */
          &messlen,                                        /* message length                */
          &compCode,                                       /* completion code               */
          &reason);                                        /* reason code                   */
    if (reason == 0)
    {
        if (pWS->nDebugFlag >= LOGDEBUG1)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"MQGET status queue. Reason=%ld",(long)reason);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
        }
    }
    else
    {
        if (reason == MQRC_NO_MSG_AVAILABLE)
        {
            if (what2do== (long*)-1) return 1;
            lRc=1;                                         /* insert status record          */
        }
        else
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Error MQGET status queue Reason=%ld",(long)reason);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA, pWS->szTmpBufff, pWS);
            return -1;
        }
    }
    b=(struct CURRCHLDATA*)&buffer;
    if (pWS->nDebugFlag >= LOGDEBUG1)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Buffer was lim=%ld curr=%ld (Update)",b->limCurrChl,b->CurrChannels);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
    }
    if (what2do != (long*)-1 && reason == MQRC_NO_MSG_AVAILABLE) { /* init structure        */
        if (pWS->nDebugFlag >= LOGDEBUG8)
            LogData(LOGDEBUG8, "Init structure....", pWS);
        memset(buffer,'\00',sizeof(buffer));
        b->limCurrChl   = pWS->lMaxChannelActive;
        b->CurrChannels = (long)what2do;                   /* insert monitored value        */
        b->MaxChannels = (long)what2do;
        md.Expiry = pWS->lExpireValue;
    }
    if (pWS->nDebugFlag >= LOGDEBUG8)
        LogData(LOGDEBUG8, "Updating status Q started....", pWS);
    if (b->CurrChannels+1 > b->limCurrChl)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Channel limit reached lim=%ld curr=%ld",b->limCurrChl,b->CurrChannels);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDATA, pWS->szTmpBufff, pWS);
        lRc=-1;
    }
    else
    {
        b->CurrChannels++;                                 /* Adjust counter                */
        lRc=0;
        if (b->CurrChannels > b->MaxChannels)
        {
            b->MaxChannels = b->CurrChannels;
        }
    }
    messlen=sizeof(currChl_type);
    pmo.Options = MQPMO_NO_SYNCPOINT;                      /* no cyncpointing..             */
    if (pWS->nDebugFlag >= LOGDEBUG8)
        LogData(LOGDEBUG8, "MQPUT next....", pWS);
    MQPUT(pWS->hConn,
          pWS->HobjStat,                                   /* object handle                 */
          &md,                                             /* message descriptor            */
          &pmo,                                            /* put message options           */
          messlen,                                         /* buffer length                 */
          buffer,                                          /* message buffer                */
          &compCode,                                       /* completion code               */
          &reason);                                        /* reason code                   */
    if (pWS->nDebugFlag >= LOGDEBUG1)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"MQPUT status queue Reason=%ld",(long)reason);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
    }
    return lRc;
}
#endif

long deductStatusQ(MQCD *pChannelDefinition, workingStorage_type *pWS)
{
#if defined UNIX || defined (OS400)
#define __DEDUCTSTATUSQ__
#endif
#if defined WIN32_SHM
#define __DEDUCTSTATUSQ__
#endif

#ifndef __DEDUCTSTATUSQ__
    long     lRc = -1;

    MQLONG   buflen;
    MQLONG   messlen;

    MQLONG   reason;
    MQLONG   compCode;
    char     buffer[30];
    currChl_type *b;

    MQGMO    gmo = {MQGMO_DEFAULT};
    MQPMO    pmo = {MQPMO_DEFAULT};
    MQMD     md = {MQMD_DEFAULT};
#endif

    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1, "deductStatusQ() Entered", pWS);

#if defined UNIX || defined (OS400)
    UpdateShrMem(pChannelDefinition,(long*)-1,pWS);
#else
#if defined WIN32_SHM
    UpdateShrMem(pChannelDefinition,(long*)-1,pWS);
#else
#ifndef __DEDUCTSTATUSQ__
    if (connectToQMgr(pWS) != MXC_RC_NONE)
        return 0;
    if (openStatusQ(pWS) != MXC_RC_NONE)
        return 0;

    gmo.Options = MQGMO_NO_WAIT |                          /* wait for new messages         */
                  MQGMO_NO_SYNCPOINT;                      /* no cyncpointing..             */
    gmo.WaitInterval = 50;                                 /* nil second limit for waiting  */

    memcpy(md.MsgId, pChannelDefinition->ChannelName, sizeof(pChannelDefinition->ChannelName));
    memcpy(md.CorrelId, MQCI_NONE, sizeof(md.CorrelId));
    buflen = sizeof(buffer) - 1;                           /* buffer size available for GET */

    if (pWS->nDebugFlag >= LOGDEBUG8)
        LogData(LOGDEBUG8, "Before MQ Get....", pWS);
    MQGET(pWS->hConn,                                      /* connection handle             */
          pWS->HobjStat,                                   /* object handle                 */
          &md,                                             /* message descriptor            */
          &gmo,                                            /* get message options           */
          buflen,                                          /* buffer length                 */
          buffer,                                          /* message buffer                */
          &messlen,                                        /* message length                */
          &compCode,                                       /* completion code               */
          &reason);                                        /* reason code                   */
    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"MQGET status queue. Reason=%ld [%s]",reason,buffer);
    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
    LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
    if (reason == 0)
    {
        if (pWS->nDebugFlag >= LOGDEBUG1)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"MQGET status queue success. Reason=%ld",reason);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
        }
    }
    else
    {
        if (closeStatusQ(pWS) != MXC_RC_NONE)
        {
            return 0;
        }
        return 0;
    }
    /*    b=&buffer;*/
    b=(struct CURRCHLDATA*)&buffer;
    if (pWS->nDebugFlag >= LOGDEBUG1)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Buffer was lim=%ld curr=%ld",b->limCurrChl,b->CurrChannels);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
    }

    if (b->CurrChannels>0)
        b->CurrChannels--;                                 /* Adjust counter                */
    messlen=sizeof(currChl_type);
    pmo.Options = MQPMO_NO_SYNCPOINT;                      /* no cyncpointing..             */
    if (pWS->nDebugFlag >= LOGDEBUG8)
        LogData(LOGDEBUG8, "MQPUT next....", pWS);
    MQPUT(pWS->hConn,
          pWS->HobjStat,                                   /* object handle                 */
          &md,                                             /* message descriptor            */
          &pmo,                                            /* put message options           */
          messlen,                                         /* buffer length                 */
          buffer,                                          /* message buffer                */
          &compCode,                                       /* completion code               */
          &reason);                                        /* reason code                   */
    if (pWS->nDebugFlag >= LOGDEBUG1)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"MQPUT status queue Reason=%ld",reason);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
    }
    if (closeStatusQ(pWS) != MXC_RC_NONE)
    {
        if (pWS->nDebugFlag >= LOGDEBUG1)
            LogData(LOGDEBUG1, "deductStatusQ() Done", pWS);
        return 0;
    }
#endif
#endif
#endif
    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1, "deductStatusQ() Done", pWS);
    return 0;
}

#endif

#if defined(__MVS__)

/*
**********************************************************************************************************
* Update status queue
*
**********************************************************************************************************
*/
long UpdateStatusQ(MQCD *pChannelDefinition,  long* what2do, workingStorage_type *pWS)
{
    long     lRc = -1;
    bi2ds_type  dsdata;
    bi2ds_type  * pdsdata;

    pdsdata = &dsdata;

    memset(pdsdata -> channel,' ',MQ_CHANNEL_NAME_LENGTH);
    memcpy(pdsdata -> channel,
           pChannelDefinition->ChannelName,
           MQ_CHANNEL_NAME_LENGTH) ;
    memset(pdsdata->function,'+',1);

    pdsdata->limit   = pWS->lMaxChannelActive;
    pdsdata->current = (long)what2do;                      /* insert monitored value        */
    pdsdata->maxrunning = (long)what2do;

    BLOCKIPS(pdsdata);
    if (pWS->nDebugFlag >= LOGDEBUG1)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"BLOCKIPS gave= [%.*s] %ld",8, pdsdata->status,strcmp(pdsdata->status,"OK"));
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
    }
    if (strcmp(pdsdata->status,"OK") == -1)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Channel limit reached lim=%ld curr=%ld",pdsdata->limit,pdsdata->current);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDATA, pWS->szTmpBufff, pWS);
        return lRc;
    }
    return pdsdata->current;
}


long deductStatusQ(MQCD *pChannelDefinition, workingStorage_type *pWS)
{
    bi2ds_type  dsdata;
    bi2ds_type  * pdsdata;

    pdsdata = &dsdata;

    memset(pdsdata -> channel,' ',MQ_CHANNEL_NAME_LENGTH);
    memcpy(pdsdata -> channel,
           pChannelDefinition->ChannelName,
           MQ_CHANNEL_NAME_LENGTH) ;
    memset(pdsdata->function,'-',1);
    BLOCKIPS(pdsdata);
    return 0;
}
#endif

/*
**********************************************************************************************************
* Interrogate a fixed size array. Replace the first space with a null delimiter.  If no space exists,
* set the byte following the array to null.
**********************************************************************************************************
*/
void nullDelimit(char arg_szString[],char arg_acArray[],long arg_lMaxBufferSize )
{
    char    *pcTemp;

    memcpy(arg_szString,arg_acArray,arg_lMaxBufferSize-1);
    pcTemp = strstr(arg_szString," ");
    if (pcTemp)
    {
        *pcTemp=0;
    }else
    {
        arg_szString[arg_lMaxBufferSize-1] = 0;
    }
}

/*
**********************************************************************************************************
* Obtain the userid executing the task on z/OS
**********************************************************************************************************
*/
#if defined(__MVS__)

void GetZOSUserId(char *buffer)
{
    typedef struct
    {
        char UidLen;
        char UserId[8];
    } AceeUid;

    char * loc;
    long * addr;
    long * addr2;
    AceeUid * UID;

    addr=   *((long*)(0X00000224));
    addr2= addr + (0X0000006C/4);
    addr= *addr2 ;
    addr2= addr + (0X000000C8/4);
    addr= *addr2 ;
    loc = addr;
    loc= loc+20;
    UID=loc;
    memcpy(buffer,UID->UserId,8);
}
#endif

char * MQExitReason(PMQCXP pChannelExitParms)
{
    switch(pChannelExitParms->ExitReason)
    {
    case(MQXR_INIT):            return("MQXR_INIT");
    case(MQXR_TERM):            return("MQXR_TERM");
    case(MQXR_MSG):             return("MQRX_MSG");
    case(MQXR_XMIT):            return("MQXR_XMIT");
    case(MQXR_SEC_MSG):         return("MQXR_SEC_MSG");
    case(MQXR_INIT_SEC):        return("MQXR_INIT_SEC");
    case(MQXR_RETRY):           return("MQXR_RETRY");
    case(MQXR_AUTO_CLUSSDR):    return("MQXR_AUTO_CLUSSDR");
    case(MQXR_AUTO_RECEIVER):   return("MQXR_AUTO_RECEIVER");
    case(MQXR_AUTO_SVRCONN):    return("MQXR_AUTO_SVRCONN");
    case(MQXR_AUTO_CLUSRCVR):   return("MQXR_AUTO_CLUSRCVR");
#ifdef MQXR_SEC_PARMS
    case(MQXR_SEC_PARMS):       return("MQXR_SEC_PARMS");
#endif
    default:                    return("Unknown ExitReason");
    }                                                      /* end switch                    */
}

char * MQExitId(PMQCXP pChannelExitParms)
{
    switch(pChannelExitParms->ExitId)
    {
    case(MQXT_CHANNEL_SEC_EXIT):         return("MQXT_CHANNEL_SEC_EXIT");
    case(MQXT_CHANNEL_MSG_EXIT):         return("MQXT_CHANNEL_MSG_EXIT");
    case(MQXT_CHANNEL_SEND_EXIT):        return("MQXT_CHANNEL_SEND_EXIT");
    case(MQXT_CHANNEL_RCV_EXIT):         return("MQXT_CHANNEL_RCV_EXIT");
    case(MQXT_CHANNEL_MSG_RETRY_EXIT):   return("MQXT_CHANNEL_MSG_RETRY_EXIT");
    case(MQXT_CHANNEL_AUTO_DEF_EXIT):    return("MQXT_CHANNEL_AUTO_DEF_EXIT");
    default:                             return("Unknown ExitID");
    }                                                      /* end switch                    */
}

char * MQChannelType(PMQCD pChannelDefinition)
{
    switch(pChannelDefinition->ChannelType)
    {
    case(MQCHT_SENDER):         return("MQCHT_SENDER");
    case(MQCHT_SERVER):         return("MQCHT_SERVER");
    case(MQCHT_RECEIVER):       return("MQCHT_RECEIVER");
    case(MQCHT_REQUESTER):      return("MQCHT_REQUESTER");
    case(MQCHT_CLNTCONN):       return("MQCHT_CLNTCONN");
    case(MQCHT_SVRCONN):        return("MQCHT_SVRCONN");
    case(MQCHT_CLUSSDR):        return("MQCHT_CLUSSDR");
    case(MQCHT_CLUSRCVR):       return("MQCHT_CLUSRCVR");
    default:                    return("Unknown ChannelType");
    }                                                      /* end switch                    */
}

char * MQExitResponseType(PMQCXP pChannelExitParams)
{
    switch(pChannelExitParams->ExitResponse)
    {
    case(MQXCC_OK):                       return("MQXCC_OK");
    case(MQXCC_SUPPRESS_FUNCTION):        return("MQXCC_SUPPRESS_FUNCTION");
    case(MQXCC_SKIP_FUNCTION):            return("MQXCC_SKIP_FUNCTION");
    case(MQXCC_SEND_AND_REQUEST_SEC_MSG): return("MQXCC_SEND_AND_REQUEST_SEC_MSG");
    case(MQXCC_SEND_SEC_MSG):             return("MQXCC_SEND_SEC_MSG");
    case(MQXCC_SUPPRESS_EXIT):            return("MQXCC_SUPPRESS_EXIT");
    case(MQXCC_CLOSE_CHANNEL):            return("MQXCC_CLOSE_CHANNEL");
    case(MQXCC_FAILED):                   return("MQXCC_FAILED");
    default:                              return("Unknown ExitResponse");
    } /* end switch */
}

char * MQAuthenticationType(PMQCXP pChannelExitParams)
{
    switch(pChannelExitParams->SecurityParms->AuthenticationType)
    {
    case(MQCSP_AUTH_NONE):               return("MQCSP_AUTH_NONE");
    case(MQCSP_AUTH_USER_ID_AND_PWD):    return("MQCSP_AUTH_USER_ID_AND_PWD");
    default:                             return("Unknown AuthenticationType");
    } /* end switch */
}


#if defined(WIN32)
long StartShrMemServer(workingStorage_type *pWS)
{
    char     filename[128] = "";
    STARTUPINFO si;
    PROCESS_INFORMATION pi;
    SECURITY_ATTRIBUTES  sec_attr;
    char     szCommand2do[256];
    char     szSemKey[256];
    HANDLE   MutexBlockIP2S;
    HANDLE   hShrMemFile;
    HANDLE   semaphore;
    shmem_type * shm;
    DWORD    retVal;
    long     lCount;
    LPCTSTR  pBuf=NULL;
    char     errMsg[256];
    DWORD    eNum;

    if (pWS->nDebugFlag>=LOGDEBUG1)
        LogData(LOGDEBUG1,"    StartShrMemServer() Entered", pWS);
    /* Generate the name of the file */
    strcpy(filename,"Global\\BlockIP2-");
    strcat(filename,pWS->szQMgr);
    strcat(filename,"-shm");              /* shared mem */

    if (pWS->nDebugFlag>=LOGDEBUG1)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"    StartShrMemServer() shmem filename [%s]", filename);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
    }

    /* This part must be encapsulated in a mutex/semaphore */
    /* Set mutex here and evt. semaphore here */

    OpenMTXBlockIP2S(pWS, &MutexBlockIP2S);
    if (pWS->nDebugFlag>=LOGDEBUG1)
        LogData(LOGDEBUG1,"    After OpenMTXBlockIP2S() Entered", pWS);

    LockBlockIP2S(pWS, &MutexBlockIP2S);
    if (pWS->nDebugFlag>=LOGDEBUG1)
        LogData(LOGDEBUG1,"    After LockBlockIP2S() Entered", pWS);

    /* See if another thread have started the SHared Memory Server(SMS) */

    hShrMemFile = OpenFileMapping(FILE_MAP_ALL_ACCESS,     /* read/write access        */
                                  FALSE,                   /* Do not inherit the name  */
                                  filename);               /* Name of mapping object   */


    if (pWS->nDebugFlag>=LOGDEBUG1)
        LogData(LOGDEBUG1,"    StartShrMemServer() Open  hShrMemFile",pWS);

    if (pWS->nDebugFlag>=LOGDEBUG1)
    {
/*        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"    StartShrMemServer() hShrMemFile [0x%lx] file[%s]", hShrMemFile,filename);*/
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"    StartShrMemServer() hShrMemFile [0x%p] file[%s]", hShrMemFile,filename);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
    }

    /* Allocate the Connection name table */
    strcpy(filename,"Global\\BlockIP2-");
    strcat(filename,pWS->szQMgr);
    strcat(filename,"-shm2");

    if (hShrMemFile == NULL)
    {
        /* Start the shared memory server */
        if (pWS->nDebugFlag>=LOGDEBUG1)
            LogData(LOGDEBUG1,"    StartShrMemServer() Start to create SHM",pWS);

        /* Initialise si and pi */
        ZeroMemory( &si, sizeof(si) );
        si.cb = sizeof(si);
        ZeroMemory( &pi, sizeof(pi) );

        strcpy(szCommand2do,pWS->ExitPath);
        strcat(szCommand2do,"\\BlockIP2S.exe ");
        strcat(szCommand2do,pWS->szQMgr);

        if (pWS->nDebugFlag>=LOGDEBUG1)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"    StartShrMemServer() CreateProcess [%s]", szCommand2do);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
        }

        /* create sem2. here */

        sec_attr.nLength              = sizeof(SECURITY_ATTRIBUTES);
        sec_attr.lpSecurityDescriptor = NULL;
        sec_attr.bInheritHandle       = TRUE;

        /* Call function to set the DACL. The DACL is set in the SECURITY_ATTRIBUTES */
        /* lpSecurityDescriptor member.                                              */
        if (!CreateMyDACL(&sec_attr,pWS))
        {
            /* Error encountered; generate message and exit. */
            eNum = GetLastError();
            WINFmtMsg(eNum, errMsg, sizeof(errMsg));
            if (pWS->nDebugFlag>=LOGDEBUG1)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"    StartShrMemServer() Failed CreateMyDACL %d %s",eNum,errMsg);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
            }
            return -1;
        }

        if (hShrMemFile == NULL || hShrMemFile == INVALID_HANDLE_VALUE)
        {
            /* Generate the name of the file */
            strcpy(filename,"Global\\BlockIP2-");
            strcat(filename,pWS->szQMgr);
            strcat(filename,"-shm");              /* shared mem */
            if (pWS->nDebugFlag>=LOGDEBUG1)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"    StartShrMemServer() shmkey1 %s",filename);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
            }

            hShrMemFile  = CreateFileMapping(
                               INVALID_HANDLE_VALUE,    /* use paging file        */
                               &sec_attr,               /* special security       */
                               PAGE_READWRITE,          /* read/write access      */
                               0,                       /* max. object size       */
                               SHMSZ,                   /* buffer size            */
                               filename);               /* name of mapping object */

            if (hShrMemFile == NULL || hShrMemFile == INVALID_HANDLE_VALUE)
            {
                /* Error encountered; generate message and exit. */
                eNum = GetLastError();
                WINFmtMsg(eNum, errMsg, sizeof(errMsg));
                if (pWS->nDebugFlag>=LOGDEBUG)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"    StartShrMemServer() Could not create file mapping object  %d %s\n",eNum,errMsg);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDEBUG,pWS->szTmpBufff,pWS);
                }
                return -1;
            }
            if (pWS->nDebugFlag>=LOGDEBUG1)
                LogData(LOGDEBUG1,"    StartShrMemServer() Created file mapping object (1)",pWS);

            if (pWS->nDebugFlag>=LOGDEBUG1)
                LogData(LOGDEBUG1,"    StartShrMemServer() Create  hShrMemFile",pWS);

            pBuf = (LPTSTR) MapViewOfFile(hShrMemFile,          /* handle to map object */
                                          FILE_MAP_ALL_ACCESS,  /* read/write permission */
                                          0,
                                          0,
                                          SHMSZ);

            if (pBuf == NULL)
            {
                /* Error encountered; generate message and exit. */
                eNum = GetLastError();
                WINFmtMsg(eNum, errMsg, sizeof(errMsg));
                if (pWS->nDebugFlag>=LOGDEBUG)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"    StartShrMemServer() Could not map view of file  %d %s",eNum,errMsg);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDEBUG,pWS->szTmpBufff,pWS);
                }
                return -1;
            }

            if (pWS->nDebugFlag>=LOGDEBUG1)
                LogData(LOGDEBUG1,"    StartShrMemServer() Map  hShrMemFile / pBuf",pWS);

            shm = (shmem_type *)pBuf;
            shm->iNumOfChannels=0;  /* No channes yet */

            if (pWS->nDebugFlag>=LOGDEBUG1)
                LogData(LOGDEBUG1,"    StartShrMemServer() Initialized file mapping object (1)",pWS);
        }
        else
        {
            CloseHandle(hShrMemFile);
            if (pWS->nDebugFlag>=LOGDEBUG1)
                LogData(LOGDEBUG1,"    StartShrMemServer() Close  hShrMemFile",pWS);
        }

        strcpy(szSemKey,"Global\\BlockIP2S-");
        strcat(szSemKey,pWS->szQMgr);
        strcat(szSemKey,"-ctrl");

        if (pWS->nDebugFlag>=LOGDEBUG1)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"    StartShrMemServer() Semaphore key is [%s].",  szSemKey);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
        }
        semaphore = CreateSemaphore( &sec_attr, 0, 65536, szSemKey );
        if (semaphore == (HANDLE) NULL)
        {
            /* Error encountered; generate message and exit. */
            eNum = GetLastError();
            WINFmtMsg(eNum, errMsg, sizeof(errMsg));
            if (pWS->nDebugFlag>=LOGDEBUG1)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"    StartShrMemServer() Failed CreateSemaphore() %d %s",eNum,errMsg);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
            }
            return -1;
        }

        if (pWS->nDebugFlag>=LOGDEBUG1)
        {
/*            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"    StartShrMemServer() Semaphore opened. (%lx)", semaphore);*/
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"    StartShrMemServer() Semaphore opened. (0x%p)", semaphore);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
        }

        /* Create a childProcess */
        if( !CreateProcess( NULL,             /* No module name (use command line).*/
                            szCommand2do,     /* Command to execute                */
                            NULL,             /* Process handle not inheritable.   */
                            NULL,             /* Thread handle not inheritable.    */
                            FALSE,            /* Set handle inheritance to FALSE.  */
                            DETACHED_PROCESS, /* fdwCreate                         */
                            NULL,             /* Use parent's environment block.   */
                            NULL,             /* Use parent's starting directory.  */
                            &si,              /* Pointer to STARTUPINFO structure. */
                            &pi )             /* Pointer to PROCESS_INFORMATION structure. */
          )
        {
            LogData(LOGDATA,"    StartShrMemServer() CreateProcess failed.",pWS);
        }
        else
        {
            if (pWS->nDebugFlag>=LOGDEBUG1)
                LogData(LOGDEBUG1,"    StartShrMemServer() BlockIP2S Shared memory controller detached.",pWS);
        }
        lCount = 10L;
        while( (( retVal = WaitForSingleObject( semaphore, 250 ) ) == WAIT_TIMEOUT) && lCount > 0) {
            lCount--;
            if (pWS->nDebugFlag>=LOGDEBUG1)
                LogData(LOGDEBUG1, "      Timeout (250ms).", pWS);
        }
        if (lCount ==0)
        {
            LogData(LOGDATA,"    StartShrMemServer() BlockIP2S didn't respond correctly.",pWS);
            if( ! ReleaseSemaphore(semaphore, 1, NULL) )      /* clean up */
                LogData(LOGDATA, "    ReleaseSemaphore() Failed.", pWS);
            CloseHandle(semaphore);
            return -1;
        }

        if (! ReleaseSemaphore(semaphore, 1, NULL))
        {
            LogData(LOGDATA, "    StartShrMemServer() ReleaseSemaphore() Failed.", pWS);
            CloseHandle(semaphore);  /* Clean up */
            return -1;
        }
        CloseHandle(semaphore);
        if (pWS->nDebugFlag>=LOGDEBUG1)
            LogData(LOGDEBUG1, "    StartShrMemServer() CloseHandle.", pWS);

        if (pBuf != NULL)
            UnmapViewOfFile(pBuf);

        if (pWS->nDebugFlag>=LOGDEBUG1)
        {
            LogData(LOGDEBUG1,"    StartShrMemServer() UnMap  pBuf",pWS);
            LogData(LOGDEBUG1,"    StartShrMemServer() Close  hShrMemFile",pWS);
        }

        CloseHandle(hShrMemFile);

    }
    /* release mutex here */
    FreeBlockIP2S(pWS, &MutexBlockIP2S);
    if (pWS->nDebugFlag>=LOGDEBUG1)
        LogData(LOGDEBUG1,"    StartShrMemServer() Leaving-2", pWS);
    return 0;
}


/********************************************************************************************
*
*   Update the shared table entry for the passed channel.
*
********************************************************************************************/
long UpdateShrMem(MQCD *pChannelDefinition,  long* what2do, workingStorage_type *pWS)
{
    char filename[128] = "";

    int channelix=-1;
    int found=-1;
    HANDLE hShrMemFile;
    LPCTSTR pBuf;
    long CurrentAc;

    HANDLE MutexBlockIP2S;

    shmem_type * shm;

    if (pWS->nDebugFlag>=LOGDEBUG1)
        LogData(LOGDEBUG1,"  UpdateShrMem() Entered", pWS);

    CurrentAc=-1;

    /* Generate the name of the file */
    strcpy(filename,"Global\\BlockIP2-");
    strcat(filename,pWS->szQMgr);
    strcat(filename,"-shm");              /* shared mem */

    if (pWS->nDebugFlag>=LOGDEBUG1)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() shmem filename [%s] CurrentAc %ld", filename,CurrentAc);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
    }

    hShrMemFile = OpenFileMapping(FILE_MAP_ALL_ACCESS,     /* read/write access        */
                                  FALSE,                   /* Do not inherit the name  */
                                  filename);               /* Name of mapping object   */

    if (pWS->nDebugFlag>=LOGDEBUG1)
    {
/*        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() hShrMemFile [0x%lx]", hShrMemFile);*/
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() hShrMemFile [0x%p]", hShrMemFile);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
    }

    if (hShrMemFile == NULL)
    {
        /* This part must be encapsulated in a mutex/semaphore */
        /*Set mutex here and evt. semaphore here */

        OpenMTXBlockIP2S(pWS, &MutexBlockIP2S);
        if (pWS->nDebugFlag>=LOGDEBUG1)
            LogData(LOGDEBUG1,"  After OpenMTXBlockIP2S() Entered", pWS);

        LockBlockIP2S(pWS, &MutexBlockIP2S);
        if (pWS->nDebugFlag>=LOGDEBUG1)
            LogData(LOGDEBUG1,"  After LockBlockIP2S() Entered", pWS);

        /* See if another thread have started the SHared Memory Server(SMS) */

        hShrMemFile = OpenFileMapping(FILE_MAP_ALL_ACCESS,     /* read/write access        */
                                      FALSE,                   /* Do not inherit the name  */
                                      filename);               /* Name of mapping object   */

        if (hShrMemFile == NULL)
        {
            if (StartShrMemServer(pWS) !=0)
            {
                if (pWS->nDebugFlag>=LOGDEBUG)
                    LogData(LOGDEBUG,"  UpdateShrMem() Start of Shared memory server failed.",pWS);
                return -1;
            }
            /* Obtain file again */

            hShrMemFile = OpenFileMapping(FILE_MAP_ALL_ACCESS,     /* read/write access        */
                                          FALSE,                   /* Do not inherit the name  */
                                          filename);               /* Name of mapping object   */

            if (pWS->nDebugFlag>=LOGDEBUG1)
            {
/*                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() hShrMemFile [0x%lx]", hShrMemFile);*/
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() hShrMemFile [0x%p]", hShrMemFile);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
            }

            if (hShrMemFile == NULL)
            {
                LogData(LOGDATA, "  UpdateShrMem() Shared memory could not be allocated.",pWS);
                /* release mutex here */
                FreeBlockIP2S(pWS, &MutexBlockIP2S);
                LogData(LOGDATA,"  UpdateShrMem() leaving", pWS);
                return -1;
            }

            /* release mutex here */
            FreeBlockIP2S(pWS, &MutexBlockIP2S);
            if (pWS->nDebugFlag>=LOGDEBUG1)
                LogData(LOGDEBUG1,"  After FreeBlockIP2S() Entered", pWS);
        }
    }
    /* eod */

    pBuf=MapViewOfFile(hShrMemFile,                        /* handle to mapping object */
                       FILE_MAP_ALL_ACCESS,                /* Read/write access        */
                       0,
                       0,
                       SHMSZ);
    if (pBuf == NULL)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() Could not map view of file [%s] error %d\n",filename,GetLastError());
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDATA, pWS->szTmpBufff, pWS);
        if (pWS->nDebugFlag>=LOGDEBUG1)
            LogData(LOGDEBUG1,"  UpdateShrMem() leaving", pWS);
        return -1;
    }

    shm = (struct _shmem_dataspace*) pBuf;

    if (pWS->nDebugFlag>=LOGDEBUG1)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() Memory table contains %ld channel entries",shm->iNumOfChannels);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
    }

    /*
    *  We're ready to locate the channel entry
    */
    for (channelix=0; channelix < shm->iNumOfChannels; channelix++)
    {
        if (!strcmp(shm->chnl[channelix].szChannelName, pWS->szChannelName))
        {
            found=channelix;
            if (pWS->nDebugFlag>=LOGDEBUG1)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() Channel [%s] located at %ld", pWS->szChannelName,found);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
            }
            channelix=shm->iNumOfChannels;   /* Stop loop */
        }
    }

    /*
    *  if we didn't locate the channel insert channel entry
    */
    if (found==-1)
    {
        found=channelix;
        if (shm->iNumOfChannels<SHMMAX_CHL-1)
        {
            DBGPRINTF((OST, "  LCS -- Debugger %d\n",__LINE__));
            LockChannelSem(pWS);
            /* see if another have inserted before us (just to be shure) */
            found=-1;
            for (channelix=0; channelix < shm->iNumOfChannels; channelix++)
            {
                if (!strcmp(shm->chnl[channelix].szChannelName, pWS->szChannelName))
                {
                    found=channelix;
                    if (pWS->nDebugFlag>=LOGDEBUG1)
                    {
                        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() Channel [%s] located at %ld", pWS->szChannelName,found);
                        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
                    }
                    channelix=shm->iNumOfChannels;   /* Stop loop */
                }
            }
            if (found == -1)
            {
                if ((shm->iNumOfChannels<SHMMAX_CHL))   /* Still, just to be shure 2.82 */
                {
                    channelix=shm->iNumOfChannels++;
                    found=channelix;
                    shm->chnl[channelix].iNextChl=0;
                    shm->chnl[channelix].iActiveChannelLimit=pWS->lMaxChannelActive;
                    shm->chnl[channelix].iCurrentActiveConnections=0;
                    shm->chnl[channelix].iMaxActiveConnections=0;
                    strcpy(shm->chnl[channelix].szChannelName,pWS->szChannelName);
                    if (pWS->nDebugFlag>=LOGDEBUG1)
                    {
                        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() Channel [%s] added to channel tabel at %ld", pWS->szChannelName,channelix);
                        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
                    }
                }
            }
            DBGPRINTF((OST, "  FCS -- Debugger %d\n",__LINE__));
            FreeChannelSem(pWS);
        }
        else
        {
            UnmapViewOfFile(pBuf);         /* detaches the shared memory segment */
            CloseHandle(hShrMemFile);
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() No more room in channel table limit is: %ld",SHMMAX_CHL);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA, pWS->szTmpBufff, pWS);
            if (pWS->nDebugFlag>=LOGDEBUG1)
                LogData(LOGDEBUG1,"  UpdateShrMem() leaving", pWS);
            return CurrentAc;
        }

    }

    /*
    *  now it's time to see what we're doing.....
    */

    channelix=found; /* Setup pointer */

    if(shm->chnl[channelix].iActiveChannelLimit != pWS->lMaxChannelActive)
    {
        DBGPRINTF((OST, "  LCS -- Debugger %d\n",__LINE__));
        LockChannelSem(pWS);
        shm->chnl[channelix].iActiveChannelLimit=pWS->lMaxChannelActive;
        if (pWS->nDebugFlag>=LOGDEBUG1)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() Channel [%s] limit changed to %ld", pWS->szChannelName,
                     pWS->lMaxChannelActive);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
        }
        DBGPRINTF((OST, "  FCS -- Debugger %d\n",__LINE__));
        FreeChannelSem(pWS);
    }

    if (shm->chnl[channelix].iCurrentActiveConnections+1>shm->chnl[channelix].iActiveChannelLimit)
    {
        if (pWS->nDebugFlag>=LOGDEBUG1)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() Channel [%s] max lim was reached %ld", pWS->szChannelName,
                     shm->chnl[channelix].iActiveChannelLimit);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
        }
    }
    else
    {
        CurrentAc=0;      /* value below limit */
    }
    DBGPRINTF((OST, "  LCS -- Debugger %d\n",__LINE__));
    LockChannelSem(pWS);
    if (what2do != (long*)-1)
    {
        shm->chnl[channelix].iCurrentActiveConnections++;
        if (shm->chnl[channelix].iCurrentActiveConnections > shm->chnl[channelix].iMaxActiveConnections)
            shm->chnl[channelix].iMaxActiveConnections=shm->chnl[channelix].iCurrentActiveConnections;
        if (pWS->nDebugFlag>=LOGDEBUG1)
            LogData(LOGDEBUG1, "  UpdateShrMem() Increment counter", pWS);
    }
    else
    {
        if (shm->chnl[channelix].iCurrentActiveConnections > 0)
            shm->chnl[channelix].iCurrentActiveConnections--;
        if (pWS->nDebugFlag>=LOGDEBUG1)
            LogData(LOGDEBUG1, "  UpdateShrMem() Decrement counter", pWS);
    }
    pWS->iCurrentActiveConnections =shm->chnl[channelix].iCurrentActiveConnections;
    DBGPRINTF((OST, "  FCS -- Debugger %d\n",__LINE__));
    FreeChannelSem(pWS);
    if (pWS->nDebugFlag>=LOGDEBUG1)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() status return %ld (0=below limit)", CurrentAc);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
    }

    UnmapViewOfFile(pBuf);

    CloseHandle(hShrMemFile);

    if (pWS->nDebugFlag>=LOGDEBUG1)
        LogData(LOGDEBUG1,"  UpdateShrMem() leaving", pWS);
    return CurrentAc;
}

/*********************************************************************************************
* PrintSid: function format a windows Security Identifier Structure (SID)
*
* @Return 0 if failure, 1 succesfull print
*********************************************************************************************/
int PrintSid( SID *ps, char * resstring, int len )
{
    PSID_IDENTIFIER_AUTHORITY psia;
    DWORD subAuthCount, i;
    char work[64], result[255];

    strcat( result,"");

    psia = GetSidIdentifierAuthority( ps );
    subAuthCount = *GetSidSubAuthorityCount( ps );
    snprintf(result,sizeof(result), "S-%lu-", SID_REVISION );
    result[sizeof(result)-1]='\00';

    if ( ( psia->Value[0] != 0 ) || ( psia->Value[1] != 0 ) )
        snprintf(work,sizeof(work), "0x%02hx%02hx%02hx%02hx%02hx%02hx", (USHORT) psia->Value[0], (USHORT) psia->Value[1],
                 (USHORT) psia->Value[2], (USHORT) psia->Value[3], (USHORT) psia->Value[4], (USHORT) psia->Value[5] );
    else
        snprintf(work,sizeof(work), "%lu", (ULONG) ( psia->Value[5] ) + (ULONG) ( psia->Value[4] << 8 ) +
                 (ULONG) ( psia->Value[3] << 16 ) + (ULONG) ( psia->Value[2] << 24 ) );
    work[sizeof(work)-1]='\00';
    strcat(result,work);

    for ( i = 0; i < subAuthCount; ++ i )
    {
        snprintf(work,sizeof(work), "-%lu", *GetSidSubAuthority( ps, i ) );
        work[sizeof(work)-1]='\00';
        strcat(result,work);
    }
    if (strlen(resstring)<=len) {
        strcpy(resstring,result);
        return 1;
    }
    else
        return 0;
}

PSID SpcLookupName(LPCTSTR lpszSystemName, LPCTSTR lpszAccountName, workingStorage_type *pWS)
{
    PSID         Sid;
    DWORD        cbReferencedDomainName, cbSid;
    LPTSTR       ReferencedDomainName;
    SID_NAME_USE eUse;

    cbReferencedDomainName = cbSid = 0;
    if (LookupAccountName(lpszSystemName, lpszAccountName, 0, &cbSid,
                          0, &cbReferencedDomainName, &eUse)) {
        SetLastError(ERROR_NONE_MAPPED);
        return 0;
    }
    if (GetLastError(  ) != ERROR_INSUFFICIENT_BUFFER) return 0;

    if (!(Sid = (PSID)LocalAlloc(LMEM_FIXED, cbSid))) return 0;
    ReferencedDomainName = (LPTSTR)LocalAlloc(LMEM_FIXED, cbReferencedDomainName);
    if (!ReferencedDomainName) {
        LocalFree(Sid);
        return 0;
    }

    if (!LookupAccountName(lpszSystemName, lpszAccountName, Sid, &cbSid,
                           ReferencedDomainName, &cbReferencedDomainName, &eUse)) {
        LocalFree(ReferencedDomainName);
        LocalFree(Sid);
        return 0;
    }

    LocalFree(ReferencedDomainName);
    return Sid;
}


/* CreateMyDACL.
*    Create a security descriptor that contains the DACL
*    you want.
*    This function uses SDDL to make Deny and Allow ACEs.
*
* Parameter:
*    SECURITY_ATTRIBUTES * pSA
*    Pointer to a SECURITY_ATTRIBUTES structure. It is your
*    responsibility to properly initialize the
*    structure and to free the structure's
*    lpSecurityDescriptor member when you have
*    finished using it. To free the structure's
*    lpSecurityDescriptor member, call the
*    LocalFree function.
*
* Return value:
*    FALSE if the address to the structure is NULL.
*    Otherwise, this function returns the value from the
*    ConvertStringSecurityDescriptorToSecurityDescriptor
*    function.
* msdn ref.
*/

BOOL CreateMyDACL(SECURITY_ATTRIBUTES * pSA, workingStorage_type *pWS)
{
    /* Define the SDDL for the DACL.

    To be changed in the future

    We're granting group mqm access to the object
    */

    char work[256],w2[256];
    SID *sid;

    TCHAR * szSD = TEXT("D:")                   /* Discretionary ACL               */
                   TEXT("(A;OICI;GA;;;BG)")     /* Allow access to                 */
                   /* built-in guests                 */
                   TEXT("(A;OICI;GA;;;AN)")     /* Allow access to                 */
                   /* anonymous logon                 */
                   TEXT("(A;OICI;GRGWGX;;;AU)") /* Allow                           */
                   /* read/write/execute              */
                   /* to authenticated                */
                   /* users                           */
                   TEXT("(A;OICI;GA;;;BA)")     /* Allow full control              */
                   /* to administrators               */
                   ;

    sid=SpcLookupName("","mqm",pWS);
    PrintSid(sid,work,sizeof(work));
    snprintf(w2,sizeof(w2),"%s(A;OICI;GRGWGXGA;;;%s)",szSD,work);
    w2[sizeof(w2)-1]='\00';
    if (pWS->nDebugFlag>=LOGDEBUG)
    {
        char buf[300];
        snprintf(buf,sizeof(buf), "      CreateMyDACL() %s", w2);
        buf[sizeof(buf)-1]='\00';
        LogData(LOGDEBUG, buf, pWS);
    }
    if (NULL == pSA)
        return FALSE;

    return ConvertStringSecurityDescriptorToSecurityDescriptor(
               w2,
               SDDL_REVISION_1,
               &(pSA->lpSecurityDescriptor),
               NULL);
}
#endif    /* exit for WIN32 */



#ifdef WIN32
/*   To be deleted when new is tested.
 *   Update the shadred table entry for the passed channel.
 */
long UpdateShrMem_old(MQCD *pChannelDefinition,  long* what2do, workingStorage_type *pWS)
{
    char     filename[128] = "";

    int      channelix=-1;
    int      found=-1;
    HANDLE   hShrMemFile;
    LPCTSTR  pBuf;
    long     CurrentAc;

    HANDLE   MutexBlockIP2S;

    shmem_type * shm;

#define BUFSIZE 4096

    STARTUPINFO si;
    PROCESS_INFORMATION pi;

    char szCommand2do[256];

    HANDLE   semaphore;
    DWORD    retVal;
    SECURITY_ATTRIBUTES  sec_attr;
    char     szSemKey[256];
    long     lCount;

    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1,"  UpdateShrMem() Entered", pWS);

    CurrentAc=-1;

    /* Generate the name of the file */
    strcpy(filename,"Global\\BlockIP2-");
    strcat(filename,pWS->szQMgr);
    strcat(filename,"-shm");                               /* shared mem                    */

    if (pWS->nDebugFlag >= LOGDEBUG1)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() shmem filename [%s] CurrentAc %ld", filename,CurrentAc);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
    }

    hShrMemFile = OpenFileMapping(FILE_MAP_ALL_ACCESS,     /* read/write access             */
                                  FALSE,                   /* Do not inherit the name       */
                                  filename);               /* Name of mapping object        */

    if (pWS->nDebugFlag >= LOGDEBUG1)
    {
/*        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() hShrMemFile [0x%lx]", hShrMemFile);*/
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() hShrMemFile [0x%p]", hShrMemFile);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
    }

    if (hShrMemFile == NULL)
    {
        /* This part must be encapsulated in a mutex/semaphore */
        /*Set mutex here and evt. semaphore here */

        OpenMTXBlockIP2S(pWS, &MutexBlockIP2S);
        if (pWS->nDebugFlag >= LOGDEBUG8)
            LogData(LOGDEBUG8,"  After OpenMTXBlockIP2S() Entered", pWS);

        LockBlockIP2S(pWS, &MutexBlockIP2S);
        if (pWS->nDebugFlag >= LOGDEBUG8)
            LogData(LOGDEBUG8,"  After LockBlockIP2S() Entered", pWS);

        /* See if another thread have started the SHared Memory Server(SMS) */

        hShrMemFile = OpenFileMapping(FILE_MAP_ALL_ACCESS, /* read/write access             */
                                      FALSE,               /* Do not inherit the name       */
                                      filename);           /* Name of mapping object        */

        if (hShrMemFile == NULL)
        {
            /* Start the shared memory server */
            if (LoadRegistry(pWS) == 0)
            {
                if (pWS->nDebugFlag >= LOGDEBUG1)
                    LogData(LOGDEBUG1,"  UpdateShrMem() LoadRegistry() OK",pWS);

                /* Initialise si and pi */
                ZeroMemory( &si, sizeof(si) );
                si.cb = sizeof(si);
                ZeroMemory( &pi, sizeof(pi) );

                strcpy(szCommand2do,pWS->ExitPath);
                strcat(szCommand2do,"\\BlockIP2S.exe ");
                strcat(szCommand2do,pWS->szQMgr);

                if (pWS->nDebugFlag >= LOGDEBUG1)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() CreateProcess [%s]", szCommand2do);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
                }
                /* create sem2. here */

                sec_attr.nLength              = sizeof( SECURITY_ATTRIBUTES );
                sec_attr.lpSecurityDescriptor = NULL;
                sec_attr.bInheritHandle       = TRUE;

                strcpy(szSemKey,"Global\\BlockIP2S-");
                strcat(szSemKey,pWS->szQMgr);
                strcat(szSemKey,"-ctrl");

                if (pWS->nDebugFlag >= LOGDEBUG1)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  Semaphore key is [%s].",  szSemKey);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
                }

                semaphore = CreateSemaphore( &sec_attr, 0, 65536, szSemKey );
                if( semaphore == (HANDLE) NULL )
                    return -1;

                if (pWS->nDebugFlag >= LOGDEBUG1)
                {
/*                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() Semaphore opened. (%lx)", semaphore);*/
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() Semaphore opened. (0x%p)", semaphore);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
                }

                /* Create a childProcess */
                if( !CreateProcess( NULL,                  /* No module name (use command line).*/
                                    szCommand2do,          /* Command to execute                */
                                    NULL,                  /* Process handle not inheritable.   */
                                    NULL,                  /* Thread handle not inheritable.    */
                                    FALSE,                 /* Set handle inheritance to FALSE.  */
                                    DETACHED_PROCESS,      /* fdwCreate                         */
                                    NULL,                  /* Use parent's environment block.   */
                                    NULL,                  /* Use parent's starting directory.  */
                                    &si,                   /* Pointer to STARTUPINFO structure. */
                                    &pi )                  /* Pointer to PROCESS_INFORMATION structure. */
                  )
                {
                    LogData(LOGDATA,"UpdateShrMem() CreateProcess failed.",pWS);
                }
                else
                {
                    LogData(LOGDATA,"UpdateShrMem() BlockIP2S Shared memory controller detached.",pWS);
                }
                lCount = 10L;
                while( (( retVal = WaitForSingleObject( semaphore, 250 ) ) == WAIT_TIMEOUT) && lCount > 0)
                {
                    lCount--;
                    if (pWS->nDebugFlag >= LOGDEBUG1)
                        LogData(LOGDEBUG1, "Timeout (250ms).", pWS);
                }
                if (lCount ==0)
                {
                    LogData(LOGDATA,"UpdateShrMem() BlockIP2S didn't respond correctly.",pWS);
                    if( ! ReleaseSemaphore(semaphore, 1, NULL) )      /* clean up           */
                        LogData(LOGDATA, "ReleaseSemaphore() Failed.", pWS);
                    CloseHandle(semaphore);
                    return -1;
                }

                if( ! ReleaseSemaphore(semaphore, 1, NULL) )
                {
                    LogData(LOGDATA, "ReleaseSemaphore() Failed.", pWS);
                    CloseHandle(semaphore);                /* Clean up                      */
                    return -1;
                }
                CloseHandle(semaphore);
                if (pWS->nDebugFlag >= LOGDEBUG1)
                    LogData(LOGDEBUG1, "  CloseHandle.", pWS);

                /* Obtain file again */

                hShrMemFile = OpenFileMapping(FILE_MAP_ALL_ACCESS,    /* read/write access        */
                                              FALSE,                  /* Do not inherit the name  */
                                              filename);              /* Name of mapping object   */

                if (pWS->nDebugFlag >= LOGDEBUG1)
                {
/*                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() hShrMemFile [0x%lx]", hShrMemFile);*/
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() hShrMemFile [0x%p]", hShrMemFile);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
                }

                if (hShrMemFile == NULL)
                {
                    LogData(LOGDEBUG1, "  UpdateShrMem() Shared memory was not allocated yet.",pWS);
                    /* release mutex here */
                    FreeBlockIP2S(pWS, &MutexBlockIP2S);
                    if (pWS->nDebugFlag >= LOGDEBUG1)
                        LogData(LOGDEBUG1,"  UpdateShrMem() Done", pWS);
                    return -1;
                }
            }
        }
        /* release mutex here */
        FreeBlockIP2S(pWS, &MutexBlockIP2S);
        if (pWS->nDebugFlag >= LOGDEBUG8)
            LogData(LOGDEBUG8,"  After FreeBlockIP2S() Entered", pWS);
    }

    pBuf=MapViewOfFile(hShrMemFile,                        /* handle to mapping object      */
                       FILE_MAP_ALL_ACCESS,                /* Read/write access             */
                       0,
                       0,
                       SHMSZ);
    if (pBuf == NULL)
    {

        /* here was BlockIP2S tried to get started */

        {
            DWORD eNum;
            char  errMsg[256];

            /* Error encountered; generate message and exit. */
            eNum = GetLastError();
            WINFmtMsg(eNum, errMsg, sizeof(errMsg));
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() Could not map view of file [%s]  %d %s",filename,eNum,errMsg);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA, pWS->szTmpBufff, pWS);
        }

        if (pWS->nDebugFlag >= LOGDEBUG1)
            LogData(LOGDEBUG1,"UpdateShrMem() Done", pWS);
        return -1;
    }

    shm = (struct _shmem_dataspace*)pBuf;

    if (pWS->nDebugFlag >= LOGDEBUG1)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() Memory table contains %ld channel entries",shm->iNumOfChannels);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
    }

    /*
     *  We're ready to locate the channel entry
     */
    for (channelix=0; channelix < shm->iNumOfChannels; channelix++)
    {
        if (!strcmp(shm->chnl[channelix].szChannelName, pWS->szChannelName))
        {
            found=channelix;
            if (pWS->nDebugFlag >= LOGDEBUG1)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() Channel [%s] located at %ld", pWS->szChannelName,found);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
            }
            channelix=shm->iNumOfChannels;                 /* Stop loop                     */
        }
    }

    /*
     *  if we didn't locate the channel insert channel entry
     */
    if (found==-1)
    {
        found=channelix;
        if (shm->iNumOfChannels<SHMMAX_CHL)
        {
            DBGPRINTF((OST, "  LCS -- Debugger %d\n",__LINE__));
            LockChannelSem(pWS);
            /* see if another have inserted before us (just to be shure) */
            found=-1;
            for (channelix=0; channelix < shm->iNumOfChannels; channelix++)
            {
                if (!strcmp(shm->chnl[channelix].szChannelName, pWS->szChannelName))
                {
                    found=channelix;
                    if (pWS->nDebugFlag >= LOGDEBUG1)
                    {
                        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() Channel [%s] located at %ld", pWS->szChannelName,found);
                        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
                    }
                    channelix=shm->iNumOfChannels;         /* Stop loop                     */
                }
            }
            if (found == -1)
            {
                channelix=shm->iNumOfChannels++;
                shm->chnl[channelix].iNextChl=0;
                shm->chnl[channelix].iActiveChannelLimit=pWS->lMaxChannelActive;
                shm->chnl[channelix].iCurrentActiveConnections=0;
                shm->chnl[channelix].iMaxActiveConnections=0;
                strcpy(shm->chnl[channelix].szChannelName,pWS->szChannelName);
                if (pWS->nDebugFlag >= LOGDEBUG1)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() Channel [%s] added to channel tabel at %ld", pWS->szChannelName,channelix);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
                }
            }
            DBGPRINTF((OST, "  FCS -- Debugger %d\n",__LINE__));
            FreeChannelSem(pWS);
        }
        else
        {
            UnmapViewOfFile(pBuf);                         /* detaches the shared memory segment */
            CloseHandle(hShrMemFile);
            if (pWS->nDebugFlag >= LOGDEBUG1)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() No more room in channel table limit is: %ld",SHMMAX_CHL);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDATA, pWS->szTmpBufff, pWS);                  /* 2.74 */
                LogData(LOGDEBUG1,"  UpdateShrMem() Done", pWS);
            }
            return CurrentAc;
        }

    }

    /*
     *  now it's time to see what we're doing.....
     */

    channelix=found;                                       /* Setup pointer                 */

    if(shm->chnl[channelix].iActiveChannelLimit != pWS->lMaxChannelActive)
    {
        DBGPRINTF((OST, "  LCS -- Debugger %d\n",__LINE__));
        LockChannelSem(pWS);
        shm->chnl[channelix].iActiveChannelLimit=pWS->lMaxChannelActive;
        if (pWS->nDebugFlag >= LOGDEBUG1)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() Channel [%s] limit changed to %ld", pWS->szChannelName,
                     pWS->lMaxChannelActive);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
        }
        DBGPRINTF((OST, "  FCS -- Debugger %d\n",__LINE__));
        FreeChannelSem(pWS);
    }

    if (shm->chnl[channelix].iCurrentActiveConnections+1>shm->chnl[channelix].iActiveChannelLimit)
    {
        if (pWS->nDebugFlag >= LOGDEBUG1)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() Channel [%s] max lim was reached %ld", pWS->szChannelName,
                     shm->chnl[channelix].iActiveChannelLimit);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
        }
    }
    else
    {
        CurrentAc=0;                                       /* value below limit             */
    }
    DBGPRINTF((OST, "  LCS -- Debugger %d\n",__LINE__));
    LockChannelSem(pWS);
    if (what2do != (long*)-1)
    {
        shm->chnl[channelix].iCurrentActiveConnections++;
        if (shm->chnl[channelix].iCurrentActiveConnections > shm->chnl[channelix].iMaxActiveConnections)
            shm->chnl[channelix].iMaxActiveConnections=shm->chnl[channelix].iCurrentActiveConnections;
        if (pWS->nDebugFlag >= LOGDEBUG1)
            LogData(LOGDEBUG1, "  UpdateShrMem() Increment counter", pWS);
    }
    else
    {
        if (shm->chnl[channelix].iCurrentActiveConnections > 0)
            shm->chnl[channelix].iCurrentActiveConnections--;
        if (pWS->nDebugFlag >= LOGDEBUG1)
            LogData(LOGDEBUG1, "  UpdateShrMem() Decrement counter", pWS);
    }
    pWS->iCurrentActiveConnections =shm->chnl[channelix].iCurrentActiveConnections;
    DBGPRINTF((OST, "  FCS -- Debugger %d\n",__LINE__));
    FreeChannelSem(pWS);
    if (pWS->nDebugFlag >= LOGDEBUG1)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() status return %ld (0=below limit) current count %ld", CurrentAc,
                 pWS->iCurrentActiveConnections);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
    }

    UnmapViewOfFile(pBuf);

    CloseHandle(hShrMemFile);

    if (pWS->nDebugFlag >= LOGDEBUG8)
        LogData(LOGDEBUG8,"UpdateShrMem() Done", pWS);
    return CurrentAc;
}

#endif


#if defined (UNIX) || defined (OS400)

/*
 *   Update the shared table entry for the passed channel.
 */
long UpdateShrMem(MQCD *pChannelDefinition,  long* what2do, workingStorage_type *pWS)
{
    long CurrentAc=-1;
    char filename[128] = "";
    FILE *fptr;
    int  shmid;
    int iRCMaster;                                              /* reasoncode from master file       2.90 */
    int channelix=-1;
    int found=-1;
    int iId=676735;

    shmem_type * shm;
    char master_time[15];

    struct tm *timenow;
#ifdef _THREAD_SAFE_JHP
    struct tm  wk_localtime;
#endif
    time_t longtime;
    long lTimeNow;

    if (pWS->KeyShmem==-1)
    {    /* key not obtained yet */

        if (pWS->lTmpPath==0)                                   /* mode added under                  2.90 */
        {
            strcpy(filename,pWS->szQmgrPrefix);
            strcat(filename,"/qmgrs/");
            strcat(filename,pWS->szQmgrDirectory);
#if defined (MQCD_VERSION_8)                                    /* New stuff only                    2.90 */
            if (pWS->Version>MQCD_VERSION_8)                    /* moved shmem folder ?              2.90 */
                strcat(filename,"/@ipcc/shmem/BlockIP2.2a");    /* new shared mem                    2.90 */
            else
                strcat(filename,"/shmem/BlockIP2.2a");          /* old shared mem                    2.74 */
#else                                                           /* Old stuff                         2.90 */
            strcat(filename,"/shmem/BlockIP2.2a");              /* shared mem                        2.74 */
#endif
        }
        else
        {
            strcpy(filename,"/tmp/.");                          /* Move anchor to /tmp./ added under 2.90 */
            strcat(filename,pWS->szQMgr);
            strcat(filename,"_BlockIP2.2");                     /* shared mem                        2.74 */
        }
        if (pWS->nDebugFlag >= LOGDEBUG1)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() - generate shm shmem [%s] MQCD version %d", filename,pWS->Version);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
        }

        if((pWS->KeyShmem=ftok(filename,iId)) == -1)
        {
            if((fptr = fopen( filename,"a+"))!=NULL)
                fclose(fptr);
            if((pWS->KeyShmem=ftok(filename,iId)) == -1)
            {
                LogData(LOGDATA,"ftok error -1", pWS);
                perror("ftok");
                return -1;
            }
        }
    }

    if ((shmid = shmget(pWS->KeyShmem, SHMSZ, 0660)) < 0)
    {
        DBGPRINTF((OST, "  LAS -- Debugger %d\n",__LINE__));
        DBGPRINTF((OST, "UpdateShrMem() Create shared memory area...\n"));
        LockAllSem(pWS);
        if (pWS->nDebugFlag >= LOGDEBUG1)
            LogData(LOGDEBUG1,"UpdateShrMem() Create shared memory area...", pWS);
        if ((shmid = shmget(pWS->KeyShmem, SHMSZ, IPC_CREAT | 0660)) < 0)
        {
            LogData(LOGDATA,"Creation failed(shmget) -1", pWS);
            DBGPRINTF((OST, "  FAS -- Debugger %d\n",__LINE__));
            FreeAllSem(pWS);
            return -1;
        }

        shm = shmat(shmid, NULL, 0);
        DBGPRINTF((OST, "  shmat() -- Debugger %d\n",__LINE__));
        if ((size_t)shm  == (size_t) -1)
        {
            LogData(LOGDATA,"  UpdateShrMem() Addressing of shmem failed(shmat) -1", pWS);
            DBGPRINTF((OST, "  FAS -- Debugger %d\n",__LINE__));
            FreeAllSem(pWS);
            perror("shmat");
            return(-1);
        }

        strcpy(shm->master_time,"");                            /* of time field                 2.90 */
        shm->iPid= iId;                                         /* Say it's mine                      */
        shm->iNumOfChannels=0;                                  /* No channes yet                     */

        if (pWS->nDebugFlag >= LOGDEBUG0)
            LogData(LOGDEBUG0, "   StbyMode check.." ,pWS);
        DBGPRINTF((OST, "   StbyMode  %s\n",master_time));
        if (pWS->iStbyMode==1)                                  /* version 7.0.1  ?              2.90 */
        {
            if (pWS->nDebugFlag >= LOGDEBUG0)
                LogData(LOGDEBUG0, "   StbyMode enabled" ,pWS);

            time(&longtime);                                    /* insert timestamp              2.90 */
#ifdef _THREAD_SAFE_JHP
            memset(&wk_localtime, '\0', sizeof(wk_localtime));  /*                               2.90 */
            timenow = localtime_r(&longtime,&wk_localtime);     /*                               2.90 */
#else
            timenow = localtime(&longtime);                     /*                               2.90 */
#endif
                                                                /* Time to live before we check  2.90 */
            shm->lTmpTime=(timenow->tm_hour*3600)+(timenow->tm_min*60)+timenow->tm_sec;
            strcpy(master_time,"");
            HasMasterChanged(master_time, pWS);                /* Obtain info from master file  2.90 */
            DBGPRINTF((OST, "   HMC  %s\n",master_time));
            strcpy(shm->master_time,master_time);

            DBGPRINTF((OST, "  1-shm->iNumOfChannels %d\n",shm->iNumOfChannels));
        }

        shm->iMaxChlInTable=SHMMAX_CHL;                         /* keep info for later versions       */
        if (pWS->nDebugFlag >= LOGDEBUG1)
            LogData(LOGDEBUG1,"  UpdateShrMem() Shared memory initialized.", pWS);

        shmdt((char*)shm);                                      /* detaches the shared memory segment */
        DBGPRINTF((OST, "  shmdt() -- Debugger %d\n",__LINE__));
        DBGPRINTF((OST, "  FAS -- Debugger %d\n",__LINE__));
        FreeAllSem(pWS);
    }
    if (pWS->nDebugFlag >= LOGDEBUG8)
        LogData(LOGDEBUG8,"  UpdateShrMem() Working",pWS);
    /*
     * Now we attach the segment to our data space.
     */
    shm = shmat(shmid, NULL, 0);
    DBGPRINTF((OST, "  shmat() -- Debugger %d\n",__LINE__));
    if ((size_t)shm == (size_t) -1)
    {
        LogData(LOGDATA,"  UpdateShrMem() Addressing of shmem failed(shmat) -1", pWS);
        perror("shmat");
        return(-1);
    }

    /*
     *  See if it's my own
     */
    if (shm->iPid != iId)
    {    /* Nope, then it's time to init */
        DBGPRINTF((OST, "  LAS -- Debugger %d\n",__LINE__));
        LockAllSem(pWS);
        if (shm->iPid != iId)
        {                                                  /* Just to be shure....          */
            shm->iPid= iId;                                /* Say it's mine                 */
            shm->iNumOfChannels=0;                         /* No channes yet                */
            if (pWS->nDebugFlag >= LOGDEBUG1)
                LogData(LOGDEBUG1,"  UpdateShrMem() Shared memory initialized.", pWS);
        }
        DBGPRINTF((OST, "  FAS -- Debugger %d\n",__LINE__));
        FreeAllSem(pWS);
    }

    /*
     *  Check to see if it's time to look on the master file....
     *  Added under                                                                              2.90
     */

     if (pWS->iStbyMode==1)                                     /* version 7.0.1  ?              2.90 */
     {
       time(&longtime);                                         /* insert timestamp              2.90 */
#ifdef _THREAD_SAFE_JHP
       memset(&wk_localtime, '\0', sizeof(wk_localtime));       /*                               2.90 */
       timenow = localtime_r(&longtime,&wk_localtime);          /*                               2.90 */
#else
       timenow = localtime(&longtime);                          /*                               2.90 */
#endif
        lTimeNow=(timenow->tm_hour*3600)+(timenow->tm_min*60)+timenow->tm_sec;

        if (pWS->nDebugFlag >= LOGDEBUG1)
        {
           snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  lTmpTime %08lX  lTimeNow %08lX TTL: %d MQCD.Version %d", shm->lTmpTime,lTimeNow,500,pWS->Version);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
        }

        if ((lTimeNow < shm->lTmpTime) || (lTimeNow > (shm->lTmpTime+20L)))
        {
            if (pWS->ExitReason==MQXR_INIT)                     /* Just do it on INIT            2.90 */
            {
                strcpy(master_time,shm->master_time);           /* Pass over new time            2.90 */
                iRCMaster=HasMasterChanged(master_time, pWS);   /* Obtain info from master file  2.90 */
                if (iRCMaster==4)                               /* file was changed....          2.90 */
                {
                    if (pWS->nDebugFlag >= LOGDEBUG1)
                       LogData(LOGDEBUG1, "  Update master_time..." ,pWS);
                    DBGPRINTF((OST, "  LAS -- Debugger %d\n",__LINE__));
                    LockAllSem(pWS);
                    strcpy(shm->master_time,master_time);       /* Update file time in struct.   2.90 */
                    shm->iPid= iId;                             /* Say it's mine                 2.90 */
                    shm->iNumOfChannels=0;                      /* No channes yet                2.90 */

                    DBGPRINTF((OST, "  FAS -- Debugger %d\n",__LINE__));
                    FreeAllSem(pWS);
                    DBGPRINTF((OST, "  1-shm->iNumOfChannels %d\n",shm->iNumOfChannels));
                }

                if (pWS->nDebugFlag>=LOGDEBUG1)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  pmaster_time %s shm %s",master_time, shm->master_time);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDEBUG1, pWS->szTmpBufff,pWS);
                }
                shm->lTmpTime=lTimeNow;                         /* Update Time before recheck    2.90 */
            }
        }
    }

    /*
     *  We're ready to locate the channel entry
     */

    for (channelix=0; channelix < shm->iNumOfChannels; channelix++)
    {
        if (!strcmp(shm->chnl[channelix].szChannelName, pWS->szChannelName))
        {
            found=channelix;
            if (pWS->nDebugFlag >= LOGDEBUG1)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() Channel [%s] located at %d", pWS->szChannelName,found);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
            }
            channelix=shm->iNumOfChannels;                 /* Stop loop                     */
        }
    }

    /*
     *  if we didn't locate the channel insert channel entry
     */
    if (found==-1) {
        found=channelix;
        if ((shm->iNumOfChannels<SHMMAX_CHL) && (shm->iNumOfChannels<shm->iMaxChlInTable))   /* Just to be shure 2.74 */
        {
            DBGPRINTF((OST, "  LCS -- Debugger %d\n",__LINE__));
            LockChannelSem(pWS);
            /* Start 2.82 */
            /* see if another have added the entry before us */
            found = -1;
            for (channelix=0; channelix < shm->iNumOfChannels; channelix++)
            {
                if (!strcmp(shm->chnl[channelix].szChannelName, pWS->szChannelName))
                {
                    found=channelix;
                    if (pWS->nDebugFlag>=LOGDEBUG1)
                    {
                        sprintf(pWS->szTmpBufff,"  UpdateShrMem() Channel [%s] located at %d", pWS->szChannelName,found);
                        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
                    }
                    channelix=shm->iNumOfChannels;   /* Stop loop */
                }
            }
            if (found == -1)  /* We are the first on the channel */
            {
                if ((shm->iNumOfChannels<SHMMAX_CHL) && (shm->iNumOfChannels<shm->iMaxChlInTable))   /* Still, just to be shure 2.82 */
                {
                    /* end 2.82 */
                    channelix=shm->iNumOfChannels++;
                    shm->chnl[channelix].iNextChl=0;
                    shm->chnl[channelix].iActiveChannelLimit=pWS->lMaxChannelActive;
                    shm->chnl[channelix].iCurrentActiveConnections=0;
                    shm->chnl[channelix].iMaxActiveConnections=0;
                    strcpy(shm->chnl[channelix].szChannelName,pWS->szChannelName);

                    DBGPRINTF((OST, "UpdateShrMem() added %s at pos %d lim %ld\n",pWS->szChannelName,channelix,pWS->lMaxChannelActive));
                    DBGPRINTF((OST, "USM(1) iCAC %d iACL %d\n",shm->chnl[channelix].iCurrentActiveConnections,shm->chnl[channelix].iActiveChannelLimit));

                    if (pWS->nDebugFlag >= LOGDEBUG1)
                    {
                        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() Channel [%s] added to channel tabel at %d", pWS->szChannelName,channelix);
                        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
                    }
                }
            }
            DBGPRINTF((OST, "  FCS -- Debugger %d\n",__LINE__));
            FreeChannelSem(pWS);
        }
        else
        {
            shmdt((char*)shm);                             /* detaches the shared memory segment */
            DBGPRINTF((OST, "  shmdt() -- Debugger %d\n",__LINE__));
            if (pWS->nDebugFlag >= LOGDEBUG1)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() No more room in channel table limit is: %ld",(long)SHMMAX_CHL);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
                LogData(LOGDEBUG8,"  UpdateShrMem() Done", pWS);
            }
            return CurrentAc;
        }
    }

    /*
     *  now it's time to see what we're doing.....
     */

    channelix=found;                                       /* Setup pointer                 */

    if(shm->chnl[channelix].iActiveChannelLimit != pWS->lMaxChannelActive)
    {
        DBGPRINTF((OST, "  LCS -- Debugger %d\n",__LINE__));
        LockChannelSem(pWS);
        shm->chnl[channelix].iActiveChannelLimit=pWS->lMaxChannelActive;
        if (pWS->nDebugFlag >= LOGDEBUG1)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() Channel [%s] limit changed to %ld", pWS->szChannelName,
                     pWS->lMaxChannelActive);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
        }
        DBGPRINTF((OST, "  FCS -- Debugger %d\n",__LINE__));
        FreeChannelSem(pWS);
    }

    DBGPRINTF((OST, "USM(2) iCAC %d iACL %d\n",shm->chnl[channelix].iCurrentActiveConnections,shm->chnl[channelix].iActiveChannelLimit));
    if (shm->chnl[channelix].iCurrentActiveConnections+1>shm->chnl[channelix].iActiveChannelLimit)
    {
        if (pWS->nDebugFlag >= LOGDEBUG1)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() Channel [%s] max lim was reached %d", pWS->szChannelName,
                     shm->chnl[channelix].iActiveChannelLimit);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
        }
    }
    else
    {
        CurrentAc=0;                                       /* value below limit             */
    }
    DBGPRINTF((OST, "  LCS -- Debugger %d\n",__LINE__));
    LockChannelSem(pWS);
    if (what2do != (long*)-1)
    {
        shm->chnl[channelix].iCurrentActiveConnections++;
        if (shm->chnl[channelix].iCurrentActiveConnections > shm->chnl[channelix].iMaxActiveConnections)
            shm->chnl[channelix].iMaxActiveConnections=shm->chnl[channelix].iCurrentActiveConnections;
        if (pWS->nDebugFlag >= LOGDEBUG1)
            LogData(LOGDEBUG1, "  UpdateShrMem() Increment counter", pWS);
    }
    else
    {
        if (shm->chnl[channelix].iCurrentActiveConnections > 0)
            shm->chnl[channelix].iCurrentActiveConnections--;
        LogData(LOGDEBUG1, "  UpdateShrMem() Decrement counter", pWS);
    }
    pWS->iCurrentActiveConnections =shm->chnl[channelix].iCurrentActiveConnections;
    DBGPRINTF((OST, "  FCS -- Debugger %d\n",__LINE__));
    FreeChannelSem(pWS);
    if (pWS->nDebugFlag >= LOGDEBUG1)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  UpdateShrMem() status return %ld (0=below limit) current count %ld", CurrentAc,
                 pWS->iCurrentActiveConnections);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
    }

    shmdt((char*)shm); 	                                   /* detaches the shared memory segment */
    DBGPRINTF((OST, "  shmdt() -- Debugger %d\n",__LINE__));
    return CurrentAc;
}
/*
    key_t     KeySemaphores;
    key_t     KeyShmem;
    LockPWS1Sem();
#define SEM_SHM1 0
#define SEM_CHL1 1
#define SEM_PWS1 2
    */

int LockAllSem(workingStorage_type *pWS)
{
    int      nsops; /* number of operations to do */
    struct   sembuf *sops = (struct sembuf *) malloc(6*sizeof(struct sembuf));
    int      rc, workint;
    char     filename[128] = "";
    FILE    *fptr;
    int      semid;
    int      iId=676735;


    rc=16;

    if (pWS->nDebugFlag >= LOGDEBUG1)
    {
        LogData(LOGDEBUG1, "LockAllSem() Entered", pWS);
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  KeySemaphores %d",pWS->KeySemaphores);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
    }


    DBGPRINTF((OST,"LockAllSem() Entered, KeySemaphores %d\n",pWS->KeySemaphores));

/*    if (pWS->KeySemaphores==-1)*/
    if (1==1)
    {                                                           /* key not obtained yet                   */
        if (pWS->nDebugFlag >= LOGDEBUG1)
            LogData(LOGDEBUG1, "  Prepare sems", pWS);

        if (pWS->lTmpPath==0)                                   /* mode added under                  2.90 */
        {
            strcpy(filename,pWS->szQmgrPrefix);
            strcat(filename,"/qmgrs/");
            strcat(filename,pWS->szQmgrDirectory);
#if defined (MQCD_VERSION_8)                                    /* New stuff only                    2.90 */
            if (pWS->Version>MQCD_VERSION_8)                    /* moved shmem folder ?              2.90 */
                strcat(filename,"/@ipcc/shmem/BlockIP2.1");     /* new shared sem                    2.90 */
            else
                strcat(filename,"/shmem/BlockIP2.1");           /* old shared sem                         */
#else                                                           /* Old stuff                         2.90 */
            strcat(filename,"/shmem/BlockIP2.1");               /* shared sem                             */
#endif
        }
        else
        {
            strcpy(filename,"/tmp/.");                          /* Move anchor to /tmp./ added under 2.90 */
            strcat(filename,pWS->szQMgr);
            strcat(filename,"_BlockIP2.1");                     /* shared sem                             */
        }

        if (pWS->nDebugFlag >= LOGDEBUG1)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  generate semaphores shmem [%s]", filename);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
        }

        if((pWS->KeySemaphores=ftok(filename,iId)) == -1)
        {
            if((fptr = fopen( filename,"a+"))!=NULL)
                fclose(fptr);
            if((pWS->KeySemaphores=ftok(filename,iId)) == -1)
            {
                LogData(LOGDATA,"ftok error -1", pWS);
                perror("ftok");
                free(sops);                  /* 2.65 fix */
                return rc;
            }
        }
        LogData(LOGDEBUG8, "  ftok1 %ld ok", pWS);

        /* open semaphore set with 6 semaphore: */
        if ((semid = semget(pWS->KeySemaphores, 6, 0660 )) == -1)
        {
            /* create a semaphore set with 6 semaphore: */
            if (pWS->nDebugFlag >= LOGDEBUG1)
                LogData(LOGDEBUG1,"  Create semaphores...", pWS);

            if ((semid = semget(pWS->KeySemaphores, 6, 0660 | IPC_CREAT)) == -1)
            {
                LogData(LOGDATA,"  semget error -1", pWS);
                free(sops);                  /* 2.65 fix */
                return rc;
            }
            if (pWS->nDebugFlag >= LOGDEBUG1)
                LogData(LOGDEBUG1,"  Create semaphores... Done", pWS);
        }
    }


    nsops = 4;                                             /* Just lock Master and Channel            2.90 */

    /* SHM Master lock */
    sops[0].sem_num = 0;                                   /* We only use one track                        */
    sops[0].sem_op = 0;                                    /* wait for semaphore flag to become zero       */
    sops[0].sem_flg = SEM_UNDO;                            /* take off semaphore asynchronous              */

    sops[1].sem_num = 0;
    sops[1].sem_op = 1;                                    /* increment semaphore -- take control of track */
    sops[1].sem_flg = SEM_UNDO | IPC_NOWAIT;               /* take off semaphore                           */

    /* SHM Channel lock */
    sops[2].sem_num = 1;                                   /* We only use one track                        */
    sops[2].sem_op = 0;                                    /* wait for semaphore flag to become zero       */
    sops[2].sem_flg = SEM_UNDO;                            /* take off semaphore asynchronous              */


    sops[3].sem_num = 1;
    sops[3].sem_op = 1;                                    /* increment semaphore -- take control of track */
    sops[3].sem_flg = SEM_UNDO | IPC_NOWAIT;               /* take off semaphore                           */


    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1,"  executing semop...", pWS);

    DBGPRINTF((OST,"LockAllSem() executing semop...\n"));

    /* Make the semop() call and report the results. */
    if ((workint = semop(semid, sops, nsops)) == -1)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"LockAllSem() semop: semop failed rc: %d errno: %d %s", workint, errno, strerror(errno) );
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDATA, pWS->szTmpBufff, pWS);
    }
    else
    {
        rc=0;
    }

    free(sops);                  /* 2.65 fix */
    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1, "LockAllSem() Done", pWS);
    DBGPRINTF((OST,"LockAllSem() Done, KeySemaphores %d\n",pWS->KeySemaphores));
    return rc;
}  /* LockAllSem() */

int FreeAllSem(workingStorage_type *pWS)
{
    int nsops;                                             /* number of operations to do    */
    struct sembuf *sops = (struct sembuf *) malloc(3*sizeof(struct sembuf));
    int rc, workint;
    int semid;

    rc=16;
    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1, "FreeAllSem() Entered", pWS);

    DBGPRINTF((OST,"FreeAllSem() Entered, KeySemaphores %d\n",pWS->KeySemaphores));

    /* open semaphore set with 6 semaphore: */
    if ((semid = semget(pWS->KeySemaphores, 6, 0660 )) == -1)
    {
        /* create a semaphore set with 6 semaphore: */
        if (pWS->nDebugFlag >= LOGDEBUG1)
            LogData(LOGDEBUG1,"FreeAllSem() semget failed ...", pWS);
        DBGPRINTF((OST,"FreeAllSem() semget failed\n"));
        free(sops);                  /* 2.65 fix */
        return rc;
    }

    nsops = 2;                                                  /* Just lock Master and Channel      2.90 */

    /* wait for semaphore to reach zero */
    /* SHM Master lock */
    sops[0].sem_num = 0;
    sops[0].sem_op = -1;                                        /* Give UP Control of track               */
    sops[0].sem_flg = SEM_UNDO | IPC_NOWAIT;                    /* take off semaphore, asynchronous       */

    /* SHM Channel lock */
    sops[1].sem_num = 1;
    sops[1].sem_op = -1;                                        /* Give UP COntrol of track               */
    sops[1].sem_flg = SEM_UNDO | IPC_NOWAIT;                    /* take off semaphore, asynchronous       */

    if ((workint = semop(semid, sops, nsops)) == -1)
    {
        LogData(LOGDATA,"FreeAllSem() semop: semop failed", pWS);
        free(sops);                  /* 2.65 fix */
        return rc;
    }
    else
        rc=0;
    free(sops);                  /* 2.65 fix */
    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1, "FreeAllSem() Done", pWS);
    DBGPRINTF((OST,"FreekAllSem() Done\n"));
    return rc;
}  /* FreeAllSem() */


int LockSpecSem(int sem_number, workingStorage_type *pWS)
{
    int      nsops;                                        /* number of operations to do    */
    struct   sembuf *sops = (struct sembuf *) malloc(2*sizeof(struct sembuf));
    int      rc, workint;
    char     filename[128] = "";
    FILE    *fptr;
    int      semid;
    int      iId=676735;

    if (pWS->nDebugFlag >= LOGDEBUG1 && sem_number != SEM_LOG)
        LogData(LOGDEBUG1, "      LockSpecSem() Entered", pWS);

    if (sem_number != SEM_LOG)
    {
    DBGPRINTF((OST,"LockSpecSem() Entered, Sem %d\n",sem_number));
    }

    rc=16;

    if (pWS->lTmpPath==0)                                       /* mode added under                  2.90 */
    {
        strcpy(filename,pWS->szQmgrPrefix);
        strcat(filename,"/qmgrs/");
        strcat(filename,pWS->szQmgrDirectory);

#if defined (MQCD_VERSION_8)                                    /* New stuff only                    2.90 */
        if (pWS->Version>MQCD_VERSION_8)                        /* moved shmem folder ?              2.90 */
            strcat(filename,"/@ipcc/shmem/BlockIP2.1");         /* new shared sem                    2.90 */
        else
            strcat(filename,"/shmem/BlockIP2.1");               /* old shared sem                         */
#else                                                           /* Old stuff                         2.90 */
        strcat(filename,"/shmem/BlockIP2.1");                   /* shared mem                             */
#endif
    }
    else
    {
        strcpy(filename,"/tmp/.");                              /* Move anchor to /tmp./ added under 2.90 */
        strcat(filename,pWS->szQMgr);
        strcat(filename,"_BlockIP2.1");                         /* shared sem                             */
    }

    if (pWS->nDebugFlag >= LOGDEBUG1 && sem_number != SEM_LOG)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"      LockSpecSem() - generate sem shmem [%s]", filename);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG1, pWS->szTmpBufff, pWS);
    }

    if((pWS->KeySemaphores=ftok(filename,iId)) == -1)
    {
        if((fptr = fopen( filename,"a+"))!=NULL)
            fclose(fptr);
        if((pWS->KeySemaphores=ftok(filename,iId)) == -1)
        {
            if (sem_number != SEM_LOG)
            {
                LogData(LOGDATA,"      LockSpecSem() - ftok error -1", pWS);
                DBGPRINTF((OST, "      LockSpecSem() - ftok error -1\n"));

            }
            perror("ftok");
            free(sops);                  /* 2.65 fix */
            return rc;
        }
    }
    if (pWS->nDebugFlag >= LOGDEBUG1 && sem_number != SEM_LOG)
        LogData(LOGDEBUG1, "      LockSpecSem() - ftok1 %ld ok", pWS);

    /* open semaphore set with 6 semaphore: */
    if ((semid = semget(pWS->KeySemaphores, 6, 0660 )) == -1)
    {
        /* create a semaphore set with 6 semaphore: */
        if (pWS->nDebugFlag >= LOGDEBUG1 && sem_number != SEM_LOG)
            LogData(LOGDEBUG1,"      LockSpecSem() - Create semaphores...", pWS);

        if ((semid = semget(pWS->KeySemaphores, 6, 0660 | IPC_CREAT)) == -1)
        {
            if (sem_number != SEM_LOG)
            {
                LogData(LOGDATA,"      LockSpecSem() - semget error -1", pWS);
                DBGPRINTF((OST,"LockSpecSem() semget error -1\n"));
            }
            free(sops);                  /* 2.65 fix */
            return rc;
        }
    }

    nsops = 2;

    sops[0].sem_num = sem_number;                          /* We only use one track                        */
    sops[0].sem_op = 0;                                    /* wait for semaphore flag to become zero       */
    sops[0].sem_flg = SEM_UNDO;                            /* take off semaphore asynchronous              */

    sops[1].sem_num = sem_number;
    sops[1].sem_op = 1;                                    /* increment semaphore -- take control of track */
    sops[1].sem_flg = SEM_UNDO | IPC_NOWAIT;               /* take off semaphore                           */

    /* Make the semop() call and report the results. */
    if ((workint = semop(semid, sops, nsops)) == -1)
    {
        if (sem_number != SEM_LOG)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"      LockSpecSem() semop: semop failed # %d rc: %d errno: %d %s", sem_number, workint, errno, strerror(errno) );
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA, pWS->szTmpBufff, pWS);
        }
    }
    else
    {
        rc=0;
    }
    free(sops);                  /* 2.65 fix */
    if (pWS->nDebugFlag >= LOGDEBUG1 && sem_number != SEM_LOG)
        LogData(LOGDEBUG1, "      LockSpecSem() Done", pWS);
    if (sem_number != SEM_LOG)
    {
        DBGPRINTF((OST,"LockSpecSem() Done\n"));
    }
    return rc;
}  /* LockSpecSem() */

int FreeSpecSem(int sem_number, workingStorage_type *pWS)
{
    int      nsops;                                        /* number of operations to do    */
    struct   sembuf *sops = (struct sembuf *) malloc(1*sizeof(struct sembuf));
    int      rc, workint;
    int      semid;

    if (pWS->nDebugFlag >= LOGDEBUG1  && sem_number != SEM_LOG)
        LogData(LOGDEBUG1, "      FreeSpecSem() Entered", pWS);
    if (sem_number != SEM_LOG)
    {
        DBGPRINTF((OST,"FreeSpecSem() Entered, Sem %d\n",sem_number));
    }
    rc=16;
    /* open semaphore set with 6 semaphore: */
    if ((semid = semget(pWS->KeySemaphores, 6, 0660 )) == -1)
    {
        /* create a semaphore set with 6 semaphore: */
        if (sem_number != SEM_LOG)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"      FreeSpecSem() semget failed # %d", sem_number);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA, pWS->szTmpBufff, pWS);
        }
        DBGPRINTF((OST,"FreeSpecSem() Failed, Sem %d\n",sem_number));
        free(sops);                  /* 2.65 fix */
        return rc;
    }

    nsops = 1;

    /* wait for semaphore to reach zero */
    sops[0].sem_num = sem_number;
    sops[0].sem_op = -1;                                   /* Give UP COntrol of track          */
    sops[0].sem_flg = SEM_UNDO | IPC_NOWAIT;               /* take off semaphore, asynchronous  */

    if ((workint = semop(semid, sops, nsops)) == -1)
    {
        if (sem_number != SEM_LOG)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"      FreeSpecSem() semop failed # %d", sem_number);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA, pWS->szTmpBufff, pWS);
        }
        free(sops);                  /* 2.65 fix */
        return rc;
    }
    else
        rc=0;
    free(sops);                  /* 2.65 fix */
    if (pWS->nDebugFlag >= LOGDEBUG1  && sem_number != SEM_LOG)
        LogData(LOGDEBUG1, "      FreeSpecSem() Done", pWS);
    if (sem_number != SEM_LOG)
    {
        DBGPRINTF((OST,"FreeSpecSem() Done, Sem %d\n",sem_number));
    }
    return rc;
}  /* FreeSpecSem() */

int LockChannelSem(workingStorage_type *pWS)
{
    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1, "    LockChannelSem() Entered", pWS);
    return LockSpecSem(SEM_CHL, pWS);
}

int FreeChannelSem(workingStorage_type *pWS)
{
    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1, "    FreeChannelSem() Entered", pWS);
    return FreeSpecSem(SEM_CHL, pWS);
}

int LockRenameSem(workingStorage_type *pWS)
{
    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1, "    LockRenameSem() Entered", pWS);
    return LockSpecSem(SEM_RENAME, pWS);
}

int FreeRenameSem(workingStorage_type *pWS)
{
    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1, "    FreeRenameSem() Entered", pWS);
    return FreeSpecSem(SEM_RENAME, pWS);
}



int LockLogSem(workingStorage_type *pWS)
{
    return LockSpecSem(SEM_LOG, pWS);
}

int FreeLogSem(workingStorage_type *pWS)
{
    return FreeSpecSem(SEM_LOG, pWS);
}

#endif    /* exit for UNIX */

#ifdef WIN32
void OpenMTX(workingStorage_type *pWS, char *szLogFile)
{
    FILE    *fptrx;                                        /* Declare file handles for input and output   */
    char szFile[256];
    char szFile2[256];

    strcpy(szFile2,szLogFile);
    translate2(szFile2,"","");                             /* remove \\, : and ' '          */
    strcpy(szFile,"Global\\Ms0o-");
    strcat(szFile,szFile2);
    if ((pWS->hMutex == (void*)ERROR_INVALID_HANDLE) || !strcmp(szFile,pWS->szMutexKey))
    {
        strcpy(pWS->szMutexKey,szFile);
        pWS->hMutex = CreateMutex( NULL, FALSE, szFile );
        if (pWS->hMutex == (void*)ERROR_INVALID_HANDLE)
        {
            if((fptrx = fopen( "c:\\BlockIP2x.log",LOGMODE))!=NULL)
            {
                fprintf(fptrx,"  CreateMutex failed Error code %ld", GetLastError());
                fclose(fptrx);
            }
        }
    }
}

void OpenMTXShm(workingStorage_type *pWS)
{
    char     szMutexKey[256];
    char     szMutexKey2[256];

    if (pWS->hMutexShm == (void*)ERROR_INVALID_HANDLE)
    {
        strcpy(szMutexKey2,pWS->szQMgr);
        strcpy(szMutexKey,"Global\\BlockIP2-");
        strcat(szMutexKey,szMutexKey2);
        strcat(szMutexKey,"-shmMutex");
        pWS->hMutexShm = CreateMutex( NULL, FALSE, szMutexKey);
        if (pWS->hMutexShm == (void*)ERROR_INVALID_HANDLE)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  CreateMutex failed Error code %ld", GetLastError());
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA, pWS->szTmpBufff, pWS);
        }
    }
}


int LockChannelSem(workingStorage_type *pWS)
{
    long     lCount=0;
    DWORD    dwWaitResult;

    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1, "    LockChannelSem() Entered", pWS);
    OpenMTXShm(pWS);
    lCount=0;
    while( ( dwWaitResult = WaitForSingleObject( pWS->hMutexShm, 5L ) ) == WAIT_TIMEOUT)
        lCount++;

    return 0;
}

int FreeChannelSem(workingStorage_type *pWS)
{
    if (pWS->nDebugFlag >= LOGDEBUG1)
        LogData(LOGDEBUG1, "    FreeChannelSem() Entered", pWS);
    ReleaseMutex(pWS->hMutexShm);
    CloseHandle(pWS->hMutexShm);
    pWS->hMutexShm = (void*)ERROR_INVALID_HANDLE;
    return 0;
}

void OpenMTXBlockIP2S(workingStorage_type *pWS, HANDLE *MutexBlockIP2S)
{
    char     szMutexKey[256];
    char     szMutexKey2[256];

    if (MutexBlockIP2S == (void*)ERROR_INVALID_HANDLE)
    {
        strcpy(szMutexKey2,pWS->szQMgr);
        strcpy(szMutexKey,"Global\\BlockIP2-");
        strcat(szMutexKey,szMutexKey2);
        strcat(szMutexKey,"-BlockIP2S");
        MutexBlockIP2S = CreateMutex( NULL, FALSE, szMutexKey);
        if (MutexBlockIP2S == (void*)ERROR_INVALID_HANDLE)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  CreateMutex failed Error code %ld", GetLastError());
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA, pWS->szTmpBufff, pWS);
        }
    }
}

int LockBlockIP2S(workingStorage_type *pWS, HANDLE *MutexBlockIP2S)
{
    long     lCount=0;
    DWORD    dwWaitResult;

    if (pWS->nDebugFlag >= LOGDEBUG8)
        LogData(LOGDEBUG8, "      LockBlockIP2S() Entered", pWS);
    OpenMTXShm(pWS);
    lCount=0;
    while( ( dwWaitResult = WaitForSingleObject( pWS->hMutexShm, 5L ) ) == WAIT_TIMEOUT)
        lCount++;

    return 0;
}

int FreeBlockIP2S(workingStorage_type *pWS, HANDLE *MutexBlockIP2S)
{
    if (pWS->nDebugFlag >= LOGDEBUG8)
        LogData(LOGDEBUG8, "      FreeBlockIP2S() Entered", pWS);
    ReleaseMutex(MutexBlockIP2S);
    CloseHandle(MutexBlockIP2S);
    MutexBlockIP2S = (void*)ERROR_INVALID_HANDLE;
    return 0;
}

#ifndef STB_DELETE
/* Replaced by STB */
int LoadRegistry(workingStorage_type *pWS) {

    HMODULE module_handle;
    int module_name_len;
    char module_name[256];
    int rc=0;   /* OK */

    /* find module handle of the executing dll (Blockip) */
    /* name has been found in the definition of the exit name */
    if ((module_handle = GetModuleHandle(pWS->szSCYNAME)) == NULL)
    {
        sprintf(pWS->szTmpBufff,"GetModuleHandle returns NULL");
        LogData(LOGDATA,pWS->szTmpBufff, pWS);
        rc=4;   /* error */
    }

    /* now that we have a handle, we can figure out the path and the name */
    module_name_len = GetModuleFileName(module_handle, module_name, sizeof(module_name));

    if (pWS->nDebugFlag >= LOGDEBUG8)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"GetModuleFileName returns %s and length %d", module_name, module_name_len);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG8, pWS->szTmpBufff, pWS);
    }

    /* find last occurence of backslash char in order get the path only (basename), we do need the remaining part */
    /* keep that name in the program workspace */
    *(strrchr(module_name, '\\')) = '\0';
    strcpy(pWS->ExitPath, module_name);

    if (pWS->nDebugFlag >= LOGDEBUG8)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"basename of GetModuleFileName is %s", module_name);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG8, pWS->szTmpBufff, pWS);
    }
    return rc;
}
/* End of replaced by STB */
#else
#define DMKEY "SOFTWARE\\IBM\\MQSeries\\CurrentVersion\\Configuration\\QueueManager\\"
#define SUBKEY "ExitsDefaultPath"


int LoadRegistry(workingStorage_type *pWS) {
    DWORD    rc;
    DWORD    len;
    HKEY     dmKey;
    DWORD    dwType;
    char     key[256];

    strcpy(key,DMKEY);
    strcat(key,pWS->szQMgr);
    strcat(key,"\\ExitPath\\");

    len = sizeof(pWS->ExitPath);

    rc = RegOpenKeyEx( HKEY_LOCAL_MACHINE,  key,   0,   KEY_QUERY_VALUE,   &dmKey   );
    if (rc != ERROR_SUCCESS)
    {
        if (pWS->nDebugFlag >= LOGDEBUG8)
        {
            DWORD eNum;
            char  errMsg[256];

            /* Error encountered; generate message and exit. */
            eNum = GetLastError();
            WINFmtMsg(eNum, errMsg, sizeof(errMsg));
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  LoadRegistry() RegOpenKeyEx failed %d %s",eNum,errMsg);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDEBUG8, pWS->szTmpBufff, pWS);
        }
        return 4;
    }

    RegQueryValueEx( dmKey, SUBKEY, 0, &dwType, (LPBYTE)pWS->ExitPath, &len );

    if (pWS->nDebugFlag >= LOGDEBUG8)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  ExitsDefaultPath for QMGR is [%s] in Registry", pWS->ExitPath);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG8,pWS->szTmpBufff,pWS);
    }

    if (RegQueryValueEx(dmKey, SUBKEY, 0, &dwType, (LPBYTE)pWS->ExitPath, &len )!=ERROR_SUCCESS)
    {
        if (pWS->nDebugFlag >= LOGDEBUG8)
        {
            DWORD eNum;
            char  errMsg[256];

            /* Error encountered; generate message and exit. */
            eNum = GetLastError();
            WINFmtMsg(eNum, errMsg, sizeof(errMsg));
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  LoadRegistry() RegQueryValueEx failed-2 %d %s",eNum,errMsg);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDEBUG8, pWS->szTmpBufff, pWS);
        }
    }
    RegCloseKey( dmKey );

    return 0;
}
#endif

#endif /* WIN32 */

#if defined (__EXT_LOG_FILENAME__)

long getFileSize(char* szFileName, workingStorage_type *pWS) {

    struct   stat buf;
    FILE    *fp;
    long     lFileSize;

    fp = fopen(szFileName, "r");
    if (fp == NULL) {
        return(-1);
    }
    if (0 == fstat(fileno(fp), &buf))
    {
        lFileSize= buf.st_size ;
    }
    else
        lFileSize=-1;
    fclose(fp);
    return lFileSize;
}


long CycleFiles(workingStorage_type *pWS) {
    char     Buffer[256];
    char     szPath[256];
    char     szFile[256];
    char     szFile1[256];
    char     szDrive[256];
    char     szFileName[128];
    char     szExt[32];
    int      i,rc;
#if defined (WIN32)
    DWORD dwWaitResult;
    long lCount;
#endif
#if defined (UNIX) || defined (OS400)
    long lCount;
#endif

    if (pWS->bFileSizeCheck==0)
    {
        strcpy(szFile,GenerateLogFileName(Buffer,pWS));
        _Lsplitpath(szFile, szDrive, szPath, szFileName, szExt);

        szFileName[strlen(szFileName)-3] = '\00';
        if (pWS->nDebugFlag >= LOGDEBUG8)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"CycleFiles: File info is drive[%s] path[%s] prefix [%s]", szDrive, szPath, szFileName);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDEBUG8,pWS->szTmpBufff,pWS);
        }
        snprintf(szFile,sizeof(szFile),"%s%s%s%03ld%s",szDrive,szPath,szFileName,pWS->lLogFileCount,szExt);
        szFile[sizeof(szFile)-1]='\00';
        if (pWS->nDebugFlag >= LOGDEBUG8)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"CycleFiles: File to delete [%s]", szFile);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDEBUG8,pWS->szTmpBufff,pWS);
        }

        if (remove(szFile)==0)
        {
            if (pWS->nDebugFlag >= LOGDEBUG8)
                LogData(LOGDEBUG8,"File removed.",pWS);
        }
        else {
            if (pWS->nDebugFlag >= LOGDEBUG8)
                LogData(LOGDEBUG8,"File was not removed.",pWS);
        }
        for (i=pWS->lLogFileCount; i>1; i--)
        {
            snprintf(szFile,sizeof(szFile),"%s%s%s%03d%s",szDrive,szPath,szFileName,i,szExt);
            szFile[sizeof(szFile)-1]='\00';
            snprintf(szFile1,sizeof(szFile1),"%s%s%s%03d%s",szDrive,szPath,szFileName,i-1,szExt);
            szFile1[sizeof(szFile1)-1]='\00';

#if defined (WIN32)
            if (i==2)
            {
                OpenMTX(pWS,szFile1);                                   /* See if the Mutex is in place */
                /* Wait for the mutex before writing the logevent, to prevent problems with the file. Just one is allowed.*/
                lCount=0;
                while(( dwWaitResult = WaitForSingleObject(pWS->hMutex, 5L)) == WAIT_TIMEOUT)
                    lCount++;
            }
#endif
#if defined (UNIX) || defined (OS400)
            if (i==2)
            {
                lCount=0;
                while((LockLogSem(pWS)!=0) && (lCount < 100))
                    lCount++;
                /* LockLogSem(pWS);*/
            }
#endif

            rc= rename(szFile1,szFile);
#if defined (UNIX) || defined (OS400)
            if (i==2)
            {
                FreeLogSem(pWS);
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"CycleFiles: Waited %d to take control of \"LogLock\"", lCount);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDEBUG,pWS->szTmpBufff,pWS);
            }
#endif

#if defined (WIN32)
            if (i==2)
            {
                ReleaseMutex(pWS->hMutex);
                CloseHandle(pWS->hMutex);
                pWS->hMutex = (void*)ERROR_INVALID_HANDLE;
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"CycleFiles: Waited %d ms to take control of mutex %s", lCount*5,szFile1);
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDEBUG,pWS->szTmpBufff,pWS);
            }
#endif
            if (rc==0)
            {
                if (pWS->nDebugFlag >= LOGDEBUG8)
                {
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"CycleFiles: File renamed [%s%03d%s]", szFileName,i-1,szExt);
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDEBUG8,pWS->szTmpBufff,pWS);
                }
            }
        }

        /*        pWS->bFileSizeCheck=1;*/
        return 0;
    }
    else
        return 0;                                          /* Allways return ok when inactive */

}

long HandleCycleFiles(workingStorage_type *pWS) {
    char     Buffer[256];
    char     szFile[256];
    long     lFileSize;

#if defined (WIN32)                                        /* add mutex for rename to prevent "small" files 2.71 */
    HANDLE   hMutex= (void*)ERROR_INVALID_HANDLE;
    DWORD dwWaitResult;
    long lCount;
#endif

    if (pWS->nDebugFlag >= LOGDEBUG8)
        LogData(LOGDEBUG8, "HandleCycleFiles() Entered",pWS);
    strcpy(szFile,GenerateLogFileName(Buffer,pWS));
    lFileSize=getFileSize(szFile,pWS);
    if (pWS->nDebugFlag >= LOGDEBUG8)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  File size of %s is %ld (limit %ld)", szFile, lFileSize,pWS->lMaxlogFileSize);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG8, pWS->szTmpBufff,pWS);
    }
    if (lFileSize > pWS->lMaxlogFileSize)
    {
#if defined (WIN32)                                        /* add mutex for rename to prevent "small" files 2.71 */
        strcpy(Buffer,"Global\\Ms0o-Rename-File");	   /* Use global rename key, doesn't matter */
        hMutex = CreateMutex(NULL, FALSE, Buffer);
        if (hMutex == (void*)ERROR_INVALID_HANDLE)
        {
            LogData(LOGDATA,"Failed to create MUTEX-Rename",pWS);
            if (pWS->nDebugFlag >= LOGDEBUG8)
                LogData(LOGDEBUG8, "HandleCycleFiles() Done",pWS);
            return 0;
        }
        /* Wait for the mutex before writing the logevent, to prevent problems with the file. Just one is allowed.*/
        lCount=0;
        while((dwWaitResult = WaitForSingleObject(hMutex, 5L)) == WAIT_TIMEOUT)
            lCount++;
#endif
#if defined (UNIX) || defined (OS400)
        DBGPRINTF((OST, "  LRS -- Debugger %d\n",__LINE__));
        LockRenameSem(pWS);
#endif
        lFileSize=getFileSize(szFile,pWS);   /* See if we still need to change the files... 2.71 */
        if (pWS->nDebugFlag >= LOGDEBUG8)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"  File size of %s is %ld (limit %ld)", szFile, lFileSize,pWS->lMaxlogFileSize);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDEBUG8, pWS->szTmpBufff,pWS);
        }
        if (lFileSize > pWS->lMaxlogFileSize)
        {
            if (pWS->nDebugFlag >= LOGDEBUG8)
                LogData(LOGDEBUG8, "  Cycle of logfiles is required.",pWS);
            if (CycleFiles(pWS)==0)
                LogData(LOGDEBUG8,"  LogFiles was changed.",pWS);
            else
                LogData(LOGDATA,"LogFile change Failed.",pWS);
        }
        else
        {
            LogData(LOGDEBUG8, "  LogFile change was done by another thread.",pWS);
        }
#if defined (UNIX) || defined (OS400)
        DBGPRINTF((OST, "  FRS -- Debugger %d\n",__LINE__));
        FreeRenameSem(pWS);
#endif

#if defined (WIN32)                                        /* add mutex for rename to prevent "small" files 2.71 */
        ReleaseMutex(hMutex);
        CloseHandle(hMutex);
        hMutex = (void*)ERROR_INVALID_HANDLE;
#endif
    }
    if (pWS->nDebugFlag >= LOGDEBUG8)
        LogData(LOGDEBUG8, "HandleCycleFiles() Done",pWS);
    return 0;
}


#endif /* __EXT_LOG_FILENAME__ */


void _Lsplitpath(const char* path, char* drv, char* dir, char* name, char* ext)
{
    const char* end;                                       /* end of processed string       */
    const char* p;                                         /* search pointer                */
    const char* s;                                         /* copy pointer                  */

    /* extract drive name */
    if (path[0] && path[1]==':')
    {
        if (drv) {
            *drv++ = *path++;
            *drv++ = *path++;
            *drv = '\0';
        }
    }
    else
        if (drv)
            *drv = '\0';

    /* search for end of string or stream separator */
    for(end=path; *end && *end!=':'; )
        end++;

    /* search for begin of file extension */
    for(p=end; p>path && *--p!='\\' && *p!='/'; )
        if (*p == '.')
        {
            end = p;
            break;
        }

    if (ext)
        for(s=end; (*ext=*s++); )
            ext++;

    /* search for end of directory name */
    for(p=end; p>path; )
        if (*--p=='\\' || *p=='/')
        {
            p++;
            break;
        }

    if (name)
    {
        for(s=p; s<end; )
            *name++ = *s++;

        *name = '\0';
    }

    if (dir)
    {
        for(s=path; s<p; )
            *dir++ = *s++;

        *dir = '\0';
    }
}


#ifdef HNLUP
/*
**********************************************************************************************************
* GetHostIP
* convert name into ip address
* return 0 on success
* 1 on failure
*
* Due to the various OS implementations are the logic dedicated to the platforms
**********************************************************************************************************
*/


#if defined WIN32

#ifdef HNLUP2
/*
**********************************************************************************************************
* convert hostname on Windows
**********************************************************************************************************
*/
#define __GetHostIP 1
MQLONG GetHostIP(char *connectionName, DNSReply *DNSReplyPtr, workingStorage_type *pWS)
{
    struct   hostent *hpPtr;
    int      wsaError;
    WORD     wVersionRequested;
    WSADATA  wsaData;
    int      err;
    char    *pIP;
    struct   in_addr address;
    int      i=0;

    DNSReplyPtr->replies=0;		/* nothing yet */
    strcpy(DNSReplyPtr->szIpAddr[0],"");

    if (isalpha(connectionName[0]))
    {   /* host address is a name */
        wVersionRequested = MAKEWORD( 2, 2 );
        err = WSAStartup( wVersionRequested, &wsaData );
        if ( err != 0 )
        {
            LogData(LOGDATA,"GetHostIP(): Found no usable WinSock DLL",pWS);
            return 1;
        }
        if ( LOBYTE( wsaData.wVersion ) != 2 ||
                HIBYTE( wsaData.wVersion ) != 2 )
        {
            LogData(LOGDATA,"GetHostIP(): Found no usable 2.2 WinSock DLL",pWS);
            WSACleanup();
            return 1;
        }
        if ( (hpPtr = gethostbyname(connectionName)) == NULL)
        {
            wsaError = WSAGetLastError();
            if(wsaError == WSAHOST_NOT_FOUND || wsaError == WSANO_DATA) {
            }
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"GetHostIP(): Unknown host/IP for: [%s] wsaError %d", connectionName, wsaError);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA, pWS->szTmpBufff,pWS);
            WSACleanup();
            return 1;
        }
        else
        {
            if (hpPtr->h_addrtype == AF_INET)
            {
                while (hpPtr->h_addr_list[i] != 0)
                {
                    address.s_addr = *(u_long *) hpPtr->h_addr_list[i++];
                    if (pWS->nDebugFlag >= LOGDEBUG0)
                    {
                        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"IPv4 Address #%d: %s", i, inet_ntoa(address));
                        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                        LogData(LOGDEBUG0, pWS->szTmpBufff,pWS);
                    }
                    pIP = inet_ntoa(address);
                    strcpy(DNSReplyPtr->szIpAddr[DNSReplyPtr->replies++],pIP);
                    if (DNSReplyPtr->replies >= 15)
                        break;
                }
            } else if (hpPtr->h_addrtype == AF_INET6)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Remotehost is an IPv6 address");
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDATA, pWS->szTmpBufff,pWS);
            }
            WSACleanup();
        }
    }
    else
    {
        if (pWS->nDebugFlag >= LOGDEBUG0)
            LogData(LOGDEBUG0, "GetHostIP(): Connection name must be alpha",pWS);
        return 1;
    }

    if (pWS->nDebugFlag >= LOGDEBUG0)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"GetHostIP(): Entered with [%s] IP Address was found [%s] registered name [%s]",connectionName,pIP,hpPtr->h_name);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG0, pWS->szTmpBufff,pWS);
    }
    return 0;
}

#else
/*
**********************************************************************************************************
* convert hostname on Windows
**********************************************************************************************************
*/
#define __GetHostIP 1
MQLONG GetHostIP(char *connectionName, char *connectionAddrPtr, workingStorage_type *pWS)
{
    struct   hostent *hpPtr;
    int      wsaError;
    WORD     wVersionRequested;
    WSADATA  wsaData;
    int      err;
    char    *pIP;
    struct   in_addr address;

    memset(connectionAddrPtr,'\00',1);

    if (isalpha(connectionName[0]))
    {   /* host address is a name */
        wVersionRequested = MAKEWORD( 2, 2 );
        err = WSAStartup( wVersionRequested, &wsaData );
        if ( err != 0 )
        {
            LogData(LOGDATA,"GetHostIP(): Found no usable WinSock DLL",pWS);
            return 1;
        }
        if ( LOBYTE( wsaData.wVersion ) != 2 ||
                HIBYTE( wsaData.wVersion ) != 2 )
        {
            LogData(LOGDATA,"GetHostIP(): Found no usable 2.2 WinSock DLL",pWS);
            WSACleanup();
            return 1;
        }
        if ( (hpPtr = gethostbyname(connectionName)) == NULL)
        {
            wsaError = WSAGetLastError();
            if(wsaError == WSAHOST_NOT_FOUND || wsaError == WSANO_DATA) {
            }
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"GetHostIP(): Unknown host/IP for: [%s] wsaError %d", connectionName, wsaError);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA, pWS->szTmpBufff,pWS);
            WSACleanup();
            return 1;
        }
        else
        {
            memcpy(&address,hpPtr->h_addr,hpPtr->h_length);
            pIP = inet_ntoa(address);
            memcpy(connectionAddrPtr,pIP,strlen(pIP));
            memset(connectionAddrPtr+strlen(pIP),'\00',1);
            WSACleanup();
        }
    }
    else
    {
        if (pWS->nDebugFlag >= LOGDEBUG0)
            LogData(LOGDEBUG0, "GetHostIP(): Connection name must be alpha",pWS);
        return 1;
    }

    if (pWS->nDebugFlag >= LOGDEBUG0)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"GetHostIP(): Entered with [%s] IP Address was found [%s] registered name [%s]",connectionName,pIP,hpPtr->h_name);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG0, pWS->szTmpBufff,pWS);
    }
    return 0;
}
#endif
#endif  /* WIN32 */

#if defined UNIX || defined (OS400)

#ifdef HNLUP2

/*
**********************************************************************************************************
* convert hostname on UNIX implementation
**********************************************************************************************************
*/
#define __GetHostIP 1
MQLONG GetHostIP(char *connectionName, DNSReply *DNSReplyPtr, workingStorage_type *pWS)
{
#if defined AIX || defined SOLARIS || defined OS400
#define _GetHostByName_ 1
    struct   hostent *hpPtr;
#if defined _THREAD_SAFE_JHP && !defined AIX && !defined OS400
    struct   hostent hp;
    char     hpBuffer[1024];
    int      hpError;
#endif    /* THREAD SAFE */
#endif    /* AIX/SOLARIS */
    /*
    #if defined AIX
        int    res;
    #endif
    */
#if !defined _GetHostByName_
    int      res;
    struct   hostent *hpPtr, hp;
    char     hpBuffer[1024];
    int      hpError;
#endif  /* _GetHostByName_ */
    char    *pIP;
    struct   in_addr address;
    int      i=0;

    DNSReplyPtr->replies=0;		/* nothing yet */
    strcpy(DNSReplyPtr->szIpAddr[0],"");

    if (isalpha(connectionName[0]))
    {   /* host address is a name */

#if defined(AIX) || defined (OS400)
        hpPtr = gethostbyname(connectionName);
        LogData(LOGDATA, "Loop4b",pWS);
        if (hpPtr == (struct hostent *) 0)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"GetHostIP(): Unknown host/IP for: [%s]", connectionName);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA, pWS->szTmpBufff,pWS);
            return 1;
        }
#endif      /* AIX */

#ifdef SOLARIS
#ifdef _THREAD_SAFE_JHP
        hpPtr = gethostbyname_r(connectionName, &hp, hpBuffer, sizeof(hpBuffer), &hpError);

        if (!hpPtr || !hp.h_addr_list || !hp.h_addr_list[0])
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"GetHostIP(): Unknown host/IP for: [%s] status %d", connectionName, hpError);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA, pWS->szTmpBufff,pWS);
            return 1;
        }
#else
        hpPtr = gethostbyname(connectionName);
        if (hpPtr == (struct hostent *) 0)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"GetHostIP(): Unknown host/IP for: [%s]", connectionName);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA, pWS->szTmpBufff,pWS);
            return 1;
        }
#endif  /* _THREAD_SAFE_JHP */
#endif  /* solaris */

        /* default t.x. linux */
#ifndef _GetHostByName_
        res  = gethostbyname_r (connectionName, &hp, hpBuffer, sizeof(hpBuffer), &hpPtr, &hpError);
        if (res || !hpPtr || !hp.h_addr_list || !hp.h_addr_list[0])
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"GetHostIP(): Unknown host/IP for: [%s] status %d", connectionName, hpError);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA, pWS->szTmpBufff,pWS);
            return 1;
        }
#endif
        else
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"GetHostIP(): hpPtr->h_addrtype %d AF_INET %d sizeof(1) %d %d ", hpPtr->h_addrtype, AF_INET,sizeof(hpPtr->h_addrtype), sizeof(AF_INET));
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA, pWS->szTmpBufff,pWS);
            if (hpPtr->h_addrtype == AF_INET)
            {
                while (hpPtr->h_addr_list[i] != 0)
                {
                    address.s_addr = *(unsigned long *) hpPtr->h_addr_list[i++];
                    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"IPv4 Address #%d: %s", i, inet_ntoa(address));
                    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                    LogData(LOGDATA, pWS->szTmpBufff,pWS);

                    pIP = inet_ntoa(address);
                    strcpy(DNSReplyPtr->szIpAddr[DNSReplyPtr->replies++],pIP);
                    if (DNSReplyPtr->replies >= 15)
                        break;
                }
            } else if (hpPtr->h_addrtype == AF_INET6)
            {
                snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"Remotehost is an IPv6 address");
                pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
                LogData(LOGDATA, pWS->szTmpBufff,pWS);
            }
        }
    }
    else
    {
        if (pWS->nDebugFlag >= LOGDEBUG0)
            LogData(LOGDEBUG0, "GetHostIP(): Connection name must be alpha",pWS);
        return 1;
    }
    if (pWS->nDebugFlag >= LOGDEBUG0)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"GetHostIP(): Entered with [%s] IP Address was found [%s] registered name [%s]",connectionName,pIP,hpPtr->h_name);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG0, pWS->szTmpBufff,pWS);
    }
    return 0;
}

#else
/*
**********************************************************************************************************
* convert hostname on UNIX implementation
**********************************************************************************************************
*/
#define __GetHostIP 1
MQLONG GetHostIP(char *connectionName, char *connectionAddrPtr, workingStorage_type *pWS)
{
#if defined AIX || defined SOLARIS || defined OS400
#define _GetHostByName_ 1
    struct   hostent *hpPtr;
#if defined _THREAD_SAFE_JHP && !defined AIX && !defined OS400
    struct   hostent hp;
    char     hpBuffer[1024];
    int      hpError;
#endif    /* THREAD SAFE */
#endif    /* AIX/SOLARIS */
    /*
    #if defined AIX
    int    res;
    #endif
    */
#if !defined _GetHostByName_
    int      res;
    struct   hostent *hpPtr, hp;
    char     hpBuffer[1024];
    int      hpError;
#endif  /* _GetHostByName_ */
    char    *pIP;
    struct   in_addr address;

    memset(connectionAddrPtr,'\00',1);

    if (isalpha(connectionName[0]))
    {   /* host address is a name */

#if defined(AIX) || defined (OS400)
        /*
        Not needed on Windows, HP-UX 11 and AIX 5 (since gethostbyaddr() is then thread-safe on its own).
        #ifdef _THREAD_SAFE_JHP
        res = gethostbyname_r(connectionName, &hp, (struct hostent_data *) &hpBuffer);
        LogData(LOGDATA,"Back...",pWS);
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"GetHostIP(): Unknown host/IP for: [%s] status %d %d", connectionName, errno,res);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDATA, pWS->szTmpBufff,pWS);
        if (res || !hp.h_addr_list || !hp.h_addr_list[0])
        {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"GetHostIP(): Unknown host/IP for: [%s] status %d %d", connectionName, errno,res);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDATA, pWS->szTmpBufff,pWS);
        return 1;
        }
        #else
*/
hpPtr = gethostbyname(connectionName);
if (hpPtr == (struct hostent *) 0)
{
    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"GetHostIP(): Unknown host/IP for: [%s]", connectionName);
    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
    LogData(LOGDATA, pWS->szTmpBufff,pWS);
    return 1;
}
/*#endif*/  /* _THREAD_SAFE_JHP */
#endif      /* AIX */

#ifdef SOLARIS
#ifdef _THREAD_SAFE_JHP
        hpPtr = gethostbyname_r(connectionName, &hp, hpBuffer, sizeof(hpBuffer), &hpError);

        if (!hpPtr || !hp.h_addr_list || !hp.h_addr_list[0])
        {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"GetHostIP(): Unknown host/IP for: [%s] status %d", connectionName, hpError);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDATA, pWS->szTmpBufff,pWS);
        return 1;
        }
#else
hpPtr = gethostbyname(connectionName);
if (hpPtr == (struct hostent *) 0)
{
    snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"GetHostIP(): Unknown host/IP for: [%s]", connectionName);
    pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
    LogData(LOGDATA, pWS->szTmpBufff,pWS);
    return 1;
}
#endif  /* _THREAD_SAFE_JHP */
#endif  /* solaris */

        /* default t.x. linux */
#ifndef _GetHostByName_
        res  = gethostbyname_r (connectionName, &hp, hpBuffer, sizeof(hpBuffer), &hpPtr, &hpError);
        if (res || !hpPtr || !hp.h_addr_list || !hp.h_addr_list[0])
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"GetHostIP(): Unknown host/IP for: [%s] status %d", connectionName, hpError);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA, pWS->szTmpBufff,pWS);
            return 1;
        }
#endif
        else
        {
            memcpy(&address,hpPtr->h_addr,hpPtr->h_length);
            pIP = (char*)inet_ntoa(address);
            memcpy(connectionAddrPtr,pIP,strlen(pIP));
            memset(connectionAddrPtr+strlen(pIP),'\00',1);
        }
    }
    else
    {
        if (pWS->nDebugFlag >= LOGDEBUG0)
            LogData(LOGDEBUG0, "GetHostIP(): Connection name must be alpha",pWS);
        return 1;
    }
    if (pWS->nDebugFlag >= LOGDEBUG0)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"GetHostIP(): Entered with [%s] IP Address was found [%s] registered name [%s]",connectionName,pIP,hpPtr->h_name);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG0, pWS->szTmpBufff,pWS);
    }
    return 0;
}
#endif   /* !HNLUP2 */
#endif   /* UNIX */

#if defined __MVS__
/*
**********************************************************************************************************
* convert hostname on z/OS systems
**********************************************************************************************************
*/
#define __GetHostIP 1
MQLONG GetHostIP(char *connectionName, char *connectionAddrPtr, workingStorage_type *pWS)
{
    struct   hostent *hpPtr;
    long     hp;
    char     hpBuffer[1024];
    int      hpError;
    long     status;
    char    *cp;
    long     len;

    memset(connectionAddrPtr,'\00',1);

    if (isalpha(connectionName[0]))
    {   /* host address is a name */

        len=strlen(connectionName);
        BLOCKIPN(connectionName,len,(char *)&hp,&status);

        hpPtr=(struct hostent*)hp;
        if (hpPtr == (struct hostent *) 0)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"GetHostIP(): Unknown host/IP for: [%s] errno returned by gethostbyname %d %d", connectionName,errno,hp);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDATA, pWS->szTmpBufff,pWS);
            return 1;
        }
        else
        {
            inet_ntop(hpPtr->h_addrtype, hpPtr->h_addr_list[0], (char*) &hpBuffer, sizeof(hpBuffer));
            memcpy(connectionAddrPtr,hpBuffer,strlen(hpBuffer));
            memset(connectionAddrPtr+strlen(hpBuffer),'\00',1);
        }
    }
    else
    {
        if (pWS->nDebugFlag >= LOGDEBUG0)
            LogData(LOGDEBUG0, "GetHostIP(): Connection name must be alpha",pWS);
        return 1;
    }
    if (pWS->nDebugFlag >= LOGDEBUG0)
    {
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"GetHostIP(): Entered with [%s] IP Address was found [%s] registered name [%s]",connectionName,hpBuffer,hpPtr->h_name);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDEBUG0, pWS->szTmpBufff,pWS);
    }
    return 0;
}
#endif   /* __MVS__ */

#ifndef __GetHostIP
/*
**********************************************************************************************************
* Handle unsupported platforms
**********************************************************************************************************
*/
MQLONG GetHostIP(char *connectionName, char *connectionAddrPtr, workingStorage_type *pWS)
{
    memset(connectionAddrPtr,'\00',1);
    LogData(LOGDATA, "GetHostIP(): hostname resolution is not implemented.",pWS);
    return 1;
}
#endif   /* __GetHosIP */

#endif   /* HNLUP */

#ifdef HNLUP
#if defined __MVS__
/* char *
 * inet_ntop(af, src, dst, size)
 *      convert a network format address to presentation format.
 * return:
 *      pointer to presentation format address (`dst'), or NULL (see errno).
 * author:
 *      Paul Vixie, 1996.
 */
char * inet_ntop(int af, char *src, char *dst, int size)
{
    switch (af)
    {
    case AF_INET:
        return (inet_ntop4(src, dst, size));
    case AF_INET6:
        return (inet_ntop6(src, dst, size));
    default:
        /*  errno = EAFNOSUPPORT; */
        printf("Error....\n");
        errno = 1;
        return (NULL);
    }
    /* NOTREACHED */
}

/* const char *
 * inet_ntop4(src, dst, size)
 *      format an IPv4 address, more or less like inet_ntoa()
 * return:
 *      `dst' (as a const)
 * notes:
 *      (1) uses no statics
 *      (2) takes a u_char* not an in_addr as input
 * author:
 *      Paul Vixie, 1996.
 */
char * inet_ntop4(char *src, char *dst, int size)
{
    static const char fmt[] = "%u.%u.%u.%u";
    char tmp[sizeof "255.255.255.255"];

    if (SPRINTF((tmp, fmt, src[0], src[1], src[2], src[3])) > size)
    {
        errno = 1;
        return (NULL);
    }
    strcpy(dst, tmp);
    return (dst);
}

/* const char *
 * inet_ntop6(src, dst, size)
 *      convert IPv6 binary address into presentation (printable) format
 * author:
 *      Paul Vixie, 1996.
 */
char * inet_ntop6(char *src, char *dst, int size)
{
    /*
     * Note that int32_t and int16_t need only be "at least" large enough
     * to contain a value of the specified size.  On some systems, like
     * Crays, there is no such thing as an integer variable with 16 bits.
     * Keep this in mind if you think this function should have been coded
     * to use pointer overlays.  All the world's not a VAX.
     */
    char tmp[sizeof "ffff:ffff:ffff:ffff:ffff:ffff:255.255.255.255"], *tp;
    struct { int base, len; } best, cur;
    /*  u_int words[NS_IN6ADDRSZ / NS_INT16SZ];*/
    int words[NS_IN6ADDRSZ / NS_INT16SZ];
    int i;

    /*
     * Preprocess:
     *  Copy the input (bytewise) array into a wordwise array.
     *  Find the longest run of 0x00's in src[] for :: shorthanding.
     */
    memset(words, '\0', sizeof words);
    /*  printf("here we are2 NS_IN6ADDRSZ %d NS_INT16SZ %d\n",NS_IN6ADDRSZ,NS_INT16SZ);*/
    for (i = 0; i < NS_IN6ADDRSZ; i++)
        words[i / 2] |= (src[i] << ((1 - (i % 2)) << 3));
    best.base = -1;
    cur.base = -1;
    for (i = 0; i < (NS_IN6ADDRSZ / NS_INT16SZ); i++)
    {
        if (words[i] == 0)
        {
            if (cur.base == -1)
                cur.base = i, cur.len = 1;
            else
                cur.len++;
        } else
        {
            if (cur.base != -1)
            {
                if (best.base == -1 || cur.len > best.len)
                    best = cur;
                cur.base = -1;
            }
        }
    }
    if (cur.base != -1)
    {
        if (best.base == -1 || cur.len > best.len)
            best = cur;
    }
    if (best.base != -1 && best.len < 2)
        best.base = -1;

    /*
     * Format the result.
     */
    tp = tmp;
    for (i = 0; i < (NS_IN6ADDRSZ / NS_INT16SZ); i++)
    {
        /* Are we inside the best run of 0x00's? */
        if (best.base != -1 && i >= best.base &&
                i < (best.base + best.len))
        {
            if (i == best.base)
                *tp++ = ':';
            continue;
        }
        /* Are we following an initial run of 0x00s or any real hex? */
        if (i != 0)
            *tp++ = ':';
        /* Is this address an encapsulated IPv4? */
        if (i == 6 && best.base == 0 &&
                (best.len == 6 || (best.len == 5 && words[5] == 0xffff)))
        {
            if (!inet_ntop4(src+12, tp, sizeof tmp - (tp - tmp)))
                return (NULL);
            tp += strlen(tp);
            break;
        }
        tp += SPRINTF((tp, "%x", words[i]));
    }
    /* Was it a trailing run of 0x00's? */
    if (best.base != -1 && (best.base + best.len) == (NS_IN6ADDRSZ / NS_INT16SZ))
        *tp++ = ':';
    *tp++ = '\0';

    /*
     * Check for overflow, copy, and we're done.
     */
    if ((int)(tp - tmp) > size)
    {
        errno = 1;
        return (NULL);
    }
    strcpy(dst, tmp);
    return (dst);
}

#endif /* __MVS__ */
#endif  /*HNLUP  inet_ntop include */


#if defined(WIN32)
void WINFmtMsg(DWORD eNum, char *ErrorMsg, int lenErrMsg)
{
    /*    DWORD eNum;*/
    TCHAR sysMsg[256];
    TCHAR* p;


    /*        eNum = GetLastError( );*/
    FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM |
                   FORMAT_MESSAGE_IGNORE_INSERTS,
                   NULL, eNum,
                   MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                   sysMsg, sizeof(sysMsg)-1, NULL );

    /* Trim the end of the line and terminate it with a null*/
    p = sysMsg;
    while( ( *p > 31 ) || ( *p == 9 ) )
        ++p;

do { *p-- = 0; } while( ( p >= sysMsg ) && ( ( *p == '.' ) || ( *p < 33 ) ) );

    strncpy(ErrorMsg,sysMsg,lenErrMsg);
}
#endif

/*********************************************************************************************
* HexDump: function dump a storage area for investigations
*
* Remark:
*
* @Return void
*********************************************************************************************/
void HexDump(void *pAddressIn, long  lSize, workingStorage_type *pWS)
{
    char szBuf[100];
    long lIndent = 1;
    long lOutLen, lIndex, lIndex2, lOutLen2;
    long lRelPos;
    struct { char *pData; unsigned long lSize; } buf;
    unsigned char *pTmp,ucTmp;
    unsigned char *pAddress = (unsigned char *)pAddressIn;

    buf.pData   = (char *)pAddress;
    buf.lSize   = lSize;

    while (buf.lSize > 0)
    {
        pTmp     = (unsigned char *)buf.pData;
        lOutLen  = (int)buf.lSize;
        if (lOutLen > 16)
            lOutLen = 16;

        /* create a 66-character formatted output line: */
        sprintf(szBuf, " >                            "
                "         |            "
                "    | %08lX(+%04lX)", (long unsigned int)pTmp, pTmp-pAddress);
        lOutLen2 = lOutLen;

        for(lIndex = 1+lIndent, lIndex2 = 54-15+lIndent, lRelPos = 0;
                lOutLen2;
                lOutLen2--, lIndex += 2, lIndex2++
           )
        {
            ucTmp = *pTmp++;

            sprintf(szBuf + lIndex, "%02X ", (unsigned short)ucTmp);
            if(!isprint(ucTmp))  ucTmp = '.'; /* nonprintable char */
            szBuf[lIndex2] = ucTmp;

            if (!(++lRelPos & 3))     /* extra blank after 4 bytes   */
            {  lIndex++; szBuf[lIndex+2] = ' '; }
        }

        if (!(lRelPos & 3)) lIndex--;

        szBuf[lIndex  ]   = '<';
        szBuf[lIndex+1]   = ' ';

        /*        LogData(LOGINTERNAL, szBuf, pWS);*/
        LogData(LOGDATA, szBuf, pWS);

        buf.pData   += lOutLen;
        buf.lSize   -= lOutLen;
    }
}

/*********************************************************************************************
* HasMasterChanged: function decide if master file has been changed (shift of qmgr)
*
* Remark:
*
* @Return int
*         0:  No change needed.
*         4:  Master file has changed.
*********************************************************************************************/
#if defined(UNIX) || defined (OS400)
int HasMasterChanged(char *ts, workingStorage_type *pWS)
{
    int rc=0;

    struct stat statb;
    struct tm  *ts1;
    char   buf[100];

    if (pWS->nDebugFlag>=LOGDEBUG1)
        LogData(LOGDEBUG1, "HasMasterChanged() Entered" ,pWS);

    DBGPRINTF((OST,"HasMasterChanged() Entered\n"));


    if (stat(pWS->szMasterFile, &statb)<0)    /*        _stat() failed_ */
    {
        /* Just do nothing for a start..*/
        snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"HasMasterChanged(): Failed to locate [%s]",pWS->szMasterFile);
        pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
        LogData(LOGDATA, pWS->szTmpBufff,pWS);

    DBGPRINTF((OST,"HasMasterChanged(): Failed to locate [%s]",pWS->szMasterFile));

    }
    else
    {
	ts1 = localtime(&statb.st_mtime);
	strftime(buf, sizeof(buf), "%Y%m%d%H%M%S", ts1);

        DBGPRINTF((OST,"HasMasterChanged() stat %s \n", buf));

        if (pWS->nDebugFlag>=LOGDEBUG0)
        {
            snprintf(pWS->szTmpBufff,sizeof(pWS->szTmpBufff),"HasMasterChanged(): Was found [%s] ts1 %s ts %s",pWS->szMasterFile,buf,ts);
            pWS->szTmpBufff[sizeof(pWS->szTmpBufff)-1]='\00';
            LogData(LOGDEBUG0, pWS->szTmpBufff,pWS);
        }

        if (strcmp(ts,buf))
        {
            if (pWS->nDebugFlag>=LOGDEBUG1)
                LogData(LOGDEBUG1, "  master was changed. Reinit..." ,pWS);

                DBGPRINTF((OST,"HasMasterChanged() master was changed. Reinit...\n"));

            /* time has changed..... update it.. */

            rc=4;
            strcpy(ts,buf);
        }

    }
    if (pWS->nDebugFlag>=LOGDEBUG1)
        LogData(LOGDEBUG1, "HasMasterChanged() Left" ,pWS);

    DBGPRINTF((OST,"HasMasterChanged() Left rc %d\n",rc));

    return rc;
}

#endif

/* Insert long MCAUser into MQCD 2.94 */
void InsertLongMCAUser(PMQCD pChannelDefinition, workingStorage_type *pWS, ws_type *pWS1)
{
#ifdef MQCD_VERSION_6                                              /* for compilers */
    if (pWS->nDebugFlag>=LOGDEBUG8)
    {
        LogData( LOGDEBUG8, "InsertLongMCAUser() Entered", pWS);
        sprintf(pWS->szTmpBufff,"LongMCAUser %s",pWS1->LongMCAUser);
        LogData( LOGDEBUG8,pWS->szTmpBufff, pWS);
    }
    if ((strlen(pWS1->LongMCAUser)>0))
    {
        memset(pChannelDefinition->MCAUserIdentifier,' ',sizeof(pChannelDefinition->MCAUserIdentifier));
        strcpy(pChannelDefinition->MCAUserIdentifier,"NoBody");
        memset(pChannelDefinition->RemoteUserIdentifier,' ',sizeof(pChannelDefinition->RemoteUserIdentifier));
        strcpy(pChannelDefinition->RemoteUserIdentifier," ");
        pChannelDefinition->LongMCAUserIdLength     = (int)strlen(pWS1->LongMCAUser);
        pChannelDefinition->LongMCAUserIdPtr        = pWS1->LongMCAUser;
        pChannelDefinition->LongRemoteUserIdLength  = (int)strlen(pWS1->LongMCAUser);
        pChannelDefinition->LongRemoteUserIdPtr     = pWS1->LongMCAUser;
        if (pWS->nDebugFlag>=LOGDEBUG8)
            LogData( LOGDEBUG, "InsertLongMCAUser() Done ", pWS);
    }
#endif
    return;
}


/* Insert MCA Userid into persisistent storage  2.94 */
void SetLongMCAUser(char *szMCAUser, workingStorage_type *pWS)
{
    if (pWS->nDebugFlag>=LOGDEBUG)
    {
        LogData( LOGDEBUG, "SetLongMCAUser() Entered", pWS);
        sprintf(pWS->szTmpBufff,"LongMCAUser kept [%s]",szMCAUser);
        LogData( LOGDEBUG,pWS->szTmpBufff, pWS);
    }
    strncpy(pWS->ppWS1->LongMCAUser,szMCAUser,sizeof(pWS->ppWS1->LongMCAUser));
    return;
}


/* End of file */

